!
! Module: exposure
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This module contains routines used in exposure calculations.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module exposure

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile:
  use cal	! contains 'use caloaldefs' and 'use caltypes'
  use dal
  use param  
  use errorHandling
  use dss
!  use eventAttributes ! in saslib. For checking which flags have been applied.
!  use dataFilter      ! in selectlib.

!  use fits_utils
  use ssc_misc    ! for fracVal
  use constants
  use dss_aux     ! for copyRange
  use array_utils ! for trapezoidSum
!  use wcs_aux
!  use coordinate ! for instToTan
!  use math_utils ! cubicSpline routines
!use dump
!use plot_utils

  implicit none 

  private 

  public sampleEnergyRange
  public getEffectiveArea
  public getFilterTransmission
  public localIntegrate
  public constructQeArray
  public getQe

  interface sampleEnergyRange
    module procedure sampleEnergyRangeDss, sampleEnergyRangeNormal
  end interface

  interface localIntegrate
    module procedure localIntegrateVector, localIntegrateScalar
  end interface

  real(double), parameter, public ::&
    minEeV = 50.0,&
    maxEeV = 20000.0,&
    binSizeInLnEeV = 0.3,&
    coarseVigThetaArcsecBinSize = 60.0,&
    coarseVigPhiBinSize = Pi / 18.0

contains 

  !---------------------------------------------------------------------
  subroutine getQe(EeVsamples, usePattern, qe)
  !---------------------------------------------------------------------
  ! NOTE! The cal MUST be set up for the correct instrument and CCD number
  ! before this function is called.
  !
  !*** NOTE! This calculates the QE at pixel (50,100) of the currently
  ! selected CCD chip.
  !.....................................................................

    implicit none

    real(single), intent(in)  :: EeVsamples(:)
    logical, intent(in) :: usePattern(:)
    real(single), intent(out) :: qe(size(EeVsamples))

    real(single) :: qeArray(size(usePattern),size(EeVsamples))
    integer(int32) :: Ei

    !-------------------------------------------------------------------
    ! Start

    call constructQeArray(EeVsamples, qeArray)
    do Ei = 1, size(EeVsamples)
      qe(Ei) = sum(qeArray(:,Ei), usePattern)
    end do

  end subroutine getQe

  !---------------------------------------------------------------------
  subroutine constructQeArray(EeVsamples, qeArray)
  !---------------------------------------------------------------------
  ! NOTE! The cal MUST be set up for the correct instrument and CCD number
  ! before this function is called.
  !
  !*** NOTE! This calculates the QE at pixel (50,100) of the currently
  ! selected CCD chip.
  !.....................................................................

    implicit none

    real(single), intent(out) :: qeArray(:,:)
    real(single), intent(in)  :: EeVsamples(size(qeArray, 2))

    integer(int32) :: numPatterns, numEsamples, patternId, eI
logical :: doDummy = .false.

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "  Constructing array (with &
      &pattern-energy axes) of samples of the quantum efficiency for &
      &this CCD.")

if (doDummy) then
  call warning('constructQeArray', "This is a dummy routine! It doesn't return the real QE function.")
end if

    numPatterns = size(qeArray, 1)
    numEsamples = size(qeArray, 2)

    do patternId = 0, numPatterns - 1
      do eI = 1, numEsamples
if (doDummy) then
  qeArray(patternId + 1, eI) = 1.0 / real(numPatterns)
else
        qeArray(patternId + 1, eI)&
          = CAL_getQuantumEfficiency(real(EeVsamples(eI), double)&
          , 50_int16, 100_int16, patternId)
end if
      end do
    end do

  end subroutine constructQeArray

  !---------------------------------------------------------------------
  subroutine sampleEnergyRangeNormal(EeVlo, EeVhi, EeVsamples)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: EeVlo, EeVhi
    real(single), intent(out) :: EeVsamples(:)

    type(RangeT)  :: EeVrange

    !-------------------------------------------------------------------
    ! Start

    EeVrange%lower%type  = INCLUSIVE
    EeVrange%lower%value = EeVlo
    EeVrange%upper%type  = INCLUSIVE
    EeVrange%upper%value = EeVhi

    call sampleEnergyRangeDss(EeVrange, EeVsamples)

  end subroutine sampleEnergyRangeNormal

  !---------------------------------------------------------------------
  subroutine sampleEnergyRangeDss(EeVrange, EeVsamples)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in)  :: EeVrange
    real(single), intent(out) :: EeVsamples(:)

    integer(int32) :: i, numEsamples
!    real(single) :: logEmin, logEmax, frac
    real(single) :: frac

    !-------------------------------------------------------------------
    ! Start

    numEsamples = size(EeVsamples)

    ! Construct the set of energy samples. Take samples such that their logs are distributed evenly across the full energy range.

    if  (EeVrange%lower%type == UNDEFINED&
    .or. EeVrange%upper%type == UNDEFINED)&
call error('','')

if (EeVrange%lower%value <= 0.0) call error('','')

!    logEmin = log(EeVrange%lower%value)
!    logEmax = log(EeVrange%upper%value)
    do i = 1, numEsamples
      frac = (i - 1) / real(numEsamples - 1)
!      EeVsamples(i) = exp(fracVal(logEmin, logEmax, frac))
EeVsamples(i) = (EeVrange%lower%value**(1.0-frac))*(EeVrange%upper%value**frac)
    end do

  end subroutine sampleEnergyRangeDss

  !---------------------------------------------------------------------
  subroutine getEffectiveArea(EeVsamples, effArea)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: EeVsamples(:)
    real(single), intent(out) :: effArea(size(EeVsamples))

    integer(int32) :: numEsamples, eI

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Getting effective area curve.")

    numEsamples = size(EeVsamples)
    do eI = 1, numEsamples
      effArea(eI) = CAL_getEffectiveArea(real(EeVsamples(eI)&
        , double), 0.0d0, 0.0d0)
    end do

  end subroutine getEffectiveArea

  !---------------------------------------------------------------------
  subroutine getFilterTransmission(EeVsamples, filterTrans)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: EeVsamples(:)
    real(single), intent(out) :: filterTrans(size(EeVsamples))

    integer(int32) :: numEsamples, eI

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Getting filter transmission curve.")

    numEsamples = size(EeVsamples)
    do eI = 1, numEsamples
      filterTrans(eI) = CAL_getFilterTransmission(real(EeVsamples(eI)&
        , double), 0.0d0, 0.0d0)
    end do

  end subroutine getFilterTransmission

  !---------------------------------------------------------------------
  function localIntegrateScalar(x, y, xRange) result(approxIntegral)
  !---------------------------------------------------------------------
  ! This function takes a set of samples of y as a function of x, plus a set of ranges of integration, and returns a trapezoidal approximation to the integral of y(x) dx within these ranges.
  !
  ! NOTE! (i) The values in x() should be in increasing order. (ii) The ranges are assumed to be non-overlapping, but may occur in any order.

    implicit none

    real(single), intent(in) :: x(:), y(size(x))
    type(RangeT), intent(in) :: xRange
    real(single) :: approxIntegral

    type(RangeT) :: xRanges(1)

    !-------------------------------------------------------------------
    ! Start

    xRanges(1) = copyRange(xRange)
    approxIntegral = localIntegrateVector(x, y, xRanges)

  end function localIntegrateScalar

  !---------------------------------------------------------------------
  function localIntegrateVector(x, y, xRanges) result(approxIntegral)
  !---------------------------------------------------------------------
  ! This function takes a set of samples of y as a function of x, plus a set of ranges of integration, and returns a trapezoidal approximation to the integral of y(x) dx within these ranges.
  !
  ! NOTE! (i) The values in x() should be in increasing order. (ii) The ranges are assumed to be non-overlapping, but may occur in any order.

    implicit none

    real(single), intent(in) :: x(:), y(size(x))
    type(RangeT), intent(in) :: xRanges(:)
    real(single) :: approxIntegral

    integer(int16) :: numSamples, numRanges, rangeNum, maxI, minI
    real(single) :: minX, maxX, term, rangeMaxX, rangeMinX, yAtRangeMaxX, yAtRangeMinX, summ
!integer :: i

    !-------------------------------------------------------------------
    ! Start

    approxIntegral = 0.0 ! default
  if (size(x) <= 1 .or. size(xRanges) <= 0) return

    numSamples = size(x)
    minX = minval(x)
    maxX = maxval(x)

    numRanges = size(xRanges)
    do rangeNum = 1, numRanges
      select case(xRanges(rangeNum)%lower%type)
      case(UNDEFINED)
        select case(xRanges(rangeNum)%upper%type)
        case(UNDEFINED)
          summ = trapezoidSum(x, y) ! in ssclib/array_utils

        case(INCLUSIVE,EXCLUSIVE)
          rangeMaxX = xRanges(rangeNum)%upper%value
          if (rangeMaxX <= minX) then
            summ = 0.0

          else
            maxI = largestIndex(x <= rangeMaxX)
            summ = trapezoidSum(x(:maxI), y(:maxI)) ! in ssclib/array_utils

            if (maxI < numSamples) then ! add the fraction at the end:
              yAtRangeMaxX = fracVal(y(maxI), y(maxI + 1)&
                , ((rangeMaxX - x(maxI)) / (x(maxI + 1) - x(maxI))))
              term = 0.5 * (y(maxI) + yAtRangeMaxX)&
                * (rangeMaxX - x(maxI))
              summ = summ + term
            end if
          end if

        case default
call error('badUpperRangeType','')
        end select

      case(INCLUSIVE,EXCLUSIVE)
        select case(xRanges(rangeNum)%upper%type)
        case(UNDEFINED)
          rangeMinX = xRanges(rangeNum)%lower%value
          if (rangeMinX >= maxX) then
            summ = 0.0

          else
            minI = smallestIndex(x >= rangeMinX)
            summ = trapezoidSum(x(minI:), y(minI:)) ! in ssclib/array_utils

            if (minI > 1) then ! add the fraction at the end:
              yAtRangeMinX = fracVal(y(minI - 1), y(minI)&
                , ((x(minI) - rangeMinX) / (x(minI) - x(minI - 1))))
              term = 0.5 * (yAtRangeMinX + y(minI))&
                * (x(minI) - rangeMinX)
              summ = summ + term
            end if
          end if

        case(INCLUSIVE,EXCLUSIVE)
          rangeMinX = xRanges(rangeNum)%lower%value
          rangeMaxX = xRanges(rangeNum)%upper%value
          if (rangeMaxX <= minX .or. rangeMinX >= maxX) then
            summ = 0.0

          else
            minI = smallestIndex(x >= rangeMinX)
            maxI = largestIndex(x <= rangeMaxX)
            summ = trapezoidSum(x(minI:maxI), y(minI:maxI))
            ! in ssclib/array_utils
!print *, 'minI, maxI, numSamples, trapezoidSum: ', minI, maxI, numSamples, summ
!do i = minI, maxI
!  print *, '  i, x(i), y(i): ', i, x(i), y(i)
!end do
            if (minI > 1) then ! add the fraction at the end:
              yAtRangeMinX = fracVal(y(minI - 1), y(minI)&
                , ((x(minI) - rangeMinX) / (x(minI) - x(minI - 1))))
              term = 0.5 * (yAtRangeMinX + y(minI))&
                * (x(minI) - rangeMinX)
              summ = summ + term
!print *, 'bottom end: rangeMinX, x(minI), yAtRangeMinX, y(minI), term: '&
!, rangeMinX, x(minI), yAtRangeMinX, y(minI), term
            end if
            if (maxI < numSamples) then ! add the fraction at the end:
              yAtRangeMaxX = fracVal(y(maxI), y(maxI + 1)&
                , ((rangeMaxX - x(maxI)) / (x(maxI + 1) - x(maxI))))
              term = 0.5 * (y(maxI) + yAtRangeMaxX)&
                * (rangeMaxX - x(maxI))
              summ = summ + term
!print *, '   top end: x(maxI), rangeMaxX, y(maxI), yAtRangeMaxX, term: '&
!, x(maxI), rangeMaxX, y(maxI), yAtRangeMaxX, term
            end if
          end if

        case default
call error('badUpperRangeType','')
        end select

      case default
call error('badLowerRangeType','')
      end select

      approxIntegral = approxIntegral + summ
    end do

  end function localIntegrateVector

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module exposure
