!+
! Module: fftw_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains a header for fftw and one or two other helpful routines.
! 
! Comments: TODO:
! 
! $Id: fftw_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module fftw_aux

  use dal
  use errorHandling
  use constants

  use ssc_misc
  use math_utils

  implicit none 

  public

!XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
!     This file contains PARAMETER statements for various constants
!     that can be passed to FFTW routines.  You should include
!     this file in any FORTRAN program that calls the fftw_f77
!     routines (either directly or with an #include statement
!     if you use the C preprocessor).

      integer, parameter, public ::&
        FFTW_FORWARD          =  -1,&
        FFTW_BACKWARD         =   1,&
        FFTW_REAL_TO_COMPLEX  =  -1,&
        FFTW_COMPLEX_TO_REAL  =   1,&
        FFTW_ESTIMATE         =   0,&
        FFTW_MEASURE          =   1,&
        FFTW_OUT_OF_PLACE     =   0,&
        FFTW_IN_PLACE         =   8,&
        FFTW_USE_WISDOM       =  16,&
        FFTW_THREADSAFE       = 128

!     Constants for the MPI wrappers:
      integer, parameter, public ::&
        FFTW_TRANSPOSED_ORDER =     1,&
        FFTW_NORMAL_ORDER     =     0,&
        FFTW_SCRAMBLED_INPUT  =  8192,&
        FFTW_SCRAMBLED_OUTPUT = 16384

!XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

  ! Arrangements so that fftw calls work on all platforms:
  type FftwPlanType
    integer, pointer :: cPointer
  end type
!  type(FftwPlanType) :: fftForwardPlanNum, fftReversePlanNum

  private template
  private calcPowerSpectrum

  interface recenter
    module procedure recenterVectorSingle
    module procedure recenterArraySingle
    module procedure recenterArrayComplex
  end interface recenter

!  interface calcPowerSpectrum
!    module procedure calcPsVectorSingle, calcPsArraySingle
!  end interface calcPowerSpectrum

contains 

  !---------------------------------------------------------------------
  function findNextHighest2357multiple(i) result(result)
  !---------------------------------------------------------------------
  ! This is designed to look for the smallest integer that satisfies the
  ! following conditions: (i) it is greater than or equal to the argument
  ! i; (ii) it is a product only of the numbers 2, 3, 5 and 7. The library
  ! fftw works most efficiently on arrays that have the second property,
  ! so images are padded in size (i) to prevent any 'wrapping' of the
  ! convolver and (ii) so they have dimensions divisable by 2, 3, 5 and/or 7.

    implicit none

    integer, intent(in) ::&
      i
    integer(int16) ::&
      result,&
      maxa,&
      maxb,&
      maxc,&
      maxd,&
      a,&
      b,&
      c,&
      d,&
      besta,&
      bestb,&
      bestc,&
      bestd
    real(single) ::&
      lni,&
      ln2,&
      ln3,&
      ln5,&
      ln7,&
      term,&
      bestTerm
    logical ::&
      firstOver

    !-------------------------------------------------------------------
    ! Start:

    lni = log(real(i))
    ln2 = log(2.0)
    ln3 = log(3.0)
    ln5 = log(5.0)
    ln7 = log(7.0)

    maxa = ceiling(lni / ln2)
    maxb = ceiling(lni / ln3)
    maxc = ceiling(lni / ln5)
    maxd = ceiling(lni / ln7)

    firstOver = .true.
    bestTerm = 0.0 ! just to define it, so the inner 'if' generates no
    ! complaints from the compiler.
    do a = 0, maxa
      do b = 0, maxb
        do c = 0, maxc
          DLOOP: do d = 0, maxd
            term = a * ln2 + b * ln3 + c * ln5 + d * ln7
            if (term .ge. lni .and. (term .lt. bestTerm .or. firstOver)) then
              bestTerm = term
              besta = a
              bestb = b
              bestc = c
              bestd = d

              if (firstOver) firstOver = .false.

          exit DLOOP
            end if
          end do DLOOP
        end do
      end do
    end do

    result = (2**besta) * (3**bestb) * (5**bestc) * (7**bestd)

  end function findNextHighest2357multiple

  !---------------------------------------------------------------------
  subroutine recenterVectorSingle(vector, ctrX)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(inout) :: vector(:)
    integer, intent(in), optional :: ctrX

    integer(int16) :: xsize, xi, localCtrX
    real(single) :: tempVector(size(vector))

    !-------------------------------------------------------------------
    ! Start

    xsize = size(vector)

    if (present(ctrX)) then
      if (ctrX .lt. 1 .or. ctrX .gt. xsize) then
call error('ctrOutOfBounds', '')
      end if
      localCtrX = ctrX
    else
      localCtrX = xsize / 2 + 1
    end if

  if (localCtrX .eq. 1) return

    do xi = 1, xsize + 1 - localCtrX
      tempVector(xi) = vector(xi + localCtrX - 1)
    end do
    do xi = xsize + 2 - localCtrX, xsize
      tempVector(xi) = vector(xi + localCtrX - 1 - xsize)
    end do

    vector = tempVector

  end subroutine recenterVectorSingle

  !---------------------------------------------------------------------
  subroutine recenterArraySingle(array, ctrX, ctrY)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(inout) :: array(:, :)
    integer, intent(in), optional :: ctrX, ctrY

    integer(int16) :: xsize, ysize, xi, yi, localCtrX, localCtrY
    real(single) :: tempArray(size(array, 1), size(array, 2))

    !-------------------------------------------------------------------
    ! Start

    xsize = size(array, 1)
    ysize = size(array, 2)

    if (present(ctrX)) then
      if (ctrX .lt. 1 .or. ctrX .gt. xsize) then
call error('ctrXoutOfBounds', '')
      end if
      localCtrX = ctrX
    else
      localCtrX = xsize / 2 + 1
    end if
    if (present(ctrY)) then
      if (ctrY .lt. 1 .or. ctrY .gt. ysize) then
call error('ctrYoutOfBounds', '')
      end if
      localCtrY = ctrY
    else
      localCtrY = ysize / 2 + 1
    end if

  if (localCtrX .eq. 1 .and. localCtrY .eq. 1) return

    do xi = 1, xsize + 1 - localCtrX
      do yi = 1, ysize + 1 - localCtrY
        tempArray(xi, yi) = array(xi+localCtrX-1, yi+localCtrY-1)
      end do
      do yi = ysize + 2 - localCtrY, ysize
        tempArray(xi, yi) = array(xi+localCtrX-1, yi+localCtrY-1-ysize)
      end do
    end do
    do xi = xsize + 2 - localCtrX, xsize
      do yi = 1, ysize + 1 - localCtrY
        tempArray(xi, yi) = array(xi+localCtrX-1-xsize, yi+localCtrY-1)
      end do
      do yi = ysize + 2 - localCtrY, ysize
        tempArray(xi, yi) = array(xi+localCtrX-1-xsize, yi+localCtrY-1-ysize)
      end do
    end do

    array = tempArray

  end subroutine recenterArraySingle

  !---------------------------------------------------------------------
  subroutine recenterArrayComplex(array, ctrX, ctrY)
  !---------------------------------------------------------------------

    implicit none

    complex(single), intent(inout) :: array(:, :)
    integer, intent(in), optional :: ctrX, ctrY

    integer(int16) :: xsize, ysize, xi, yi, localCtrX, localCtrY
    complex(single) :: tempArray(size(array, 1), size(array, 2))

    !-------------------------------------------------------------------
    ! Start

    xsize = size(array, 1)
    ysize = size(array, 2)

    if (present(ctrX)) then
      if (ctrX .lt. 1 .or. ctrX .gt. xsize) then
call error('ctrXoutOfBounds', '')
      end if
      localCtrX = ctrX
    else
      localCtrX = xsize / 2 + 1
    end if
    if (present(ctrY)) then
      if (ctrY .lt. 1 .or. ctrY .gt. ysize) then
call error('ctrYoutOfBounds', '')
      end if
      localCtrY = ctrY
    else
      localCtrY = ysize / 2 + 1
    end if

  if (localCtrX .eq. 1 .and. localCtrY .eq. 1) return

    do xi = 1, xsize + 1 - localCtrX
      do yi = 1, ysize + 1 - localCtrY
        tempArray(xi, yi) = array(xi+localCtrX-1, yi+localCtrY-1)
      end do
      do yi = ysize + 2 - localCtrY, ysize
        tempArray(xi, yi) = array(xi+localCtrX-1, yi+localCtrY-1-ysize)
      end do
    end do
    do xi = xsize + 2 - localCtrX, xsize
      do yi = 1, ysize + 1 - localCtrY
        tempArray(xi, yi) = array(xi+localCtrX-1-xsize, yi+localCtrY-1)
      end do
      do yi = ysize + 2 - localCtrY, ysize
        tempArray(xi, yi) = array(xi+localCtrX-1-xsize, yi+localCtrY-1-ysize)
      end do
    end do

    array = tempArray

  end subroutine recenterArrayComplex

  !---------------------------------------------------------------------
  subroutine calcPowerSpectrum(fftArray, psArray)
  !---------------------------------------------------------------------

    implicit none

    complex(single), intent(in) :: fftArray(:, :)
    real(single), intent(out) :: psArray(:, :)

    !-------------------------------------------------------------------
    ! Start

    psArray = fftArray * conjg(fftArray)

  end subroutine calcPowerSpectrum

  !---------------------------------------------------------------------
  subroutine symmetrise(array)
  !---------------------------------------------------------------------
  ! The purpose of this subroutine is manipulate a real array (of size (I, J)) which has the property that
  !
  !	array(i,j) = array(i,J+2-j) = array(I+2-i,j) = array(I+2-i,J+2-j)
  !
  ! for i and j > 1. In other words, it should be symmetrical about reflection in either the i = I/2+1 or j = J/2+1 axes, and therefore give rise to no imaginary terms upon back-fft.
  !
  ! The object is accomplished by averaging the contributions from the four corners of the array.

    implicit none

    real(single), intent(inout) :: array(:, :)

    integer(int16) :: xsize, ysize, xi, yi

    !-------------------------------------------------------------------
    ! Start

    xsize = size(array, 1)
    ysize = size(array, 2)

    do xi = 2, xsize/2 + 1
      do yi = 2, ysize/2 + 1
        array(xi, yi) = (array(xi, yi) + array(xi, ysize + 2 - yi)&
        + array(xsize + 2 - xi, yi) + array(xsize + 2 - xi, ysize + 2 - yi))&
        / 4.0
        array(xi,             ysize + 2 - yi) = array(xi, yi)
        array(xsize + 2 - xi, yi)             = array(xi, yi)
        array(xsize + 2 - xi, ysize + 2 - yi) = array(xi, yi)
      end do
    end do

  end subroutine symmetrise

  !---------------------------------------------------------------------
  subroutine fastHankel(vector, transform, status)
  !---------------------------------------------------------------------
  ! This routine employs the algorithm of A E Siegman, Opt Lett 1 13-15 (1977). The input vector should be a function of the radial coordinate, with the radius of the first value equal to zero.

    implicit none

    real(single), intent(in) :: vector(:)
    real(single), intent(out) :: transform(size(vector))
    integer, intent(out), optional :: status

    type(FftwPlanType) :: fftForwardPlanNum, fftReversePlanNum
    integer :: twoNplusPad

    integer :: vsize, n, i, localStatus
    real :: rMin, rMax, alphaTimesN, alpha, vMin, vMax, vectorR(size(vector)), transformV(size(vector)), expAlpha
    real, allocatable :: fR(:), gV(:), f(:)
    complex, allocatable :: cmplxG(:), j(:), cmplxF(:)
    real, parameter :: k1 = 4.0, k2 = 4.0
integer, pointer :: vi(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0

    vsize = size(vector)

!vvv
    rMin = 1.0
    rMax = real(vsize - 1)
    alphaTimesN = log(rMax / rMin)
    alpha = (k1 / k2) * (rMax / rMin)
    n = ceiling(alphaTimesN / alpha)
    vMin = 1.0 / k1 / rMax
    vMax = vMin * (rMax / rMin)

    twoNplusPad = findNextHighest2357multiple(2 * n)
    allocate(cmplxG(twoNplusPad), j(twoNplusPad), fR(n), gV(n))
    allocate(f(twoNplusPad), cmplxF(twoNplusPad))
!    vectorR    = real(indicize(vsize) - 1)
vi => indicize(vsize)
vectorR    = real(vi - 1)
deallocate(vi)
    transformV = vectorR
    do i = 1, twoNplusPad
      expAlpha = exp(alpha * i)
      fR(i) = rMin * expAlpha
      gV(i) = vMin * expAlpha
      j(i) = cmplx(alpha * TwoPi * fR(i) * gV(i)&
      * besselJ0(TwoPi * real(fR(i) * gV(i), double)))
    end do

    call fftw_f77_create_plan(fftForwardPlanNum, twoNplusPad&
    , FFTW_FORWARD, FFTW_ESTIMATE + FFTW_IN_PLACE)
    call fftw_f77_create_plan(fftReversePlanNum, twoNplusPad&
    , FFTW_BACKWARD, FFTW_ESTIMATE + FFTW_IN_PLACE)

    call fftw_f77_one(fftReversePlanNum, j)
!^^^ all between can be reused for multiple vectors.

    f = 0.0
    call cleverInterpolate(vector, vectorR, f(:n), fR(:n), style=0)
    cmplxF(:n) = cmplx(f(:n) * fR(:n))

    call fftw_f77_one(fftForwardPlanNum, cmplxF)
    cmplxG = cmplxF * j
    call fftw_f77_one(fftReversePlanNum, cmplxG)

    call cleverInterpolate(real(cmplxG(:n)) / gV(:n), gV(:n)&
    , transform, transformV, style=0)

!???    call fftw_f77_destroy_plan(fftForwardPlanNum)
!???    call fftw_f77_destroy_plan(fftReversePlanNum)

    if (present(status)) status = localStatus

  end subroutine fastHankel

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module fftw_aux
