!+
! Module: histogram
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some assorted routines.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module histogram

  use dal
!  use param  
!  use errorHandling

!  use constants
!  use sort_mod
!  use random
!  use ssc_misc

  implicit none 

  private 

  public makeHistogram

  interface makeHistogram
    module procedure makeHistogramEdgesReal, makeHistogramBoundsReal
    module procedure makeHistogramEdgesDble, makeHistogramBoundsDble
  end interface

  contains

  !---------------------------------------------------------------------
  subroutine makeHistogramBoundsReal(inVector, binContents, lEdge, rEdge&
  , underflows, overflows, status)
  !---------------------------------------------------------------------

    implicit none

    real,    intent(in)            :: inVector(:)
    integer, intent(out)           :: binContents(:)
    real,    intent(in)            :: lEdge, rEdge
    integer, intent(out), optional :: underflows, overflows, status

    integer :: hsize, i, localUnderflows, localOverflows, localStatus
    real :: frac
    real :: binEdges(size(binContents) + 1)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default.

    hsize = size(binContents)

    do i = 1, hsize + 1
      frac = (i - 1) / real(hsize)
      binEdges(i) = lEdge * (1.0 - frac) + rEdge * frac
    end do

    call makeHistogramKernel(dble(inVector), binContents, dble(binEdges)&
    , localUnderflows, localOverflows, localStatus)

    if (present(underflows)) underflows = localUnderflows
    if (present(overflows)) overflows = localOverflows
    if (present(status)) status = localStatus

  end subroutine makeHistogramBoundsReal

  !---------------------------------------------------------------------
  subroutine makeHistogramEdgesReal(inVector, binContents, binEdges&
  , underflows, overflows, status)
  !---------------------------------------------------------------------

    implicit none

    real,    intent(in)            :: inVector(:)
    integer, intent(out)           :: binContents(:)
    real,    intent(in),  optional :: binEdges(:)
    integer, intent(out), optional :: underflows, overflows, status

    integer :: hsize, binEdge, localUnderflows, localOverflows, localStatus
    real :: frac, minx, maxx
    real, allocatable :: myBinEdges(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default.

    hsize = size(binContents)
    minx = minval(inVector)
    maxx = maxval(inVector)

    allocate(myBinEdges(hsize + 1))
    if (present(binEdges)) then
      myBinEdges = binEdges(1:hsize + 1)

    else
      do binEdge = 1, hsize + 1
        frac = (binEdge - 1) / real(hsize)
        myBinEdges(binEdge) = minx * (1.0 - frac) + maxx * frac
      end do
    end if

    call makeHistogramKernel(dble(inVector), binContents, dble(myBinEdges)&
    , localUnderflows, localOverflows, localStatus)

    if (present(underflows)) underflows = localUnderflows
    if (present(overflows)) overflows = localOverflows
    if (present(status)) status = localStatus

  end subroutine makeHistogramEdgesReal

  !---------------------------------------------------------------------
  subroutine makeHistogramBoundsDble(inVector, binContents, lEdge, rEdge&
  , underflows, overflows, status)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)),  intent(in)            :: inVector(:)
    integer, intent(out)           :: binContents(:)
    real(kind(1d0)),  intent(in)            :: lEdge, rEdge
    integer, intent(out), optional :: underflows, overflows, status

    integer :: hsize, i, localUnderflows, localOverflows, localStatus
    real(kind(1d0)) :: frac
    real(kind(1d0)) :: binEdges(size(binContents) + 1)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default.

    hsize = size(binContents)

    do i = 1, hsize + 1
      frac = (i - 1) / real(hsize)
      binEdges(i) = lEdge * (1.0 - frac) + rEdge * frac
    end do

    call makeHistogramKernel(inVector, binContents, binEdges&
    , localUnderflows, localOverflows, localStatus)

    if (present(underflows)) underflows = localUnderflows
    if (present(overflows)) overflows = localOverflows
    if (present(status)) status = localStatus

  end subroutine makeHistogramBoundsDble

  !---------------------------------------------------------------------
  subroutine makeHistogramEdgesDble(inVector, binContents, binEdges&
  , underflows, overflows, status)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)),  intent(in)            :: inVector(:)
    integer, intent(out)           :: binContents(:)
    real(kind(1d0)),  intent(in),  optional :: binEdges(:)
    integer, intent(out), optional :: underflows, overflows, status

    integer :: hsize, binEdge, localUnderflows, localOverflows, localStatus
    real(kind(1d0)) :: frac, minx, maxx
    real(kind(1d0)), allocatable :: myBinEdges(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default.

    hsize = size(binContents)
    minx = minval(inVector)
    maxx = maxval(inVector)

    allocate(myBinEdges(hsize + 1))
    if (present(binEdges)) then
      myBinEdges = binEdges(1:hsize + 1)

    else
      do binEdge = 1, hsize + 1
        frac = (binEdge - 1) / real(hsize)
        myBinEdges(binEdge) = minx * (1.0 - frac) + maxx * frac
      end do
    end if

    call makeHistogramKernel(inVector, binContents, myBinEdges&
    , localUnderflows, localOverflows, localStatus)

    if (present(underflows)) underflows = localUnderflows
    if (present(overflows)) overflows = localOverflows
    if (present(status)) status = localStatus

  end subroutine makeHistogramEdgesDble

  !---------------------------------------------------------------------
  subroutine makeHistogramKernel(inVector, binContents, binEdges&
  , underflows, overflows, status)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)),  intent(in)  :: inVector(:)
    integer, intent(out) :: binContents(:)
    real(kind(1d0)),  intent(in)  :: binEdges(:)
    integer, intent(out) :: underflows, overflows, status

    integer :: vsize, hsize, i, bin
    real(kind(1d0)) :: minx, maxx, x

    !-------------------------------------------------------------------
    ! Start

    status = 0 ! default

    vsize = size(inVector)
    hsize = size(binContents)

    if (vsize .le. 1) then
      status = -1 ! only 1 value.
  return
    end if

    if (hsize .le. 1) then
      status = -2 ! only 1 bin.
  return
    end if

    if (size(binEdges) .lt. hsize + 1) then
      status = -3
  return
    else if (size(binEdges) .gt. hsize + 1) then
      status = status + 1
    end if

    do i = 1, hsize
      if (binEdges(i) .ge. binEdges(i + 1)) then
        status = -4
  return
      end if
    end do

    minx = binEdges(1)
    maxx = binEdges(hsize + 1)

    binContents = 0
    underflows = 0
    overflows = 0
    do i = 1, vsize
      x = inVector(i)
      if (x .lt. minx) then
        if (underflows .ge. huge(underflows)) then
          status = -5
  return
        end if
        underflows = underflows + 1

      else if (x .gt. maxx) then
        if (overflows .ge. huge(overflows)) then
          status = -6
  return
        end if
        overflows = overflows + 1

      else
        do bin = hsize, 1, -1
          if (x .ge. binEdges(bin)) then
            if (binContents(bin) .ge. huge(binContents)) then
              status = -7
  return
            end if
            binContents(bin) = binContents(bin) + 1
        exit
          end if
        end do
      end if
    end do

  end subroutine makeHistogramKernel

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module histogram
