!+
! Module: math_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some assorted routines.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module math_utils

  use dal
!  use param  
  use errorHandling

  use constants
  use sort_mod
!  use random
  use ssc_misc

  implicit none 

  private 

!  public makeHistogram
  public rebin
  public prepareForInterpolate
  public interpolate
  public pointInEllipseWithAngle
  public pointInEllipseWithPhase
  public calcEllipseFromAngle
  public calcEllipseFromPhase
  public ellipseAngleToPhase
  public calcAitoff
  public median
  public valueAtBinFraction
  public meanToBinFraction
  public fracVal
  public isOdd
  public normalise
  public vectorCross
!  public integrate
  public newtonsMethod
  public hybridNewtonsMethod
!  public makeSourceList
  public cyclicConvolution
  public preciseSum
  public findEdges
  public robustMean
public temp_robustMean
  public incompleteGammaP
  public incompleteGammaQ
  public lnGamma
  public poisson
  public factorial
  public chi2Prob
  public poissonProb
  public gaussProb
  public erf
  public erfc
  public besselJ0
  public sourceOrientation
  public bilinearValue
  public cleverInterpolate
!  public cubicSpline

  public mergeStatuses ! really only designed to be of use to these ssc
  ! routines and so won't document it.

!  interface makeHistogram
!    module procedure makeHistogramEdgesReal, makeHistogramBoundsReal
!    module procedure makeHistogramEdgesDble, makeHistogramBoundsDble
!  end interface

  interface rebin
    module procedure rebinBasicReal, rebinWithWidthsReal, rebinWithEdgesReal
    module procedure rebinBasicDble, rebinWithWidthsDble, rebinWithEdgesDble
  end interface

  interface prepareForInterpolate
    module procedure prepareForInterpolate1d, prepareForInterpolate2d
  end interface

  interface interpolate
    module procedure interpolateSingle, interpolateVector
  end interface

  interface fracVal
    module procedure fracValScalar, fracValVector
  end interface

  interface valueAtBinFraction
    module procedure valueAtBinFractionVector, valueAtBinFractionArray
  end interface

  interface meanToBinFraction
    module procedure meanToBinFractionVector, meanToBinFractionArray
  end interface

  interface normalise
    module procedure normaliseReal
    module procedure normaliseDble
  end interface

  interface vectorCross
    module procedure vectorCrossReal
    module procedure vectorCrossDble
  end interface

  interface integrate
    module procedure integrate1dReal, integrateNdReal
!    module procedure integrate1dDble, integrateNdDble
  end interface

!  interface hybridNewtonsMethod
!    module procedure hybridNewtonsMethod_real
!    module procedure hybridNewtonsMethod_dble
!  end interface

  interface robustMean
    module procedure robustMean_1d
    module procedure robustMean_2d
  end interface

!  interface cubicSpline
!    module procedure cubicSplineWithX
!    module procedure cubicSplineOhneX
!  end interface

  integer, parameter :: defaultNumIntegralSamples = 100
  type :: boundType
    integer :: type
    real(kind(1d0)) :: value
  end type boundType
  type :: intervalType
    type(boundType) :: lower, upper
  end type intervalType
  type :: intervalsType
    type(boundType) :: lower, upper
    integer :: index
  end type intervalsType
  type :: sequenceType
    type(boundType) :: bound
    logical         :: isLower
    integer         :: index
  end type sequenceType
!  integer, parameter :: OPEN = 1	! From dssutil. Check that the
!  integer, parameter :: CLOSED = 2	! match-up is as I have guessed it.
  integer, parameter :: INCLUSIVE = 1
  integer, parameter :: EXCLUSIVE = 2
  integer, parameter :: UNDEFINED = 3

  type, public :: sourceType
    real :: x, y, flux, temperature
  end type sourceType

  type, public :: statusType
    integer :: status, maxNeg, numPosBits
  end type

logical :: testFlag = .false.

contains 

  !---------------------------------------------------------------------
  subroutine ellipsePhaseToCartesian(xAmp, yAmp, phase&
  , longSemiAxis, shortSemiAxis, rotatedAngle)
  !---------------------------------------------------------------------
  ! This and the associated subroutine ellipseCartesianToPhase() change the parameterisation of an ellipse. In the 'cartesian' form, the ellipse is defined by a long and a short semiaxis, plus an angle about which the whole ellipse is rotated. The short axis is conceived of as originally aligned with the x axis of the coordinate system, small positive values of rotatedAngle being towards the y axis. In terms of the 'cartesian' parameters, the points (x,y) on the ellipse are best obtained from the following equations:
  !
  !	      x0^2                y0^2
  !	----------------- + ---------------- = 1		1(a)
  !	 shortSemiAxis^2     longSemiAxis^2
  !
  !	x = x0 * cos(rotatedAngle) - y0 * sin(rotatedAngle)	1(b)
  !
  !	y = y0 * cos(rotatedAngle) + x0 * sin(rotatedAngle)	1(c)
  !
  ! The 'phase' form of the ellipse defines (x,y) in terms of a 'dummy' parameter theta as follows:
  !
  !	x = xAmp * cos(theta)					2(a)
  !
  !	y = yAmp * cos(theta + phase)				2(b)
  !
  ! Each format is more convenient that the other for certain applications, hence the usefulness of routines to convert between them.
  !
  !.....................................................................
  !
  ! The ellipsePhaseToCartesian transformation is achieved conceptually by first combining parametric equations 2(a) and 2(b) into a single equation giving the radius of points on the ellipse as a function of theta:
  !
  !	r^2 = xAmp^2 * cos^2(theta) + yAmp^2 * cos^2(theta + phase)	3(a)
  !
  ! thus
  !
  !	r^2 = 0.5 * (xAmp^2 + yAmp^2
  !	    + Re[xAmp^2 * exp{i*2theta}
  !	    + yAmp^2 * exp{i*(2theta + 2phase)}])			3(b)
  !
  ! which can be further manipulated to
  !
  !	r^2 = 0.5 * (xAmp^2 + yAmp^2
  !	    + Re[Z * exp{i*(2theta + alpha)}])				4(a)
  !
  ! where
  !
  !	Z^2 = xAmp^4 + yAmp^4 + xAmp^2 * yAmp^2 * cos(2phase)	4(b)
  !
  ! (note the powers of 4) and
  !
  !	                 yAmp^2 * sin(2phase)
  !	tan(alpha) = ----------------------------------.		4(c)
  !	              xAmp^2 + yAmp^2 * cos(2phase)
  !
  ! Consider just the vector
  !
  !	Z * exp{i*(2theta + alpha)}
  !	 = xAmp^2 * exp{i*2theta} + yAmp^2 * exp{i*(2theta + 2phase)}
  !
  ! rotating on the complex plane (Argand diagram). From equation 3(a), clearly
  !
  !	shortSemiAxis^2 = xAmp^2 * cos^2(-alpha/2)
  !	                + yAmp^2 * cos^2(-alpha/2 + phase)		5(a)
  ! and
  !
  !	 longSemiAxis^2 = xAmp^2 * cos^2(pi/2 - alpha/2)
  !	                + yAmp^2 * cos^2(pi/2 - alpha/2 + phase);	5(b)
  !
  ! However it is probably easier to use equations 2 to get
  !
  !	x_short = xAmp * cos(-alpha/2)					6(a)
  !
  !	y_short = yAmp * cos(-alpha/2 + phase)				6(b)
  !
  !	x_long  = xAmp * cos(pi/2 - alpha/2)				6(c)
  !
  !	y_long  = yAmp * cos(pi/2 - alpha/2 + phase),			6(d)
  !
  ! and thence
  !
  !	shortSemiAxis^2 = x_short * x_short + y_short * y_short		7(a)
  !
  !	longSemiAxis^2  = x_long  * x_long  + y_long  * y_long		7(b)
  !
  ! because then also
  !
  !	rotatedAngle = atan2(y_short, x_short).				8
  !
  !.....................................................................

    implicit none

    real, intent(in)  :: xAmp, yAmp, phase
    real, intent(out) :: longSemiAxis, shortSemiAxis, rotatedAngle

    real :: xAmp2, yAmp2, alpha, xShort, yShort, xLong, yLong

    !-------------------------------------------------------------------
    ! Start

    xAmp2 = xAmp * xAmp
    yAmp2 = yAmp * yAmp
    alpha = atan2(yAmp2 * sin(2.0 * phase), xAmp2 + yAmp2 * cos(2.0 * phase))

    xShort = xAmp * cos(-alpha / 2.0)
    yShort = yAmp * cos(-alpha / 2.0 + phase)
    xLong  = xAmp * cos(PiOn2 - alpha / 2.0)
    yLong  = yAmp * cos(PiOn2 - alpha / 2.0 + phase)

    shortSemiAxis = sqrt(xShort * xShort + yShort * yShort)
    longSemiAxis  = sqrt(xLong  * xLong  + yLong  * yLong)

    rotatedAngle = atan2(yShort, xShort)

  end subroutine ellipsePhaseToCartesian

  !---------------------------------------------------------------------
  subroutine ellipseCartesianToPhase(longSemiAxis, shortSemiAxis, rotatedAngle&
  , xAmp, yAmp, phase)
  !---------------------------------------------------------------------
  ! This and the associated subroutine ellipsePhaseToCartesian() change the parameterisation of an ellipse. In the 'cartesian' form, the ellipse is defined by a long and a short semiaxis, plus an angle about which the whole ellipse is rotated. The short axis is conceived of as originally aligned with the x axis of the coordinate system, small positive values of rotatedAngle being towards the y axis. In terms of the 'cartesian' parameters, the points (x,y) on the ellipse are best obtained from the following equations (repeated from ellipsePhaseToCartesian()):
  !
  !	      x0^2                y0^2
  !	----------------- + ---------------- = 1		1(a)
  !	 shortSemiAxis^2     longSemiAxis^2
  !
  !	x = x0 * cos(rotatedAngle) - y0 * sin(rotatedAngle)	1(b)
  !
  !	y = y0 * cos(rotatedAngle) + x0 * sin(rotatedAngle)	1(c)
  !
  ! The 'phase' form of the ellipse defines (x,y) in terms of a 'dummy' parameter theta as follows:
  !
  !	x = xAmp * cos(theta)					2(a)
  !
  !	y = yAmp * cos(theta + phase)				2(b)
  !
  ! Each format is more convenient that the other for certain applications, hence the usefulness of routines to convert between them.
  !
  !.....................................................................
  !
  ! The ellipseCartesianToPhase transformation is achieved conceptually by first changing equation 1(a) to parametric form, giving
  !
  !	x0(beta) = shortSemiAxis * cos(beta)			9(a)
  !
  !	y0(beta) =  longSemiAxis * sin(beta).			9(a)
  !
  ! Thus equations 1(b) and 1(c) become
  !
  !	x(beta) = shortSemiAxis * cos(beta) * cos(rotatedAngle)
  !	        -  longSemiAxis * sin(beta) * sin(rotatedAngle)	10(a)
  !
  !	y(beta) =  longSemiAxis * sin(beta) * cos(rotatedAngle)
  !	        + shortSemiAxis * cos(beta) * sin(rotatedAngle)	10(b)
  !
  ! Differentiating these yields
  !
  !	                    longSemiAxis
  !	tan(beta_xAmp) = - --------------- tan(rotatedAngle)	11(a)
  !	                    shortSemiAxis
  !
  !	                    longSemiAxis
  !	tan(beta_yAmp) =   --------------- cot(rotatedAngle),	11(a)
  !	                    shortSemiAxis
  !
  ! where beta_xAmp is the value of beta at which x = +-xAmp and beta_yAmp is the value of beta at which y = +-yAmp. Substitution of beta_xAmp and beta_yAmp into equations 10 provides the values respectively of xAmp and yAmp.
  !
  ! It can also be shown (again, how I am not sure!) that
  !
  !	phase = (beta_yAmp - beta_xAmp) modulo pi.		12
  !
  !.....................................................................

    implicit none

    real, intent(in)  :: longSemiAxis, shortSemiAxis, rotatedAngle
    real, intent(out) :: xAmp, yAmp, phase

    real :: beta_xAmp, beta_yAmp

    !-------------------------------------------------------------------
    ! Start

    beta_xAmp = atan2(-longSemiAxis * sin(rotatedAngle)&
    , shortSemiAxis * cos(rotatedAngle))

    beta_yAmp = atan2(longSemiAxis * cos(rotatedAngle)&
    , shortSemiAxis * sin(rotatedAngle))

    xAmp = abs(shortSemiAxis * cos(beta_xAmp) * cos(rotatedAngle)&
              - longSemiAxis * sin(beta_xAmp) * sin(rotatedAngle))

    yAmp = abs(longSemiAxis * sin(beta_yAmp) * cos(rotatedAngle)&
            + shortSemiAxis * cos(beta_yAmp) * sin(rotatedAngle))

    phase = modulo(beta_yAmp - beta_xAmp, real(Pi))

  end subroutine ellipseCartesianToPhase

  !---------------------------------------------------------------------
  function pointInEllipseWithPhase(x, y, xAmp, yAmp, phase)
  !---------------------------------------------------------------------
  ! The ellipse is assumed to be centred about the origin. It is defined
  ! by two amplitudes and a phase.
  !
  ! The function returns -1 if the point (x, y) lies fully within the
  ! ellipse; 0 if it is on the border; and 1 if it is fully outside the
  ! ellipse.

    implicit none

    real, intent(in) ::&
      x,&
      y,&
      xAmp,&
      yAmp,&
      phase

    integer :: pointInEllipseWithPhase
    real :: relX, termA, termB, topY, botY

    !-------------------------------------------------------------------
    ! Start

    relX = x / xAmp

    if (abs(relX) > 1.0) then
      pointInEllipseWithPhase = 1
  return
    else if (abs(relX) == 1.0) then
      pointInEllipseWithPhase = 0
  return
    end if

    termA = relX * cos(phase)
    termB = sqrt(1 - relX * relX) * sin(phase)
    topY = yAmp * (termA + termB)
    botY = yAmp * (termA - termB)

    if (y > topY .or. y < botY) then
      pointInEllipseWithPhase = 1
    else if (y == topY .or. y == botY) then
      pointInEllipseWithPhase = 0
    else
      pointInEllipseWithPhase = -1
    end if

  end function pointInEllipseWithPhase

  !---------------------------------------------------------------------
  function pointInEllipseWithAngle(x, y, xSemiAxis, ySemiAxis, rotatedAngle)
  !---------------------------------------------------------------------
  ! The ellipse is assumed to be centred about the origin. It is defined
  ! by two semi-axes. (The short and long semi-axes are defined respectively
  ! as half the shortest and longest diameters of the ellipse. They are
  ! necessarily orthogonal to each other.) The ellipse is to be thought
  ! of as initially having its semiaxes aligned with the x- and y-axes of
  ! the coordinate system, subsequently having been rotated in the direction
  ! +ve x-axis toward +ve y-axis by an angle rotatedAngle. NOTE that this
  ! angle is must be specified in RADIANS.
  !
  ! The function returns -1 if the point (x, y) lies fully within the
  ! ellipse; 0 if it is on the border; and 1 if it is fully outside the
  ! ellipse.
  !
  ! The algorithm used is to find the coordinates (x1, y1) of the specified
  ! point in the coordinate frame rotated to align with the ellipse
  ! semiaxes, then to insert (x1, y1) into the cartesian equation for
  ! the ellipse in this frame.

    implicit none

    real, intent(in) ::&
      x,&
      y,&
      xSemiAxis,&
      ySemiAxis,&
      rotatedAngle

    integer :: pointInEllipseWithAngle
    real :: sinAngle, cosAngle, x1, y1, gauge

    !-------------------------------------------------------------------
    ! Start

    sinAngle = sin(rotatedAngle)
    cosAngle = cos(rotatedAngle)
    x1 = x * cosAngle + y * sinAngle
    y1 = y * cosAngle - x * sinAngle

    gauge = (x1 / xSemiAxis) * (x1 / xSemiAxis)&
    + (y1 / ySemiAxis) * (y1 / ySemiAxis)

    if (gauge .gt. 1.0) then
      pointInEllipseWithAngle = 1
    else if (gauge .eq. 1.0) then
      pointInEllipseWithAngle = 0
    else
      pointInEllipseWithAngle = -1
    end if

  end function pointInEllipseWithAngle

  !---------------------------------------------------------------------
  subroutine calcEllipseFromAngle(xSemi, ySemi, angleDeg, xVals, yVals)
  !---------------------------------------------------------------------
  ! This sub calculates coordinates of samples of an ellipse. The larger
  ! the size of the vectors xVals and yVals, the more accurately is the
  ! ellipse sampled. The intention is, after scaling and addition of
  ! offsets, to send the resulting polynomial to the pgplot routine pgpoly,
  ! viz:
  !
  !   call pgpoly(size(xVals), xVals, yVals)
  !
  ! The ellipse is constructed orthogonal to the x-y axes, then rotated by
  ! angle 'angleDeg'. xSemi is defined as half the diameter of the
  ! (unrotated) ellipse in the direction of the x axis; similarly for ySemi.
  ! The rotation is positive-valued in the direction from +ve x to +ve y axis.

    implicit none 

    real,  intent(in) :: xSemi,&
                         ySemi,&
                         angleDeg
    real, intent(out) :: xVals(:),&
                         yVals(size(xVals))

    integer :: point, numPoints
    real :: pointAngle, cosWholeAngle, sinWholeAngle, x, y

    !-------------------------------------------------------------------
    ! Start:

    cosWholeAngle = cos(angleDeg * PiOn180)
    sinWholeAngle = sin(angleDeg * PiOn180)

    numPoints = size(xVals)

    xVals = 0.0
    yVals = 0.0

    do point = 1, numPoints
      pointAngle = (point / real(numPoints)) * TwoPi
      x = xSemi * cos(pointAngle)
      y = ySemi * sin(pointAngle)
      xVals(point) = x * cosWholeAngle - y * sinWholeAngle
      yVals(point) = y * cosWholeAngle + x * sinWholeAngle
    end do

  end subroutine calcEllipseFromAngle

  !---------------------------------------------------------------------
  subroutine calcEllipseFromPhase(xAmp, yAmp, phase, xVals, yVals)
  !---------------------------------------------------------------------
  ! This sub calculates coordinates of samples of an ellipse. The larger
  ! the size of the vectors xVals and yVals, the more accurately is the
  ! ellipse sampled. The intention is, after scaling and addition of
  ! offsets, to send the resulting polynomial to the pgplot routine pgpoly,
  ! viz:
  !
  !   call pgpoly(size(xVals), xVals, yVals)
  !

    implicit none 

    real,  intent(in) :: xAmp,&
                         yAmp,&
                         phase
    real, intent(out) :: xVals(:),&
                         yVals(size(xVals))

    integer :: point, numPoints
    real :: pointAngle

    !-------------------------------------------------------------------
    ! Start:

    numPoints = size(xVals)

    xVals = 0.0
    yVals = 0.0

    do point = 1, numPoints
      pointAngle = (point / real(numPoints)) * TwoPi
      xVals(point) = xAmp * cos(pointAngle)
      yVals(point) = yAmp * cos(pointAngle + phase)
    end do

  end subroutine calcEllipseFromPhase

  !---------------------------------------------------------------------
  subroutine ellipseAngleToPhase(xSemi, ySemi, angleDeg, xAmp, yAmp, phase)
  !---------------------------------------------------------------------
  ! Converts between the two ellipse representations.

    implicit none 

    real,  intent(in)  :: xSemi,&
                          ySemi,&
                          angleDeg
    real,  intent(out) :: xAmp,&
                          yAmp,&
                          phase

    real :: cosAngle, sinAngle

    !-------------------------------------------------------------------
    ! Start:

    cosAngle = cos(angleDeg * PiOn180)
    sinAngle = sin(angleDeg * PiOn180)
    xAmp = sqrt(xSemi * xSemi * cosAngle * cosAngle&
              + ySemi * ySemi * sinAngle * sinAngle)

    yAmp = sqrt(xSemi * xSemi * sinAngle * sinAngle&
              + ySemi * ySemi * cosAngle * cosAngle)

    if (ySemi .ne. 0.0 .and. xSemi .ne. 0.0) then
      phase = PiOn2 - atan((ySemi * ySemi - xSemi * xSemi)&
        * sinAngle * cosAngle / ySemi / xSemi)

    else if (ySemi == 0.0 .and. xSemi == 0.0) then
      phase = PiOn2

    else
      if (ySemi == 0.0) then ! but xSemi \= 0.0
        phase = Pi
      else ! xSemi == 0.0 but ySemi \= 0.0
        phase = 0.0
      end if
    end if

  end subroutine ellipseAngleToPhase

  !---------------------------------------------------------------------
  subroutine calcEllipseFromAngle_old(xOffset, yOffset, xSemi, ySemi, angleDeg&
  , xVals, yVals)
  !---------------------------------------------------------------------
!*** from swcat.
! follow by call pgpoly(numPoints, xvals, yvals) to draw the ellipse.

    implicit none 

    real,  intent(in) :: xOffset,&
                         yOffset,&
                         xSemi,&
                         ySemi,&
                         angleDeg
    real, intent(out) :: xVals(:),&
                         yVals(:)

    integer :: point, numPoints
    real :: pointAngle, cosWholeAngle, sinWholeAngle, x, y

    !-------------------------------------------------------------------
    ! Start:

!    cosWholeAngle = cos(angleDeg * myPiOn180)
!    sinWholeAngle = sin(angleDeg * myPiOn180)
    cosWholeAngle = cos(angleDeg * PiOn180)
    sinWholeAngle = sin(angleDeg * PiOn180)

    numPoints = min(size(xVals), size(yVals))

    xVals = 0.0
    yVals = 0.0

    do point = 1, numPoints
!      pointAngle = (point / real(numPoints)) * myTwoPi
      pointAngle = (point / real(numPoints)) * TwoPi
      x = xSemi * cos(pointAngle)
      y = ySemi * sin(pointAngle)
      xVals(point) = xOffset + x * cosWholeAngle + y * sinWholeAngle
      yVals(point) = yOffset + y * cosWholeAngle - x * sinWholeAngle
    end do

  end subroutine calcEllipseFromAngle_old

  !---------------------------------------------------------------------
  subroutine calcAitoff(raDeg, decDeg, x, y)
  !---------------------------------------------------------------------
!*** could make the aitoff calc quicker by precalculating sin and cos dec etc.
!*** from swcat.
!*** include also drawAitoffGrid?

    implicit none 

    real, intent(in) :: raDeg, decDeg
    real, intent(out) :: x, y
    real :: ra, dec, r1, denom, cosDec

    !-------------------------------------------------------------------
    ! Start:

!    ra  = raDeg  * myPiOn180
!    dec = decDeg * myPiOn180
    ra  = raDeg  * PiOn180
    dec = decDeg * PiOn180

!    r1 = (ra - myPi) / 2.0
    r1 = (ra - Pi) / 2.0
    cosDec = cos(dec)
    denom = sqrt(1.0 + cosDec * cos(r1))
    x = 2.0 * cosDec * sin(r1) / denom
    y = sin(dec) / denom

  end subroutine calcAitoff

  !---------------------------------------------------------------------
  function median(vector)
  !---------------------------------------------------------------------
  ! NOTE that for EVEN-numbered size(vector) the function returns the value in the lower central bin of the sorted vector, rather than the (more correct) mean of the lower and upper central bins. For ODD-numbered size the function returns the value in the central bin of the sorted vector, as expected.

    implicit none

    real, intent(in) :: vector(:)
    real :: median

    !-------------------------------------------------------------------
    ! Start

    median = valueAtBinFractionVector(vector, 0.5)

  end function median

  !---------------------------------------------------------------------
  function median_old(vector)
  !---------------------------------------------------------------------
  ! Adapted from the Numerical Recipes function select().
  !
  ! NOTE that for EVEN vsize the function returns the value in the lower
  ! central bin of the sorted vector, rather than the (more correct) mean
  ! of the lower and upper central bins. For ODD vsize the function returns
  ! the value in the central bin of the sorted vector, as expected.

    implicit none

    real, intent(in) :: vector(:)
    real :: median_old

    integer :: vsize, medianBin, i, ir, j, l, mid
    real :: tempVec(size(vector)), tempVal

    !-------------------------------------------------------------------
    ! Start

    vsize = size(vector)
    medianBin = ceiling(vsize / 2.0)
    tempVec = vector

    l = 1
    ir = vsize

    do
      if (ir - l .le. 1) then
        if (ir - l .eq. 1) then
          if (tempVec(l) .gt. tempVec(ir)) call swap(tempVec(l), tempVec(ir))
        end if
        median_old = tempVec(medianBin)
    exit

      else
        mid = (l + ir) / 2
        call swap(tempVec(mid), tempVec(l + 1))

        if (tempVec(l)  .gt. tempVec(ir))  call swap(tempVec(l),   tempVec(ir))
        if (tempVec(l+1).gt. tempVec(ir))  call swap(tempVec(l+1), tempVec(ir))
        if (tempVec(l)  .gt. tempVec(l+1)) call swap(tempVec(l),   tempVec(l+1))

        i = l + 1
        j = ir
        tempVal = tempVec(l + 1)
        do
          do
            i = i + 1
          if (tempVec(i) .ge. tempVal) exit
          end do

          do
            j = j - 1
          if (tempVec(j) .le. tempVal) exit
          end do

        if (j .lt. i) exit

          call swap(tempVec(i), tempVec(j))
        end do

        tempVec(l + 1) = tempVec(j)
        tempVec(j) = tempVal
        if (j .ge. medianBin) ir = j - 1
        if (j .le. medianBin) l = i
      end if
    end do

  end function median_old

  !---------------------------------------------------------------------
  function valueAtBinFractionVector(vector, binFraction)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: vector(:), binFraction
    real :: valueAtBinFractionVector

    integer :: nearestBin
    real :: adjustedBinFraction

    !-------------------------------------------------------------------
    ! Start

    adjustedBinFraction = min(1.0, max(0.0, binFraction))
    nearestBin = max(1, ceiling(size(vector) * adjustedBinFraction))
    valueAtBinFractionVector = valueAtBinFractionKernel(vector, nearestBin)

  end function valueAtBinFractionVector

  !---------------------------------------------------------------------
  function valueAtBinFractionArray(array, binFraction, maskArray)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: array(:,:), binFraction
    logical, intent(in), optional :: maskArray(:,:)
    real :: valueAtBinFractionArray

    integer :: xsize, ysize, vsize, numMasked, nearestBin
!    real :: vector(size(array, 1) * size(array, 2))
real, allocatable :: vector(:)
!    real :: tempArray(size(array, 1), size(array, 2))
    real :: adjustedBinFraction
    logical :: doMasking

    !-------------------------------------------------------------------
    ! Start

    xsize = size(array, 1)
    ysize = size(array, 2)
    vsize = xsize * ysize

    doMasking = .false. ! default
    if (present(maskArray)) then
      numMasked = count(maskArray)

      if (xsize .eq. size(maskArray, 1)&
      .and. ysize .eq. size(maskArray, 2)&
      .and. numMasked .gt. 0) then
        doMasking = .true.
      end if
    end if

    if (doMasking) then
!      where(maskArray)
!        tempArray = array
!      elsewhere
!        tempArray = minval(array)
!      end where
!      vector = reshape(tempArray, (/vsize/))
allocate(vector(numMasked))
vector = pack(array, maskArray)

    else
      numMasked = vsize
allocate(vector(numMasked))
      vector = reshape(array, (/vsize/))
    end if

    adjustedBinFraction = min(1.0, max(0.0, binFraction))
    nearestBin = max(1, ceiling(numMasked * adjustedBinFraction))
!    nearestBin = nearestBin + (vsize - numMasked)

    valueAtBinFractionArray = valueAtBinFractionKernel(vector, nearestBin)

  end function valueAtBinFractionArray

  !---------------------------------------------------------------------
  function valueAtBinFractionKernel(vector, nearestBin)
  !---------------------------------------------------------------------
  ! Adapted from the Numerical Recipes function select().

!    use sscmisc

    implicit none

    real, intent(in) :: vector(:)
    integer, intent(in) :: nearestBin
    real :: valueAtBinFractionKernel

    integer :: vsize, i, ir, j, l, mid
    real :: tempVec(size(vector)), tempVal

    !-------------------------------------------------------------------
    ! Start

    vsize = size(vector)
    tempVec = vector

    l = 1
    ir = vsize

    do
      if (ir - l .le. 1) then
        if (ir - l .eq. 1) then
          if (tempVec(l) .gt. tempVec(ir)) call swap(tempVec(l), tempVec(ir))
        end if
        valueAtBinFractionKernel = tempVec(nearestBin)
    exit

      else
        mid = (l + ir) / 2
        call swap(tempVec(mid), tempVec(l + 1))

        if (tempVec(l)  .gt. tempVec(ir))  call swap(tempVec(l),   tempVec(ir))
        if (tempVec(l+1).gt. tempVec(ir))  call swap(tempVec(l+1), tempVec(ir))
        if (tempVec(l)  .gt. tempVec(l+1)) call swap(tempVec(l),   tempVec(l+1))

        i = l + 1
        j = ir
        tempVal = tempVec(l + 1)
        do
          do
            i = i + 1
          if (tempVec(i) .ge. tempVal) exit
          end do

          do
            j = j - 1
          if (tempVec(j) .le. tempVal) exit
          end do

        if (j .lt. i) exit

          call swap(tempVec(i), tempVec(j))
        end do

        tempVec(l + 1) = tempVec(j)
        tempVec(j) = tempVal
        if (j .ge. nearestBin) ir = j - 1
        if (j .le. nearestBin) l = i
      end if
    end do

  end function valueAtBinFractionKernel

  !---------------------------------------------------------------------
  real function fracValScalar(min, max, frac)
  !---------------------------------------------------------------------

    implicit none 

    real, intent(in) :: min, max, frac

    !-------------------------------------------------------------------
    ! Start:

    fracValScalar = (1.0 - frac) * min + frac * max

  end function fracValScalar

  !---------------------------------------------------------------------
  function fracValVector(min, max, frac) result(fracValues)
  !---------------------------------------------------------------------

    implicit none 

    real, intent(in) :: min, max, frac(:)
    real :: fracValues(size(frac))

    !-------------------------------------------------------------------
    ! Start:

    fracValues = (1.0 - frac) * min + frac * max

  end function fracValVector

  !---------------------------------------------------------------------
  logical function isOdd(i)
  !---------------------------------------------------------------------

    implicit none 

    integer, intent(in) :: i

    !-------------------------------------------------------------------
    ! Start:

    if (real(floor(i / 2.0)) .eq. i / 2.0) then
      isOdd = .false.
    else
      isOdd = .true.
    end if

  end function isOdd

  !---------------------------------------------------------------------
  real function meanToBinFractionVector(vector, binFraction)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: vector(:), binFraction

    real :: cutoff
    logical :: maskVector(size(vector))
    integer :: numMasked

    !-------------------------------------------------------------------
    ! Start

    cutoff = valueAtBinFractionVector(vector, binFraction)

    where(vector .le. cutoff)
      maskVector = .true.
    elsewhere
      maskVector = .false.
    end where

    numMasked = count(maskVector)
    if (numMasked .gt. 0) then
      meanToBinFractionVector = sum(vector, mask=maskVector) / numMasked
    else
      meanToBinFractionVector = minval(vector)
    end if

  end function meanToBinFractionVector

  !---------------------------------------------------------------------
  real function meanToBinFractionArray(array, binFraction, maskArray)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: array(:,:), binFraction
    logical, intent(in), optional :: maskArray(:,:)

    real :: cutoff
    logical :: localMaskArray(size(array, 1), size(array, 2))
    integer :: numMasked
!integer :: xi, yi, counts
!real(double) :: summ

    !-------------------------------------------------------------------
    ! Start

    if (present(maskArray)) then
      cutoff = valueAtBinFractionArray(array, binFraction, maskArray)
!print *, 'cutoff =', cutoff, 'frac masked pixels =', count(maskArray) / size(array, 1) / real(size(array, 2))
    else
      cutoff = valueAtBinFractionArray(array, binFraction)
    end if

    where(array .le. cutoff)
      localMaskArray = .true.
    elsewhere
      localMaskArray = .false.
    end where

    if (present(maskArray)) then
      localMaskArray = localMaskArray .and. maskArray
    end if

    numMasked = count(localMaskArray)
!print *, 'binFraction check:', numMasked, count(maskArray), numMasked / real(count(maskArray))
    if (numMasked .gt. 0) then
      meanToBinFractionArray = sum(array, mask=localMaskArray) / numMasked
    else
      meanToBinFractionArray = minval(array)
    end if

!print *, ''
!print *, ''
!summ = 0.0
!counts = 0
!do xi = 1, size(array, 1)
!  do yi = 1, size(array, 2)
!    if (localMaskArray(xi, yi)) then
!      summ = summ + array(xi, yi)
!      counts = counts + 1
!      if (array(xi, yi) .gt. 1.0) then
!print *, 'Hey!!', xi, yi, array(xi, yi)
!      end if
!    end if
!  end do
!end do
!print *, counts, numMasked, summ / counts

  end function meanToBinFractionArray

  !---------------------------------------------------------------------
  subroutine rebinBasicReal(oldYvalues, newYvalues, status)
  !---------------------------------------------------------------------

    implicit none 

    real, intent(in) ::&
      oldYvalues(:)
    real, intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins
    integer ::&
      oldBin, newBin, localStatus

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    do oldBin = 1, numOldBins + 1
      oldXedges(oldBin) = (oldBin - 1) / real(numOldBins)
    end do
    do newBin = 1, numNewBins + 1
      newXedges(newBin) = (newBin - 1) / real(numNewBins)
    end do

    call rebinWithEdgesReal(oldYvalues, newYvalues, oldXedges, newXedges&
    , localStatus)

    if (present(status)) status = localStatus

  end subroutine rebinBasicReal

  !---------------------------------------------------------------------
  subroutine rebinBasicDble(oldYvalues, newYvalues, status)
  !---------------------------------------------------------------------

    implicit none 

    real(kind(1d0)), intent(in) ::&
      oldYvalues(:)
    real(kind(1d0)), intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real(kind(1d0)) ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins
    integer ::&
      oldBin, newBin, localStatus

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    do oldBin = 1, numOldBins + 1
      oldXedges(oldBin) = (oldBin - 1) / real(numOldBins)
    end do
    do newBin = 1, numNewBins + 1
      newXedges(newBin) = (newBin - 1) / real(numNewBins)
    end do

    call rebinWithEdgesDble(oldYvalues, newYvalues, oldXedges, newXedges&
    , localStatus)

    if (present(status)) status = localStatus

  end subroutine rebinBasicDble

  !---------------------------------------------------------------------
  subroutine rebinWithWidthsReal(oldYvalues, newYvalues, oldXwidths&
  , newXwidths, oldXcentres, newXcentres, status)
  !---------------------------------------------------------------------

    implicit none 

    real, intent(in), dimension(:) ::&
      oldXwidths, oldXcentres, oldYvalues, newXwidths, newXcentres
    real, intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins,&
      localStatus, foreignStatus
    integer ::&
      oldBin, newBin

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    if (size(oldXwidths) .lt. numOldBins) then
      if(present(status)) status = -1
  return
    else if (size(oldXwidths) .gt. numOldBins) then
      localStatus = localStatus + 1
    end if
    if (size(newXwidths) .lt. numNewBins) then
      if(present(status)) status = -2
  return
    else if (size(newXwidths) .gt. numNewBins) then
      localStatus = localStatus + 2
    end if

    if (size(oldXcentres) .lt. numOldBins) then
      if(present(status)) status = -3
  return
    else if (size(oldXcentres) .gt. numOldBins) then
      localStatus = localStatus + 4
    end if
    if (size(newXcentres) .lt. numNewBins) then
      if(present(status)) status = -4
  return
    else if (size(newXcentres) .gt. numNewBins) then
      localStatus = localStatus + 8
    end if

    do oldBin = 1, numOldBins
      oldXedges(oldBin) = oldXcentres(oldBin) - 0.5 * oldXwidths(oldBin)
    end do
    oldXedges(numOldBins + 1) = oldXcentres(numOldBins) + 0.5&
    * oldXwidths(numOldBins)

    do newBin = 1, numNewBins
      newXedges(newBin) = newXcentres(newBin) - 0.5 * newXwidths(newBin)
    end do
    newXedges(numNewBins + 1) = newXcentres(numNewBins) + 0.5&
    * newXwidths(numNewBins)

    call rebinWithEdgesReal(oldYvalues, newYvalues, oldXedges, newXedges&
    , foreignStatus)
    if (foreignStatus .lt. 0) then
      if(present(status)) status = foreignStatus - 4
  return
    end if

    localStatus = localStatus + foreignStatus * 16
    if(present(status)) status = localStatus

  end subroutine rebinWithWidthsReal

  !---------------------------------------------------------------------
  subroutine rebinWithWidthsDble(oldYvalues, newYvalues, oldXwidths&
  , newXwidths, oldXcentres, newXcentres, status)
  !---------------------------------------------------------------------

    implicit none 

    real(kind(1d0)), intent(in), dimension(:) ::&
      oldXwidths, oldXcentres, oldYvalues, newXwidths, newXcentres
    real(kind(1d0)), intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real(kind(1d0)) ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins,&
      localStatus, foreignStatus
    integer ::&
      oldBin, newBin

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    if (size(oldXwidths) .lt. numOldBins) then
      if(present(status)) status = -1
  return
    else if (size(oldXwidths) .gt. numOldBins) then
      localStatus = localStatus + 1
    end if
    if (size(newXwidths) .lt. numNewBins) then
      if(present(status)) status = -2
  return
    else if (size(newXwidths) .gt. numNewBins) then
      localStatus = localStatus + 2
    end if

    if (size(oldXcentres) .lt. numOldBins) then
      if(present(status)) status = -3
  return
    else if (size(oldXcentres) .gt. numOldBins) then
      localStatus = localStatus + 4
    end if
    if (size(newXcentres) .lt. numNewBins) then
      if(present(status)) status = -4
  return
    else if (size(newXcentres) .gt. numNewBins) then
      localStatus = localStatus + 8
    end if

    do oldBin = 1, numOldBins
      oldXedges(oldBin) = oldXcentres(oldBin) - 0.5d0 * oldXwidths(oldBin)
    end do
    oldXedges(numOldBins + 1) = oldXcentres(numOldBins) + 0.5d0&
    * oldXwidths(numOldBins)

    do newBin = 1, numNewBins
      newXedges(newBin) = newXcentres(newBin) - 0.5d0 * newXwidths(newBin)
    end do
    newXedges(numNewBins + 1) = newXcentres(numNewBins) + 0.5d0&
    * newXwidths(numNewBins)

    call rebinWithEdgesDble(oldYvalues, newYvalues, oldXedges, newXedges&
    , foreignStatus)
    if (foreignStatus .lt. 0) then
      if(present(status)) status = foreignStatus - 4
  return
    end if

    localStatus = localStatus + foreignStatus * 16
    if(present(status)) status = localStatus

  end subroutine rebinWithWidthsDble

  !---------------------------------------------------------------------
  subroutine rebinWithEdgesReal(oldYvalues, newYvalues, oldXedges, newXedges&
  , status)
  !---------------------------------------------------------------------

    implicit none 

    real, intent(in) ::&
      oldXedges(:), oldYvalues(:), newXedges(:)
    real, intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    integer :: foreignStatus
    real(kind(1d0)) :: newYvaluesDble(size(newYvalues))

    !-------------------------------------------------------------------
    ! Start:

    call rebinKernel(dble(oldYvalues), newYvaluesDble, dble(oldXedges)&
    , dble(newXedges), foreignStatus)

    if (present(status)) status = foreignStatus
    newYvalues = real(newYvaluesDble)

  end subroutine rebinWithEdgesReal

  !---------------------------------------------------------------------
  subroutine rebinWithEdgesDble(oldYvalues, newYvalues, oldXedges, newXedges&
  , status)
  !---------------------------------------------------------------------

    implicit none 

    real(kind(1d0)), intent(in) ::&
      oldXedges(:), oldYvalues(:), newXedges(:)
    real(kind(1d0)), intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start:

    call rebinKernel(oldYvalues, newYvalues, oldXedges&
    , newXedges, foreignStatus)

    if (present(status)) status = foreignStatus

  end subroutine rebinWithEdgesDble

  !---------------------------------------------------------------------
  subroutine rebinKernel(oldYvalues, newYvalues, oldXedges, newXedges&
  , status)
  !---------------------------------------------------------------------
  ! This is designed to be a pretty general and robust rebinning routine.
  ! Neither the old or new series is constrained to have bins of equal width.

    implicit none 

    real(kind(1d0)), intent(in) ::&
      oldXedges(:), oldYvalues(:), newXedges(:)
    real(kind(1d0)), intent(out) ::&
      newYvalues(:)
    integer, intent(out) :: status

    integer ::&
      numOldBins,&
      numNewBins
    integer ::&
      oldBin, oldStartBin, oldFinisBin, newBin
    logical ::&
      foundOverlappingBin
    real(kind(1d0)) ::&
      overlap, binWidth

    !-------------------------------------------------------------------
    ! Start:

    status = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    if (size(oldXedges) .lt. numOldBins + 1) then
      status = -1
  return
    else if (size(oldXedges) .gt. numOldBins + 1) then
      status = status + 1
    end if
    if (size(newXedges) .lt. numNewBins + 1) then
      status = -2
  return
    else if (size(newXedges) .gt. numNewBins + 1) then
      status = status + 2
    end if

    do oldBin = 1, numOldBins
      binWidth = oldXedges(oldBin + 1) - oldXedges(oldBin)
      if (binWidth .le. 0.0) then
        status = -3
  return
      end if
    end do

    do newBin = 1, numNewBins
      binWidth = newXedges(newBin + 1) - newXedges(newBin)
      if (binWidth .le. 0.0) then
        status = -4
  return
      end if
    end do

    newYvalues = 0.0 ! vector op, default.

    oldStartBin = 1
    NEW: do newBin = 1, numNewBins
      ! Find first old bin that overlaps newBin:
      foundOverlappingBin = .false. ! default
      OLD1: do oldBin = oldStartBin, numOldBins
        if (oldXedges(oldBin + 1) .gt. newXedges(newBin)&
        .and. oldXedges(oldBin) .lt. newXedges(newBin + 1)) then
          foundOverlappingBin = .true.
          oldStartBin = oldBin
      exit OLD1
        end if
      end do OLD1

      if (.not. foundOverlappingBin) then ! have either reached the end of the
      ! old series, or not reached the start. Leave new values at default 0.0.
    exit NEW
      end if

      ! Find the last old bin that overlaps newBin:
      foundOverlappingBin = .true. ! default
      oldFinisBin = numOldBins ! default
      OLD2: do oldBin = oldStartBin + 1, numOldBins
        if (oldXedges(oldBin) .ge. newXedges(newBin + 1)) then
          foundOverlappingBin = .false.
          oldFinisBin = oldBin - 1
      exit OLD2
        end if
      end do OLD2

      if (foundOverlappingBin) then ! have reached the end of the
      ! old series without leaving the new bin.
        oldFinisBin = numOldBins
      end if

      ! Now work though all the old bins that overlap the new, and pour in
      ! the correct proportional amounts from them into newBin:
      OLD3: do oldBin = oldStartBin, oldFinisBin
        if (oldXedges(oldBin + 1) .lt. newXedges(newBin + 1)) then
          if (oldXedges(oldBin) .gt. newXedges(newBin)) then !B - all
          ! of old bin goes into new:
            overlap = oldXedges(oldBin + 1) - oldXedges(oldBin)

          else !A - upper part of old goes into new:
            overlap = oldXedges(oldBin + 1) - newXedges(newBin)
          end if

        else
          if (oldXedges(oldBin) .gt. newXedges(newBin)) then !D - lower
          ! part of old goes into new:
            overlap = newXedges(newBin + 1) - oldXedges(oldBin)

          else !C - middle part of old goes into new:
            overlap = newXedges(newBin + 1) - newXedges(newBin)
          end if
        end if
        newYvalues(newBin) = newYvalues(newBin) + oldYvalues(oldBin)&
        * overlap
      end do OLD3

      oldStartBin = oldFinisBin
    end do NEW

    ! Divide by the new bin width:
    do newBin = 1, numNewBins
      binWidth = newXedges(newBin + 1) - newXedges(newBin)
      newYvalues(newBin) = newYvalues(newBin) / binWidth
    end do

  end subroutine rebinKernel

  !---------------------------------------------------------------------
  function normaliseReal(argument)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: argument(:)
    real :: normaliseReal(size(argument)), vectorLength2
    integer :: i

    !-------------------------------------------------------------------
    ! Start

    vectorLength2 = 0.0
    do i = 1, size(argument)
      vectorLength2 = vectorLength2 + argument(i) * argument(i)
    end do
    if (vectorLength2 .le. 0.0) then
      normaliseReal = 0.0
    else
      normaliseReal = argument / sqrt(vectorLength2)
    end if

  end function normaliseReal

  !---------------------------------------------------------------------
  function normaliseDble(argument)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)), intent(in) :: argument(:)
    real(kind(1d0)) :: normaliseDble(size(argument)), vectorLength2
    integer :: i

    !-------------------------------------------------------------------
    ! Start

    vectorLength2 = 0.0d0
    do i = 1, size(argument)
      vectorLength2 = vectorLength2 + argument(i) * argument(i)
    end do
    if (vectorLength2 .le. 0.0) then
      normaliseDble = 0.0d0
    else
      normaliseDble = argument / sqrt(vectorLength2)
    end if

  end function normaliseDble

  !---------------------------------------------------------------------
  function vectorCrossReal(vectorA, vectorB)
  !---------------------------------------------------------------------

    implicit none

    real, dimension(3), intent(in) :: vectorA, vectorB
    real :: vectorCrossReal(3)

    !-------------------------------------------------------------------
    ! Start

    vectorCrossReal(1) = vectorA(2) * vectorB(3) - vectorA(3) * vectorB(2)
    vectorCrossReal(2) = vectorA(3) * vectorB(1) - vectorA(1) * vectorB(3)
    vectorCrossReal(3) = vectorA(1) * vectorB(2) - vectorA(2) * vectorB(1)

  end function vectorCrossReal

  !---------------------------------------------------------------------
  function vectorCrossDble(vectorA, vectorB)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)), dimension(3), intent(in) :: vectorA, vectorB
    real(kind(1d0)) :: vectorCrossDble(3)

    !-------------------------------------------------------------------
    ! Start

    vectorCrossDble(1) = vectorA(2) * vectorB(3) - vectorA(3) * vectorB(2)
    vectorCrossDble(2) = vectorA(3) * vectorB(1) - vectorA(1) * vectorB(3)
    vectorCrossDble(3) = vectorA(1) * vectorB(2) - vectorA(2) * vectorB(1)

  end function vectorCrossDble

  !---------------------------------------------------------------------
  function bias(sag, x) result(y)
  !---------------------------------------------------------------------
  ! The action of this little function is to bias the x values (which should
  ! be between 0 and 1) either closer to 0 or closer to 1, depending on
  ! the value of 'sag'. For sag = -1, the 'upward' bias is at its limit:
  ! whatever the value of x, y will be set to 1. For sag = +1 the bias is
  ! at the opposite extreme, in which the function returns 0 for any value
  ! of x. For sag = 0, y = x. The interesting things happen when sag !=
  ! either -1, 0 or +1. In these cases the relation of y to x is a
  ! hyperbolic function that generates the required bias.
  !
  ! This function gives accurate results except for x values close to 1.
  ! A call such as
  !
  !	y - 1.0 - bias(-sag, 1.0-x)
  !
  ! is better in this instance.

    implicit none

    real(kind=single), intent(in) :: sag, x
    real(kind=single) :: y

    !-------------------------------------------------------------------
    ! Start:

    if (x .gt. 1.0 .or. x .lt. 0.0) then
      write(errstr, '(a, f10.7, a)') 'The argument ', x, ' is outside &
      &the range 0 to 1.'
call error('biasArgOutOfRange') ! should return a status really, and let the calling routine work out what to do.
    end if

    if (abs(sag) .gt. 1.0) then
      write(errstr, '(a, f10.7, a)') 'The sag value ', sag, ' is outside &
      &the range -1 to 1.'
call error('biasSagOutOfRange') ! should return a status really, and let the calling routine work out what to do.
    end if

    if (sag .eq. -1.0) then
      y = 0.0
    else if (sag .eq. 1.0) then
      y = 1.0
    else
      y = x * (1.0 + sag) / (1.0 + sag * (2.0 * x - 1.0))
    end if

  end function bias

  !---------------------------------------------------------------------
  function spread(stretch, x) result(y)
  !---------------------------------------------------------------------
  ! The action of this little function is to stretch or compress the x
  ! values (which should be between 0 and 1) about 0, depending on the
  ! value of 'stretch'. For stretch = -1, all the points are compressed
  ! into one: whatever the value of x, y will be set to 0. For stretch
  ! = +1 the spread is at its maximum, in which the function returns -1
  ! for negative values of x and +1 for positive values. At stretch = 0,
  ! y = x. The interesting things happen when stretch != either -1, 0
  ! or +1. In these cases the relation of y to x is a doUble hyperbolic
  ! function (generated by the appropriate call to the function bias)
  ! that generates the required spread.

    implicit none

    real(kind=single), intent(in) :: stretch, x
    real(kind=single) :: y

    !-------------------------------------------------------------------
    ! Start:

    if (abs(x) .gt. 1.0) then
      write(errstr, '(a, f10.7, a)') 'The argument ', x, ' is outside &
      &the range 0 to 1.'
call error('spreadArgOutOfRange') ! should return a status really, and let the calling routine work out what to do.
    end if

    if (abs(stretch) .gt. 1.0) then
      write(errstr, '(a, f10.7, a)') 'The stretch value ', stretch, ' is outside &
      &the range -1 to 1.'
call error('stretchOutOfRange') ! should return a status really, and let the calling routine work out what to do.
    end if

    if (stretch .eq. -1.0) then
      y = 0.0
    else if (stretch .eq. 1.0) then
      y = real(sign(1.0, x))
    else
      if (x .gt. 0.0) then
        y = bias(-stretch, x)
      else
        y = -bias(-stretch, -x)
      end if
    end if

  end function spread

  !---------------------------------------------------------------------
  function bbFluxPerHertz_old(photonEnergyKeV, tempMegaKelvin) result(flux)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: photonEnergyKeV, tempMegaKelvin
    real :: flux, arg

    !-------------------------------------------------------------------
    ! Start:

!    arg = photonEnergyKeV * qeOnBoltzmann / tempMegaKelvin / 1000.0
    arg = photonEnergyKeV * ElectronQ / Boltzmann / tempMegaKelvin / 1000.0
    if (arg .gt. 20.0) then
      flux = 0.0
    else if (arg .lt. 0.01) then
      flux = 2.0 * arg * arg / (2.0 + arg) ! long-wavelength approx. using
      ! the first three terms of the exp function.
    else
      flux = arg * arg * arg / (exp(arg) - 1.0)
    end if

    ! Scaling is ignored because it will
    ! be normalised anyway by the calling routine.

  end function bbFluxPerHertz_old

  !---------------------------------------------------------------------
  subroutine cmplxArrayDivideDble(srcImageA, srcImageB, xsize, ysize)
  !---------------------------------------------------------------------

! Divides A by B.

    implicit none 

    integer, intent(in) :: xsize, ysize
!    complex(kind=real(kind(1d0))), dimension(xsize, ysize), intent(inout) ::&
    complex(kind(1d0)), dimension(xsize, ysize), intent(inout) ::&
      srcImageA
!    complex(kind=real(kind(1d0))), dimension(xsize, ysize), intent(in) ::&
    complex(kind(1d0)), dimension(xsize, ysize), intent(in) ::&
      srcImageB
    real(kind(1d0)) :: denom, rPartA, iPartA, rPartB, iPartB, rPartC, iPartC
    integer :: xi, yi

!------------------------------------------------------------------------
! Start:

    do xi = 1, xsize
      do yi = 1, ysize
        rPartA = real(srcImageA(xi, yi))
        iPartA = aimag(srcImageA(xi, yi))
        rPartB = real(srcImageB(xi, yi))
        iPartB = aimag(srcImageB(xi, yi))
        denom = rPartB * rPartB + iPartB * iPartB
        if (denom .eq. 0.0) then
          srcImageA(xi, yi) = (0d0, 0d0)
        else
          rPartC = rPartA * rPartB + iPartA * iPartB
          iPartC = iPartA * rPartB - rPartA * iPartB
!          srcImageA(xi, yi) = cmplx(rPartC, iPartC, real(kind(1d0)))
          srcImageA(xi, yi) = cmplx(rPartC, iPartC, kind(1d0))
        end if
      end do
    end do

  end subroutine cmplxArrayDivideDble

  !---------------------------------------------------------------------
  subroutine cmplxArrayDivideReal(srcImageA, srcImageB, xsize, ysize)
  !---------------------------------------------------------------------

! Divides A by B.

    implicit none 

    integer, intent(in) :: xsize, ysize
    complex, dimension(xsize, ysize), intent(inout) ::&
      srcImageA
    complex, dimension(xsize, ysize), intent(in) ::&
      srcImageB
    real :: denom, rPartA, iPartA, rPartB, iPartB, rPartC, iPartC
    integer :: xi, yi

!------------------------------------------------------------------------
! Start:

    do xi = 1, xsize
      do yi = 1, ysize
        rPartA = real(srcImageA(xi, yi))
        iPartA = aimag(srcImageA(xi, yi))
        rPartB = real(srcImageB(xi, yi))
        iPartB = aimag(srcImageB(xi, yi))
        denom = rPartB * rPartB + iPartB * iPartB
        if (denom .eq. 0.0) then
          srcImageA(xi, yi) = (0.0, 0.0)
        else
          rPartC = rPartA * rPartB + iPartA * iPartB
          iPartC = iPartA * rPartB - rPartA * iPartB
          srcImageA(xi, yi) = cmplx(rPartC, iPartC)
        end if
      end do
    end do

  end subroutine cmplxArrayDivideReal

  !---------------------------------------------------------------------
  subroutine cmplxVectorDivideReal(vectorNum, vectorDen, vectorRes, size)
  !---------------------------------------------------------------------

! Divides vectorNum by vectorDen and leaves result in vectorRes.

    implicit none 

    integer, intent(in) :: size
    complex, dimension(size), intent(in) ::&
      vectorNum, vectorDen
    complex, dimension(size), intent(out) ::&
      vectorRes
    real :: denom, rPartA, iPartA, rPartB, iPartB, rPartC, iPartC
    integer :: i

!------------------------------------------------------------------------
! Start:

    do i = 1, size
      rPartA = real(vectorNum(i))
      iPartA = aimag(vectorNum(i))
      rPartB = real(vectorDen(i))
      iPartB = aimag(vectorDen(i))
      denom = rPartB * rPartB + iPartB * iPartB
      if (denom .eq. 0.0) then
        vectorRes(i) = (0.0, 0.0)
      else
        rPartC = rPartA * rPartB + iPartA * iPartB
        iPartC = iPartA * rPartB - rPartA * iPartB
        vectorRes(i) = cmplx(rPartC, iPartC)
      end if
    end do

  end subroutine cmplxVectorDivideReal

  !---------------------------------------------------------------------
  subroutine integrate1dRealIntervals(y, intervals, n, result, status)
  !---------------------------------------------------------------------

    implicit none

    type(intervalsType)            :: intervals(:)
    integer, intent(in),  optional :: n
    real,    intent(out)           :: result
    integer, intent(out), optional :: status

    real :: realNumMidSamples, frac,&
      x, midIntegral, localLo, localHi
    real, allocatable :: inInterval(:)
    integer :: i, localStatus, localN, numMidSamples, loN, hiN

    interface
      real function y(x)
        real, intent(in) :: x
      end function y
    end interface

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    if (present(n)) then
      localN = n
    else
      localN = defaultNumIntegralSamples
    end if

    !**** should check intervals are in order, and don't overlap.

    !**** can't yet handle infinite bounds.

localLo = minval(intervals%lower%value)
localHi = maxval(intervals%upper%value)
allocate(inInterval(localN))
inInterval = 0.0 ! default
do i = 1, size(intervals)
  loN = 1 + nint(localN * ((intervals(i)%lower%value - localLo) / (localHi - localLo)))
  hiN = 1 + nint(localN * ((intervals(i)%upper%value - localLo) / (localHi - localLo)))
  loN = min(loN, localN)
  hiN = min(hiN, localN)
  inInterval(loN:hiN) = 1.0
end do
numMidSamples = localN
    realNumMidSamples = real(numMidSamples)

    midIntegral = 0.0
    do i = 1, numMidSamples
      frac = (i - 0.5) / realNumMidSamples
      x = localLo * (1.0 - frac) + localHi * frac
midIntegral = midIntegral + y(x) * inInterval(i)
    end do
    midIntegral = midIntegral * (localHi - localLo) / realNumMidSamples

    result = midIntegral

    if (present(status)) status = localStatus

  end subroutine integrate1dRealIntervals

  !---------------------------------------------------------------------
  subroutine integrate1dReal(y, lo, hi, loInf, hiInf, n, result, status)
  !---------------------------------------------------------------------

    implicit none

    real,    intent(in),  optional :: lo, hi
    logical, intent(in),  optional :: loInf, hiInf
    integer, intent(in),  optional :: n
    real,    intent(out)           :: result
    integer, intent(out), optional :: status

    real :: realNumLoSamples, realNumMidSamples, realNumHiSamples, frac,&
      x, loIntegral, midIntegral, hiIntegral, localLo, localHi, u
    integer :: i, localStatus, localN, numLoSamples, numMidSamples, numHiSamples
    logical :: localLoInf, localHiInf

    interface
      real function y(x)
        real, intent(in) :: x
      end function y
    end interface

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    if (present(lo)) then
      localLo = lo
      if (present(loInf)) then
        localLoInf = loInf
      else
        localLoInf = .false.
      end if
      if (localLoInf .and. localLo .gt. 0.0) then
        if (present(status)) status = -1
  return
      end if
    else
      localLoInf = .true.
      localLo = -1.0
    end if

    if (present(hi)) then
      localHi = hi
      if (present(hiInf)) then
        localHiInf = hiInf
      else
        localHiInf = .false.
      end if
      if (localHiInf .and. localHi .lt. 0.0) then
        if (present(status)) status = -2
  return
      end if
    else
      localHiInf = .true.
      localHi = 1.0
    end if

    if (present(n)) then
      localN = n
    else
      localN = defaultNumIntegralSamples
    end if

    if (present(lo) .and. present(hi)) then
      if (lo .gt. hi) then
        if (present(status)) status = -3
  return
      else if (lo .eq. hi) then
        result = 0.0
        localStatus = localStatus + 1
        if (present(status)) status = localStatus
  return
      end if
    end if

    if (localLoInf) then
      if (localHiInf) then
        numLoSamples = floor(localN / 3.0)
        numHiSamples = floor(localN / 3.0)
        if (numLoSamples .lt. 1) then
          if (present(status)) status = -4
  return
        end if
      else
        numHiSamples = 0
        numLoSamples = floor(localN / 2.0)
        if (numLoSamples .lt. 1) then
          if (present(status)) status = -5
  return
        end if
      end if
    else
      numLoSamples = 0
      if (localHiInf) then
        numHiSamples = floor(localN / 2.0)
        if (numHiSamples .lt. 1) then
          if (present(status)) status = -6
  return
        end if
      else
        numHiSamples = 0
      end if
    end if
    numMidSamples = localN - numLoSamples - numHiSamples
    if (numMidSamples .lt. 1) then
      if (present(status)) status = -7
  return
    end if

    realNumLoSamples = real(numLoSamples)
    realNumMidSamples = real(numMidSamples)
    realNumHiSamples = real(numHiSamples)

    loIntegral = 0.0
    if (realNumLoSamples .gt. 0) then
      do i = 1, numLoSamples
        frac = (i - 0.5) / realNumLoSamples
        u = frac / localLo
        x = 1.0 / u
        loIntegral = loIntegral + y(x) / u / u
      end do
      loIntegral = loIntegral / realNumLoSamples / abs(localLo)
    end if

    midIntegral = 0.0
    do i = 1, numMidSamples
      frac = (i - 0.5) / realNumMidSamples
      x = localLo * (1.0 - frac) + localHi * frac
      midIntegral = midIntegral + y(x)
    end do
    midIntegral = midIntegral * (localHi - localLo) / realNumMidSamples

    hiIntegral = 0.0
    if (realNumHiSamples .gt. 0) then
      do i = 1, numHiSamples
        frac = (i - 0.5) / realNumHiSamples
        u = frac / localHi
        x = 1.0 / u
        hiIntegral = hiIntegral + y(x) / u / u
      end do
      hiIntegral = hiIntegral / realNumHiSamples / localHi
    end if

    result = loIntegral + midIntegral +  hiIntegral

    if (present(status)) status = localStatus

  end subroutine integrate1dReal

  !---------------------------------------------------------------------
  subroutine integrateNdReal(y, args, freeIndices, lo, hi, loInf, hiInf, n&
  , dims, result, status)
  !---------------------------------------------------------------------
! traps: actual y called with x(some number) instead of x(:); setting all lo/hiinf() but not all lo/hi().

    implicit none

    real,    intent(in),  optional :: lo(:), hi(:), args(:)
    logical, intent(in),  optional :: loInf(:), hiInf(:), freeIndices(:)
    integer, intent(in),  optional :: n(:)
    integer, intent(in)            :: dims
    real,    intent(out)           :: result
    integer, intent(out), optional :: status

    real :: localLo(dims), localHi(dims), localArgs(dims)
    integer :: localStatus, numLoSamples(dims), numMidSamples(dims),&
      numHiSamples(dims), localN(dims), i
    logical :: localLoInf(dims), localHiInf(dims), localFreeIndices(dims),&
      infClashFlag

    interface
      real function y(x)
        real, intent(in) :: x(:)
      end function y
    end interface

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

!***** check all sizes are equal to dims

    if (present(lo)) then
      localLo = lo
      if (present(loInf)) then
        localLoInf = loInf
      else
        localLoInf = .false.
      end if
      do i = 1, dims
        if (localLoInf(i) .and. localLo(i) .ge. 0.0) then
          if (present(status)) status = -1
  return
        end if
      end do
    else
      localLo = -1.0
      localLoInf = .true.
      if (present(loInf)) then
        infClashFlag = .false. ! default
        do i = 1, dims
          if (.not. loInf(i)) then
            infClashFlag = .true.
        exit
          end if
        end do
        if (infClashFlag .and. present(status)) status = status + 2
      end if
    end if

    if (present(hi)) then
      localHi = hi
      if (present(hiInf)) then
        localHiInf = hiInf
      else
        localHiInf = .false.
      end if
      do i = 1, dims
        if (localHiInf(i) .and. localHi(i) .le. 0.0) then
          if (present(status)) status = -2
  return
        end if
      end do
    else
      localHi = 1.0
      localHiInf = .true.
      if (present(hiInf)) then
        infClashFlag = .false. ! default
        do i = 1, dims
          if (.not. hiInf(i)) then
            infClashFlag = .true.
        exit
          end if
        end do
        if (infClashFlag .and. present(status)) status = status + 4
      end if
    end if

    if (present(args)) then
      localArgs = args
      if (present(freeIndices)) then
        localFreeIndices = freeIndices
      else
        localFreeIndices = .true.
      end if
    else
      localArgs = 0.0
      localFreeIndices = .true.
    end if

    if (present(n)) then
      localN = n
    else
      localN = defaultNumIntegralSamples
    end if

    if (present(lo) .and. present(hi)) then
      do i = 1, dims
        if (lo(i) .gt. hi(i)) then
        if (present(status)) status = -3
  return
        else if (lo(i) .eq. hi(i)) then
          result = 0.0
          localStatus = localStatus + 1
          if (present(status)) status = localStatus
  return
        end if
      end do
    end if

    do i = 1, dims
      if (localLoInf(i)) then
        if (localHiInf(i)) then
          numLoSamples(i) = floor(localN(i) / 3.0)
          numHiSamples(i) = floor(localN(i) / 3.0)
          if (numLoSamples(i) .lt. 1) then
            if (present(status)) status = -4
  return
          end if
        else
          numHiSamples(i) = 0
          numLoSamples(i) = floor(localN(i) / 2.0)
          if (numLoSamples(i) .lt. 1) then
            if (present(status)) status = -5
  return
          end if
        end if
      else
        numLoSamples(i) = 0
        if (localHiInf(i)) then
          numHiSamples(i) = floor(localN(i) / 2.0)
          if (numHiSamples(i) .lt. 1) then
            if (present(status)) status = -6
  return
          end if
        else
          numHiSamples(i) = 0
        end if
      end if
      numMidSamples(i) = localN(i) - numLoSamples(i) - numHiSamples(i)
      if (numMidSamples(i) .lt. 1) then
        if (present(status)) status = -7
  return
      end if
    end do

    call integrateKernel(y, localArgs, localFreeIndices, localLo&
    , localHi, numLoSamples, numMidSamples, numHiSamples, result&
    , localStatus)

    if (present(status)) status = localStatus

  end subroutine integrateNdReal

  !---------------------------------------------------------------------
  recursive subroutine integrateKernel(y, args, freeIndices, lo, hi&
  , numLoSamples, numMidSamples, numHiSamples, result, status)
  !---------------------------------------------------------------------

    implicit none

    real,    intent(in)  :: lo(:), hi(:), args(:)
    logical, intent(in)  :: freeIndices(:)
    integer, intent(in)  :: numLoSamples(:), numMidSamples(:), numHiSamples(:)
    real,    intent(out) :: result
    integer, intent(out) :: status

    real :: realNumLoSamples, realNumMidSamples, realNumHiSamples, frac,&
      x(size(args)), u, localLo, localHi, loIntegral, midIntegral, hiIntegral,&
      localResult, term
    integer :: i, dims, firstFreeIndex, lastFreeIndex
    logical :: localFreeIndices(size(args))

    interface
      real function y(x)
        real, intent(in) :: x(:)
      end function y
    end interface

    !-------------------------------------------------------------------
    ! Start

    status = 0 ! default
    dims = size(args)

    ! Find index of last free index:
    do i = dims, 1, -1
      if (freeIndices(i)) then
        lastFreeIndex = i
    exit
      end if
    end do

    ! Find index of first free index and integrate over that:
    do i = 1, dims
      if (freeIndices(i)) then
        firstFreeIndex = i
    exit
      end if
    end do

    realNumLoSamples = real(numLoSamples(firstFreeIndex))
    realNumMidSamples = real(numMidSamples(firstFreeIndex))
    realNumHiSamples = real(numHiSamples(firstFreeIndex))
    localLo = lo(firstFreeIndex)
    localHi = hi(firstFreeIndex)

    x = args

    if (firstFreeIndex .eq. lastFreeIndex) then ! direct integration:
      loIntegral = 0.0
      if (realNumLoSamples .gt. 0.0) then
        do i = 1, numLoSamples(firstFreeIndex)
          frac = (i - 0.5) / realNumLoSamples
          u = frac / localLo
          x(firstFreeIndex) = 1.0 / u
          term = y(x) / u / u
          loIntegral = loIntegral + term
        end do
        loIntegral = loIntegral / realNumLoSamples / abs(localLo)
      end if

      midIntegral = 0.0
      do i = 1, numMidSamples(firstFreeIndex)
        frac = (i - 0.5) / realNumMidSamples
        x(firstFreeIndex) = localLo * (1.0 - frac) + localHi * frac
        term = y(x)
        midIntegral = midIntegral + term
      end do
      midIntegral = midIntegral * (localHi - localLo) / realNumMidSamples

      hiIntegral = 0.0
      if (realNumHiSamples .gt. 0) then
        do i = 1, numHiSamples(firstFreeIndex)
          frac = (i - 0.5) / realNumHiSamples
          u = frac / localHi
          x(firstFreeIndex) = 1.0 / u
          term = y(x) / u / u
          hiIntegral = hiIntegral + term
        end do
        hiIntegral = hiIntegral / realNumHiSamples / localHi
      end if

    else ! recursive call:
      localFreeIndices = freeIndices
      localFreeIndices(firstFreeIndex) = .false.

      loIntegral = 0.0
      if (realNumLoSamples .gt. 0.0) then
        do i = 1, numLoSamples(firstFreeIndex)
          frac = (i - 0.5) / realNumLoSamples
          u = frac / localLo
          x(firstFreeIndex) = 1.0 / u
          call integrateKernel(y, x, localFreeIndices, lo&
          , hi, numLoSamples, numMidSamples, numHiSamples, localResult&
          , status)

          loIntegral = loIntegral + localResult / u / u
        end do
        loIntegral = loIntegral / realNumLoSamples / abs(localLo)
      end if

      midIntegral = 0.0
      do i = 1, numMidSamples(firstFreeIndex)
        frac = (i - 0.5) / realNumMidSamples
        x(firstFreeIndex) = localLo * (1.0 - frac) + localHi * frac
        call integrateKernel(y, x, localFreeIndices, lo&
        , hi, numLoSamples, numMidSamples, numHiSamples, localResult&
        , status)

        midIntegral = midIntegral + localResult
      end do
      midIntegral = midIntegral * (localHi - localLo) / realNumMidSamples

      hiIntegral = 0.0
      if (realNumHiSamples .gt. 0) then
        do i = 1, numHiSamples(firstFreeIndex)
          frac = (i - 0.5) / realNumHiSamples
          u = frac / localHi
          x(firstFreeIndex) = 1.0 / u
          call integrateKernel(y, x, localFreeIndices, lo&
          , hi, numLoSamples, numMidSamples, numHiSamples, localResult&
          , status)

          hiIntegral = hiIntegral + localResult / u / u
        end do
        hiIntegral = hiIntegral / realNumHiSamples / localHi
      end if
    end if

    result = loIntegral + midIntegral +  hiIntegral

  end subroutine integrateKernel

  !---------------------------------------------------------------------
  subroutine newtonsMethod(startX, y, dy, solutionX, status)
  !---------------------------------------------------------------------
  ! This subroutine solves the equation y(x) = 0 by Newton's Method. The
  ! user must construct two functions in the calling module, one of which
  ! returns y and the other dy/dx for a given x. These functions (and their
  ! single argument x) must be of type real. The user should pass their
  ! names in the call to newtonsMethod.

    implicit none

    real, intent(in) :: startX
    real, intent(out) :: solutionX
    integer, intent(out) :: status

    interface
      real function y(x)
        real, intent(in) :: x
      end function y
      real function dy(x)
        real, intent(in) :: x
      end function dy
    end interface

    real :: yValue, yDeriv, x, oldXstep, xStep, myPrecision

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, ''
if (testFlag) print *, 'Newtons method'

    status = 0 ! default

    myPrecision = 10.0 ** (-precision(x))

    yValue = y(startX)
    yDeriv = dy(startX)
    if (yDeriv .eq. 0.0) then
      status = -1
  return
    end if
    Xstep = -Yvalue / yDeriv
    if (abs(startX) .gt. 0.0) then
      if (abs(Xstep / startX) .le. myPrecision) then
        solutionX = startX + Xstep
  return
      end if
    end if
if (testFlag) print *, x
    x = startX + Xstep
    oldXstep = Xstep

    do
if (testFlag) print *, x
      yValue = y(x)
      yDeriv = dy(x)
      if (yDeriv .eq. 0.0) then
        status = -1
  return
      end if

      Xstep = -yValue / yDeriv

      if (abs(x) .gt. 0.0) then
        if (abs(Xstep / x) .le. myPrecision) then
          solutionX = x + Xstep
if (testFlag) print *, solutionX
  return
        end if
      end if

      if (abs(Xstep) .gt. abs(oldXstep)) then
        status = -2
  return
      end if

      x = x + Xstep
      oldXstep = Xstep
    end do

  end subroutine newtonsMethod

  !---------------------------------------------------------------------
  subroutine hybridNewtonsMethod(startX0, startX1, y, dy, solutionX&
  , maxNumIterations, convergenceCriterion, stat, status)
  !---------------------------------------------------------------------
  ! Adapted from the hybrid bisection/Newton-Raphson method in Numerical
  ! Methods.
  !
  ! This subroutine solves the equation y(x) = 0 by Newton's Method. The
  ! user must construct two functions in the calling module, one of which
  ! returns y and the other dy/dx for a given x. These functions (and their
  ! single argument x) must be of type real. The user should pass their
  ! names in the call to hybridNewtonsMethod.
  !
  ! The starting values startX0 and startX1 should bracket the zero point
  ! but it is not necessary that they be in any order.

    implicit none

    real,    intent(in)            :: startX0, startX1
    real,    intent(out)           :: solutionX
    integer, intent(in),  optional :: maxNumIterations
    real,    intent(in),  optional :: convergenceCriterion
    integer, intent(out), optional :: status
    type(statusType), intent(out), optional :: stat

    interface
      real function y(x)
        real, intent(in) :: x
      end function y
      real function dy(x)
        real, intent(in) :: x
      end function dy
    end interface

    real :: yValue, yDeriv, localConvergenceCriterion, startY0, startY1,&
      xLo, xHi, dx, dxOld, oldSolutionX
    integer :: localStatus, localMaxNumIter, j
    type(statusType) :: localStat

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    localStat%status = 0 ! default
    localStat%numPosBits = 1
    localStat%maxNeg = -2
    if (present(stat)) stat = localStat

    solutionX = 0.0 ! Don't like to leave it with no value, even if the routine
    ! can't calculate it.

    if (present(maxNumIterations)) then
      localMaxNumIter = maxNumIterations
    else
      localMaxNumIter = 100
!*** can do better than this.
    end if

    startY0 = y(startX0)
    startY1 = y(startX1)

    if (startY0 .eq. 0.0) then
      solutionX = startX0
  return
    end if
    if (startY1 .eq. 0.0) then
      solutionX = startX1
  return
    end if

    if (startX0 .eq. startX1) then
      if (present(status)) status = -1 ! no space between bounds.
      if (present(stat)) stat%status = -1 ! no space between bounds.
  return
    end if

    if (startY0 .gt. 0.0 .and. startY1 .gt. 0.0&
    .or. startY0 .lt. 0.0 .and. startY1 .lt. 0.0) then
      if (present(status)) status = -2 ! startX0 and
      if (present(stat)) stat%status = -2 ! startX0 and
      ! startX1 don't bracket the root.
  return
    end if

    ! Orient bounds such that y(xLo) < 0:
    if (startY0 .lt. 0.0) then
      xLo = startX0
      xHi = startX1
    else
      xLo = startX1
      xHi = startX0
    end if

    solutionX = 0.5 * (xLo + xHi)
    dxOld = xHi - xLo
    dx = dxOld

    if (present(convergenceCriterion)) then
      localConvergenceCriterion = abs(convergenceCriterion)
    else
      localConvergenceCriterion = 0.5 * abs(xLo + xHi)&
      * (10.0 ** (1-precision(dx)))
    end if

    yValue = y(solutionX)
    yDeriv = dy(solutionX)
    do j = 1, localMaxNumIter
      dxOld = dx

      if (((solutionX - xHi) * yDeriv - yValue)&
      * ((solutionX - xLo) * yDeriv - yValue) .gt. 0.0&
      .or. abs(2.0 * yValue) .gt. abs(dxOld * yDeriv)) then
        dx = 0.5 * (xHi - xLo)
        solutionX = xLo + dx
    if (solutionX .eq. xLo) exit

      else
        dx = yValue / yDeriv
        oldSolutionX = solutionX
        solutionX = solutionX - dx
    if (solutionX .eq. oldSolutionX) exit
      end if

    if (abs(dx) .lt. localConvergenceCriterion) exit

      yValue = y(solutionX)
      yDeriv = dy(solutionX)

      if (yValue .lt. 0.0) then
        xLo = solutionX
      else
        xHi = solutionX
      end if
    end do
    if (j .ge. localMaxNumIter) then
      localStatus = 1
      localStat%status = 1
    end if

    if (present(status)) status = localStatus
    if (present(stat)) stat = localStat

  end subroutine hybridNewtonsMethod

  !---------------------------------------------------------------------
  subroutine prepareForInterpolate1d(xVector, yVector, lastIndex, status)
  !---------------------------------------------------------------------

    implicit none

    real,    intent(inout)         :: xVector(:), yVector(:)
    integer, intent(out)           :: lastIndex
    integer, intent(out), optional :: status

    integer :: localStatus, ySize
    real, allocatable :: yVector2d(:,:)

    !-------------------------------------------------------------------
    ! Start:

    ySize = size(yVector)
    allocate(yVector2d(ySize, 1))
    call prepareForInterpolate2d(xVector, yVector2d, lastIndex, localStatus)
    yVector = yVector2d(:, 1)

    if (present(status)) status = localStatus

  end subroutine prepareForInterpolate1d

  !---------------------------------------------------------------------
  subroutine prepareForInterpolate2d(xVector, yVector, lastIndex, status)
  !---------------------------------------------------------------------
  ! In order to preserve portability, the interpolate() subroutine has been left a bit fragile, in that it demands that the values of xVector should be unique, and also sorted in order of increasing value. The prepareForInterpolate() interface can be used at the expense of its own portability (it uses the ssclib routine quick_sort, which is not guaranteed to be generally available) to remove this requirement. This precursor to interpolate() takes any series of values, sorts them, then compresses all stretches of the same x values into one. The new y value for each of the new representative x values is the average of all the y values that corresponded to the same previous x value.
  !
  ! The sorted, unique values of both x and y are crowded into the bottom half of their respective arrays. The last element of this sequence is returned in the integer lastIndex. The two routines should therefore be called in a similar way to the following example:
  !
  !	call prepareForInterpolate(xVector, yVector, lastIndex)
  !	call interpolate(xVector(1:lastIndex), yVector(1:lastIndex), x, y)

    use sort_mod

    implicit none

    real,    intent(inout)         :: xVector(:), yVector(:,:)
    integer, intent(out)           :: lastIndex
    integer, intent(out), optional :: status

    integer :: localStatus, ySize1, ySize2, i, j, startOfSameX, endOfSameX
    integer, allocatable :: indices(:)
    real, allocatable :: xSortedYvector(:,:)
    real :: lastXvalue

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    lastIndex = 0

    ySize1 = size(yVector, 1)
    if (size(xVector) .ne. ySize1) then
      if (present(status)) status = -1
  return
    end if

    allocate(indices(ySize1))
    call quick_sort(xVector, indices)

    ySize2 = size(yVector, 2)
    allocate(xSortedYvector(ySize1, ySize2))
    do i = 1, ySize1
      do j = 1, ySize2
        xSortedYvector(i, j) = yVector(indices(i), j)
      end do
    end do

    startOfSameX = 1
    lastXvalue = xVector(1)
    do i = 2, ySize1 - 1
      if (xVector(i) .gt. lastXvalue) then
        endOfSameX = i - 1
        lastIndex = lastIndex + 1
        if (endOfSameX .gt. startOfSameX) then
          do j = 1, ySize2
            yVector(lastIndex, j)&
            = sum(xSortedYvector(startOfSameX:endOfSameX, j))&
            / real(1 + endOfSameX - startOfSameX)
          end do
        else
          yVector(lastIndex, :) = xSortedYvector(startOfSameX, :)
        end if
        xVector(lastIndex) = xVector(startOfSameX)

        startOfSameX = i
        lastXvalue = xVector(i)
      end if
    end do
    if (xVector(ySize1) .gt. lastXvalue) then
      endOfSameX = ySize1 - 1
      lastIndex = lastIndex + 1
      if (endOfSameX .gt. startOfSameX) then
        do j = 1, ySize2
          yVector(lastIndex, j)&
          = sum(xSortedYvector(startOfSameX:endOfSameX, j))&
          / real(1 + endOfSameX - startOfSameX)
        end do
      else
        yVector(lastIndex, :) = xSortedYvector(startOfSameX, :)
      end if
      xVector(lastIndex) = xVector(startOfSameX)

      startOfSameX = ySize1
    end if
    endOfSameX = ySize1
    lastIndex = lastIndex + 1
    if (endOfSameX .gt. startOfSameX) then
      do j = 1, ySize2
        yVector(lastIndex, j)&
        = sum(xSortedYvector(startOfSameX:endOfSameX, j))&
        / real(1 + endOfSameX - startOfSameX)
      end do
    else
      yVector(lastIndex, :) = xSortedYvector(startOfSameX, :)
    end if
    xVector(lastIndex) = xVector(startOfSameX)

    if (present(status)) status = localStatus

  end subroutine prepareForInterpolate2d

  !---------------------------------------------------------------------
  subroutine interpolateSingle(xVector, yVector, x, y, status)
  !---------------------------------------------------------------------

    implicit none

    real,    intent(in)            :: xVector(:), yVector(:), x
    real,    intent(out)           :: y
    integer, intent(out), optional :: status

    integer :: localStatus, numPoints, i, j
    logical :: xPointsAreInOrder

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    numPoints = size(xVector)

    if (size(yVector) .ne. numPoints) then
      localStatus = -1

    else
      xPointsAreInOrder = .true.
      do i = 1, numPoints - 1
        if (xVector(i) .ge. xVector(i + 1)) then
          xPointsAreInOrder = .false.
      exit
        end if
      end do
      if (.not. xPointsAreInOrder) then
        localStatus = -2

      else
        if (x .lt. xVector(1)) then
          localStatus = localStatus + 1
          y = yVector(1)

        else if (x .gt. xVector(numPoints)) then
          localStatus = localStatus + 2
          y = yVector(numPoints)

        else if (numPoints .lt. 2) then
          localStatus = localStatus + 4
          y = yVector(1)

        else
          j = numPoints
          do i = 2, numPoints
            if (x .lt. xVector(i)) then
              j = i
          exit
            end if
          end do

          y = yVector(j - 1) + (x - xVector(j - 1))&
          * (yVector(j) - yVector(j - 1)) / (xVector(j) - xVector(j - 1))
        end if
      end if
    end if

    if (present(status)) status = localStatus

  end subroutine interpolateSingle

  !---------------------------------------------------------------------
  subroutine interpolateVector(oldXvector, oldYvector, newXvector&
  , newYvector, nulls, status)
  !---------------------------------------------------------------------

    implicit none

    real,    intent(in)            :: oldXvector(:),&
                                      oldYvector(:),&
                                      newXvector(:)
    real,    intent(out)           :: newYvector(:)
    logical, intent(out), optional :: nulls(:)
    integer, intent(out), optional :: status

    integer :: localStatus, numOldPoints, numNewPoints, i, j,&
      startNewPoint, finisNewPoint, loOldPoint, hiOldPoint
    logical :: xPointsAreInOrder, allNewXtooSmall, allNewXtooLarge, localNulls(size(newXvector))

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    numOldPoints = size(oldXvector)
    numNewPoints = size(newXvector)

    if (numOldPoints .lt. 2) then
      if (present(status)) status = -1
  return
    end if
    if (numNewPoints .lt. 1) then
      if (present(status)) status = -2
  return
    end if
    if (size(oldYvector) .ne. numOldPoints) then
      if (present(status)) status = -3
  return
    end if
    if (size(newYvector) .ne. numNewPoints) then
      if (present(status)) status = -4
  return
    end if
    if (present(nulls)) then
      if (size(nulls) .ne. numNewPoints) then
        if (present(status)) status = -5
  return
      end if
    end if

    xPointsAreInOrder = .true. ! default
    do i = 1, numOldPoints - 1
!print *, i, numOldPoints, oldXvector(i), oldXvector(i + 1)
      if (oldXvector(i) .ge. oldXvector(i + 1)) then
        xPointsAreInOrder = .false.
    exit
      end if
    end do
    if (.not. xPointsAreInOrder) then
      if (present(status)) status = -6
  return
    end if
    xPointsAreInOrder = .true. ! default
    do i = 1, numNewPoints - 1
      if (newXvector(i) .ge. newXvector(i + 1)) then
        xPointsAreInOrder = .false.
    exit
      end if
    end do
    if (.not. xPointsAreInOrder) then
      if (present(status)) status = -7
  return
    end if

!do i = 1, numOldPoints
!  print *, i, oldXvector(i), oldYvector(i)
!end do
!print *, ''
!print *, ''
!print *, ''

    newYvector = 0.0 ! default
    localNulls = .false. ! vector default

    allNewXtooSmall = .true. ! default
    startNewPoint = 1
    do i = 1, numNewPoints
      if (newXvector(i) .ge. oldXvector(1)) then
        startNewPoint = i
        allNewXtooSmall = .false.
    exit
      end if
    end do
    if (allNewXtooSmall) then
      if (present(nulls)) nulls = .true.
      if (present(status)) status = -8
  return
    end if

    if (startNewPoint .gt. 1) then
      localStatus = localStatus + 1
      localNulls(1:startNewPoint - 1) = .true.
    end if

    allNewXtooLarge = .true. ! default
    finisNewPoint = numNewPoints
    do i = numNewPoints, 1, -1
      if (newXvector(i) .lt. oldXvector(numOldPoints)) then
        finisNewPoint = i
        allNewXtooLarge = .false.
    exit
      end if
    end do
    if (allNewXtooLarge) then
      if (present(nulls)) nulls = .true.
      if (present(status)) status = -9
  return
    end if

    if (finisNewPoint .lt. numNewPoints) then
      localStatus = localStatus + 2
      localNulls(finisNewPoint + 1:numNewPoints) = .true.
    end if

    loOldPoint = 1
    do i = startNewPoint, finisNewPoint
      do j = loOldPoint, numOldPoints - 1
        if (oldXvector(j + 1) .gt. newXvector(i)) then
          loOldPoint = j
          hiOldPoint = j + 1
      exit
        end if
      end do

      newYvector(i) = oldYvector(loOldPoint)&
      + (newXvector(i) - oldXvector(loOldPoint))&
      * (oldYvector(hiOldPoint) - oldYvector(loOldPoint))&
      / (oldXvector(hiOldPoint) - oldXvector(loOldPoint))

!print *, i, loOldPoint, hiOldPoint,newYvector(i) 
    end do

    if (present(status)) status = localStatus
    if (present(nulls)) nulls = localNulls

  end subroutine interpolateVector

  !---------------------------------------------------------------------
  subroutine getIntervalsANDandOR(intervals, ANDedIntervals, ORRedIntervals, status)
  !---------------------------------------------------------------------

    implicit none

    type(intervalsType), intent(in)            :: intervals(:)
    type(intervalType), dimension(:), pointer :: ANDedIntervals, ORRedIntervals
    integer,            intent(out), optional :: status

    type(intervalType) :: tempAI(size(intervals)), tempOI(size(intervals))
    integer :: i, j, minIndex, maxIndex, maxNumIndices, numIndices,&
      numEvents, numANDedIntervals, numORRedIntervals, numSimultaneousIntervals
    type(sequenceType) :: sequence(2*size(intervals))
    integer, allocatable :: numOverlaps(:), indicesFound(:)
    logical :: indexNotSeenAtPreviousI
!  type :: boundType
!    integer :: type
!    real(kind(1d0)) :: value
!  end type boundType
!  type :: intervalType
!    type(boundType) :: lower, upper
!    integer :: index
!  end type intervalType
!  type :: sequenceType
!    type(boundType) :: bound
!    logical         :: isLower
!    integer         :: index
!  end type sequenceType
    !-------------------------------------------------------------------
    ! Start:

!    emptyANDlist = .false. ! default
    status = 0 ! default
    if (size(intervals) .le. 0) then
      allocate(ANDedIntervals(1:0)) ! associate pointer with a zero-size array
      allocate(ORRedIntervals(1:0)) ! associate pointer with a zero-size array
  return
    end if

    ! Find out how many different index values there are:
    minIndex = minval(intervals%index)
    maxIndex = maxval(intervals%index)
    maxNumIndices = maxIndex - minIndex + 1
    allocate(numOverlaps(minIndex:maxIndex), indicesFound(maxNumIndices))
    numIndices = 0
    do i = 1, size(intervals)
      indexNotSeenAtPreviousI = .true.
      do j = 1, numIndices
        if (intervals(i)%index .eq. indicesFound(j)) then
          indexNotSeenAtPreviousI = .false.
      exit
        end if
      end do
      numIndices = numIndices + 1
      indicesFound(numIndices) = intervals(i)%index
    end do

    ! Load sequence:
    do i = 1, size(intervals)
      sequence(2*i-1)%bound%type = intervals(i)%lower%type
      sequence(2*i-1)%bound%value = intervals(i)%lower%value
      sequence(2*i-1)%isLower = .true.
      sequence(2*i-1)%index = intervals(i)%index

      sequence(2*i)%bound%type = intervals(i)%upper%type
      sequence(2*i)%bound%value = intervals(i)%upper%value
      sequence(2*i)%isLower = .false.
      sequence(2*i)%index = intervals(i)%index
    end do

    ! Sort sequence elements in order of increasing time (all isLower = .true. UNDEFINEDs at the start, all isLower = .false. UNDEFINEDs at the end):
    call sortSequence(sequence)

    ! Get rid of overlaps for each index:
    numOverlaps = 0 ! default
    numEvents = 0
    do i = 1, size(sequence)
      if (sequence(i)%isLower) then
        if (numOverlaps(sequence(i)%index) .eq. 0) then
          numEvents = numEvents + 1
          sequence(numEvents) = sequence(i)
        end if
        numOverlaps(sequence(i)%index) = numOverlaps(sequence(i)%index) + 1

      else ! sequence(i)%isUpper
        if (numOverlaps(sequence(i)%index) .eq. 1) then
          numEvents = numEvents + 1
          sequence(numEvents) = sequence(i)
        end if
        numOverlaps(sequence(i)%index) = numOverlaps(sequence(i)%index) - 1
      end if
    end do
!**** put in some bug detection for a while?

    ! Now the meat of the routine - calculation of the anded and orred sets of intervals:
!    tempAI%index = 0
!    tempOI%index = 0
    numANDedIntervals = 0
    numORRedIntervals  = 0
    numSimultaneousIntervals = 0
    do i = 1, numEvents
      if (sequence(i)%isLower) then
        numSimultaneousIntervals = numSimultaneousIntervals + 1
        if (numSimultaneousIntervals .gt. numIndices) then
!          if (present(status)) status = -200 ! actually a bug - shouldn't
!          ! happen.
!  return
        end if

        if (numSimultaneousIntervals .eq. 1) then
          numORRedIntervals = numORRedIntervals + 1
          tempOI(numORRedIntervals)%lower = sequence(i)%bound
!          tempOI(numORRedIntervals)%index = sequence(i)%index

        else if (numSimultaneousIntervals .eq. numIndices) then
          numANDedIntervals = numANDedIntervals + 1
          tempAI(numANDedIntervals)%lower = sequence(i)%bound
!          tempAI(numANDedIntervals)%index = sequence(i)%index
        end if

      else
        if (numSimultaneousIntervals .eq. 1) then
!          numORRedIntervals = numORRedIntervals + 1
          tempOI(numORRedIntervals)%upper = sequence(i)%bound
!          tempOI(numORRedIntervals)%index = sequence(i)%index

        else if (numSimultaneousIntervals .eq. numIndices) then
!          numANDedIntervals = numANDedIntervals + 1
          tempAI(numANDedIntervals)%upper = sequence(i)%bound
!          tempAI(numANDedIntervals)%index = sequence(i)%index
        end if

        numSimultaneousIntervals = numSimultaneousIntervals - 1
        if (numSimultaneousIntervals .lt. 0) then
!          if (present(status)) status = -300 ! actually a bug - shouldn't
!          ! happen.
!  return
        end if
      end if
    end do

    ! Get rid of abutting intervals:
    j = 1
    do i = 1, numANDedIntervals - 1
      if (.not. ((tempAI(j)%upper%type .eq. UNDEFINED&
      .or. tempAI(i + 1)%upper%type .eq. UNDEFINED)&
      .or. (tempAI(j)%upper%type .eq. EXCLUSIVE&
      .and. tempAI(i + 1)%upper%type .eq. EXCLUSIVE))&
      .and. tempAI(j)%upper%value .eq. tempAI(i + 1)%lower%value) then
      ! Delete the pair of equal times:
        tempAI(j)%upper = tempAI(i + 1)%upper
      else
        j = j + 1
        tempAI(j)%lower = tempAI(i + 1)%lower
        tempAI(j)%upper = tempAI(i + 1)%upper
      end if
    end do

!    if (present(ANDedIntervals)) then
      allocate(ANDedIntervals(j))
      ANDedIntervals = tempAI(1:j)
!    end if

    j = 1
    do i = 1, numORRedIntervals - 1
      if (.not. ((tempOI(j)%upper%type .eq. UNDEFINED&
      .or. tempOI(i + 1)%upper%type .eq. UNDEFINED)&
      .or. (tempOI(j)%upper%type .eq. EXCLUSIVE&
      .and. tempOI(i + 1)%upper%type .eq. EXCLUSIVE))&
      .and. tempOI(j)%upper%value .eq. tempOI(i + 1)%lower%value) then
      ! Delete the pair of equal times:
        tempOI(j)%upper = tempOI(i + 1)%upper
      else
        j = j + 1
        tempOI(j)%lower = tempOI(i + 1)%lower
        tempOI(j)%upper = tempOI(i + 1)%upper
      end if
    end do

!    if (present(ORRedIntervals)) then
      allocate(ORRedIntervals(j))
      ORRedIntervals = tempOI(1:j)
!    end if

  end subroutine getIntervalsANDandOR

  !---------------------------------------------------------------------
  subroutine sortSequence(sequence)
  !---------------------------------------------------------------------

    implicit none

    type(sequenceType), intent(inout) :: sequence(:)

    integer :: numDefinedBounds, numLowerUndefinedBounds, i, j, indices(size(sequence))
    type(sequenceType) :: tempSequence(size(sequence))
    real(kind(1d0)) :: events(size(sequence))

!  type :: boundType
!    integer :: type
!    real(kind(1d0)) :: value
!  end type boundType
!  type :: intervalType
!    type(boundType) :: lower, upper
!    integer :: index
!  end type intervalType
!  type :: sequenceType
!    type(boundType) :: bound
!    logical         :: isLower
!    integer         :: index
!  end type sequenceType
    !-------------------------------------------------------------------
    ! Start:

    numDefinedBounds = 0
    numLowerUndefinedBounds = 0
    do i = 1, size(sequence)
      if (sequence(i)%bound%type .ne. UNDEFINED) then
        numDefinedBounds = numDefinedBounds + 1
        indices(numDefinedBounds) = i
        events(numDefinedBounds) = sequence(i)%bound%value
      else if (sequence(i)%isLower) then
        numLowerUndefinedBounds = numLowerUndefinedBounds + 1
        tempSequence(numLowerUndefinedBounds) = sequence(i)
      end if
    end do

    if (numDefinedBounds .gt. 0) then
      call quick_sort(events(1:numDefinedBounds), indices(1:numDefinedBounds))
    end if

    do i = 1, numDefinedBounds
      tempSequence(numLowerUndefinedBounds + i) = sequence(indices(i))
    end do

    j = numLowerUndefinedBounds + numDefinedBounds
    do i = 1, size(sequence)
      if (sequence(i)%bound%type .eq. UNDEFINED&
      .and. .not.sequence(i)%isLower) then
        j = j + 1
        tempSequence(j) = sequence(i)
      end if
    end do

    sequence = tempSequence

  end subroutine sortSequence

!  !---------------------------------------------------------------------
!  subroutine makeSourceList(coneRadius, minFlux, sourceData)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    real, intent(in) :: coneRadius, minFlux
!    type(sourceType), intent(out) :: sourceData(:)
!
!    integer :: numSources, src
!    real :: radii(size(sourceData)), theta(size(sourceData))
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    numSources = size(sourceData)
!
!    call getSineRandoms(coneRadius, radii)
!    call getRandomNumbers(theta)
!    theta = twoPi * theta
!    do src = 1, numSources
!      sourceData(src)%x = radii(src) * cos(theta(src))
!      sourceData(src)%y = radii(src) * sin(theta(src))
!    end do
!    call getMinus5on2Randoms(sourceData%flux)
!    sourceData%flux = sourceData%flux * minFlux
!    call getRandomNumbers(sourceData%temperature)
!
!  end subroutine makeSourceList

  !---------------------------------------------------------------------
  subroutine cyclicConvolution(vector, convolver, mask, padPlease&
  , normalise, interpolationLimit, interpolationMask, uncerts, status)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in)               :: convolver(:)
    real, intent(inout)            :: vector(:)
    real, intent(inout),  optional :: uncerts(:)
    logical, intent(in),  optional :: mask(:), padPlease, normalise
    integer, intent(in),  optional :: interpolationLimit
    integer, intent(out), optional :: status
    logical, intent(out), optional :: interpolationMask(:)

    integer :: vsize, csize, n, i, j, localStatus, psize, cHalfWidth,&
      interpolateHolesSmallerThan, holeStart, holeFinis
    real, allocatable :: paddedVector(:), output(:)
    real :: summ, weight, norm
    logical :: localPadPlease, localNormalise, inNullPatch
    logical, allocatable :: localMask(:), localInterpolationMask(:)

    !-------------------------------------------------------------------
    ! Start

!print *, 'aaa'

    localStatus = 0 ! default.

    vsize = size(vector)
    csize = size(convolver)
    cHalfWidth = csize / 2

    if (present(padPlease)) then
      localPadPlease = padPlease
    else
      localPadPlease = .true.
    end if
    if (present(normalise)) then
      localNormalise = normalise
    else
      localNormalise = .true.
    end if
    if (present(interpolationLimit)) then
!print *, 'present(interpolationLimit)'
      interpolateHolesSmallerThan = interpolationLimit
    else
      interpolateHolesSmallerThan = 0
    end if

    if (interpolateHolesSmallerThan .gt. csize) then
      if (present(status)) status = -1
  return
    end if

    if (localPadPlease) then
      psize = vsize + csize ! Pad doesn't need to be quite this big, but
      ! it won't hurt.
    else
      psize = vsize
    end if

    if (localNormalise) then
      norm = 1.0
    else
      norm = sum(convolver)
    end if

    allocate(paddedVector(psize), output(psize), localMask(psize)&
    , localInterpolationMask(psize))

    localMask = .false.
    if (present(mask)) then
!print *, 'present(mask)'
      if (size(mask) .ne. vsize) then
        if (present(status)) status = -3
        deallocate(paddedVector, output, localMask, localInterpolationMask)
  return
      end if
      localMask(1:vsize) = mask

    else
      localMask(1:vsize) = .true.
    end if

    ! Check all the holes to see if any are small enough to be interpolated:
    localInterpolationMask = localMask ! vector default
    if (.not. localMask(1)) then
      inNullPatch = .true.
      holeStart = 1
    else
      inNullPatch = .false.
    end if
    do i = 2, vsize
      if (inNullPatch .and. localMask(i)) then
        holeFinis = i - 1
        if (holeFinis - holeStart + 1 .lt. interpolateHolesSmallerThan) then
          localInterpolationMask(holeStart:holeFinis) = .true.
        end if
        inNullPatch = .false.

      else if (.not. inNullPatch .and. .not. localMask(i)) then
        holeStart = i
        inNullPatch = .true.
      end if
    end do
    if (inNullPatch) then
      holeFinis = vsize
      if (holeFinis - holeStart + 1 .lt. interpolateHolesSmallerThan) then
        localInterpolationMask(holeStart:holeFinis) = .true.
      end if
    end if

    paddedVector(1:vsize) = vector
    output = 0.0 ! vector default.
    do n = 1, vsize
    if (.not. localInterpolationMask(n)) cycle
      summ = 0.0
      weight = 0.0
      do i = 1, csize
        j = n - cHalfWidth + i - 1
        if (j .le. 0)     j = j + psize
        if (j .gt. psize) j = j - psize
      if (.not. localMask(j)) cycle
        summ = summ + convolver(i) * paddedVector(j)
        weight = weight + convolver(i)
      end do
      if (weight .le. 0.0)  then
        if (present(status)) status = -4
        deallocate(paddedVector, output, localMask, localInterpolationMask)
  return
      end if
      output(n) = summ * norm / weight
    end do

    vector = output(1:vsize)

    if (present(uncerts)) then
!print *, 'present(uncerts)'
      paddedVector(1:vsize) = uncerts * uncerts
      output = 0.0 ! vector default.
      do n = 1, vsize
      if (.not. localInterpolationMask(n)) cycle
        summ = 0.0
        weight = 0.0
        do i = 1, csize
          j = n - cHalfWidth + i - 1
          if (j .le. 0)     j = j + psize
          if (j .gt. psize) j = j - psize
!if(n .eq. 50) print *, i, j
        if (.not. localMask(j)) cycle
          summ = summ + convolver(i) * convolver(i) * paddedVector(j)
          weight = weight + convolver(i)
!if(n .eq. 50) print *, summ, weight
        end do
        if (weight .le. 0.0)  then
          output(n) = 0.0
        else
          output(n) = sqrt(summ) * norm / weight
        end if
!if(n .eq. 50) print *, output(n)
      end do

      uncerts = output(1:vsize)
    end if

    if (present(interpolationMask))&
      interpolationMask = localInterpolationMask(1:vsize)

    deallocate(paddedVector, output, localMask, localInterpolationMask)

    if (present(status)) status = localStatus

  end subroutine cyclicConvolution

  !---------------------------------------------------------------------
  subroutine preciseSum(inVector, summ, status)
  !---------------------------------------------------------------------

    use sort_mod

    implicit none

    real,    intent(in)            :: inVector(:)
    real,    intent(out)           :: summ
    integer, intent(out), optional :: status

    integer :: localStatus, numValues, pair, numPairs, endOfList
!    real(kind(1d0)), allocatable :: tempVector(:)
    real, allocatable :: tempVector(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0

    summ = 0.0

    numValues = size(inVector)

    if (numValues .le. 0) then
      if (present(status)) status = 1
  return
    end if

    allocate(tempVector(numValues))
    tempVector = inVector

    endOfList = numValues
    do while(endOfList .gt. 1)
      call quick_sort(tempVector(1:endOfList))
      numPairs = endOfList / 2
      do pair = 1, numPairs
        tempVector(pair) = tempVector(2 * pair - 1) + tempVector(2 * pair)
      end do
      if (endOfList .eq. 2 * numPairs + 1) then
        tempVector(numPairs + 1) = tempVector(endOfList)
        endOfList = numPairs + 1
      else
        endOfList = numPairs
      end if
    end do
    summ = tempVector(1)

    deallocate(tempVector)

    if (present(status)) status = localStatus

  end subroutine preciseSum

  !---------------------------------------------------------------------
  subroutine findEdges(mask, figureEdges, groundEdges, status)
  !---------------------------------------------------------------------
  ! Given a logical array mask(), this subroutine returns an array
  ! figureEdges which
  ! is true at all pixels in mask() which are _not_ surrounded by nearest
  ! neighbours which are all true. The 'nearest neighbours' of a pixel at
  ! (x,y) are defined as those positions (x',y') for which x-1<=x'<=x+1
  ! and y-1<=y'<=y+1. However, values of (x',y') that fall outside the
  ! boundaries of mask() are not counted as nearest neighbours: hence,
  ! no pixel on the boundary of mask() is set as an edge pixel just because
  ! it is on the boundary.
  !
  ! A second array groundEdges is also returned, which is equivalent
  ! to figureEdges calculated for .not.mask.
  !
  ! Example input mask:
  !
  !	11110000111
  !	11111000011
  !	11111100011
  !	11111100001
  !	00000000000
  !	11111110000
  !	11101111100
  !
  ! The returned array figureEdges should look as follows:
  !	
  !	00010000110
  !	00011000010
  !	00001100011
  !	11111100001
  !	00000000000
  !	11111110000
  !	00101011100
  !
  ! The returned array groundEdges should look as follows:
  !	
  !	00001101000
  !	00000111100
  !	00000010100
  !	00000010110
  !	11111111011
  !	00000001110
  !	00010000010

    implicit none

    logical, intent(in)  :: mask(:,:)
    logical, intent(out), optional :: figureEdges(:,:), groundEdges(:,:)
    integer, intent(out), optional :: status

    logical ::&
      edgeMaskBig(0:size(mask,1)+1,0:size(mask,2)+1),&
      edgeMaskSmall(size(mask,1), size(mask,2))
    integer :: xsize, ysize

    !-------------------------------------------------------------------
    ! Start

    xsize = size(mask, 1)
    ysize = size(mask, 2)

    if (present(figureEdges)) then
      if (size(figureEdges, 1) .ne. xsize&
      .or. size(figureEdges, 2) .ne. ysize) then
        if (present(status)) status = -1
  return
      end if
    end if

    if (present(groundEdges)) then
      if (size(groundEdges, 1) .ne. xsize&
      .or. size(groundEdges, 2) .ne. ysize) then
        if (present(status)) status = -2
  return
      end if
    end if

    if (.not. present(figureEdges) .and. .not. present(groundEdges)) then
      if (present(status)) status = 1
  return
    end if

    edgeMaskSmall = .false. ! default

    edgeMaskBig = .false. ! default
    edgeMaskBig(0:xsize-1, 0:ysize-1) = mask
    edgeMaskBig(0:xsize-1, 1:ysize)&
      = edgeMaskBig(0:xsize-1, 1:ysize)   .or. mask
    edgeMaskBig(0:xsize-1, 2:ysize+1)&
      = edgeMaskBig(0:xsize-1, 2:ysize+1) .or. mask
    edgeMaskBig(1:xsize,   0:ysize-1)&
      = edgeMaskBig(1:xsize  , 0:ysize-1) .or. mask
    edgeMaskBig(1:xsize,   1:ysize)&
      = edgeMaskBig(1:xsize  , 1:ysize)   .or. mask
    edgeMaskBig(1:xsize,   2:ysize+1)&
      = edgeMaskBig(1:xsize  , 2:ysize+1) .or. mask
    edgeMaskBig(2:xsize+1, 0:ysize-1)&
      = edgeMaskBig(2:xsize+1, 0:ysize-1) .or. mask
    edgeMaskBig(2:xsize+1, 1:ysize)&
      = edgeMaskBig(2:xsize+1, 1:ysize)   .or. mask
    edgeMaskBig(2:xsize+1, 2:ysize+1)&
      = edgeMaskBig(2:xsize+1, 2:ysize+1) .or. mask

    edgeMaskSmall = edgeMaskBig(1:xsize, 1:ysize)

    edgeMaskBig = .false. ! default
    edgeMaskBig(0:xsize-1, 0:ysize-1) = .not. mask
    edgeMaskBig(0:xsize-1, 1:ysize)&
      = edgeMaskBig(0:xsize-1, 1:ysize)   .or. .not. mask
    edgeMaskBig(0:xsize-1, 2:ysize+1)&
      = edgeMaskBig(0:xsize-1, 2:ysize+1) .or. .not. mask
    edgeMaskBig(1:xsize,   0:ysize-1)&
      = edgeMaskBig(1:xsize  , 0:ysize-1) .or. .not. mask
    edgeMaskBig(1:xsize,   1:ysize)&
      = edgeMaskBig(1:xsize  , 1:ysize)   .or. .not. mask
    edgeMaskBig(1:xsize,   2:ysize+1)&
      = edgeMaskBig(1:xsize  , 2:ysize+1) .or. .not. mask
    edgeMaskBig(2:xsize+1, 0:ysize-1)&
      = edgeMaskBig(2:xsize+1, 0:ysize-1) .or. .not. mask
    edgeMaskBig(2:xsize+1, 1:ysize)&
      = edgeMaskBig(2:xsize+1, 1:ysize)   .or. .not. mask
    edgeMaskBig(2:xsize+1, 2:ysize+1)&
      = edgeMaskBig(2:xsize+1, 2:ysize+1) .or. .not. mask

    edgeMaskSmall = edgeMaskSmall .and. edgeMaskBig(1:xsize, 1:ysize)

    if (present(figureEdges)) then
      figureEdges = .false.
      figureEdges = edgeMaskSmall .and. mask
    end if
    if (present(groundEdges)) then
      groundEdges = .false.
      groundEdges = edgeMaskSmall .and. .not. mask
    end if

    if (present(status)) status = 0

  end subroutine findEdges

  !---------------------------------------------------------------------
  subroutine findEdges_old(mask, edgeMask, status)
  !---------------------------------------------------------------------
  ! Given a logical array mask(), this subroutine returns an array which
  ! is true at all pixels in mask() which are _not_ surrounded by nearest
  ! neighbours of the same value. The 'nearest neighbours' of a pixel at
  ! (x,y) are defined as those positions (x',y') for which x-1<=x'<=x+1
  ! and y-1<=y'<=y+1. However, values of (x',y') that fall outside the
  ! boundaries of mask() are not counted as nearest neighbours: hence,
  ! no pixel on the boundary of mask() is set as an edge pixel just because
  ! it is on the boundary.
  !
  ! Example input mask:
  !
  !	111100001
  !	111110000
  !	111111000
  !	111111000
  !	000000000
  !	111111100
  !	111011111
  !
  ! The returned array for this should look as follows:
  !	
  !	000111011
  !	000111111
  !	000011100
  !	111111100
  !	111111110
  !	111111110
  !	001110111

    implicit none

    logical, intent(in)  :: mask(:,:)
    logical, intent(out) :: edgeMask(:,:)
    integer, intent(out), optional :: status

    logical :: edgeMaskMaster(0:size(mask,1)+1,0:size(mask,2)+1)
    integer :: xsize, ysize

    !-------------------------------------------------------------------
    ! Start

    xsize = size(mask, 1)
    ysize = size(mask, 2)

    edgeMask = .false. ! default

    if (size(edgeMask, 1) .ne. xsize .or. size(edgeMask, 2) .ne. ysize) then
      if (present(status)) status = -1
  return
    end if

    edgeMaskMaster = .false. ! default
    edgeMaskMaster(0:xsize-1, 0:ysize-1)&
      = edgeMaskMaster(0:xsize-1, 0:ysize-1) .and. mask
    edgeMaskMaster(0:xsize-1, 1:ysize)&
      = edgeMaskMaster(0:xsize-1, 1:ysize)   .and. mask
    edgeMaskMaster(0:xsize-1, 2:ysize+1)&
      = edgeMaskMaster(0:xsize-1, 2:ysize+1) .and. mask
    edgeMaskMaster(1:xsize,   0:ysize-1)&
      = edgeMaskMaster(1:xsize  , 0:ysize-1) .and. mask
    edgeMaskMaster(1:xsize,   1:ysize)&
      = edgeMaskMaster(1:xsize  , 1:ysize)   .and. mask
    edgeMaskMaster(1:xsize,   2:ysize+1)&
      = edgeMaskMaster(1:xsize  , 2:ysize+1) .and. mask
    edgeMaskMaster(2:xsize+1, 0:ysize-1)&
      = edgeMaskMaster(2:xsize+1, 0:ysize-1) .and. mask
    edgeMaskMaster(2:xsize+1, 1:ysize)&
      = edgeMaskMaster(2:xsize+1, 1:ysize)   .and. mask
    edgeMaskMaster(2:xsize+1, 2:ysize+1)&
      = edgeMaskMaster(2:xsize+1, 2:ysize+1) .and. mask

    edgeMask = edgeMaskMaster(1:xsize, 1:ysize)

    edgeMaskMaster = .false. ! default
    edgeMaskMaster(0:xsize-1, 0:ysize-1)&
      = edgeMaskMaster(0:xsize-1, 0:ysize-1) .and. .not. mask
    edgeMaskMaster(0:xsize-1, 1:ysize)&
      = edgeMaskMaster(0:xsize-1, 1:ysize)   .and. .not. mask
    edgeMaskMaster(0:xsize-1, 2:ysize+1)&
      = edgeMaskMaster(0:xsize-1, 2:ysize+1) .and. .not. mask
    edgeMaskMaster(1:xsize,   0:ysize-1)&
      = edgeMaskMaster(1:xsize  , 0:ysize-1) .and. .not. mask
    edgeMaskMaster(1:xsize,   1:ysize)&
      = edgeMaskMaster(1:xsize  , 1:ysize)   .and. .not. mask
    edgeMaskMaster(1:xsize,   2:ysize+1)&
      = edgeMaskMaster(1:xsize  , 2:ysize+1) .and. .not. mask
    edgeMaskMaster(2:xsize+1, 0:ysize-1)&
      = edgeMaskMaster(2:xsize+1, 0:ysize-1) .and. .not. mask
    edgeMaskMaster(2:xsize+1, 1:ysize)&
      = edgeMaskMaster(2:xsize+1, 1:ysize)   .and. .not. mask
    edgeMaskMaster(2:xsize+1, 2:ysize+1)&
      = edgeMaskMaster(2:xsize+1, 2:ysize+1) .and. .not. mask

    edgeMask = .not. (edgeMask .or. edgeMaskMaster(1:xsize, 1:ysize))

    if (present(status)) status = 0

  end subroutine findEdges_old

  !---------------------------------------------------------------------
  subroutine robustMean_2d(array, varianceArray, largestDevianceRatio&
  , maskedMean, mask, status)
  !---------------------------------------------------------------------

    implicit none

    real,    intent(in)              :: array(:,:), varianceArray(:,:),&
      largestDevianceRatio
    real,    intent(out)             :: maskedMean
    logical, intent(inout), optional :: mask(:,:)
    integer, intent(out),   optional :: status

    integer :: localStatus, vectorShape(1), arrayShape(2)
    logical :: localMask(size(array, 1) * size(array, 2))
    real :: vector(size(array, 1) * size(array, 2)),&
      varianceVector(size(array, 1) * size(array, 2))

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    maskedMean = 0.0

    vectorShape(1) = size(array, 1) * size(array, 2)
    arrayShape(1) = size(array, 1)
    arrayShape(2) = size(array, 2)
    vector = reshape(array, vectorShape)

    if (size(varianceArray, 1) .ne. size(array, 1)&
    .or. size(varianceArray, 2) .ne. size(array, 2)) then
      if (present(status)) status = -2
  return
    end if
    if (present(mask)) then
      if (size(mask, 1) .ne. size(array, 1)&
      .or. size(mask, 2) .ne. size(array, 2)) then
        mask = .false.
        if (present(status)) status = -3
  return
      end if
    end if

    varianceVector = reshape(varianceArray, vectorShape)
    if (present(mask)) then
      localMask = reshape(mask, vectorShape)
      call robustMean_1d(vector, varianceVector, largestDevianceRatio&
      , maskedMean, mask=localMask, status=localStatus)
      mask = reshape(localMask, arrayShape)
    else
      call robustMean_1d(vector, varianceVector, largestDevianceRatio&
      , maskedMean, status=localStatus)
    end if

    if(present(status)) status = localStatus

  end subroutine robustMean_2d

  !---------------------------------------------------------------------
  subroutine robustMean_1d(vector, implicitVariance, largestDevianceRatio&
  , weightedMaskedMean, mask, outVar, status)
  !---------------------------------------------------------------------
  ! Usually when one has a sequence of values which is known to include some
  ! 'outliers', the quickest way to find a sort of average which doesn't
  ! include the outliers is to take the median of the sequence. However if
  ! most of the sequence values are either 0 or 1, there is not much use in
  ! taking a median since it too will be either 0 or 1. It is the purpose
  ! of this routine is to return a mean value of a masked subset of such a
  ! sequence, the mask being chosen such that no masked value departs from
  ! the mean of the masked values by more than largestDevianceRatio times the
  ! standard deviation of the masked values. In other words, the explicit
  ! variance of any data point i, calculated as (y_i - mean_y)^2, should not
  ! exceed the implicit variance, which is the square of the uncertainty in
  ! y_i, by more than largestVarianceRatio = largestDevianceRatio^2.

    implicit none

    real,    intent(in)              :: vector(:), implicitVariance(:),&
      largestDevianceRatio
    real,    intent(out)             :: weightedMaskedMean
    real,    intent(out),   optional :: outVar!!!, reducedChiSquared
    logical, intent(inout), optional :: mask(size(vector))
    integer, intent(out),   optional :: status

    integer :: vsize, localStatus
    logical :: localMask(size(vector))
    real :: numMasked, largestVarianceRatio, deviance(size(vector)),&
      explicitVariance(size(vector)), maskedMean, varInMaskedMean, varInWeightedMaskedMean

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    maskedMean = 0.0
    if (present(outVar)) outVar = 0.0
!    if (present(reducedChiSquared)) reducedChiSquared = 0.0

    vsize = size(vector)
    if (vsize .le. 0) then
      if (present(status)) status = -1
  return
    end if
    if (size(implicitVariance) .ne. vsize) then
      if (present(status)) status = -2
  return
    end if
    if (present(mask)) then
      if (size(mask) .ne. vsize) then
        mask = .false.
        if (present(status)) status = -3
  return
      end if
    end if
    if (largestDevianceRatio .le. 1.0) then
      if (present(status)) status = -4
  return
    end if

    localMask = .true.

    largestVarianceRatio = largestDevianceRatio * largestDevianceRatio
    do
      numMasked = real(count(localMask))
      if (numMasked .le. 0) then
        if (present(status)) status = -2
  return
      end if

      maskedMean = sum(vector, localMask) / numMasked
      weightedMaskedMean = sum(vector / implicitVariance, localMask)&
        / sum(1.0 / implicitVariance, localMask)

!      deviance = vector - maskedMean
deviance = vector - weightedMaskedMean
      explicitVariance = deviance * deviance

      if (.not. any(explicitVariance / implicitVariance&
      .gt. largestVarianceRatio .and. localMask)) exit ! because no new
      ! pixels have been masked in this iteration: hence algorithm has
      ! converged.

      where(explicitVariance / implicitVariance .gt. largestVarianceRatio)&
        localMask = .false.
    end do

    numMasked = real(count(localMask))
    varInMaskedMean = sum(implicitVariance, localMask) / numMasked / numMasked
    varInWeightedMaskedMean = 1.0 / sum(1.0 / implicitVariance, localMask)

    if (present(mask)) mask = localMask
    if (present(outVar)) outVar = varInWeightedMaskedMean
    if(present(status)) status = localStatus

  end subroutine robustMean_1d

  !---------------------------------------------------------------------
  subroutine robustMean_1d_old(vector, largestDevianceRatio&
  , maskedMean, mask, status)
  !---------------------------------------------------------------------
  ! Usually when one has a sequence of values which is known to include some
  ! 'outliers', the quickest way to find a sort of average which doesn't
  ! include the outliers is to take the median of the sequence. However if
  ! most of the sequence values are either 0 or 1, there is not much use in
  ! taking a median since it too will be either 0 or 1. It is the purpose
  ! of this routine is to return a mean value of a masked subset of such a
  ! sequence, the mask being chosen such that no masked value departs from
  ! the mean of the masked values by more than largestDeviation times the
  ! standard deviation of the masked values.

    implicit none

    real,    intent(in)              :: vector(:), largestDevianceRatio
    real,    intent(out)             :: maskedMean
    logical, intent(inout), optional :: mask(:)
    integer, intent(out),   optional :: status

    integer :: vsize, localStatus
    logical :: localMask(size(vector))
    real :: variance, numUnMasked, largestVarianceRatio, deviance(size(vector))

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    maskedMean = 0.0
    vsize = size(vector)
    if (present(mask)) then
      if (size(mask) .ne. vsize) then
        mask = .false.
        if (present(status)) status = -1
  return
      end if
    end if

    localMask = .true.

    if (vsize .gt. 0) then
      largestVarianceRatio = largestDevianceRatio * largestDevianceRatio
      do
        numUnMasked = real(count(localMask))
        if (numUnMasked .le. 0) then
          if (present(status)) status = -2
  return
        end if

        maskedMean = sum(vector, localMask) / numUnMasked
        variance = (sum(vector*vector, localMask) - maskedMean * maskedMean)&
        / numUnMasked

        deviance = vector - maskedMean

      if (.not. any(deviance * deviance / variance .gt. largestVarianceRatio&
      .and. localMask)) exit ! because no new pixels have been masked in
      ! this iteration: hence algorithm has converged.

        where(deviance * deviance / variance .gt. largestVarianceRatio)&
          localMask = .false.
      end do
    end if

    if (present(mask)) mask = localMask
    if(present(status)) status = localStatus

  end subroutine robustMean_1d_old

  !---------------------------------------------------------------------
  subroutine temp_robustMean(y, yUncert, largestDevianceRatio&
  , meanY, mask, meanYuncert, reducedChiSquared, status)
  !---------------------------------------------------------------------
  ! Usually when one has a sequence of values which is known to include some
  ! 'outliers', the quickest way to find a sort of average which doesn't
  ! include the outliers is to take the median of the sequence. However if
  ! most of the sequence values are either 0 or 1, there is not much use in
  ! taking a median since it too will be either 0 or 1. It is the purpose
  ! of this routine is to return a mean value of a masked subset of such a
  ! sequence, the mask being chosen such that no masked value departs from
  ! the mean of the masked values by more than largestDevianceRatio times the
  ! standard deviation of the masked values. In other words, the explicit
  ! variance of any data point i, calculated as (y_i - mean_y)^2, should not
  ! exceed the implicit variance, which is the square of the uncertainty in
  ! y_i, by more than largestVarianceRatio = largestDevianceRatio^2.
  !
  ! Note that the y values in the calculation of the mean are weighted
  ! by 1 / yUncert^2, which weighting is known to give the lowest uncertainty
  ! in the result.

    implicit none

    real,    intent(in)            :: y(:), yUncert(size(y)),&
                                      largestDevianceRatio
    real,    intent(out)           :: meanY
    real,    intent(out), optional :: meanYuncert, reducedChiSquared
    logical, intent(out), optional :: mask(size(y))
    integer, intent(out), optional :: status

    integer :: vsize, localStatus, degFree
    logical :: localMask(size(y))
    real :: largestVarianceRatio, deviance(size(y)),&
      variance(size(y)), scatterVariance(size(y)), meanYvariance

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    meanY = 0.0
    if (present(meanYuncert)) meanYuncert = 0.0
    if (present(reducedChiSquared)) reducedChiSquared = 0.0

    vsize = size(y)
    if (vsize <= 0) then
      if (present(status)) status = -1
  return
    end if
    if (largestDevianceRatio <= 1.0) then
      if (present(status)) status = -2
  return
    end if
    if (any(yUncert <= 0)) then
      if (present(status)) status = -3
  return
    end if

    localMask = .true.
    variance = yUncert * yUncert

    largestVarianceRatio = largestDevianceRatio * largestDevianceRatio
    do
      if (count(localMask) <= 0) then
        if (present(status)) status = -4
  return
      end if

      meanY = sum(y / variance, localMask) / sum(1.0 / variance, localMask)
      deviance = y - meanY
      scatterVariance = deviance * deviance

      if (.not. any(scatterVariance / variance  > largestVarianceRatio&
      .and. localMask))&
    exit ! because no new pixels have been masked in this iteration:
         ! hence the algorithm has converged.

      where(scatterVariance / variance > largestVarianceRatio)&
        localMask = .false.
    end do

    meanYvariance = 1.0 / sum(1.0 / variance, localMask)

    if (present(mask)) mask = localMask
    if (present(meanYuncert)) meanYuncert = sqrt(meanYvariance)

    if (present(reducedChiSquared)) then
      degFree = count(localMask) - 1
      if (degFree <= 0) then
        if (present(status)) status = -5
  return
      end if

      reducedChiSquared = sum(scatterVariance / variance, localMask)&
        / real(degFree)
    end if

    if(present(status)) status = localStatus

  end subroutine temp_robustMean

  !---------------------------------------------------------------------
  function mergeStatuses(statA, statB) result(mergedStat)
  !---------------------------------------------------------------------
  ! This function is designed to be used where a subroutine A which returns
  ! a status value to the user call internally a second subroutine B which
  ! also returns a status value (designated here by the variable
  ! 'foreignStatus'). In this case it is desirable to merge these values.
  ! The present function is designed to accomplish this in a sensible
  ! manner. It should be called straight after each call to subroutine B
  ! and before the merged status is tested for a fatal value.
  !
  ! The conception of status values in the ssclib is as follows: the default
  ! value is zero; negative values represent fatal errors and should result
  ! in immediate return of control to the calling program; positive values
  ! are non-fatal errors or warnings and should not result in premature
  ! termination of the called subprogram. Because of this distinction, the
  ! status value needs to be able to record more than one warning state.
  ! To accomplish this, warning states are recorded as bits set in the
  ! status integer. (Careful note must be taken by the designer of further
  ! routines in ssclib that the available number of bits in the data type
  ! of 'status' is not exceeded.)
  !
  ! The merging of values proceeds broadly as follows. The status of
  ! subroutine A ought not to be negative at the time that mergeStatuses()
  ! is called, but if it is, no action is taken, the merged status value
  ! is given that same negative value and presumably subroutine A will
  ! exit soon after the (pointless) call to mergeStatuses. Likewise if
  ! foreignStatus equals 0, the merged status value again simply takes on
  ! the value of status.
  !
  ! Matters now become more interesting. If status >= 0 but foreignStatus
  ! < 0, the returned value must be a negative number, but not one of those
  ! which subroutine A has already assigned to its own error conditions.
  ! For this purpose, subroutine A supplies to mergeStatuses() the maximum
  ! negative status value defined in A. Function mergeStatuses() adds this
  ! (note that maxNeg must be a non-positive number) to foreignStatus and
  ! returns the result. The second non-trivial case is where status >= 0
  ! and foreignStatus > 0. Again it is necessary to ensure that bits
  ! reserved for use by subroutine A are not overwritten. The method is
  ! similar to the first case: subroutine A supplies in the variable
  ! 'maxPos' the largest bit it uses itself for positive status values
  ! (counting the first bit as number 1), and foreignStatus is left-shifted
  ! by this amount before being added to status. NOTE! In both cases it is
  ! up to the subroutine designer to (i) ensure that maxPos and maxNeg are
  ! correct and (ii) correctly document all the meanings of the returned
  ! status values.
  !
  ! Example: suppose that subroutine A defined negative values up to -5
  ! and positive values up to 8. A call to mergeStatuses() after a call
  ! to subroutine B should be of the form
  !
  !	status = mergeStatuses(status, foreignStatus, 4, -5)
  !
  ! If status and foreignStatus had, say, values 3 (bits 1 and 2 set) and 4
  ! (bit 3 set) before the call, status should have the value 67 after the
  ! call (bits 1, 2 and 4+3=7 set).

    implicit none

    type(statusType), intent(in) :: statA, statB
    type(statusType) :: mergedStat

    !-------------------------------------------------------------------
    ! Start

    mergedStat%status = statA%status ! default

    mergedStat%maxNeg = statA%maxNeg + statB%maxNeg
    mergedStat%numPosBits = statA%numPosBits + statB%numPosBits

    if (statA%status .ge. 0) then
      if (statB%status .lt. 0) then
        mergedStat%status = statB%status + statA%maxNeg
      else if (statB%status .gt. 0) then
!        mergedStat%status = iand(statA%status&
!        , ishft(statB%status, statA%numPosBits))
        mergedStat%status = statA%status + ishft(statB%status, statA%numPosBits)
      end if
    end if

  end function mergeStatuses

  !---------------------------------------------------------------------
  integer function mergeStatuses_old(status, foreignStatus, maxPos, maxNeg)
  !---------------------------------------------------------------------
  ! This function is designed to be used where a subroutine A which returns
  ! a status value to the user call internally a second subroutine B which
  ! also returns a status value (designated here by the variable
  ! 'foreignStatus'). In this case it is desirable to merge these values.
  ! The present function is designed to accomplish this in a sensible
  ! manner. It should be called straight after each call to subroutine B
  ! and before the merged status is tested for a fatal value.
  !
  ! The conception of status values in the ssclib is as follows: the default
  ! value is zero; negative values represent fatal errors and should result
  ! in immediate return of control to the calling program; positive values
  ! are non-fatal errors or warnings and should not result in premature
  ! termination of the called subprogram. Because of this distinction, the
  ! status value needs to be able to record more than one warning state.
  ! To accomplish this, warning states are recorded as bits set in the
  ! status integer. (Careful note must be taken by the designer of further
  ! routines in ssclib that the available number of bits in the data type
  ! of 'status' is not exceeded.)
  !
  ! The merging of values proceeds broadly as follows. The status of
  ! subroutine A ought not to be negative at the time that mergeStatuses()
  ! is called, but if it is, no action is taken, the merged status value
  ! is given that same negative value and presumably subroutine A will
  ! exit soon after the (pointless) call to mergeStatuses. Likewise if
  ! foreignStatus equals 0, the merged status value again simply takes on
  ! the value of status.
  !
  ! Matters now become more interesting. If status >= 0 but foreignStatus
  ! < 0, the returned value must be a negative number, but not one of those
  ! which subroutine A has already assigned to its own error conditions.
  ! For this purpose, subroutine A supplies to mergeStatuses() the maximum
  ! negative status value defined in A. Function mergeStatuses() adds this
  ! (note that maxNeg must be a non-positive number) to foreignStatus and
  ! returns the result. The second non-trivial case is where status >= 0
  ! and foreignStatus > 0. Again it is necessary to ensure that bits
  ! reserved for use by subroutine A are not overwritten. The method is
  ! similar to the first case: subroutine A supplies in the variable
  ! 'maxPos' the largest bit it uses itself for positive status values
  ! (counting the first bit as number 1), and foreignStatus is left-shifted
  ! by this amount before being added to status. NOTE! In both cases it is
  ! up to the subroutine designer to (i) ensure that maxPos and maxNeg are
  ! correct and (ii) correctly document all the meanings of the returned
  ! status values.
  !
  ! Example: suppose that subroutine A defined negative values up to -5
  ! and positive values up to 8. A call to mergeStatuses() after a call
  ! to subroutine B should be of the form
  !
  !	status = mergeStatuses(status, foreignStatus, 4, -5)
  !
  ! If status and foreignStatus had, say, values 3 (bits 1 and 2 set) and 4
  ! (bit 3 set) before the call, status should have the value 67 after the
  ! call (bits 1, 2 and 4+3=7 set).

    implicit none

    integer, intent(in) :: status, foreignStatus, maxPos, maxNeg

    !-------------------------------------------------------------------
    ! Start

    mergeStatuses_old = status ! default

    if (status .ge. 0) then
      if (foreignStatus .lt. 0) then
        mergeStatuses_old = foreignStatus + maxNeg
      else if (foreignStatus .gt. 0) then
        mergeStatuses_old = iand(status, ishft(foreignStatus, maxPos))
      end if
    end if

  end function mergeStatuses_old

  !---------------------------------------------------------------------
  subroutine incompleteGammaP(a, x, p, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function P(a,x), defined by
  !
  !		                 x
  !		           1    /
  !		P(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               0
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the following routine.
  !
  ! This routine was cribbed from gammp() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real, intent(in)  :: a, x
    real, intent(out) :: p
    integer, intent(out), optional :: status

    integer :: localStatus
    real :: q

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    p = 0.0

    if (x .lt. 0.0 .or. a .le. 0.0) then
      if (present(status)) status = -1
  return
    end if

    if (x .lt. a + 1) then ! use the series representation:
      call incompleteGammaPbySeries(a, x, p, localStatus)
    else ! use the continued-fraction representation:
      call incompleteGammaQbyFraction(a, x, q, localStatus)
      p = 1.0 - q
    end if

    if (present(status)) status = localStatus

  end subroutine incompleteGammaP

  !---------------------------------------------------------------------
  subroutine incompleteGammaQ(a, x, q, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function Q(a,x), defined by
  !
  !		                inf
  !		           1    /
  !		Q(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               x
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the previous routine. Note also that the
  ! _unnormalised_ incomplete gamma T(a)Q(a,x) is also available from the
  ! slatec function gamic(a,x).
  !
  ! This routine was cribbed from gammq() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real, intent(in)  :: a, x
    real, intent(out) :: q
    integer, intent(out), optional :: status

    integer :: localStatus
    real :: p

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    p = 0.0

    if (x .lt. 0.0 .or. a .le. 0.0) then
      if (present(status)) status = -1
  return
    end if

    if (x .lt. a + 1) then ! use the series representation:
      call incompleteGammaPbySeries(a, x, p, localStatus)
      q = 1.0 - p
    else ! use the continued-fraction representation:
      call incompleteGammaQbyFraction(a, x, q, localStatus)
    end if

    if (present(status)) status = localStatus

  end subroutine incompleteGammaQ

  !---------------------------------------------------------------------
  subroutine incompleteGammaPbySeries(a, x, p, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function P(a,x), defined by
  !
  !		                 x
  !		           1    /
  !		P(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               0
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the following routine.
  !
  ! The function is here evaluated from a series.
  !
  ! This routine was cribbed from gser() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real, intent(in)  :: a, x
    real, intent(out) :: p
    integer, intent(out) :: status

    integer, parameter :: maxNumIterations = 100
    real :: convergeCriterion, ap, summ, del, lnGammaVal
    integer :: i
    logical :: notConverged

    !-------------------------------------------------------------------
    ! Start

    status = 0
    convergeCriterion = 3.0 ** (-precision(x))

    p = 0.0

    if (x .lt. 0.0) then
      status = -1
  return
    else if (x .eq. 0.0) then
  return
    end if

    ap = a
    summ = 1.0 / a
    del = summ
    notConverged = .true. ! default
    do i = 1, maxNumIterations
      ap = ap + 1.0
      del = del * x / ap
      summ = summ + del

      if (abs(del) .lt. abs(summ) * convergeCriterion) then
        notConverged = .false.
    exit
      end if
    end do

    if (notConverged) then
      status = -2
  return
    end if

    if (a .eq. 0.5) then
      lnGammaVal = LnPi * 0.5
    else
      lnGammaVal = lnGamma(a)
    end if

    p = summ * exp(-x + a * log(x) - lnGammaVal)

  end subroutine incompleteGammaPbySeries

  !---------------------------------------------------------------------
  subroutine incompleteGammaQbyFraction(a, x, q, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function Q(a,x), defined by
  !
  !		                inf
  !		           1    /
  !		Q(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               x
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the previous routine. Note also that the
  ! _unnormalised_ incomplete gamma T(a)Q(a,x) is also available from the
  ! slatec routine gamic(a.x).
  !
  ! The function is evaluated by a modified Lentz's continued fraction.
  !
  ! This routine was cribbed from gcf() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real, intent(in)  :: a, x
    real, intent(out) :: q
    integer, intent(out) :: status

    integer, parameter :: maxNumIterations = 100
    real :: convergeCriterion, verySmallReal, b, c, d, h, an, del, lnGammaVal
    integer :: i
    logical :: notConverged

    !-------------------------------------------------------------------
    ! Start

    status = 0
    convergeCriterion = 3.0 ** (-precision(x))
    verySmallReal = epsilon(x)

    b = x + 1.0 - a
    c = 1.0 / verySmallReal
    d = 1.0 / b
    h = d
    notConverged = .true. ! default
    do i = 1, maxNumIterations
      an = -i * (i - a)
      b = b + 2.0
      d = an * d + b
      if (abs(d) .lt. verySmallReal) d = verySmallReal
      c = b + an / c
      if (abs(c) .lt. verySmallReal) c = verySmallReal
      d = 1.0 / d
      del = d * c
      h = h * del
      if(abs(del - 1.0) .lt. convergeCriterion) then
        notConverged = .false.
    exit
      end if
    end do

    if (notConverged) then
      status = -2
  return
    end if

    if (a .eq. 0.5) then
      lnGammaVal = LnPi * 0.5
    else
      lnGammaVal = lnGamma(a)
    end if

    q = h * exp(-x + a * log(x) - lnGammaVal)

  end subroutine incompleteGammaQbyFraction

  !---------------------------------------------------------------------
  real function lnGamma(a)
  !---------------------------------------------------------------------
  ! Returns the natural log of the absolute value of the complete gamma
  ! function T(a), which is defined by
  !
  !		       inf
  !		       /
  !		T(a) = | dt exp(-t) t^(a-1).
  !		       /
  !		      0
  !
  ! The function is evaluated by a Lanczos approximation.
  !
  ! The ln|T(a)| function can also be evaluated using the slatec function
  ! alngam(a).
  !
  ! This routine was cribbed from gammln() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real, intent(in) :: a

    integer :: j
    real(kind(1d0)) :: ser, stp, tmp, x, y, cof(6)

    save cof, stp
    data cof, stp /&
      76.18009172947146d0,&
     -86.50532032941677d0,&
      24.01409824083091d0,&
      -1.231739572450155d0,&
       0.1208650973866179d-2,&
      -0.5395239384953d-5,&
       2.5066282746310005d0/

    !-------------------------------------------------------------------
    ! Start

    x = a
    y = x
    tmp = x + 5.5d0
    tmp = (x + 0.5d0) * log(tmp) - tmp
    ser = 1.000000000190015d0
    do j = 1, 6
      y = y + 1.0d0
      ser = ser + cof(j) / y
    end do
    lnGamma = tmp + log(stp * ser / x)

  end function lnGamma

  !---------------------------------------------------------------------
  real function poisson(av, i)
  !---------------------------------------------------------------------
  ! Returns the Poisson probability of the occurrence of i events when
  ! the ensemble average = av. This probability distribution is given by
  !
  !	        av^i exp(-av)
  !	P(i) = ---------------
  !	             i!
  !
  ! This routine was adapted from material in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real, intent(in) :: av
    integer, intent(in) :: i

    logical :: firstEntry = .true.
    integer, save :: threshhold
    real :: logHugeInt, nOld, nNew

    !-------------------------------------------------------------------
    ! Start

    poisson = 0.0

  if (av <= 0) return
  if (i < 0) return

    if (firstEntry) then ! calculate the threshhold value corresponding to
    ! the max integer size of the present platform:

      firstEntry = .false.

      ! Invert Stirling's formula for factorial(n) by iteration:
      logHugeInt = log(real(huge(i)))
      nOld = logHugeInt
      do
        nNew = (logHugeInt - 0.5 * log(TwoPi * nOld)) / (log(nOld) - 1.0)
      if (abs(nNew - nOld) < 1.0) exit
        nOld = nNew
      end do

      threshhold = floor(nNew - 1) ! the minus 1 just for a bit of elbow room.
    end if

    if (i < threshhold) then
      poisson = exp(i * log(av) - av) / factorial(i)

    else
      poisson = exp(i * log(av) - av - lnGamma(i + 1.0))
    end if

  end function poisson

  !---------------------------------------------------------------------
  integer function factorial(i)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: i

    integer :: j

    !-------------------------------------------------------------------
    ! Start

    factorial = 1
    do j = 2, i
      factorial = factorial * j
    end do

  end function factorial

  !---------------------------------------------------------------------
  subroutine chi2Prob(chi2, degFree, probability, status)
  !---------------------------------------------------------------------
  ! This subroutine gives the probability that a given chi2/degFree is
  ! 'natural' - ie it gives the probability of the null hypothesis. Chi2
  ! (chi squared) is defined as follows:
  !
  !	         sum[i=1, N](y_i - e_i)^2
  !	chi^2 = --------------------------
  !	                 sigma_e_i^2
  !
  ! where y_i is the measured or data value, e_i is the model or expected
  ! value and sigma_e_i is the uncertainty in the latter.

    implicit none

    real,    intent(in)            :: chi2
    integer, intent(in)            :: degFree
    real,    intent(out)           :: probability
    integer, intent(out), optional :: status

    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start

    probability = 0.0

    if (chi2 .le. 0.0) then
      if (present(status)) status = -1
  return
    end if
    if (degFree .le. 0) then
      if (present(status)) status = -2
  return
    end if

    call incompleteGammaQ(degFree / 2.0, chi2 / 2.0, probability, foreignStatus)

    if (present(status)) status = foreignStatus

  end subroutine chi2Prob

  !---------------------------------------------------------------------
  subroutine poissonProb(testY, meanY, probability, status)
  !---------------------------------------------------------------------
  ! Returns the probability that a Poisson-distributed (integer) variable
  ! y will exceed testY.

    implicit none

    integer, intent(in) :: testY
    real, intent(in) :: meanY
    real, intent(out) :: probability
    integer, intent(out) :: status

    !-------------------------------------------------------------------
    ! Start

    status = 0
    probability = 0.0
    if (testY .lt. 0) then
      status = -1
  return
    end if
    if (meanY .lt. 0.0) then
      status = -2
  return
    end if

    call incompleteGammaP(real(testY + 1), meanY, probability, status)

  end subroutine poissonProb

  !---------------------------------------------------------------------
  function gaussProb(testY, meanY, sigma) result(probability)
  !---------------------------------------------------------------------
  ! Returns the probability that a gaussian-distributed variable y will
  ! depart from the mean by greater than abs(testY - meanY). For values
  ! obeying a gaussian distribution of given mean and
  ! standard deviation s, the probability of a given y value (or greater)
  ! occurring by chance is 1 - erf(abs(testY - meanY)/s/sqrt(2))

    implicit none

    real, intent(in) :: testY, meanY, sigma
    real :: probability

    !-------------------------------------------------------------------
    ! Start

!    probability = 1.0 - erf(real(abs(testY - meanY) / sigma / RootTwo))
    probability = erfc(real(abs(testY - meanY) / sigma / RootTwo))

  end function gaussProb

  !---------------------------------------------------------------------
  real function erf(x, style)
  !---------------------------------------------------------------------
  ! The error function
  !
  !	                     x
  !	             2      /
  !	erf(x) = ---------- | dt exp(-t^2).
  !	          sqrt(pi)  /
  !	                   0
  !

    implicit none

    real, intent(in) :: x
    integer, intent(in), optional :: style

    integer :: localStyle

    !-------------------------------------------------------------------
    ! Start

    if (present(style)) then
      localStyle = style
    else
      localStyle = 0
    end if

    select case(localStyle)
    case(0)
      call incompleteGammaP(0.5, x*x, erf)
    case(1)
      erf = erfApprox0(x)
    case(2)
      erf = erfApprox1(x)
    case default
    end select

  end function erf

  !---------------------------------------------------------------------
  real function erfc(x, style)
  !---------------------------------------------------------------------
  ! The complementary error function
  !
  !	                                   inf
  !	                           2      /
  !	erfc(x) = 1 - erf(x) = ---------- | dt exp(-t^2).
  !	                        sqrt(pi)  /
  !	                                 x
  !
  ! Note that, for values obeying a gaussian distribution of mean <y> and
  ! standard deviation s, the probability of a given y value (or greater)
  ! occurring by chance is erfc(abs(y - <y>)/s/sqrt(2))

    implicit none

    real, intent(in) :: x
    integer, intent(in), optional :: style

    integer :: localStyle

    !-------------------------------------------------------------------
    ! Start

    if (present(style)) then
      localStyle = style
    else
      localStyle = 0
    end if

    select case(localStyle)
    case(0)
      call incompleteGammaQ(0.5, x*x, erfc)
    case(1)
      erfc = 1.0 - erfApprox0(x)
    case(2)
      erfc = 1.0 - erfApprox1(x)
    case default
    end select

  end function erfc

  !---------------------------------------------------------------------
  function erfApprox0(x) result(erf)
  !---------------------------------------------------------------------
  ! Taken from Numerical Methods for Fortran.

    implicit none

    real, intent(in) :: x

    real :: t, z, erf
    real, parameter ::&
      p0 = -1.26551223,&
      p1 =  1.00002368,&
      p2 =  0.37409196,&
      p3 =  0.09678418,&
      p4 = -0.18628806,&
      p5 =  0.27886807,&
      p6 = -1.13520398,&
      p7 =  1.48851587,&
      p8 = -0.82215223,&
      p9 =  0.17087277

    !-------------------------------------------------------------------
    ! Start

    z = abs(x)
    t = 1.0 / (1.0 + 0.5 * z)
    erf = t * exp(-z * z + p0 + t * (p1 + t * (p2&
        + t * (p3 + t * (p4 + t * (p5 + t * (p6&
        + t * (p7 + t * (p8 + t * p9)))))))))

    if (x .lt. 0.0) erf = 2.0 - erf

  end function erfApprox0

  !---------------------------------------------------------------------
  function erfApprox1(x) result(erf)
  !---------------------------------------------------------------------
  ! This approx uses the formula of Busen K M, Am J Phys 49, 886 (1981).

    implicit none

    real, intent(in) :: x
    real :: erf, sign

    !-------------------------------------------------------------------
    ! Start

    sign = 1.0
    if (x < 0.0) sign = -1.0

    if (x > 1.0) then
      erf = sqrt(1.0 - exp(-4.0 * x**1.88 / Pi))
    else
      erf = sqrt(1.0 - exp(-4.0 * x * x / Pi))
    end if

    erf = erf * sign

  end function erfApprox1

  !---------------------------------------------------------------------
  function erfcApprox0(x) result(erfc)
  !---------------------------------------------------------------------
  !x    input	Deviate, range -infinity to +infinity.
  !erfc output	Area under Gaussian from X to +infinity.
  !-
  ! Author	Clive Page	1983 July 5
  ! Standard Fortran-77 version 1984 Sept 25, CGP.
  ! From algorithm AS66 by I.D.Hill, Appl Stats  22,424 (1973).
  ! Adapted to f90 by Ian Stewart 8 Aug 2000

    implicit none 

    real, intent(in) :: x
    real :: erfc, y, z

    !-------------------------------------------------------------------
    ! Start

    z = abs(x)
    if(z .gt. 13.0) then
      erfc = 0.0

    else
      y = 0.5 * z * z
      if(z .le. 1.28) then
        erfc = 0.5 - z * (0.398942280444 - 0.399903438504 * y /&
        (y + 5.75885480458 - 29.8213557808 /&
        (y + 2.62433121679 + 48.6959930692 /&
        (y + 5.92885724438))))

      else
        erfc = 0.398942280385 * exp(-y) /&
        (z - 3.8052E-8      + 1.00000615302 /&
        (z + 3.98064794E-4  + 1.98615381364 /&
        (z - 0.151689116635 + 5.29330324926 /&
        (z + 4.8385912808   - 15.1508972451 /&
        (z + 0.742380924027 + 30.789933034  /&
        (z + 3.99019417011))))))
      end if
    end if

    if(x .lt. 0.0) erfc = 1.0 - erfc

  end function erfcApprox0

  !---------------------------------------------------------------------
  real(kind(1d0)) function besselJ0(arg)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)), intent(in) :: arg

    real(kind(1d0)) :: term, mult, jd, P0, Q0
    integer :: n
!    real(kind(1d0)) :: J1, trial_J1, trial_J0
    real(kind(1d0)), parameter :: precision = 1.0d-14, alpha = 1.0d-100
    integer, parameter :: k = 45

    !-------------------------------------------------------------------
    ! Start

    if (arg .lt. 1.0) then !power series is most convenient.
      mult = arg * arg / 4.0d0
      term = 1.0d0
      jd = 1.0d0
      n = 1
      do while (abs(term) .gt. precision)
        term = -(term * mult / dble(n)) / dble(n)
        jd = jd + term
        n = n + 1
      end do
      besselJ0 = jd

    else if (arg .gt. 16.0) then !asymptotic series is better.
      call PQ0_calc(arg, P0, Q0)
      besselJ0 = (P0 * cos(arg - Pi / 4.0d0)&
                - Q0 * sin(arg - Pi / 4.0d0))&
      * sqrt(2.0d0 / Pi / arg)

    else !value from recurrence relation is best.
      besselJ0 = trial_J0(arg, k, alpha)
    end if

  end function besselJ0

  !---------------------------------------------------------------------
  subroutine PQ0_calc(arg, pSum, qSum)
  !---------------------------------------------------------------------
      
    implicit none

    real(kind(1d0)), intent(in)  :: arg
    real(kind(1d0)), intent(out) :: pSum, qSum

    real(kind(1d0)) :: seed, multiplier, oldterm, term
    integer :: n, sig

    !-------------------------------------------------------------------
    ! Start

    seed = 1.0d0
    oldterm = seed
    pSum = seed
    qSum = 0.0d0

    sig = -1
    n = 1
    multiplier = sig * dble(2 * n - 1) * dble(2 * n - 1) / dble(n)&
    / 8.0d0 / arg

    term = oldterm * multiplier
    do while (abs(term) .lt. abs(oldterm))
      oldterm = term
      qSum = qSum + term
      sig = -sig
      n = n + 1
      multiplier = sig * dble(2 * n - 1) * dble(2 * n - 1) / dble(n)&
      / 8.0d0 / arg

      term = oldterm * multiplier
    if (abs(term) .ge. abs(oldterm)) exit
      oldterm = term
      pSum = pSum + term
      sig = -sig
      n = n + 1
      multiplier = sig * dble(2 * n - 1) * dble(2 * n - 1) / dble(n)&
      / 8.0d0 / arg

      term = oldterm * multiplier
    end do

  end subroutine PQ0_calc

  !---------------------------------------------------------------------
  real(kind(1d0)) function trial_J0(arg, k, alpha)
  !---------------------------------------------------------------------
  ! Calculates J0 by the recurrence method of Goldstein M and Thaler
  ! R M 1960 (don't know the journal).

    implicit none

    real(kind(1d0)), intent(in) :: arg, alpha
    integer, intent(in) :: k

    real(kind(1d0)) :: summ, J(0:50)
    integer :: p

    !-------------------------------------------------------------------
    ! Start

    J(k + 1) = 0.0d0
    J(k) = alpha

    do p = k, 1, -1
      J(p - 1) = 2.0d0 * p * J(p) / arg - J(p + 1)
    end do

    summ = J(0)
    do p = 1, k / 2
      summ = summ + 2.0d0 * J(2 * p)
    end do

    trial_J0 = J(0) / summ

  end function trial_J0
      
  !---------------------------------------------------------------------
  subroutine sourceOrientation(image, srcX, srcY, radialFunction, maxR&
  , coeffs, status, vec)
  !---------------------------------------------------------------------
  ! This routine attempts to obtain the orientation of a source psf. This
  ! is a desirable quantity if one is attempting to estimate the location
  ! of the optical axis of the image-forming system.
  !

!    use sscmisc

    implicit none

    real,    intent(in)            :: image(:,:)
    real,    intent(in)            :: srcX, srcY, maxR
    complex, intent(out)           :: coeffs(2)
    integer, intent(out), optional :: status
real, pointer :: vec(:)

    interface
      real function radialFunction(r)
        real, intent(in) :: r
      end function radialFunction
    end interface

    integer, parameter ::&
      halfNumThetaBins = 64,&
      numThetaBins = halfNumThetaBins * 2
    integer :: numRbins, i, j
    real :: deltaR, polar(numThetaBins), cosTheta(numThetaBins),&
      sinTheta(numThetaBins), cos2Theta(numThetaBins),&
      sin2Theta(numThetaBins), theta, r, x, y,&
      xVector(size(image, 1)), yVector(size(image, 2))
integer, pointer :: xi(:), yi(:)

    !-------------------------------------------------------------------
    ! Start

    coeffs = (0.0, 0.0)

    deltaR = 1.0 / 10.0 ! arbitrary
    numRbins = ceiling(maxR / deltaR)
    if (numRbins .lt. 2) then
      if (present(status)) status = -1
  return
    end if

!    xVector = real(indicize(size(image, 1)))
!    yVector = real(indicize(size(image, 2)))
xi => indicize(size(image, 1))
xVector = real(xi)
deallocate(xi)
yi => indicize(size(image, 2))
yVector = real(yi)
deallocate(yi)

    polar = 0.0
    do i = 1, numThetaBins
      theta = fracVal(0.0, real(TwoPi), (i - 1)/real(numThetaBins))
      cosTheta(i) = cos(theta)
      sinTheta(i) = sin(theta)
      polar(i) = 0.0
      do j = 2, numRbins
        r = fracVal(0.0, maxR, (j - 1)/real(numRbins - 1))
        x = srcX + r * cosTheta(i)
        y = srcY + r * sinTheta(i)
        polar(i) = polar(i) + r * radialFunction(r)&
        * bilinearSample(image, xVector, yVector, x, y)
      end do
    end do

    do i = 1, halfNumThetaBins
      theta = fracVal(0.0, real(TwoPi), (i - 1)/real(numThetaBins))
      cos2Theta(i) = cos(2.0 * theta)
      sin2Theta(i) = sin(2.0 * theta)
      cos2Theta(i + halfNumThetaBins) = cos2Theta(i)
      sin2Theta(i + halfNumThetaBins) = sin2Theta(i)
    end do

    do i = 1, numThetaBins
      coeffs(1) = coeffs(1) + cmplx(polar(i)*cosTheta(i), polar(i)*sinTheta(i))
      coeffs(2) = coeffs(2) + cmplx(polar(i)*cos2Theta(i),polar(i)*sin2Theta(i))
    end do
    coeffs = coeffs / sum(polar)

    if (present(status)) status = 0

allocate(vec(numThetaBins))
vec = polar

  end subroutine sourceOrientation

  !---------------------------------------------------------------------
  real function bilinearSample(zGrid, xVector, yVector, x, y)
  !---------------------------------------------------------------------
!*** there's a better one in testbed.

    implicit none

    real, intent(in) ::&
      zGrid(:,:),&
      xVector(size(zGrid, 1)),&
      yVector(size(zGrid, 2)),&
      x,&
      y

    real :: fracX, fracY
    integer :: xi, xxi, yi, yyi, xsize, ysize

    !-------------------------------------------------------------------
    ! Start

    bilinearSample = 0.0 ! default
    xsize = size(zGrid, 1)
    ysize = size(zGrid, 2)
    if (x .lt. xVector(1) .or. x .ge. xVector(xsize)&
    .or. y .lt. yVector(1) .or. y .ge. yVector(ysize))&
  return

    do xi = 2, xsize
      if (x .lt. xVector(xi)) then
        xxi = xi - 1
    exit
      end if
    end do
    do yi = 2, ysize
      if (y .lt. yVector(yi)) then
        yyi = yi - 1
    exit
      end if
    end do

    fracX = (x - xVector(xxi)) / (xVector(xxi+1) - xVector(xxi))
    fracY = (y - yVector(yyi)) / (yVector(yyi+1) - yVector(yyi))
    bilinearSample = bilinearValue(zGrid(xxi:xxi+1, yyi:yyi+1), fracX, fracY)

  end function bilinearSample

  !---------------------------------------------------------------------
  real function bilinearValue(zAtCorners, fracX, fracY)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: zAtCorners(2,2), fracX, fracY

    !-------------------------------------------------------------------
    ! Start

    bilinearValue = (1.0 - fracX) * (1.0 - fracY) * zAtCorners(1, 1)&
             +             fracX  * (1.0 - fracY) * zAtCorners(1, 2)&
             +             fracX         * fracY  * zAtCorners(2, 2)&
             +      (1.0 - fracX)        * fracY  * zAtCorners(2, 1)

  end function bilinearValue

  !---------------------------------------------------------------------
  subroutine cleverInterpolate(y0, x0, y1, x1, style)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: x0(:), y0(size(x0)), x1(:)
    real, intent(out) :: y1(size(x1))
    integer, intent(in) :: style

integer :: junk

    !-------------------------------------------------------------------
    ! Start

y1 = y0 * x0 ! just to keep
junk = style ! the compiler happy

  end subroutine cleverInterpolate

!  !---------------------------------------------------------------------
!  function cubicSplineWithX(xx, yy, x) result(y)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    real, intent(in) :: xx(0:3), yy(0:3), x
!    real :: y
!
!    !-------------------------------------------------------------------
!    ! Start
!
!  end function cubicSplineWithX
!
!  !---------------------------------------------------------------------
!  function cubicSplineOhneX(yy, xFrac) result(y)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    real, intent(in) :: yy(0:3), xFrac
!    real :: y
!
!    !-------------------------------------------------------------------
!    ! Start
!
!  end function cubicSplineOhneX

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module math_utils
