!+
! Module: my_psf
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Returns semi-axes of an ellipse round each of a list of sources.
! 
! Comments: TODO:
! 
! $Id: my_psf.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module my_psf

  use dal

  use math_utils ! defines statusType.
  use constants

  implicit none 

  private 

  public getPsfEllipseAxes
  public getPsfValue
  public getTruncatedPsfValue
  public getPsfIntegral

  type :: srcEnvelopeType
    real(double) ::&
      thetaArcsec,&
      EeV
    real(single) ::&
      peakFlux,&
      bkgFlux,&
      longAxisArcsec,&
      shortAxisArcsec
    logical :: valid
  end type srcEnvelopeType

  real(single) ::&
    x0,&
    y0,&
    x0squared,&
    y0squared,&
    alpha,&
    sigmaX,&
    sigmaY,&
    twoSigmaXsquared,&
    twoSigmaYsquared,&
    fractionInWing,&
    level

  real(single), public, parameter ::&
    psfTruncRadiusArcsec = 100.0,&
    psfTruncRadiusArcsecSquared = psfTruncRadiusArcsec * psfTruncRadiusArcsec

contains 

  !---------------------------------------------------------------------
  subroutine getPsfEllipseAxes(thetaArcsec, EeV, srcPeakFlux&
  , bkgFlux, bkgFrac, longAxisArcsec, shortAxisArcsec, srcIsValid)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: EeV(:),&
                                 thetaArcsec(size(EeV))
    real(single), intent(in)  :: srcPeakFlux(size(EeV)),&
                                 bkgFlux(size(EeV))
    real(single), intent(in)  :: bkgFrac
    real(single), intent(out) :: longAxisArcsec(size(EeV)),&
                                 shortAxisArcsec(size(EeV))
    logical,      intent(out) :: srcIsValid(size(EeV))

    type(srcEnvelopeType) :: srcEnvelopeData(size(EeV))

    integer(int16) :: numSrcs

    !-------------------------------------------------------------------
    ! Start

    longAxisArcsec  = 0.0
    shortAxisArcsec = 0.0
    numSrcs = size(EeV)

    srcEnvelopeData%peakFlux    = srcPeakFlux
    srcEnvelopeData%bkgFlux     = bkgFlux
    srcEnvelopeData%thetaArcsec = thetaArcsec
    srcEnvelopeData%EeV         = EeV

    call getPsfEllipseAxes_kernel(srcEnvelopeData, bkgFrac)

    longAxisArcsec  = srcEnvelopeData%longAxisArcsec
    shortAxisArcsec = srcEnvelopeData%shortAxisArcsec
    srcIsValid      = srcEnvelopeData%valid

  end subroutine getPsfEllipseAxes

  !---------------------------------------------------------------------
  subroutine getPsfEllipseAxes_kernel(srcEnvelopeData, bkgFrac)
  !---------------------------------------------------------------------
  ! This is designed to return, for a list of XMM source positions, the
  ! long and short semi-axes of an ellipse around each source such that
  ! the value of the PSF for that source at the locus of the ellipse is
  ! equal to a specified fraction (bkgFrac, may be any value > 0) of the
  ! local background rate. The values srcRate, bkgRate and
  ! offAxisAngleArcsec of srcData should be set by the user, whereas the
  ! semi-axes are returned in longAxisArcsec and shortAxisArcsec. The
  ! long axis is assumed to lie in the azimuthal direction, the short,
  ! in the radial direction.

    implicit none

    type(srcEnvelopeType), intent(inout) :: srcEnvelopeData(:)
    real(single), intent(in) :: bkgFrac

    integer(int16) :: numSrcs, i
    real :: xLo, xHi, yLo, yHi, solutionX, solutionY
    integer, parameter :: maxPos = 0, maxNeg = 0
    type(statusType) :: foreignStat ! typedef in mathutils_mod

    !-------------------------------------------------------------------
    ! Start

    srcEnvelopeData%longAxisArcsec  = 0.0     ! default
    srcEnvelopeData%shortAxisArcsec = 0.0     ! default
    srcEnvelopeData%valid           = .false. ! default

    numSrcs = size(srcEnvelopeData)

    do i = 1, numSrcs
    if (srcEnvelopeData(i)%peakFlux .le. 0.0) cycle
      level = srcEnvelopeData(i)%bkgFlux * bkgFrac / srcEnvelopeData(i)%peakFlux
    if (level .le. 0.0) cycle
      if (level .ge. 1.0) then
        srcEnvelopeData(i)%valid = .true.
    cycle
      end if

      call setPsfParams(srcEnvelopeData(i)%thetaArcsec, srcEnvelopeData(i)%EeV)

      call preNewton(fx, xLo, xHi, sigmaX, stat=foreignStat)
    if (foreignStat%status .lt. 0) cycle

      call hybridNewtonsMethod(xLo, xHi, fx, dfx, solutionX&
        , stat=foreignStat)

    if (foreignStat%status .lt. 0) cycle

      call preNewton(fy, yLo, yHi, sigmaY, stat=foreignStat)
    if (foreignStat%status .lt. 0) cycle

      call hybridNewtonsMethod(yLo, yHi, fy, dfy, solutionY&
      , stat=foreignStat)

    if (foreignStat%status .lt. 0) cycle

      srcEnvelopeData(i)%shortAxisArcsec = solutionX
      srcEnvelopeData(i)%longAxisArcsec  = solutionY
      srcEnvelopeData(i)%valid           = .true.
    end do

  end subroutine getPsfEllipseAxes_kernel

  !---------------------------------------------------------------------
  subroutine setPsfParams(thetaArcsec, EeV)
  !---------------------------------------------------------------------
!*** add 'arcsec' to variables which are in this unit!!

    implicit none

    real(double), intent(in) :: thetaArcsec, EeV

    real(single), parameter ::&
      sigmaBaseX = 5.0,&
      sigmaBaseY = 5.0,&
      sigmaCurveX = 1.0e4,&
      sigmaCurveY = 8.0e4,&
      x0Base = 10.0,&
      y0Base = 10.0,&
      x0Curve = 0.0,&
      y0Curve = 4.0e4
    real(single) ::&
      theta
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

    theta = thetaArcsec / 3600.0 * piOn180

    sigmaX = sigmaBaseX * (1.0 + sigmaCurveX * theta * theta)
    sigmaY = sigmaBaseY * (1.0 + sigmaCurveY * theta * theta)
    twoSigmaXsquared = 2.0 * sigmaX * sigmaX
    twoSigmaYsquared = 2.0 * sigmaY * sigmaY
    x0 = x0Base * (1.0 + x0Curve * theta * theta)
    y0 = y0Base * (1.0 + y0Curve * theta * theta)
!    x0 = 10.0
!    y0 = 10.0
    x0squared = x0 * x0
    y0squared = y0 * y0
    alpha = 1.4
!!!! NOTE! alpha MUST be GREATER THAN 1. Otherwise the integral of the
!!!! psf will not converge.

    fractionInWing = 0.2

if (testFlag) print *, 'theta, sigmaY, sigmaX', theta, sigmaY, sigmaX

theta = EeV ! Just to keep the compiler from generating a warning.

  end subroutine setPsfParams

  !---------------------------------------------------------------------
  subroutine preNewton(f, xLo, xHi, xGuess, maxNumIterations, stat)
  !---------------------------------------------------------------------
  ! This routine uses a binary search to find a pair of values that
  ! straddles a zero of the function f. It is assumed that 1 and only
  ! 1 zero exists between x=0 and x=inf, and that f(0) > 0 (and hence
  ! that f(inf) < 0).

    implicit none

    real,    intent(out)                    :: xLo, xHi
    real,    intent(in),  optional          :: xGuess
    integer, intent(in),  optional          :: maxNumIterations
    type(statusType), intent(out), optional :: stat

    interface
      real function f(x)
        real, intent(in) :: x
      end function f
    end interface

    type(statusType) :: localStat
    integer :: numIter, localMaxNumIter
    real :: localXguess

    !-------------------------------------------------------------------
    ! Start

    localStat%status = 0 ! default
    localStat%maxNeg = -2
    localStat%numPosBits = 0
    if (present(stat)) stat = localStat

    xLo = 0.0
    xHi = 0.0

    if (present(xGuess)) then
      if (xGuess .le. 0.0) then
        if (present(stat)) stat%status = -1
  return
      end if
      localXguess = xGuess
    else
      localXguess = 1.0
    end if

    if (present(maxNumIterations)) then
      localMaxNumIter = maxNumIterations
    else
      localMaxNumIter = 100
!*** can do better than this.
    end if

    if (f(localXguess) .le. 0.0) then
      xLo = 0.0
      xHi = localXguess
  return

    else
      xLo = localXguess
      xHi = xLo * 2.0
      numIter = 1
      do while(f(xHi) .gt. 0.0)
        numIter = numIter + 1
        if (numIter .gt. localMaxNumIter) then
          if (present(stat)) stat%status = -2
  return
        end if

        xLo = xHi
        xHi = xLo * 2.0
      end do
    end if

  end subroutine preNewton

  !---------------------------------------------------------------------
  real function estimate(f, divX, numIterations)
  !---------------------------------------------------------------------
  ! This is used to approximate the zero of a (presumably monatonic)
  ! function by a binary search.

    implicit none

    interface
      real function f(x)
        real, intent(in) :: x
      end function f
    end interface
    real, intent(in) :: divX
    integer, intent(in) :: numIterations

    integer :: i
    real :: botX, topX, meanX
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    if (f(divX) .gt. 0.0) then
      botX = 0.0
      topX = 1.0 / divX
      meanX = (botX + topX) / 2.0
if (testFlag) print *, '<-', 1.0 / meanX, f(1.0 / meanX)
      do i = 1, numIterations
        if (f(1.0 / meanX) .gt. 0.0) then
          topX = meanX
        else
          botX = meanX
        end if
        meanX = (botX + topX) / 2.0
if (testFlag) print *, '<-', 1.0 / meanX, f(1.0 / meanX)
      end do

      estimate = 1.0 / meanX

    else
      botX = 0.0
      topX = divX
      meanX = (botX + topX) / 2.0
if (testFlag) print *, '->', meanX, f(meanX)
      do i = 1, numIterations
        if (f(meanX) .gt. 0.0) then
          botX = meanX
        else
          topX = meanX
        end if
        meanX = (botX + topX) / 2.0
if (testFlag) print *, '->', meanX, f(meanX)
      end do

      estimate = meanX
    end if

  end function estimate

  !---------------------------------------------------------------------
  real function fx(x)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: x

    !-------------------------------------------------------------------
    ! Start

    fx = fractionInWing / (1.0 + x * x / x0squared)**alpha&
    + (1.0 - fractionInWing) * exp(-x * x / twoSigmaXsquared)&
    - level

  end function fx

  !---------------------------------------------------------------------
  real function dfx(x)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: x

    !-------------------------------------------------------------------
    ! Start

    dfx = -fractionInWing * (alpha * 2.0 * x / x0squared)&
    / (1.0 + x * x / x0squared)**(alpha + 1.0)&
    - (1.0 - fractionInWing) * (x / sigmaX / sigmaX)&
    * exp(-x * x / twoSigmaXsquared)

  end function dfx

  !---------------------------------------------------------------------
  real function fy(y)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: y

    !-------------------------------------------------------------------
    ! Start

    fy = fractionInWing / (1.0 + y * y / y0squared)**alpha&
    + (1.0 - fractionInWing) * exp(-y * y / twoSigmaYsquared)&
    - level

  end function fy

  !---------------------------------------------------------------------
  real function dfy(y)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: y

    !-------------------------------------------------------------------
    ! Start

    dfy = -fractionInWing * (alpha * 2.0 * y / y0squared)&
    / (1.0 + y * y / y0squared)**(alpha + 1.0)&
    - (1.0 - fractionInWing) * (y / sigmaY / sigmaY)&
    * exp(-y * y / twoSigmaYsquared)

  end function dfy

  !---------------------------------------------------------------------
  real function gauss_x(x)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: x

    !-------------------------------------------------------------------
    ! Start

    gauss_x = (1.0 - fractionInWing) * exp(-x * x / twoSigmaXsquared)

  end function gauss_x

  !---------------------------------------------------------------------
  real function gauss_y(y)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: y

    !-------------------------------------------------------------------
    ! Start

    gauss_y = (1.0 - fractionInWing) * exp(-y * y / twoSigmaYsquared)

  end function gauss_y

  !---------------------------------------------------------------------
  real function king_x(x)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: x

    !-------------------------------------------------------------------
    ! Start

    king_x = fractionInWing / (1.0 + x * x / x0squared)**alpha

  end function king_x

  !---------------------------------------------------------------------
  real function king_y(y)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: y

    !-------------------------------------------------------------------
    ! Start

    king_y = fractionInWing / (1.0 + y * y / y0squared)**alpha

  end function king_y

  !---------------------------------------------------------------------
  function getPsfValue(thetaArcsec, phi, srcThetaArcsec, srcPhi&
  , EeV, srcExtentDeg) result(psfValue)
  !---------------------------------------------------------------------
  ! This returns the psf function (normalised so that the peak value is 1)
  ! at a given point.

    implicit none

    real(double), intent(in) :: thetaArcsec, phi, srcThetaArcsec,&
      srcPhi, EeV
    real(single), intent(in), optional :: srcExtentDeg
    real(double) :: psfValue

    real(single) :: x, y, x2, y2, srcExtentArcsec, twoSigmaXsquaredTotal,&
      twoSigmaYsquaredTotal

    !-------------------------------------------------------------------
    ! Start

    call setPsfParams(srcThetaArcsec, EeV)

    x = thetaArcsec * cos(phi - srcPhi) - srcThetaArcsec
    y = thetaArcsec * sin(phi - srcPhi)
    x2 = x * x
    y2 = y * y

!*** the following is only a bodge, because it does not also broaden the King function:
    if (present(srcExtentDeg)) then
      srcExtentArcsec = srcExtentDeg * 3600.0
      twoSigmaXsquaredTotal = twoSigmaXsquared&
      + 2.0 * srcExtentArcsec * srcExtentArcsec
      twoSigmaYsquaredTotal = twoSigmaYsquared&
      + 2.0 * srcExtentArcsec * srcExtentArcsec
    else
      twoSigmaXsquaredTotal = twoSigmaXsquared
      twoSigmaYsquaredTotal = twoSigmaYsquared
    end if

    psfValue = fractionInWing / (1.0 + x2 / x0squared + y2 / y0squared)**alpha&
    + (1.0 - fractionInWing) * exp(-x2 / twoSigmaXsquaredTotal&
    - y2 / twoSigmaYsquaredTotal)

  end function getPsfValue

  !---------------------------------------------------------------------
  function getTruncatedPsfValue(thetaArcsec, phi, srcThetaArcsec, srcPhi&
  , EeV, srcExtentDeg) result(psfValue)
  !---------------------------------------------------------------------
  ! This is designed to work with makeSimulatedImage_withList() in simulations_mod. It is similar to getPsfValue() except that the psf is multiplied by a quadratric vignetting function so that it is zero outside a certain radius and tails off smoothly as it approaches that radius.

    implicit none

    real(double), intent(in) :: thetaArcsec, phi, srcThetaArcsec,&
      srcPhi, EeV
    real(single), intent(in), optional :: srcExtentDeg
    real(double) :: psfValue

    real(single) :: x, y, x2, y2, r2, srcExtentArcsec, twoSigmaXsquaredTotal,&
      twoSigmaYsquaredTotal

    !-------------------------------------------------------------------
    ! Start

    call setPsfParams(srcThetaArcsec, EeV)

    if (abs(thetaArcsec - srcThetaArcsec) > psfTruncRadiusArcsec) then
      psfValue = 0.0
  return
    end if
    x = thetaArcsec * cos(phi - srcPhi) - srcThetaArcsec
    if (abs(x) > psfTruncRadiusArcsec) then
      psfValue = 0.0
  return
    end if
    y = thetaArcsec * sin(phi - srcPhi)
    x2 = x * x
    y2 = y * y

    r2 = x2 + y2
    if (r2 > psfTruncRadiusArcsecSquared) then
      psfValue = 0.0
  return
    end if

!*** the following is only a bodge, because it does not also broaden the King function:
    if (present(srcExtentDeg)) then
      srcExtentArcsec = srcExtentDeg * 3600.0
      twoSigmaXsquaredTotal = twoSigmaXsquared&
      + 2.0 * srcExtentArcsec * srcExtentArcsec
      twoSigmaYsquaredTotal = twoSigmaYsquared&
      + 2.0 * srcExtentArcsec * srcExtentArcsec
    else
      twoSigmaXsquaredTotal = twoSigmaXsquared
      twoSigmaYsquaredTotal = twoSigmaYsquared
    end if

    psfValue = fractionInWing / (1.0 + x2 / x0squared + y2 / y0squared)**alpha&
    + (1.0 - fractionInWing) * exp(-x2 / twoSigmaXsquaredTotal&
    - y2 / twoSigmaYsquaredTotal)

    psfValue = psfValue * (1.0 - r2 / psfTruncRadiusArcsecSquared)

  end function getTruncatedPsfValue

  !---------------------------------------------------------------------
  function getPsfIntegral(srcThetaArcsec, EeV, srcExtentDeg)&
  result(psfIntegral)
  !---------------------------------------------------------------------
  ! This returns the integral of the psf for a source of peak brightness
  ! of 1.0 counts per square arcsecond.

    implicit none

    real(double), intent(in) :: srcThetaArcsec, EeV
    real(single), intent(in), optional :: srcExtentDeg
    real(double) :: psfIntegral

    real(single) :: srcExtentArcsec, twoSigmaXsquaredTotal,&
      twoSigmaYsquaredTotal, sigmaXtotal, sigmaYtotal

    !-------------------------------------------------------------------
    ! Start

    call setPsfParams(srcThetaArcsec, EeV)

!*** the following is only a bodge, because it does not also broaden the King function:
    if (present(srcExtentDeg)) then
      srcExtentArcsec = srcExtentDeg * 3600.0
      twoSigmaXsquaredTotal = twoSigmaXsquared&
      + 2.0 * srcExtentArcsec * srcExtentArcsec
      sigmaXtotal = sqrt((twoSigmaXsquaredTotal) / 2.0)
      twoSigmaYsquaredTotal = twoSigmaYsquared&
      + 2.0 * srcExtentArcsec * srcExtentArcsec
      sigmaYtotal = sqrt((twoSigmaYsquaredTotal) / 2.0)
    else
      sigmaXtotal = sigmaX
      sigmaYTotal = sigmaY
    end if

    psfIntegral = Pi * (fractionInWing * x0 * y0 / (alpha - 1.0)&
    + (1.0 - fractionInWing) * 2.0 * sigmaXtotal * sigmaYtotal)

  end function getPsfIntegral

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module my_psf
