!
! Module: random
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines to generate random numbers with a variety of probability distributions.
! 
! Comments: TODO:
! 
! $Id: random.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module random

!  use random ! In utils.

  use constants
  use array_utils ! for getBinNum()

  implicit none 

  private 

  public poissonRandom
  public getPoissonRandoms
  public minus5on2Random
  public getMinus5on2Randoms
  public powerRandom
  public getPowerRandoms
  public getLogNlogSrandoms
  public getPowerSegmentRandoms
  public doIntPowerSegmentChecks
  public doDiffPowerSegmentChecks
!  public sineRandom
  public getSineRandoms
  public exponentialRandom
  public getExponentialRandoms
  public gaussianRandom
  public getGaussianRandoms
  public randomNumber
  public getRandomNumbers
  public randoms
  public makeRandomSequence
  public coinIsHeads

contains 

  !---------------------------------------------------------------------
  function poissonRandom(average)
  !---------------------------------------------------------------------
  ! Returns a random value with a Poissonian distribution. The algorithm is
  ! not efficient for large values of 'average'.

    implicit none

    real, intent(in)  :: average
    integer :: poissonRandom

    real :: trial

    !-------------------------------------------------------------------
    ! Start:

    poissonRandom = 0
    trial = exponentialRandom()
    do while (trial .lt. average)
      poissonRandom = poissonRandom + 1
      trial = trial + exponentialRandom()
    end do

  end function poissonRandom

  !---------------------------------------------------------------------
  subroutine getPoissonRandoms(average, numbers)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have a Poissonian distribution.

    implicit none

    real,    intent(in)  :: average
    integer, intent(out) :: numbers(:)

    integer :: n, number

    !-------------------------------------------------------------------
    ! Start:

    number = size(numbers)
    do n = 1, number
      numbers(n) = poissonRandom(average)
    end do

  end subroutine getPoissonRandoms

  !---------------------------------------------------------------------
  function minus5on2Random()
  !---------------------------------------------------------------------

    implicit none

    real :: minus5on2Random
    integer, parameter :: size = 1000
    real :: harvest(size)
    integer :: i
    save i, harvest
    data i /1/

    !-------------------------------------------------------------------
    ! Start:

    if (i .eq. 1) then
      call getMinus5on2Randoms(harvest)
    end if
    minus5on2Random = harvest(i)
    i = i + 1
    if (i .gt. size) i = 1

  end function minus5on2Random

  !---------------------------------------------------------------------
  subroutine getMinus5on2Randoms(numbers)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have a distribution P given by
  ! P(x) = x^(-5/2) for x > 1.0. To get eg luminosities that follow the
  ! same distribution, multiply the returned values by the minimum detectable
  ! luminosity. It is not possible to return a distribution of luminosities
  ! down to zero luminosity because the integral of such a function is
  ! without bound.

    implicit none

    real, intent(out) :: numbers(:)
    real, allocatable :: harvest(:)
    integer :: n
    integer :: number
    real :: minusTwoThirds
    logical :: firstCall
    save minusTwoThirds, firstCall
    data firstCall /.true./

    !-------------------------------------------------------------------
    ! Start:

    if (firstCall) then
      firstCall = .false.
      minusTwoThirds = -2.0 / 3.0
    end if

    number = size(numbers)
    allocate(harvest(number))

    call getRandomNumbers(harvest)
    do n = 1, number
      numbers(n) = (1.0 - harvest(n))**minusTwoThirds ! the subtraction of the
      ! random number from 1.0 is to prevent the occurrence of division by zero.
    end do

  end subroutine getMinus5on2Randoms

  !---------------------------------------------------------------------
  function powerRandom(power)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: power

    real :: powerRandom
    integer, parameter :: size = 1000
    real :: harvest(size)
    integer :: i
    save i, harvest
    data i /1/

    !-------------------------------------------------------------------
    ! Start:

    if (i .eq. 1) then
      call getPowerRandoms(power, harvest)
    end if
    powerRandom = harvest(i)
    i = i + 1
    if (i .gt. size) i = 1

  end function powerRandom

  !---------------------------------------------------------------------
  subroutine getPowerRandoms(power, numbers)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have a distribution P given by
  ! P(x) = x^power. If power < -1, the distribution is given for for x > 1.0;
  ! if power >= 0, the distribution is given for 0 < x <= 1.0.

    implicit none

    real, intent(in) :: power
    real, intent(out) :: numbers(:)

    real, allocatable :: harvest(:)
    integer :: n
    integer :: number
    real :: oneOverPowerPlusOne
    logical :: firstCall
    save oneOverPowerPlusOne, firstCall
    data firstCall /.true./

    !-------------------------------------------------------------------
    ! Start:

    if (firstCall) then
      firstCall = .false.
      oneOverPowerPlusOne = 1.0 / (power + 1.0)
    end if

    number = size(numbers)
    allocate(harvest(number))

    call getRandomNumbers(harvest)
    if (power < -1.0) then
      do n = 1, number
        numbers(n) = (1.0 - harvest(n))**oneOverPowerPlusOne ! the subtraction
        ! of the random number from 1.0 is to prevent the occurrence of
        ! division by zero. 1-rnd may equal 1 but cannot equal 0, hence
        ! the taking of negative powers of zero (ie, dividing by zero)
        ! is avoided.
      end do
    else if (power > -1.0) then
      do n = 1, number
        numbers(n) = harvest(n)**oneOverPowerPlusOne
      end do
    else ! power = -1, don't know how best to deal with this.
      numbers = -1.0
    end if

  end subroutine getPowerRandoms

  !---------------------------------------------------------------------
  subroutine getLogNlogSrandoms(x, yInt, hiPower, xSamples, status)
  !---------------------------------------------------------------------
  ! Note that it was decided not to implement a 'logNlogSrandom'
  ! function for this subroutine because of the need for a status return.
  ! There are too many things which can go wrong with this routine to
  ! allow a foolproof single-value return.
  !.....................................................................
  !
  ! The subroutine returns a vector of random values xSamples() which
  ! have a distribution P which, when integrated, follows a piecewise
  ! power law. The ith segment of the integrated piecewise function is
  ! defined between x(i) and x(i+1) and is of the form
  !
  !	log(yInt) = log(yInt(i)) + slope * (log(x) - log(x(i)))
  !
  ! where
  !	         log(yInt(i+1)) - log(yInt(i))
  !	slope = -------------------------------.
  !	            log(x(i+1)) - log(x(i))
  !
  ! Above x(N) the integrated function is defined by
  !
  !	log(yInt) = log(yInt(N)) + hiPower * (log(x) - log(x(N))).
  !
  ! Notes: (i) The x values must be in increasing order, and they cannot
  ! be negative. (ii) The y values must all be >0, and must never increase
  ! with increasing i. (iii) N must be >0. (iv) hiPower must be <0.
  !
  ! In the event of an error, the subroutine returns values of -1 for all
  ! elements of numbers().
  !.....................................................................

    implicit none

    real,    intent(in)            :: x(:), yInt(size(x)), hiPower
    real,    intent(out)           :: xSamples(:)
    integer, intent(out), optional :: status

    integer ::&
      localStatus, foreignStatus, numBins, numXsamples, i, bin
    real :: harvest(size(xSamples)),&
      reciprocalSlopes(size(x)), negYint(size(x))

    !-------------------------------------------------------------------
    ! Start


    localStatus = 0
    xSamples = -1

    numBins = size(x)
    numXsamples = size(xSamples)

    call doLogNlogSchecks(x, yInt, hiPower, foreignStatus)

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus
  return
    end if

    ! Calculate reciprocal slopes:

    do bin = 1, numBins - 1
      reciprocalSlopes(bin) = (log(x(   bin+1)) - log(x(   bin)))&
                            / (log(yInt(bin+1)) - log(yInt(bin)))
    end do
    reciprocalSlopes(numBins) = 1.0 / hiPower

    ! Main part of the routine:

    call getRandomNumbers(harvest)
    harvest = harvest * yInt(1)

    negYint = -yInt

    do i = 1, numXsamples
      bin = getBinNum(negYint, -harvest(i)) ! in array_utils
      xSamples(i) = x(bin) * (harvest(i) / yInt(bin))**(reciprocalSlopes(bin))
    end do

    if (present(status)) status = localStatus

  end subroutine getLogNlogSrandoms

  !---------------------------------------------------------------------
  subroutine doLogNlogSchecks(x, yInt, hiPower, status)
  !---------------------------------------------------------------------

    implicit none

    real,    intent(in)  :: x(:), yInt(size(x)), hiPower
    integer, intent(out) :: status

    integer :: numPoints

    !-------------------------------------------------------------------
    ! Start:

    numPoints = size(x)

    if (numPoints < 1) then
      status = -1
  return
    end if

    if (numPoints > 1) then ! Check order of x values:
      if (any(x(2:) <= x(:numPoints-1))) then
        status = -2
  return
      end if

      if (any(yInt(2:) > yInt(:numPoints-1))) then
        status = -3
  return
      end if
    end if

    ! Check all x values >= 0:
    if (x(1) < 0.0) then
      status = -4
  return
    end if

    ! Check all yInt values > 0:
    if (any(yInt <= 0.0)) then
      status = -5
  return
    end if

    if (hiPower >= -1.0) then
      status = -6
  return
    end if

  end subroutine doLogNlogSchecks

  !---------------------------------------------------------------------
  subroutine getPowerSegmentRandoms(x, y, powers, numbers&
  , loPower, hiPower, status)
  !---------------------------------------------------------------------
  ! Note that it was decided not to implement a 'powerSegmentRandom'
  ! function for this subroutine because of the need for a status return.
  ! There are too many things which can go wrong with this routine to
  ! allow a foolproof single-value return.
  !.....................................................................
  !
  ! The subroutine returns a vector of random values which have a
  ! distribution P which follows a piecewise power law. The ith segment
  ! of this piecewise function is defined between x(i) and x(i+1) and is
  ! of the form
  !
  !	          y(i)
  !	y = ----------------- x**powers(i).
  !	     x(i)**powers(i)
  !
  ! Notes: (i) The amplitude is tied to the value at the lower x limit.
  ! (ii) There is nothing constraining the pieces of curve to join at the
  ! x boundaries. (iii) The value y(N) (where N=size(x)) is unused unless
  ! hiPower is supplied (see below). (iv) The x values must be in
  ! increasing order, and they cannot be negative - an error status is
  ! returned otherwise. (v) The y values all must be >=0, and at least
  ! 1 must be >0. Error conditions arise otherwise. (vi) Either N > 1
  ! with neither loPower or hiPower, or loPower and/or hiPower are supplied,
  ! in which case N may be >= 1. 
  !
  ! If the value loPower is supplied, P is assumed to continue below x=x(1)
  ! until x=0, with a functional form given by
  !
  !	         y(1)
  !	y = --------------- x**loPower.					(1)
  !	     x(1)**loPower
  !
  ! Notes: (i) This lower continuation is tied to the value on its
  ! _upper_ x boundary, in contrast to all the rest. (ii) loPower must
  ! be > -1. An error status is returned otherwise. (iii) If loPower is
  ! supplied, x(1) must be > 0. If loPower is not supplied, either (a)
  ! x(1) > 0, or (b) x(1) = 0 and power(i) > -1. An error status is
  ! returned if neither is the case.
  !
  ! If hiPower is supplied, P is assumed to continue on above x(N) until
  ! infinity, with a functional form
  !
  !	         y(N)
  !	y = ----------------x**hiPower.
  !	     x(N)**hiPower
  !
  ! Note: hiPower < -1.
  !
  ! In the event of an error, the subroutine returns values of -1 for all
  ! elements of numbers().
  !.....................................................................
  !
  ! This routine works as follows. We want to form a function
  !
  !	R = J(x)/J(inf),						(2)
  !
  ! where J is basically the stepwise integral of P, viz:
  !
  !	                       x
  !	                      /
  !	J(x) = J(x(i)) + A(i) | dx x**powers(i)				(3)
  !	                      /
  !	                     x(i)
  ! where
  !	                        y(i)
  !	A(i) is usually = -----------------,
  !	                   x(i)**powers(i)
  !
  ! with an exceptional form for A_lo, as shown above in equation 1.
  ! Provided that the x, y and powers obey the strictures given earlier,
  ! J(inf) exists and thus R(0) = 0 and R(inf) = 1. To return random
  ! numbers with the distribution P, evenly-distributed randoms are
  ! chosen, each is equated with R, and equations (3) and (2) are then
  ! inverted to give the x values. The inverse equation is given by
  !
  !	    [                           J(inf)
  !	x = [{powers(i) + 1}{R - R(i)} -------- +
  !	    [                            A(i)
  !
  !
  !	                        ]    (       1       )
  !	+ x(i)**{powers(i) + 1})] ** (---------------).
  !	                        ]    ( powers(i) + 1 )
  !
  !.....................................................................

    implicit none

    real,    intent(in)            :: x(:), y(size(x)), powers(size(x)-1)
    real,    intent(out)           :: numbers(:)
    real,    intent(in),  optional :: loPower, hiPower
    integer, intent(out), optional :: status

    integer ::&
      localStatus, foreignStatus, numPoints, numNumbers, i, bin, binEdge,&
      numBins
    real :: harvest(size(numbers)), pPlus1, A(size(x)-1), Alo, Ahi, Jinf
    real, allocatable :: J(:), R(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    numbers = -1

    numPoints = size(x)
    numNumbers = size(numbers)

    call doDiffPowerSegmentChecks(x, y, powers&
      , present(loPower), present(hiPower), foreignStatus)

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus
  return
    end if

    numBins = numPoints - 1

    if (present(loPower)) then
      if (loPower <= -1.0) then
        if (present(status)) status = -10
  return
      end if

      Alo = y(1) / (x(1)**loPower)
      numBins = numBins + 1
    end if

    if (present(hiPower)) then
      if (hiPower >= -1.0) then
        if (present(status)) status = -11
  return
      end if

      Ahi = y(numPoints) / (x(numPoints)**hiPower)
      numBins = numBins + 1
    end if

    ! Calc A values:
    do i = 1, numPoints - 1
      A(i) = y(i) / (x(i)**powers(i))
    end do

    ! Calc total integral Jinf (and check > 0):
    allocate(J(numBins + 1), R(numBins + 1))
    binEdge = 1
    J(binEdge) = 0.0
    if (present(loPower)) then
      binEdge = binEdge + 1
      pPlus1 = loPower + 1.0
      J(binEdge) = J(binEdge-1) + (Alo / pPlus1) * (x(1)**pPlus1)
    end if

    do i = 1, numPoints
      binEdge = binEdge + 1
      pPlus1 = powers(i) + 1.0
      J(binEdge) = J(binEdge-1) + (A(i) / pPlus1)&
        * ((x(i+1)**pPlus1) - (x(i)**pPlus1))
    end do

    if (present(hiPower)) then
      binEdge = binEdge + 1
      pPlus1 = hiPower + 1.0
      J(binEdge) = J(binEdge-1) - (AHi / pPlus1) * (x(i)**pPlus1)

      ! The negative sign here is fine, because pPlus1
      ! must be < 0, thus the whole term ends up positive.
    end if

    Jinf = J(binEdge)
    if (Jinf <= 0.0) then
      if (present(status)) status = -12
  return
    end if

    ! Calc R boundaries (and check (tho should be needless with checks
    ! already done, still I'm a belt and braces man) that they are ordered
    ! in increasing value):

    R = J / Jinf
    if (any(R(2:) - R(:numBins) <= 0.0)) then
      status = -13
  return
    end if

    !...................................................................
    ! Main part of the routine:

    call getRandomNumbers(harvest)

    if (present(loPower)) then
      if (present(hiPower)) then
        do i = 1, numNumbers
          bin = getBinNum(R, harvest(i)) ! in array_utils

          if (bin == 1) then
            pPlus1 = loPower + 1.0
            numbers(i) = (pPlus1 * harvest(i) * Jinf / Alo)**(1.0 / pPlus1)

          else if (bin == numBins) then
            pPlus1 = hiPower + 1.0
            numbers(i) = (pPlus1 * (harvest(i) - R(bin)) * Jinf / Ahi&
              + x(bin-1)**pPlus1)**(1.0 / pPlus1)

          else
            pPlus1 = powers(bin-1) + 1.0
            numbers(i) = (pPlus1 * (harvest(i) - R(bin)) * Jinf / A(bin-1)&
              + x(bin-1)**pPlus1)**(1.0 / pPlus1)
          end if
        end do

      else ! hiPower not present
        do i = 1, numNumbers
          bin = getBinNum(R, harvest(i)) ! in array_utils

          if (bin == 1) then
            pPlus1 = loPower + 1.0
            numbers(i) = (pPlus1 * harvest(i) * Jinf / Alo)**(1.0 / pPlus1)

          else
            pPlus1 = powers(bin-1) + 1.0
            numbers(i) = (pPlus1 * (harvest(i) - R(bin)) * Jinf / A(bin-1)&
              + x(bin-1)**pPlus1)**(1.0 / pPlus1)
          end if
        end do
      end if

    else ! loPower not present
      if (present(hiPower)) then
        do i = 1, numNumbers
          bin = getBinNum(R, harvest(i)) ! in array_utils

          if (bin == numBins) then
            pPlus1 = hiPower + 1.0
            numbers(i) = (pPlus1 * (harvest(i) - R(bin)) * Jinf / Ahi&
              + x(bin)**pPlus1)**(1.0 / pPlus1)

          else
            pPlus1 = powers(bin) + 1.0
            numbers(i) = (pPlus1 * (harvest(i) - R(bin)) * Jinf / A(bin)&
              + x(bin)**pPlus1)**(1.0 / pPlus1)
          end if
        end do

      else ! hiPower not present
        do i = 1, numNumbers
          bin = getBinNum(R, harvest(i)) ! in array_utils

          pPlus1 = powers(bin) + 1.0
          numbers(i) = (pPlus1 * (harvest(i) - R(bin)) * Jinf / A(bin)&
            + x(bin)**pPlus1)**(1.0 / pPlus1)
        end do
      end if
    end if

    if (present(status)) status = localStatus

  end subroutine getPowerSegmentRandoms

  !---------------------------------------------------------------------
  subroutine doDiffPowerSegmentChecks(x, y, powers&
  , presentLoPower, presentHiPower, status)
  !---------------------------------------------------------------------
!*** bit of a waste of space here because int and diff differ only in 2 characters.

    implicit none

    real,    intent(in)  :: x(:), y(size(x)), powers(size(x)-1)
    logical, intent(in)  :: presentLoPower, presentHiPower
    integer, intent(out) :: status

    integer :: numPoints

    !-------------------------------------------------------------------
    ! Start:

    numPoints = size(x)

    if (presentLoPower .or. presentHiPower) then
      if (numPoints < 1) then
        status = -1
  return
      end if

      if (numPoints > 1) then ! Check order of x values:
        if (any(x(2:) <= x(:numPoints-1))) then
          status = -2
  return
        end if
      end if

    else
      if (numPoints < 2) then
        status = -3
  return
      end if

      ! Check order of x values:
      if (any(x(2:) <= x(:numPoints-1))) then
        status = -4
  return
      end if
    end if

    ! Check all x values >= 0:
    if (x(1) < 0.0) then
      status = -5
  return
    end if

    if (presentLoPower) then
      ! Check all x values > 0:
      if (x(1) <= 0.0) then
        status = -6
  return
      end if

    else
      if (powers(1) <= -1.0 .and. x(1) == 0.0) then
        status = -7
  return
      end if
    end if

    ! Check all y values >= 0:
    if (any(y < 0.0)) then
      status = -8
  return
    end if

    ! Check at least 1 y value > 0:
    if (all(y <= 0.0)) then
      status = -9
  return
    end if

  end subroutine doDiffPowerSegmentChecks

  !---------------------------------------------------------------------
  subroutine doIntPowerSegmentChecks(x, y, powers&
  , presentLoPower, presentHiPower, status)
  !---------------------------------------------------------------------

    implicit none

    real,    intent(in)  :: x(:), y(size(x)), powers(size(x)-1)
    logical, intent(in)  :: presentLoPower, presentHiPower
    integer, intent(out) :: status

    integer :: numPoints

    !-------------------------------------------------------------------
    ! Start:

    numPoints = size(x)

    if (presentLoPower .or. presentHiPower) then
      if (numPoints < 1) then
        status = -1
  return
      end if

      if (numPoints > 1) then ! Check order of x values:
        if (any(x(2:) <= x(:numPoints-1))) then
          status = -2
  return
        end if
      end if

    else
      if (numPoints < 2) then
        status = -3
  return
      end if

      ! Check order of x values:
      if (any(x(2:) <= x(:numPoints-1))) then
        status = -4
  return
      end if
    end if

    ! Check all x values >= 0:
    if (x(1) < 0.0) then
      status = -5
  return
    end if

    if (presentLoPower) then
      ! Check all x values > 0:
      if (x(1) <= 0.0) then
        status = -6
  return
      end if

    else
      if (powers(1) <= 0.0 .and. x(1) == 0.0) then
        status = -7
  return
      end if
    end if

    ! Check all y values >= 0:
    if (any(y < 0.0)) then
      status = -8
  return
    end if

    ! Check at least 1 y value > 0:
    if (all(y <= 0.0)) then
      status = -9
  return
    end if

  end subroutine doIntPowerSegmentChecks

!  !---------------------------------------------------------------------
!  function powerSegmentRandom(powers, lowerLimits)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    real, intent(in) :: powers(:), lowerLimits(size(powers))
!
!    real :: powerRandom
!    integer, parameter :: size = 1000
!    real :: harvest(size)
!    integer :: i
!    save i, harvest
!    data i /1/
!
!    !-------------------------------------------------------------------
!    ! Start:
!
!    if (i .eq. 1) then
!      call getPowerSegmentRandoms(powers, lowerLimits, harvest)
!    end if
!    powerSegmentRandom = harvest(i)
!    i = i + 1
!    if (i .gt. size) i = 1
!
!  end function powerSegmentRandom
!
!  !---------------------------------------------------------------------
!  subroutine getPowerSegmentRandoms(powers, lowerLimits, numbers, upperLimit)
!  !---------------------------------------------------------------------
!  ! Returns a vector of values which have a distribution P which follows a piecewise-continuous power law. The ith segment is defined by x limits, the lower being given by lowerLimits(i) and the upper by lowerLimits(i + 1) (except for the maximum i, for which an upper limit may not be necessary, see below) together with the power that applies within that segment. Ie, within each segment P(x) = x^powers(i). P is assumed to be continuous at the segment boundaries.
!  !
!  ! Is an upper limit necessary or not? If power(max_i) >= -1, then yes, because the integral to infinity of such a function does not converge. Otherwise no. The argument upperLimit has therefore been made optional.
!  !
!  ! Note that the lower limit cannot be less than 0. Indeed, for similar reasons as just discussed, if power(1) <= -1, the lower limit cannot even be equal to 0.
!  !
!  ! If either of the last two conditions are breached, the subroutine returns values of -1 for all elements of numbers().
!
!    implicit none
!
!    real, intent(in)           :: powers(:), lowerLimits(size(powers))
!    real, intent(out)          :: numbers(:)
!    real, intent(in), optional :: upperLimit
!
!    real, allocatable :: harvest(:)
!    integer :: n
!    integer :: number, numPowers
!    real :: oneOverPowerPlusOne(size(powers)), localUpperLimit
!    logical :: firstCall
!    save oneOverPowerPlusOne, firstCall
!    data firstCall /.true./
!
!    !-------------------------------------------------------------------
!    ! Start:
!
!    if (firstCall) then
!      firstCall = .false.
!      oneOverPowerPlusOne = 1.0 / (power + 1.0)
!    end if
!
!    numPowers = size(powers)
!    number = size(numbers)
!    allocate(harvest(number))
!
!    call getRandomNumbers(harvest)
!    if (power < -1.0) then
!      do n = 1, number
!        numbers(n) = (1.0 - harvest(n))**oneOverPowerPlusOne ! the subtraction
!        ! of the random number from 1.0 is to prevent the occurrence of
!        ! division by zero. 1-rnd may equal 1 but cannot equal 0, hence
!        ! the taking of negative powers of zero (ie, dividing by zero)
!        ! is avoided.
!      end do
!
!    else if (power > -1.0) then
!      do n = 1, number
!        numbers(n) = harvest(n)**oneOverPowerPlusOne
!      end do
!
!    else ! power = -1, don't know how best to deal with this.
!      numbers = -1.0
!    end if
!
!  end subroutine getPowerSegmentRandoms

  !---------------------------------------------------------------------
  subroutine getSineRandoms(maxX, numbers, status)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have a distribution P given by
  ! P(x) = sine(x) / (1 - cos(maxX)) for 0.0 < x < maxX. This can be used
  ! to get the radius or codeclination values of points which have a smooth
  ! distribution over a conical segment of a sphere of cone angle maxX radians.

    implicit none

    real, intent(in) :: maxX
    real, intent(out) :: numbers(:)
    integer, intent(out), optional :: status

    real, allocatable :: harvest(:)
    integer :: n
    integer :: number
    real :: k
    logical :: firstCall
    save k, firstCall
    data firstCall /.true./

    !-------------------------------------------------------------------
    ! Start:

    if (present(status)) status = 0
    numbers = 0.0 ! default
    if (maxX .gt. PiOverTwo) then
      if (present(status)) status = -1
  return
    end if
    if (maxX .le. 0.0) then
      if (present(status)) status = -2
  return
    end if

    if (firstCall) then
      firstCall = .false.
      k = calcK(maxX)
    end if

    number = size(numbers)
    allocate(harvest(number))

    call getRandomNumbers(harvest)
    do n = 1, number
      numbers(n) = asin(sqrt(k * harvest(n) * (2.0 - k * harvest(n))))
    end do

  end subroutine getSineRandoms

  !---------------------------------------------------------------------
  function calcK(maxX) result(k)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: maxX
    real :: myPrecision, k, sinXonTwo, kOld, kNew

    !-------------------------------------------------------------------
    ! Start:

!print *, '    0000'
    myPrecision = 50.0 ** (-precision(maxX))
!print *, '    myPrecision', myPrecision

    if (maxX .le. 0.0) then
      k = 0.0
  return
    end if

    if (maxX .lt. 0.1) then ! use an iterative expression to get k.
    ! The 0.1 is essentially arbitrary.
!print *, '    maxX', maxX
      sinXonTwo = sin(maxX) * sin(maxX) / 2.0
!print *, '    sinXonTwo', sinXonTwo
      kOld = sinXonTwo
!print *, '    kOld', kOld
      do
!print *, '    5555'
        kNew = sinXonTwo + kOld * kOld
!print *, '    kNew', kNew
!print *, '    (kNew - kOld) / kOld', (kNew - kOld) / kOld
      if ((kNew - kOld) / kOld .le. myPrecision) exit
!print *, '    7777'
        kOld = kNew
!print *, '    kOld', kOld
      end do
!print *, '    9999'
      k = kOld
!print *, '    k', k

    else
!print *, '    0202'
      k = 1.0 - sqrt(1.0 - sin(maxX) * sin(maxX))
!print *, '    k', k
    end if
!print *, '    0404'

  end function calcK

  !---------------------------------------------------------------------
  function exponentialRandom()
  !---------------------------------------------------------------------

    implicit none

    real :: exponentialRandom
    integer, parameter :: size = 1000
    real :: harvest(size)
    integer :: i
    save i, harvest
    data i /1/

    !-------------------------------------------------------------------
    ! Start:

    if (i .eq. 1) then
      call getExponentialRandoms(harvest)
    end if
    exponentialRandom = harvest(i)
    i = i + 1
    if (i .gt. size) i = 1

  end function exponentialRandom

  !---------------------------------------------------------------------
  subroutine getExponentialRandoms(numbers)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have an exponential probability
  ! distribution with a mean value of 1.0.

    implicit none

    real, intent(out) :: numbers(:)
    real, allocatable :: harvest(:)
    integer :: n
    integer :: number

    !-------------------------------------------------------------------
    ! Start:

    number = size(numbers)
    allocate(harvest(number))

    call getRandomNumbers(harvest)
    do n = 1, number
      numbers(n) = -log(1.0 - harvest(n)) ! the subtraction of the
      ! random number from 1.0 is to prevent the occurrence of zero-
      ! valued argument to log().
    end do

  end subroutine getExponentialRandoms

  !---------------------------------------------------------------------
  function gaussianRandom()
  !---------------------------------------------------------------------

    implicit none

    real :: gaussianRandom
    integer, parameter :: size = 1000
    real :: harvest(size)
    integer :: i
    save i, harvest
    data i /1/

    !-------------------------------------------------------------------
    ! Start:

    if (i .eq. 1) then
      call getGaussianRandoms(harvest)
    end if
    gaussianRandom = harvest(i)
    i = i + 1
    if (i .gt. size) i = 1

  end function gaussianRandom

  !---------------------------------------------------------------------
  subroutine getGaussianRandoms(numbers)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have an gaussian probability
  ! distribution.

    implicit none

    real, intent(out) :: numbers(:)
!    real, allocatable :: harvest(:)
    integer :: n, number
    real fac, rsq, v1, v2

    !-------------------------------------------------------------------
    ! Start:

    number = size(numbers)
!    allocate(harvest(number))

!    call getRandomNumbers(harvest)

    if (number / 2.0 .eq. floor(number / 2.0)) then ! even
      do n = 1, number - 1, 2
        do
          v1 = 2.0 * randomNumber() - 1.0
          v2 = 2.0 * randomNumber() - 1.0
          rsq = v1 * v1 + v2 * v2

        if (rsq .gt. 0.0 .and. rsq .lt. 1.0) exit
        end do

        fac = sqrt(-2.0 * log(rsq) / rsq)
        numbers(n)     = v1 * fac
!print *, fac, v1, n, numbers(n)
        numbers(n + 1) = v2 * fac
!print *, fac, v2, n + 1, numbers(n + 1)
      end do

    else ! odd
      do n = 1, number - 2, 2
        do
          v1 = 2.0 * randomNumber() - 1.0
          v2 = 2.0 * randomNumber() - 1.0
          rsq = v1 * v1 + v2 * v2

        if (rsq .gt. 0.0 .and. rsq .lt. 1.0) exit
        end do

        fac = sqrt(-2.0 * log(rsq) / rsq)
        numbers(n)     = v1 * fac
        numbers(n + 1) = v2 * fac
      end do

      do
        v1 = 2.0 * randomNumber() - 1.0
        v2 = 2.0 * randomNumber() - 1.0
        rsq = v1 * v1 + v2 * v2

      if (rsq .gt. 0.0 .and. rsq .lt. 1.0) exit
      end do

      fac = sqrt(-2.0 * log(rsq) / rsq)
      numbers(number) = v1 * fac
    end if

  end subroutine getGaussianRandoms

  !---------------------------------------------------------------------
  function randoms(number)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: number
    real :: randoms(number)

    !-------------------------------------------------------------------
    ! Start

    call getRandomNumbers(randoms)

  end function randoms

  !---------------------------------------------------------------------
  function makeRandomSequence(duration, rate) result(randomSequence)
  !---------------------------------------------------------------------
  ! This function returns a random sequence of 'times' which occur at an
  ! average rate and for a duration given in the function arguments.
  !
  ! The difficulty in implementing this function lies in the fact that
  ! the final size of the returned sequence is not known until the last
  ! time is calculated (at least, with the present algorithm). So the amount
  ! of memory needed cannot be allocated a priori. This is partially
  ! overcome by constructing an array of pointers (called here 'chunk')
  ! and allocating memory to each component of this array as it becomes
  ! necessary. The values are then copied in sequence from the allocated
  ! components into the returned pointer.

    implicit none

    real, intent(in) :: duration, rate
    real(kind(1d0)), pointer :: randomSequence(:)

    type :: bufferType
      real(kind(1d0)), pointer :: buffer(:)
    end type
    integer, parameter :: maxNumChunks = 20
    type(bufferType) :: chunk(maxNumChunks)
    integer :: bufferSize, chunkNum, n, lastFullChunk, start, finis
    real(kind(1d0)) :: time, timeIncrement

    !-------------------------------------------------------------------
    ! Start

    bufferSize = min(1048576, max(1024, nint(duration * rate)))

    chunkNum = 1
    allocate(chunk(chunkNum)%buffer(bufferSize))
    n = 0
    time = 0.0d0
    do
      timeIncrement = exponentialRandom() / rate
      time = time + timeIncrement
    if (time > duration) exit

      n = n + 1
      if (n > bufferSize) then
        chunkNum = chunkNum + 1
        if (chunkNum > maxNumChunks) then
          n = 0
    exit
        end if

        allocate(chunk(chunkNum)%buffer(bufferSize))
        n = 1
      end if

      chunk(chunkNum)%buffer(n) = time
    end do

    lastFullChunk = chunkNum - 1

    allocate(randomSequence(lastFullChunk * bufferSize + n))
    start = 1
    finis = bufferSize
    do chunkNum = 1, lastFullChunk
      randomSequence(start:finis) = chunk(chunkNum)%buffer
      start = start + bufferSize
      finis = finis + bufferSize
      deallocate(chunk(chunkNum)%buffer)
    end do
    if (n > 0) then
      randomSequence(start:start + n - 1) = chunk(lastFullChunk + 1)%buffer(:n)
      deallocate(chunk(lastFullChunk + 1)%buffer)
    end if

  end function makeRandomSequence

  !---------------------------------------------------------------------
  function randomNumber()
  !---------------------------------------------------------------------

    implicit none

    real :: randomNumber
    integer, parameter :: size = 1000
    real :: harvest(size)
    integer :: i
    save i, harvest
    data i /1/

    !-------------------------------------------------------------------
    ! Start:

    if (i .eq. 1) then
      call getRandomNumbers(harvest)
    end if
    randomNumber = harvest(i)
    i = i + 1
    if (i .gt. size) i = 1

  end function randomNumber

  !---------------------------------------------------------------------
  subroutine getRandomNumbers(numbers)
  !---------------------------------------------------------------------

    implicit none

    real, intent(out) :: numbers(:)

    !-------------------------------------------------------------------
    ! Start:

   call random_number(numbers) ! f90 intrinsic routine. Probably not very good.
!    call SAS_rand(numbers) ! in utils. Not yet available in rel track (IMS 2002-12-05).

  end subroutine getRandomNumbers

  !---------------------------------------------------------------------
  real function myran1(rseed)
  !---------------------------------------------------------------------
!**** from evarimgen. Doesn't work yet and isn't used.

    implicit none

    integer rseed, ia, im, iq, ir, ntab, ndiv
    real am, eps, rnmx
    parameter (ia = 16807,&
     		   im = 2147483647,&
     		   am = 1./im,&
     		   iq = 127773,&
     		   ir = 2836,&
     		 ntab = 32,&
     		 ndiv = 1 + (im - 1) / ntab,&
     		  eps = 1.2e-7,&
     		 rnmx = 1. - eps)
    integer j, k, iv(ntab), iy
    save iv, iy
    data iv /ntab * 0/, iy /0/

    if (rseed.le.0.or.iy.eq.0) then ! initialise:
      rseed = max(-rseed, 1) ! Prevents rseed = 0.
      do j = ntab + 8, 1, -1
        k = rseed / iq
        rseed = ia * (rseed - k * iq) - ir * k
        if (rseed.lt.0) rseed = rseed + im
        if (j.le.ntab) iv(j) = rseed
      end do
      iy = iv(1)
    end if

    k = rseed / iq
    rseed = ia * (rseed - k * iq) - ir * k
    if (rseed.lt.0) rseed = rseed + im
    j = 1 + iy / ndiv
    iy = iv(j)
    iv(j) = rseed
    myran1 = min(am * iy, rnmx)

  end function myran1

  !---------------------------------------------------------------------
  logical function coinIsHeads()
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    if (randomNumber() > 0.5) then
      coinIsHeads = .true.
    else
      coinIsHeads = .false.
    end if

  end function coinIsHeads

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module random
