!
! Module: wcs_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Returns semi-axes of an ellipse round each of a list of sources.
! 
! Comments: TODO:
! 
! $Id: wcs_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module wcs_aux

  use dal
  use errorHandling ! not needed now but should replace 'noisy' messages.

  use constants
  use coordinate ! for tanToRaDec.
  use fits_utils ! contains subs getRealKwd(), getRealKwd(), getStringKwd().

  implicit none 

  private 

  public getWcsInfo
  public easifyWcsType
  public wcsAreSame
  public dumpWcs

public getWcsInfoSourceListSet

  interface getWcsInfo
    module procedure getWcsInfo_image, getWcsInfoArray
    module procedure getWcsInfoSet, getWcsInfoSetName, getWcsInfoSetArrayName
    module procedure getWcsInfo_table
  end interface

  interface easifyWcsType
    module procedure easifyWcsType_scalar
    module procedure easifyWcsType_vector
  end interface

  type, public :: wcsAxisType
!    integer(int32) ::&
!      refPixels
    character(len=8) ::&
      ctype
    real(double) ::&
      refPixels,&
      pixelSizeWorld,&
      refWorld
    integer(int8) ::&
      pixelSizeSign
  end type wcsAxisType

  type, public :: wcsDataType
    type(wcsAxisType) :: x, y
  end type wcsDataType

contains 

  !---------------------------------------------------------------------
  subroutine dumpWcs(wcs, verbosityArgument)
  !---------------------------------------------------------------------

    implicit none

    type(wcsDataType), intent(in)           :: wcs
    integer(int32),    intent(in), optional :: verbosityArgument

    integer(int32) :: verbosity

    !-------------------------------------------------------------------
    ! Start

    if (present(verbosityArgument)) then
      verbosity = verbosityArgument
    else
      verbosity = sparse
    end if

    call message(AppMsg, verbosity, 'WCS dump:')
    call message(AppMsg, verbosity, '  X axis:')
    call message(AppMsg, verbosity, '    CTYPE: '//wcs%x%ctype)
    write(errstr, '(a, es23.16)') '    CRVAL:', wcs%x%refWorld
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)') '    CRPIX:', wcs%x%refPixels
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)') '    CDELT:', wcs%x%pixelSizeWorld
    call message(AppMsg, verbosity)

    call message(AppMsg, verbosity, '  Y axis:')
    call message(AppMsg, verbosity, '    CTYPE: '//wcs%y%ctype)
    write(errstr, '(a, es23.16)') '    CRVAL:', wcs%y%refWorld
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)') '    CRPIX:', wcs%y%refPixels
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)') '    CDELT:', wcs%y%pixelSizeWorld
    call message(AppMsg, verbosity)
!print *, wcs%y%pixelSizeWorld

  end subroutine dumpWcs

  !---------------------------------------------------------------------
  logical function wcsAreSame(wcs1, wcs2)
  !---------------------------------------------------------------------

    implicit none

    type(wcsDataType), intent(in) :: wcs1, wcs2

    !-------------------------------------------------------------------
    ! Start

    wcsAreSame = .false. ! default

  if (wcs1%x%ctype          .ne. wcs2%x%ctype         ) return
  if (wcs1%x%refPixels      .ne. wcs2%x%refPixels     ) return
  if (wcs1%x%pixelSizeWorld .ne. wcs2%x%pixelSizeWorld) return
  if (wcs1%x%refWorld       .ne. wcs2%x%refWorld      ) return

  if (wcs1%y%ctype          .ne. wcs2%y%ctype         ) return
  if (wcs1%y%refPixels      .ne. wcs2%y%refPixels     ) return
  if (wcs1%y%pixelSizeWorld .ne. wcs2%y%pixelSizeWorld) return
  if (wcs1%y%refWorld       .ne. wcs2%y%refWorld      ) return

    wcsAreSame = .true.

  end function wcsAreSame

  !---------------------------------------------------------------------
  subroutine getWcsInfoSetName(setName, wcs)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*),      intent(in)  :: setName
    type(wcsDataType), intent(out) :: wcs

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, 0, READ)

    call getRealKwd("CRPIX1",   set, inArray, wcs%X%refPixels)
    call getRealKwd("CRPIX2",   set, inArray, wcs%Y%refPixels)
    call getRealKwd("CDELT1",   set, inArray, wcs%X%pixelSizeWorld)
    call getRealKwd("CDELT2",   set, inArray, wcs%Y%pixelSizeWorld)
    call getRealKwd("CRVAL1",   set, inArray, wcs%X%refWorld)
    call getRealKwd("CRVAL2",   set, inArray, wcs%Y%refWorld)
!print *, 'aaa'
    call getStringKwd("CTYPE1", set, inArray, wcs%X%ctype)
!print *, 'bbb'
    call getStringKwd("CTYPE2", set, inArray, wcs%Y%ctype)
!print *, 'zzz'

    call release(set)

  end subroutine getWcsInfoSetName

  !---------------------------------------------------------------------
  subroutine getWcsInfoSetArrayName(setName, arrayName, wcs)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*),      intent(in)  :: setName, arrayName
    type(wcsDataType), intent(out) :: wcs

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, arrayName, READ)

    call getRealKwd("CRPIX1",   set, inArray, wcs%X%refPixels)
    call getRealKwd("CRPIX2",   set, inArray, wcs%Y%refPixels)
    call getRealKwd("CDELT1",   set, inArray, wcs%X%pixelSizeWorld)
    call getRealKwd("CDELT2",   set, inArray, wcs%Y%pixelSizeWorld)
    call getRealKwd("CRVAL1",   set, inArray, wcs%X%refWorld)
    call getRealKwd("CRVAL2",   set, inArray, wcs%Y%refWorld)
    call getStringKwd("CTYPE1", set, inArray, wcs%X%ctype)
    call getStringKwd("CTYPE2", set, inArray, wcs%Y%ctype)

    call release(set)

  end subroutine getWcsInfoSetArrayName

  !---------------------------------------------------------------------
  subroutine getWcsInfo_image(set, inArray, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set
    type(ArrayT),   intent(in) :: inArray
    type(wcsDataType), intent(out) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    call getRealKwd("CRPIX1",   set, inArray, wcs%X%refPixels)
    call getRealKwd("CRPIX2",   set, inArray, wcs%Y%refPixels)
    call getRealKwd("CDELT1",   set, inArray, wcs%X%pixelSizeWorld)
    call getRealKwd("CDELT2",   set, inArray, wcs%Y%pixelSizeWorld)
    call getRealKwd("CRVAL1",   set, inArray, wcs%X%refWorld)
    call getRealKwd("CRVAL2",   set, inArray, wcs%Y%refWorld)
    call getStringKwd("CTYPE1", set, inArray, wcs%X%ctype)
    call getStringKwd("CTYPE2", set, inArray, wcs%Y%ctype)

  end subroutine getWcsInfo_image

  !---------------------------------------------------------------------
  subroutine getWcsInfoArray(inArray, wcs, status)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT),   intent(in) :: inArray
    type(wcsDataType), intent(out) :: wcs
    integer,           intent(out), optional :: status

    integer :: localStatus

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    if (.not. (hasAttribute(inArray, 'CRPIX1')&
         .and. hasAttribute(inArray, 'CRPIX2')&
         .and. hasAttribute(inArray, 'CDELT1')&
         .and. hasAttribute(inArray, 'CDELT2')&
         .and. hasAttribute(inArray, 'CRVAL1')&
         .and. hasAttribute(inArray, 'CRVAL2')&
         .and. hasAttribute(inArray, 'CTYPE1')&
         .and. hasAttribute(inArray, 'CTYPE2'))) then
      if(present(status)) status = -1

      wcs%X%refPixels      = 0.0
      wcs%Y%refPixels      = 0.0
      wcs%X%pixelSizeWorld = 0.0
      wcs%Y%pixelSizeWorld = 0.0
      wcs%X%refWorld       = 0.0
      wcs%Y%refWorld       = 0.0
      wcs%X%ctype          = ''
      wcs%Y%ctype          = ''

  return
    end if

    wcs%X%refPixels      = real32Attribute(inArray, 'CRPIX1')
    wcs%Y%refPixels      = real32Attribute(inArray, 'CRPIX2')
    wcs%X%pixelSizeWorld = real32Attribute(inArray, 'CDELT1')
    wcs%Y%pixelSizeWorld = real32Attribute(inArray, 'CDELT2')
    wcs%X%refWorld       = real32Attribute(inArray, 'CRVAL1')
    wcs%Y%refWorld       = real32Attribute(inArray, 'CRVAL2')
    wcs%X%ctype          = stringAttribute(inArray, 'CTYPE1')
    wcs%Y%ctype          = stringAttribute(inArray, 'CTYPE2')

    if(present(status)) status = localStatus

  end subroutine getWcsInfoArray

  !---------------------------------------------------------------------
  subroutine getWcsInfoArray_old(inArray, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT),   intent(in) :: inArray
    type(wcsDataType), intent(out) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    wcs%X%refPixels      = real32Attribute(inArray, 'CRPIX1')
    wcs%Y%refPixels      = real32Attribute(inArray, 'CRPIX2')
    wcs%X%pixelSizeWorld = real32Attribute(inArray, 'CDELT1')
    wcs%Y%pixelSizeWorld = real32Attribute(inArray, 'CDELT2')
    wcs%X%refWorld       = real32Attribute(inArray, 'CRVAL1')
    wcs%Y%refWorld       = real32Attribute(inArray, 'CRVAL2')
    wcs%X%ctype          = stringAttribute(inArray, 'CTYPE1')
    wcs%Y%ctype          = stringAttribute(inArray, 'CTYPE2')

  end subroutine getWcsInfoArray_old

  !---------------------------------------------------------------------
  subroutine getWcsInfoSet(set, wcs, status)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT),    intent(in)            :: set
    type(wcsDataType), intent(out)           :: wcs
    integer,           intent(out), optional :: status

    integer :: localStatus

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    if (.not. (hasAttribute(set, 'CRPIX1')&
         .and. hasAttribute(set, 'CRPIX2')&
         .and. hasAttribute(set, 'CDELT1')&
         .and. hasAttribute(set, 'CDELT2')&
         .and. hasAttribute(set, 'CRVAL1')&
         .and. hasAttribute(set, 'CRVAL2')&
         .and. hasAttribute(set, 'CTYPE1')&
         .and. hasAttribute(set, 'CTYPE2'))) then
      if(present(status)) status = -1

      wcs%X%refPixels      = 0.0
      wcs%Y%refPixels      = 0.0
      wcs%X%pixelSizeWorld = 0.0
      wcs%Y%pixelSizeWorld = 0.0
      wcs%X%refWorld       = 0.0
      wcs%Y%refWorld       = 0.0
      wcs%X%ctype          = ''
      wcs%Y%ctype          = ''

  return
    end if

    wcs%X%refPixels      = real32Attribute(set, 'CRPIX1')
    wcs%Y%refPixels      = real32Attribute(set, 'CRPIX2')
    wcs%X%pixelSizeWorld = real32Attribute(set, 'CDELT1')
    wcs%Y%pixelSizeWorld = real32Attribute(set, 'CDELT2')
    wcs%X%refWorld       = real32Attribute(set, 'CRVAL1')
    wcs%Y%refWorld       = real32Attribute(set, 'CRVAL2')
    wcs%X%ctype          = stringAttribute(set, 'CTYPE1')
    wcs%Y%ctype          = stringAttribute(set, 'CTYPE2')

    if(present(status)) status = localStatus

  end subroutine getWcsInfoSet

  !---------------------------------------------------------------------
  subroutine getWcsInfo_set_old(set, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT),    intent(in)            :: set
    type(wcsDataType), intent(out)           :: wcs

    !-------------------------------------------------------------------
    ! Start:

    wcs%X%refPixels      = real32Attribute(set, 'CRPIX1')
    wcs%Y%refPixels      = real32Attribute(set, 'CRPIX2')
    wcs%X%pixelSizeWorld = real32Attribute(set, 'CDELT1')
    wcs%Y%pixelSizeWorld = real32Attribute(set, 'CDELT2')
    wcs%X%refWorld       = real32Attribute(set, 'CRVAL1')
    wcs%Y%refWorld       = real32Attribute(set, 'CRVAL2')
    wcs%X%ctype          = stringAttribute(set, 'CTYPE1')
    wcs%Y%ctype          = stringAttribute(set, 'CTYPE2')

  end subroutine getWcsInfo_set_old

  !---------------------------------------------------------------------
  subroutine getWcsInfo_table(tab, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    type(wcsDataType), intent(out) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    wcs%X%refPixels      = real32Attribute(tab, 'CRPIX1')
    wcs%Y%refPixels      = real32Attribute(tab, 'CRPIX2')
    wcs%X%pixelSizeWorld = real32Attribute(tab, 'CDELT1')
    wcs%Y%pixelSizeWorld = real32Attribute(tab, 'CDELT2')
    wcs%X%refWorld       = real32Attribute(tab, 'CRVAL1')
    wcs%Y%refWorld       = real32Attribute(tab, 'CRVAL2')
    wcs%X%ctype          = stringAttribute(tab, 'CTYPE1')
    wcs%Y%ctype          = stringAttribute(tab, 'CTYPE2')

  end subroutine getWcsInfo_table

  !---------------------------------------------------------------------
  subroutine getWcsInfoSourceListSet(setName, wcs)
  !---------------------------------------------------------------------
!*** These kwds should surely be stored as proper column kwds?

    implicit none

    character(*), intent(in) :: setName
    type(wcsDataType), intent(out) :: wcs

    type(DataSetT) :: set

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)

    wcs%X%refPixels      = real32Attribute(set, 'TCRPIX1')
    wcs%Y%refPixels      = real32Attribute(set, 'TCRPIX2')
    wcs%X%pixelSizeWorld = real32Attribute(set, 'TCDELT1')
    wcs%Y%pixelSizeWorld = real32Attribute(set, 'TCDELT2')
    wcs%X%refWorld       = real32Attribute(set, 'TCRVAL1')
    wcs%Y%refWorld       = real32Attribute(set, 'TCRVAL2')
    wcs%X%ctype          = stringAttribute(set, 'TCTYPE1')
    wcs%Y%ctype          = stringAttribute(set, 'TCTYPE2')

    call release(set)

  end subroutine getWcsInfoSourceListSet

  !---------------------------------------------------------------------
  subroutine easifyWcsType_scalar(ctypeX, ctypeY, wcsType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: ctypeX, ctypeY
    character(8), intent(out) :: wcsType

    !-------------------------------------------------------------------
    ! Start:

    if (  trim(ctypeX) .eq. 'DETX' .and. trim(ctypeY) .eq. 'DETY') then
      wcsType = 'DETXY'

    else&
    if (trim(ctypeX).eq.'RA---TAN' .and. trim(ctypeY) .eq. 'DEC--TAN') then
      wcsType = 'XY'
    end if

  end subroutine easifyWcsType_scalar

  !---------------------------------------------------------------------
  subroutine easifyWcsType_vector(ctype, wcsType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: ctype(2)
    character(8), intent(out) :: wcsType

    !-------------------------------------------------------------------
    ! Start:

    if (  trim(ctype(1)) .eq. 'DETX' .and. trim(ctype(2)) .eq. 'DETY') then
      wcsType = 'DETXY'

    else&
    if (trim(ctype(1)).eq.'RA---TAN' .and. trim(ctype(2)) .eq. 'DEC--TAN') then
      wcsType = 'XY'
    end if

  end subroutine easifyWcsType_vector

  !---------------------------------------------------------------------
  function pixelsToTan(pixelValue, wcsAxis) result(tangentValue)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: pixelValue
    type(wcsAxisType), intent(in) :: wcsAxis
    real(single) :: tangentValue

    !-------------------------------------------------------------------
    ! Start

    tangentValue = (pixelValue - wcsAxis%refPixels)&
!    * wcsAxis%pixelSizeWorld * PiOn180
    * abs(wcsAxis%pixelSizeWorld) * PiOn180
!*** NOTE!! Use of the abs() here defines the tan system as always increasing with pixel index in both X and Y directions. This is NOT consistent with other uses of it in this library. At some point the rest should be brought into line.

! ...actually I would prefer also to define pixelSizeWorld always as positive.

  end function pixelsToTan

  !---------------------------------------------------------------------
  subroutine pixelXYtoRaDec(wcs, xi, yStart, yFinis&
  , raDegVector, decDegVector)
  !---------------------------------------------------------------------

    implicit none

    type(wcsDataType), intent(in) :: wcs
    integer(int16), intent(in) :: xi, yStart, yFinis
    real(double), intent(out) ::&
      raDegVector(yStart:yFinis), decDegVector(yStart:yFinis)

    integer(int16) :: yi
    real(double) ::&
      tanX(yStart:yFinis), tanY(yStart:yFinis)

    !-------------------------------------------------------------------
    ! Start

    tanX = (xi - 0.5 - wcs%x%refPixels) * abs(wcs%x%pixelSizeWorld) * PiOn180
    do yi = yStart, yFinis
      tanY(yi) = (yi - 0.5 - wcs%y%refPixels) * abs(wcs%y%pixelSizeWorld) * PiOn180
!print *, yi, tanX(yi), tanY(yi)
    end do
!print *, ''

    call tanToRaDec(wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180&
    , tanX, tanY, raDegVector, decDegVector)

    raDegVector  = raDegVector  / PiOn180
    decDegVector = decDegVector / PiOn180
    do yi = yStart, yFinis
!print *, yi, raDegVector(yi), decDegVector(yi)
    end do
!print *, ''

!print *, 'wcs%x%refPixels:',      wcs%x%refPixels
!print *, 'wcs%x%pixelSizeWorld:', wcs%x%pixelSizeWorld
!print *, 'wcs%x%refWorld:',       wcs%x%refWorld
!print *, 'wcs%y%refPixels:',      wcs%y%refPixels
!print *, 'wcs%y%pixelSizeWorld:', wcs%y%pixelSizeWorld
!print *, 'wcs%y%refWorld:',       wcs%y%refWorld

!call error('endPixelXYtoRaDec','')
  end subroutine pixelXYtoRaDec

!  !---------------------------------------------------------------------
!  subroutine worldToPixelCoords_scalar(wcs, worldvalue, pixelValue)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    type(wcsDataType),   intent(in)  :: wcs
!    type(srcRegionType), intent(in)  :: worldvalue
!    type(srcRegionType), intent(out) :: pixelValue
!
!    type(srcRegionType) :: worldvalueVector(1)
!    type(srcRegionType) :: pixelValueVector(1)
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    worldvalueVector(1) = worldvalue
!
!    call worldToPixelCoords_vector(wcs, worldvalueVector, pixelValueVector)
!
!    pixelValue = pixelValueVector(1)
!
!  end subroutine worldToPixelCoords_scalar
!
!  !---------------------------------------------------------------------
!  subroutine worldToPixelCoords_vector(wcs, srcDataWorld, srcDataPixels)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    type(wcsDataType),   intent(in)  :: wcs
!    type(srcRegionType), intent(in)  :: srcDataWorld(:)
!    type(srcRegionType), intent(out) :: srcDataPixels(:)
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    call message(AppMsg, Noisy, 'ssclib: convert world to pixel coords.')
!
!    srcDataPixels%x = (srcDataWorld%x - wcs%X%refWorld)&
!    / wcs%X%pixelSizeWorld + wcs%X%refPixels
!    srcDataPixels%y = (srcDataWorld%y - wcs%Y%refWorld)&
!    / wcs%Y%pixelSizeWorld + wcs%Y%refPixels
!!*** add 0.5 later for plotting purposes? (not for the region???)
!
!  end subroutine worldToPixelCoords_vector

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module wcs_aux
