!
! Module: eimsim_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This is a collection of routines used to make simulated epic images.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eimsim_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.

  use dal
  use param  

  use eimsim_utils
  use eimsim_reduction
  use dal_aux		! for addOrOpenColumn().
  use array_utils	! for smallestTrueIndex()
  use ml_corr
  use flux_transforms	! for calcEcf

  implicit none 

  private 

  public eimsim

contains 

  !---------------------------------------------------------------------
  subroutine eimsim
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    select case(stringParameter('function'))
!    case('makeecf')
!      call makeEcfFile_interface
    case('skytemplate')				! eimsim_prepare
      call makeLocationTemplate_interface
    case('mask')				! eimsim_prepare
      call makePaddedMask_interface
    case('makebkgmap')				! eimsim_prepare
      call makeBkgMap_interface
    case('makefaintsrcbkgmap')			! nothing at present
      call makeMapFaintSrcBkg_interface

    ! eimsim_single calls:
    case('srclist')				! eimsim_single
      call makeSimSrcList_interface
    case('addcols')				! eimsim_single
      call addColumns				! eimsim_e5detect
    case('image')				! eimsim_single
      call makeRateImage_interface
    case('makectsimg')				! eimsim_single
      call makeAvCountsImage_interface
    case('poissonize')				! eimsim_single
      call poissonizeCountsImage_interface
    case('correct')				! eimsim_single
      call correctPosAndRates_interface
    case('eddington')				! eimsim_single
      call correctEddington_interface
    case('ratetofluxrand')			! eimsim_single
      call detRateToFluxRandom_interface
    case('compare')				! eimsim_single
      call compareDetWithSim_interface
    case('addsimratecols')			! eimsim_single
      call addSimRates_interface

    case('syserror')				! eimsim_finalize
      call calcSysErrors
    case('simhisto')				! eimsim_finalize
      call makeMergedSimSrcHistogram
    case('completeness1')			! eimsim_finalize
      call calcCompletenessHistogramsPart1
    case('completeness2')			! eimsim_finalize
      call calcCompletenessHistogramsPart2
    case('reliability1')			! eimsim_finalize
      call calcReliabilityHistogramPart1
    case('reliability2')			! eimsim_finalize
      call calcReliabilityHistogramPart2
    case('mlcorr')				! eimsim_finalize
      call correctAllBandMLValues
    case('nosources')				! eimsim_poisstest
						! eimsim_bkg_batch
						! eimsim_emltest
      call eventizeBlankImage_interface
    case default
call error('badFunctionValue','The value of the --function parameter &
&was not recognized.')
    end select

  end subroutine eimsim
  
!  !---------------------------------------------------------------------
!  subroutine makeEcfFile_interface
!  !---------------------------------------------------------------------
!  ! function=makeecf
!
!    implicit none
!
!    character(255) :: templateSetName, ecfFileName
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    ecfFileName = stringParameter('ecffile')
!    templateSetName = stringParameter('templateset')
!
!    call makeEcfFile(ecfFileName, templateSetName)
!
!  end subroutine makeEcfFile_interface

  !---------------------------------------------------------------------
  subroutine makeLocationTemplate_interface
  !---------------------------------------------------------------------
  ! function=skytemplate

    implicit none

    character(255), allocatable :: expMapNames(:)
    character(255) :: templateSetName
    real(double) :: padSizeArcsec
    integer(int16) :: expMapNum, numExpMaps

    !-------------------------------------------------------------------
    ! Start

    numExpMaps = parameterCount('insets')
    allocate(expMapNames(numExpMaps))
    do expMapNum = 1, numExpMaps
      expMapNames(expMapNum) = stringParameter('insets', expMapNum - 1)
    end do
    templateSetName = stringParameter('templateset')
    padSizeArcsec = realParameter('padsizearcsec')

    call makeLocationTemplate(expMapNames, templateSetName, padSizeArcsec)

  end subroutine makeLocationTemplate_interface

  !---------------------------------------------------------------------
  subroutine makePaddedMask_interface
  !---------------------------------------------------------------------
  ! function=mask

    implicit none

    character(255) :: expMapSetName, paddedMaskSetName, tempSetName
    real(double) :: padSizeArcsec

    !-------------------------------------------------------------------
    ! Start

    expMapSetName     = stringParameter('expmapset')
    paddedMaskSetName = stringParameter('paddedmaskset')
    tempSetName = stringParameter('tempset')
    padSizeArcsec = realParameter('padsizearcsec')

    call makePaddedMask(expMapSetName, paddedMaskSetName, padSizeArcsec&
      , tempSetName)

  end subroutine makePaddedMask_interface

  !---------------------------------------------------------------------
  subroutine makeBkgMap_interface
  !---------------------------------------------------------------------
  ! function=makebkgmap

    implicit none

    character(255) ::&
      expMapSetName, nonVigSetName, bkgMapSetName, tempImageName
    real(double) :: vigBkgFlux, nonVigBkgFlux

    !-------------------------------------------------------------------
    ! Start

    expMapSetName      = stringParameter('expmapset')
    nonVigSetName      = stringParameter('nonvigset')
    bkgMapSetName      = stringParameter('bkgmapset')
    tempImageName      = stringParameter('tempset')

    vigBkgFlux    = realParameter('vigbkgrate')
    nonVigBkgFlux = realParameter('nonvigrate')
    ! both in counts / arcsec^2 / s.

    call makeBkgMap(expMapSetName, nonVigSetName, bkgMapSetName&
      , tempImageName, vigBkgFlux, nonVigBkgFlux)

  end subroutine makeBkgMap_interface

  !---------------------------------------------------------------------
  subroutine makeMapFaintSrcBkg_interface
  !---------------------------------------------------------------------
  ! function=makefaintsrcbkgmap

    implicit none

    character(255) ::&
      sensitivityMapName, templateSetName, expMapSetName,&
      faintSrcBkgMapName, tempImageName
    real(double) :: EeVlo, EeVhi, ecf

    !-------------------------------------------------------------------
    ! Start

    sensitivityMapName = stringParameter('sensmapset')
    templateSetName    = stringParameter('templateset')
    expMapSetName      = stringParameter('expmapset')
    faintSrcBkgMapName = stringParameter('faintsrcbkgset')
    tempImageName      = stringParameter('tempset')

    EeVlo = realParameter('elo')
    EeVhi = realParameter('ehi')
    ecf   = realParameter('ecf')

    call makeMapFaintSrcBkg(sensitivityMapName, templateSetName&
      , expMapSetName, EeVlo, EeVhi, ecf, faintSrcBkgMapName, tempImageName)

  end subroutine makeMapFaintSrcBkg_interface




  ! eimsim_single calls:

  !---------------------------------------------------------------------
  subroutine makeSimSrcList_interface
  !---------------------------------------------------------------------
  ! function=srclist

    implicit none

    character(255) :: templateSetName, sourceListSetName
    character(10), allocatable :: obsIdList(:)
    integer :: numObsIds, obsIdNum

    !-------------------------------------------------------------------
    ! Start

    templateSetName = stringParameter('templateset')
    sourceListSetName = stringParameter('simsrclistset')

    numObsIds = parameterCount('obsids')
    allocate(obsIdList(numObsIds))
    do obsIdNum = 1, numObsIds
      obsIdList(obsIdNum) = stringParameter('obsids', obsIdNum-1)
    end do

    call makeSimSrcList(templateSetName, sourceListSetName, obsIdList)

  end subroutine makeSimSrcList_interface

  !---------------------------------------------------------------------
  subroutine addColumns
  !---------------------------------------------------------------------
  ! function=addcols

!**** generalize => include a vector of data types??
!*** replace by fcalcs??

    character(255) :: setTabName, setName, tabName
    integer(int16) :: numNewCols, newColNum

    type(DataSetT) :: set
    type(TableT) :: tab
    type(ColumnT) :: dataValCol
    integer(int32), pointer :: dataVals(:)
!    integer(int32) :: numSelRows
    character(255) :: colName
    integer(int32) :: colValue

    !-------------------------------------------------------------------
    ! Start

    setTabName = stringParameter('tab')
    numNewCols = parameterCount('newcolumnlist')
    if (numNewCols .ne. parameterCount('newcolvaluelist'))&
call error('columnListMismatch','')
  if (numNewCols <= 0) return

    call splitSetTabName(setTabName, setName, tabName)

    set = dataSet(setName, MODIFY)
    tab = table(set, tabName)

    do newColNum = 1, numNewCols
      colName = stringParameter('newcolumnlist', newColNum-1)
      colValue = intParameter('newcolvaluelist', newColNum-1)

      dataValCol = addOrOpenColumn(tab, colName, INTEGER32)
      dataVals => int32Data(dataValCol)
      dataVals = colValue
      call release(dataValCol)
    end do

    call release(set)

  end subroutine addColumns

  !---------------------------------------------------------------------
  subroutine makeRateImage_interface
  !---------------------------------------------------------------------
  ! function=image

    implicit none

    character(255) ::&
      expMapSetName, sourceListSetName, rateImageName,&
      paddedMaskSetName, tempImageName
    character(10) :: obsId
    integer(int8) :: bandId
    real(double) ::&
      psfCutoffFluxDensity,&
      psfEnergyFraction, ecf, EeVlo, EeVhi

    !-------------------------------------------------------------------
    ! Start

    expMapSetName     = stringParameter('expmapset') ! This is used to set
    ! the instrument type and also the instrument pointing direction. Both
    ! these are needed for the correct PSF useage.

    sourceListSetName = stringParameter('simsrclistset')
    psfEnergyFraction = realParameter('energyfraction')
    psfCutoffFluxDensity = realParameter('psfmaxcoarseflux')
    ecf           = realParameter('ecf')
    EeVlo         = realParameter('elo')
    EeVhi         = realParameter('ehi')
    rateImageName = stringParameter('rateimageset')
    tempImageName = stringParameter('tempset')
    paddedMaskSetName = stringParameter('paddedmaskset')
    obsId            = stringParameter('obsids', 0)
    bandId           = intParameter('idband')

    call makeRateImage(expMapSetName, sourceListSetName, paddedMaskSetName&
      , psfEnergyFraction, psfCutoffFluxDensity, ecf, EeVlo, EeVhi&
      , rateImageName, tempImageName, obsId, bandId)

  end subroutine makeRateImage_interface

  !---------------------------------------------------------------------
  subroutine makeAvCountsImage_interface
  !---------------------------------------------------------------------
  ! function=makectsimg

    implicit none

    character(255) ::&
      expMapSetName, bkgCountMapSetName, rateImageSetName, avCountsImageSetName, tempImageName

    !-------------------------------------------------------------------
    ! Start

    expMapSetName        = stringParameter('expmapset')
    bkgCountMapSetName   = stringParameter('bkgmapset')
    rateImageSetName     = stringParameter('rateimageset')
    avCountsImageSetName = stringParameter('avcountsimageset')
    tempImageName        = stringParameter('tempset')

    call makeAvCountsImage(expMapSetName, bkgCountMapSetName&
      , rateImageSetName, avCountsImageSetName, tempImageName)

  end subroutine makeAvCountsImage_interface

  !---------------------------------------------------------------------
  subroutine poissonizeCountsImage_interface
  !---------------------------------------------------------------------
  ! function=poissonize

    implicit none

    character(255) ::&
      avCountsImageName, countsImageName, tempImageName

    !-------------------------------------------------------------------
    ! Start

    avCountsImageName = stringParameter('avcountsimageset')
    countsImageName   = stringParameter('imageset')
    tempImageName     = stringParameter('tempset')

    call poissonizeCountsImage(avCountsImageName, countsImageName&
      , tempImageName)

  end subroutine poissonizeCountsImage_interface

  !---------------------------------------------------------------------
  subroutine correctEddington_interface
  !---------------------------------------------------------------------
  ! function=eddington

    implicit none

    character(255) :: detSrcSetName
    real(single) :: minLikelihood

    !-------------------------------------------------------------------
    ! Start

    detSrcSetName = stringParameter('detsrclistset')
    minLikelihood = realParameter('likemin')

    call correctEddington(detSrcSetName, minLikelihood)

  end subroutine correctEddington_interface

  !---------------------------------------------------------------------
  subroutine correctPosAndRates_interface
  !---------------------------------------------------------------------
  ! function=correct

    implicit none

    character(255) ::&
      detSrcSetName, ximaErrCorrSetName, yimaErrCorrSetName, rateCorrSetName
    logical :: useCorrected

    !-------------------------------------------------------------------
    ! Start

    detSrcSetName      = stringParameter('detsrclistset')
    ximaErrCorrSetName = stringParameter('ximaerrcorrectorset')
    yimaErrCorrSetName = stringParameter('yimaerrcorrectorset')
    rateCorrSetName    = stringParameter('ratecorrectorset')
    useCorrected      = booleanParameter('usecorrected')

    call correctPosAndRates(detSrcSetName, ximaErrCorrSetName&
      , yimaErrCorrSetName, rateCorrSetName, useCorrected)

  end subroutine correctPosAndRates_interface

  !---------------------------------------------------------------------
  subroutine detRateToFluxRandom_interface
  !---------------------------------------------------------------------
  ! function=ratetofluxrand

    implicit none

    character(255) :: detSrcListSetName!, logNlogSsetName
    logical :: useCorrected

    !-------------------------------------------------------------------
    ! Start

    detSrcListSetName = stringParameter('detsrclistset')
    useCorrected = booleanParameter('usecorrected')

    call detRateToFluxRandom(detSrcListSetName, useCorrected)

  end subroutine detRateToFluxRandom_interface

  !---------------------------------------------------------------------
  subroutine compareDetWithSim_interface
  !---------------------------------------------------------------------
  ! function=compare
  !
  ! Given RATE, X_IMA and Y_IMA values for two source lists, one simulated and the other detected, this subroutine performs two functions: (i) it attempts to deduce which of the simulated sources corresponds with each detected source; (ii) it calculates the probability of the chance occurrence of a source 'closer' than this to the detected source.

    implicit none

    character(255) ::&
      detSrcSetName, simSrcSetName, tempSetName
    character(10) :: obsId
    integer(int8) :: bandId

    !-------------------------------------------------------------------
    ! Start

    detSrcSetName = stringParameter('detsrclistset')
    simSrcSetName = stringParameter('simsrclistset')
    tempSetName   = stringParameter('tempset')
    obsId         = stringParameter('obsids', 0)
    bandId        = intParameter('idband')

    call compareDetWithSim(detSrcSetName, simSrcSetName, tempSetName&
      , obsId, bandId)

  end subroutine compareDetWithSim_interface

  !---------------------------------------------------------------------
  subroutine addSimRates_interface
  !---------------------------------------------------------------------
  ! function=addsimratecols

    implicit none

    character(255) :: detSrcSetName

    !-------------------------------------------------------------------
    ! Start

    detSrcSetName   = stringParameter('detsrclistset')

    call addSimRates(detSrcSetName)

  end subroutine addSimRates_interface




  !---------------------------------------------------------------------
  subroutine correctAllBandMLValues
  !---------------------------------------------------------------------
  ! function=mlcorr

    implicit none

    character(255) :: setTabName, dataPathName, setName, tabName
    character(20) :: colName
    type(DataSetT) :: set
    type(TableT) :: tab
    real(single), pointer :: detML(:), newDetML(:)

    !-------------------------------------------------------------------
    ! Start

    setTabName   = stringParameter('histoset')
    colName      = stringParameter('newcolumnlist', 0)
    dataPathName = stringParameter('string')

    call splitSetTabName(setTabName, setName, tabName)
    set = dataSet(setName, MODIFY)
    tab = table(set, tabName)
    detML => real32Data(column(tab, colName))
!    newDetML => addOrOpenColumnDataReal32(tab, trim(colName)//'_CORR')
newDetML => real32Data(addOrOpenColumn(tab, trim(colName)//'_CORR', REAL32))

!    allocate(newDetML(size(detML)))
    call corr_sum_band(detML, newDetML, size(detML), dataPathName)
!    detML = newDetML
!    deallocate(newDetML)

    call release(set)

  end subroutine correctAllBandMLValues

  !---------------------------------------------------------------------
  subroutine eventizeBlankImage_interface
  !---------------------------------------------------------------------
  ! function=nosources

    implicit none

    character(255) :: expMapSetName, imageName, tempImageName
!    real(single)   :: vigBkgFlux, nonVigBkgFlux
real(double)   :: vigBkgFlux, nonVigBkgFlux

    !-------------------------------------------------------------------
    ! Start

    expMapSetName = stringParameter('expmapset')
    imageName     = stringParameter('rawimageset')
    tempImageName = stringParameter('tempset')

    vigBkgFlux      = realParameter('vigbkgrate')
    nonVigBkgFlux   = realParameter('nonvigbkgrate')

    call eventizeBlankImage(expMapSetName, imageName, vigBkgFlux&
      , nonVigBkgFlux, tempImageName)

  end subroutine eventizeBlankImage_interface

  !---------------------------------------------------------------------
  subroutine correctDetectedFluxes
  !---------------------------------------------------------------------
  ! function=fluxcorrect
  !
  ! Explanation: the time-consuming step in this source simulation
  ! business is the creation of the raw image. Multiplying this by an
  ! exposure map, adding background and eventizing are relatively quick
  ! in contrast. Hence, to save time, different exposure conditions are
  ! often simulated by using the same raw image, list of sim sources
  ! and exposure map, but multiplying the exposure map * raw image
  ! by an additional factor before eventizing. However since the detect
  ! chain uses the same exposure map without this multiplication, the
  ! values of RATE etc that it finds are too large by a factor of
  ! exposureMultiplier. Therefore the RATE etc values of the simulated
  ! sources are here divided by this factor.

    implicit none

    real(single) :: exposureMultiplier
    type(DataSetT) :: set
    type(TableT) :: tab
    real(single), dimension(:), pointer ::&
      scts, sctsErr, bkgMap, flux, fluxErr, rate, rateErr

    !-------------------------------------------------------------------
    ! Start

    exposureMultiplier = realParameter('expmul')

    set = dataSet(stringParameter('detsrclistset'), MODIFY)
    tab = table(set, 'SRCLIST')

    scts    => real32Data(column(tab, 'SCTS'))
    sctsErr => real32Data(column(tab, 'SCTS_ERR'))
    bkgMap  => real32Data(column(tab, 'BG_MAP'))
    flux    => real32Data(column(tab, 'FLUX'))
    fluxErr => real32Data(column(tab, 'FLUX_ERR'))
    rate    => real32Data(column(tab, 'RATE'))
    rateErr => real32Data(column(tab, 'RATE_ERR'))

    scts    = scts / exposureMultiplier
    sctsErr = sctsErr / exposureMultiplier
    bkgMap  = bkgMap / exposureMultiplier
    flux    = flux / exposureMultiplier
    fluxErr = fluxErr / exposureMultiplier
    rate    = rate / exposureMultiplier
    rateErr = rateErr / exposureMultiplier

    call release(set)

  end subroutine correctDetectedFluxes

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module eimsim_mod
