!
! Module: eimsim_reduction
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This is a collection of routines used to reduce simulations data.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eimsim_reduction

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.

  use dal
  use cal	! contains 'use caloaldefs' and 'use caltypes'
  use param  
  use errorHandling

  use array_utils ! for getBinNumber(), median()
  use cal_aux	! for getCalInfo().
  use dal_aux	! for addOrOpenColumn()
!use histogram_utils	! for makeHistogram().
  use histogram_2	! for makeHistogram().
!  use math_utils	! for median().
  use ssc_misc	! for fracVal().
  use flux_transforms

  implicit none 

  private 

  integer(int32) ::&
    instrumentId
  type(SpacecraftAttitudeType) ::&
    scAttitude, instrumAttitude
  real(double) ::&
    expStartTimeStamp

  public calcSysErrors
  public makeMergedSimSrcHistogram
  public calcCompletenessHistogramsPart1
  public calcCompletenessHistogramsPart2
  public calcReliabilityHistogramPart1
  public calcReliabilityHistogramPart2

contains 

  !---------------------------------------------------------------------
  subroutine calcSysErrors
  !---------------------------------------------------------------------
  ! function=syserror
  !
  ! This subroutine does the following:
  !
  ! calculates the average of XIMADIFF and YIMADIFF;
  !
!  ! makes a histogram of the reduced chi2 for X_IMA and Y_IMA as a function of SIM_RATE;
  ! makes a histogram of the reduced chi2 for X_IMA and Y_IMA as a function of SIM_SCTS;
  !
  ! makes a histogram of the average RATE as a function of SCTS;
  !
  ! makes a histogram of reduced chi2 for RATE as a function of SCTS.

!**** NOTE the source list should be filtered to remove spurious sources beforehand.

    implicit none

    character(255) :: detSrcSetName, histoSetName
    type(DataSetT) :: sourceSet, histoSet
    type(TableT) :: sourceTab, histoTab
    integer(int32) :: numSrcs, row, binNum
integer(int32), pointer :: n(:)
    real(single), pointer ::&
      ximaDiff(:), yimaDiff(:),&
      sctsMid(:), binnedValue(:), scatterInBinnedVal(:), binnedErr(:),&
      uncertInBinnedVal(:), binnedRelRate(:), uncertInBinnedRelRate(:),&
      scatterInRelRate(:), binnedRelRateErr(:),&
      chiSqu(:), chiSquErr(:),&
xMid(:),&
      xImaErr(:), yImaErr(:), simRate(:),&
      detRate(:), detRateErr(:), detSrcCts(:)
    real(single) :: meanXima, meanYima, medianXima, medianYima
    real(double) :: imaBinSize, imaDiffRelBinSize, rateBinSize
    real(single), allocatable ::&
      xImaChi2Element(:), yImaChi2Element(:), binEdges(:),&
      rateChi2Element(:), log10DetSrcCts(:), xImaDiffRel(:), yImaDiffRel(:)
    type(histoType), pointer ::&
      xImaDiffHisto(:), xImaErrHisto(:), ximaChi2Histo(:), xImaRelHisto(:),&
      yImaDiffHisto(:), yImaErrHisto(:), yimaChi2Histo(:), yImaRelHisto(:),&
      rateRatioHisto(:), rateErrRatioHisto(:), rateChi2Histo(:)

    !-------------------------------------------------------------------
    ! Start

    histoSetName  = stringParameter('histoset')
    detSrcSetName = stringParameter('detsrclistset')

    histoSet = dataSet(histoSetName, CREATE)

    sourceSet = dataSet(detSrcSetName, READ)
    sourceTab = table(sourceSet, 'SRCLIST', READ)
    numSrcs = numberOfRows(sourceTab)
if (numSrcs .le. 0) call error('noSources', 'No sources in the table.')

    !...................................................................
    ! Get mean and median X/Y_IMA offsets:

    xImaDiff => real32Data(column(sourceTab, 'XIMADIFF'))
    yImaDiff => real32Data(column(sourceTab, 'YIMADIFF'))
    meanXima = sum(xImaDiff)/real(numSrcs)
    meanYima = sum(yImaDiff)/real(numSrcs)
    medianXima = median(xImaDiff) ! in ssclib/array_utils.
    medianYima = median(yImaDiff) ! in ssclib/array_utils.

print *, 'meanXima=', meanXima
print *, 'meanYima=', meanYima
print *, 'medianXima=', medianXima
print *, 'medianYima=', medianYima

    !...................................................................
    ! Histograms of xima and yima against log10DetSrcCts:

    detSrcCts => real32Data(column(sourceTab, 'SCTS'))
    allocate(log10DetSrcCts(0:numSrcs-1))
    log10DetSrcCts = log10(detSrcCts)
    imaBinSize = 0.1d0

    ! First XIMA:

    xImaErr  => real32Data(column(sourceTab, 'X_IMA_ERR'))

    call makeHistogram(log10DetSrcCts, xImaDiffHisto, xImaDiff&
      , binSize=real(imaBinSize))
    call makeHistogram(log10DetSrcCts, xImaErrHisto, xImaErr&
      , binSize=real(imaBinSize))

    allocate(xImaChi2Element(numSrcs))
    xImaChi2Element = (xImaDiff - meanXima) / xImaErr
    xImaChi2Element = xImaChi2Element * xImaChi2Element

    call makeHistogram(log10DetSrcCts, xImaChi2Histo, xImaChi2Element&
      , binSize=real(imaBinSize))

!** probably ought to multiply by numSrcs/(numSrcs-1)

    histoTab = addTable(histoSet, 'XIMA', size(xImaChi2Histo))
    sctsMid           => real32Data(addColumn(histoTab, 'SCTS_MID_BIN', REAL32&
      , 'counts', comment='Histogram X value at mid-bin.'))
    binnedValue       => real32Data(addColumn(histoTab, 'X_IMA_AVE'&
      , REAL32, 'pixels'))
    scatterInBinnedVal=> real32Data(addColumn(histoTab, 'X_IMA_STDDEV'&
      , REAL32, 'pixels'))
    uncertInBinnedVal => real32Data(addColumn(histoTab, 'X_IMA_AVE_ERR'&
      , REAL32, 'pixels'))
    binnedErr         => real32Data(addColumn(histoTab, 'X_IMA_ERR_AVE'&
      , REAL32, 'pixels'))
    chiSqu            => real32Data(addColumn(histoTab, 'CHI_SQU',      REAL32))
    chiSquErr         => real32Data(addColumn(histoTab, 'CHI_SQU_ERR',  REAL32))

    sctsMid = 10.0**((xImaChi2Histo%xLo + xImaChi2Histo%xHi)/2.0)
    binnedValue        = xImaDiffHisto%yAve
    scatterInBinnedVal = xImaDiffHisto%yStdDev
    uncertInBinnedVal  = xImaDiffHisto%yUncert
    binnedErr          = xImaErrHisto%yAve
    chiSqu             = xImaChi2Histo%yAve
    chiSquErr          = xImaChi2Histo%yUncert

    call setAttribute(histoTab, 'XIMAMEAN', meanXima)
    call setAttribute(histoTab, 'XIMAMEDN', medianXima)
    call setAttribute(histoTab, 'BIN_SIZE', imaBinSize)
    call release(histoTab)
    deallocate(xImaDiffHisto, xImaErrHisto, xImaChi2Histo)

    ! Now YIMA:

    yImaErr  => real32Data(column(sourceTab, 'Y_IMA_ERR'))

    call makeHistogram(log10DetSrcCts, yImaDiffHisto, yImaDiff&
      , binSize=real(imaBinSize))
    call makeHistogram(log10DetSrcCts, yImaErrHisto, yImaErr&
      , binSize=real(imaBinSize))

    allocate(yImaChi2Element(numSrcs))
    yImaChi2Element = (yImaDiff - meanYima) / yImaErr
    yImaChi2Element = yImaChi2Element * yImaChi2Element

    call makeHistogram(log10DetSrcCts, yImaChi2Histo, yImaChi2Element&
      , binSize=real(imaBinSize))

!** probably ought to multiply by numSrcs/(numSrcs-1)

    histoTab = addTable(histoSet, 'YIMA', size(yImaChi2Histo))
    sctsMid           => real32Data(addColumn(histoTab, 'SCTS_MID_BIN', REAL32&
      , 'counts', comment='Histogram X value at mid-bin.'))
    binnedValue       => real32Data(addColumn(histoTab, 'Y_IMA_AVE',    REAL32&
      , 'pixels'))
    scatterInBinnedVal=> real32Data(addColumn(histoTab, 'Y_IMA_STDDEV', REAL32&
      , 'pixels'))
    uncertInBinnedVal => real32Data(addColumn(histoTab, 'Y_IMA_AVE_ERR'&
      , REAL32, 'pixels'))
    binnedErr         => real32Data(addColumn(histoTab, 'Y_IMA_ERR_AVE',REAL32&
      , 'pixels'))
    chiSqu            => real32Data(addColumn(histoTab, 'CHI_SQU',      REAL32))
    chiSquErr         => real32Data(addColumn(histoTab, 'CHI_SQU_ERR',  REAL32))

    sctsMid = 10.0**((yImaChi2Histo%xLo + yImaChi2Histo%xHi)/2.0)
    binnedValue        = yImaDiffHisto%yAve
    scatterInBinnedVal = yImaDiffHisto%yStdDev
    uncertInBinnedVal  = yImaDiffHisto%yUncert
    binnedErr          = yImaErrHisto%yAve
    chiSqu             = yImaChi2Histo%yAve
    chiSquErr          = yImaChi2Histo%yUncert

    call setAttribute(histoTab, 'YIMAMEAN', meanXima)
    call setAttribute(histoTab, 'YIMAMEDN', medianXima)
    call setAttribute(histoTab, 'BIN_SIZE', imaBinSize)
    call release(histoTab)
    deallocate(yImaDiffHisto, yImaErrHisto, yImaChi2Histo)

print *, 'Total x chi2: ', sum(xImaChi2Element) / real(numSrcs-1)
print *, 'Total y chi2: ', sum(yImaChi2Element) / real(numSrcs-1)

    !...................................................................
    ! Histograms of counts against ximadiff/xima_err and yimadiff/yima_err:

    imaDiffRelBinSize = 0.25d0

    ! First XIMA:

    allocate(xImaDiffRel(numSrcs))
    xImaDiffRel = (xImaDiff - meanXima) / xImaErr

    call makeHistogram(xImaDiffRel, xImaRelHisto&
      , binSize=real(imaDiffRelBinSize))

    histoTab = addTable(histoSet, 'XIMAREL', size(xImaRelHisto))
    xMid => real32Data(addColumn(histoTab, 'D_XIMA_REL', REAL32, 'sigmas'&
      , comment='Histogram X value at mid-bin. &
      &(XIMADIFF-<XIMADIFF>)/X_IMA_ERR.'))
    n    =>  int32Data(addColumn(histoTab, 'N',          INTEGER32&
      , comment='Number of entries in the histogram bin.'))
    xMid = (xImaRelHisto%xLo + xImaRelHisto%xHi)/2.0
    n = xImaRelHisto%n

    call setAttribute(histoTab, 'BIN_SIZE', imaDiffRelBinSize)
    call release(histoTab)
    deallocate(xImaRelHisto)

    ! Now YIMA:

    allocate(yImaDiffRel(numSrcs))
    yImaDiffRel = (yImaDiff - meanYima) / yImaErr

    call makeHistogram(yImaDiffRel, yImaRelHisto&
      , binSize=real(imaDiffRelBinSize))

    histoTab = addTable(histoSet, 'YIMAREL', size(yImaRelHisto))
    xMid => real32Data(addColumn(histoTab, 'D_YIMA_REL', REAL32, 'sigmas'&
      , comment='Histogram X value at mid-bin. &
      &(YIMADIFF-<YIMADIFF>)/Y_IMA_ERR.'))
    n    =>  int32Data(addColumn(histoTab, 'N',          INTEGER32&
      , comment='Number of entries in the histogram bin.'))
    xMid = (yImaRelHisto%xLo + yImaRelHisto%xHi)/2.0
    n = yImaRelHisto%n

    call setAttribute(histoTab, 'BIN_SIZE', imaDiffRelBinSize)
    call release(histoTab)
    deallocate(yImaRelHisto)

    call release(sourceTab)
!**** huh??

    !...................................................................
    ! Histogram of detRate / simRate against log10DetSrcCts:

    sourceTab = table(sourceSet, 'SRCLIST', READ)
    simRate    => real32Data(column(sourceTab, 'SIM_RATE'))
    detRate    => real32Data(column(sourceTab, 'RATE'))
    detRateErr => real32Data(column(sourceTab, 'RATE_ERR'))

    rateBinSize = 0.05d0
    call makeHistogram(log10DetSrcCts, rateRatioHisto, detRate / simRate&
      , binSize=real(rateBinSize))
    call makeHistogram(log10DetSrcCts, rateErrRatioHisto, detRateErr / simRate&
      , binSize=real(rateBinSize))

    allocate(binEdges(size(rateRatioHisto)+1))
    allocate(rateChi2Element(numSrcs))
    binEdges(:size(rateRatioHisto))  = rateRatioHisto%xLo
    binEdges(size(rateRatioHisto)+1) = rateRatioHisto(size(rateRatioHisto))%xHi
    do row = 0, numSrcs-1
      binNum = getBinNumber(binEdges, log10DetSrcCts(row)) ! in array_utils
      rateChi2Element(row+1) = (detRate(row) / simRate(row)&
        - rateRatioHisto(binNum)%yAve) / (detRateErr(row) / simRate(row))
    end do
    rateChi2Element = rateChi2Element * rateChi2Element

    call makeHistogram(log10DetSrcCts, rateChi2Histo, rateChi2Element&
      , binSize=real(rateBinSize))

    histoTab = addTable(histoSet, 'CTSRATIO', size(rateRatioHisto))
    sctsMid               => real32Data(addColumn(histoTab, 'SCTS_MID'&
      , REAL32, 'counts', comment='Histogram X value at mid-bin.'))
    binnedRelRate         => real32Data(addColumn(histoTab, 'RATE_REL'&
      , REAL32, comment='RATE / SIM_RATE.'))
    uncertInBinnedRelRate => real32Data(addColumn(histoTab, 'RATE_REL_ERR'&
      , REAL32, comment='Uncertainty in RATE_REL.'))
    scatterInRelRate      => real32Data(addColumn(histoTab, 'RATE_REL_STDDEV'&
      , REAL32, comment='Standard deviation of RATE_REL.'))
    binnedRelRateErr      => real32Data(addColumn(histoTab, 'RATE_ERR_REL'&
      , REAL32, comment='RATE_ERR / SIM_RATE.'))
    chiSqu                => real32Data(addColumn(histoTab, 'CHI_SQU'&
      , REAL32, comment='((RATE / SIM_RATE - <RATE / SIM_RATE>)&
      &/(RATE_ERR / SIM_RATE))^2.'))
    chiSquErr             => real32Data(addColumn(histoTab, 'CHI_SQU_ERR'&
      , REAL32, comment='Uncertainty in CHI_SQU.'))

    sctsMid = 10.0**((rateRatioHisto%xLo + rateRatioHisto%xHi)/2.0)
    binnedRelRate = rateRatioHisto%yAve
    uncertInBinnedRelRate = rateRatioHisto%yUncert
    scatterInRelRate = rateRatioHisto%yStdDev
    binnedRelRateErr = rateErrRatioHisto%yAve
    chiSqu = rateChi2Histo%yAve
    chiSquErr = rateChi2Histo%yUncert

    call setAttribute(histoTab, 'BIN_SIZE', rateBinSize)

    deallocate(rateRatioHisto, rateErrRatioHisto, rateChi2Histo)
    call release(histoSet)
    call release(sourceSet)

  end subroutine calcSysErrors

  !---------------------------------------------------------------------
  subroutine makeMergedSimSrcHistogram
  !---------------------------------------------------------------------
  ! function=simhisto
  !
  ! This sub makes a histogram table which records the incidence of sim sources as a function of their rate. The bins are of equal size in log10(SIM_RATE). Columns as follows are constructed:
  !
  !	N_SIM		Number of simulated sources.
  !	LOG10_SR_LO	Lower edge of the bin in log10(SIM_RATE).
  !	LOG10_SR_HI	Upper edge of the bin in log10(SIM_RATE).
  !
  ! The sub is given the name of a master histogram set and the name of a text file which contains a 'ls' of all relevant sim source lists. If the master set exists at entry to the loop over sim source lists, its bin edges are used as a template to make a histogram of the new source list (extending the rows if necessary). The new histogram is added to the old and N_FIELDS is incremented. If the master set doesn't yet exist, the sub creates it at loop entry and constructs the histogram according to the passed parameter 'binsize'. The sub also looks for the SKYAREA keyword in each source list and compares it with that in the histo file. If they are different, an error dump ensues.

    implicit none

    real(double) :: binSize, maskAreaDeg1st, maskAreaDeg
    integer(int32) ::&
      numSimSrc, numHistoBins, numHistoRowsToInsertAtStart,&
      numHistoRowsToInsertAtFinis, bin, numFields
    character(255) :: masterHistoSetName, simSrcSetName, listOfSimSrcLists
    character(20) :: rateColumnName
    type(DataSetT) :: simSrcSet, masterHistoSet
    type(TableT) :: simSrcTab, masterHistoTab
    type(ColumnT) :: simRateCol, nCol, xLoCol, xHiCol
    real(single), pointer :: simRate(:), xLo(:), xHi(:)
    real(single), allocatable :: binEdges(:), log10SimRate(:)
    integer(int32), pointer :: n(:)
    logical :: fileExists
    type(histoType), pointer :: newHisto(:)
    integer :: ioUnit, ioStatus

    !-------------------------------------------------------------------
    ! Start

    binSize = realParameter('histobinsize')
    masterHistoSetName = stringParameter('histoset')
    listOfSimSrcLists = stringParameter('filelist')
    rateColumnName = stringParameter('newcolumnlist', 0)

    inquire(file=listOfSimSrcLists, exist=fileExists) ! tests
if (.not. fileExists) call error('noListFile','')

    call getFreeIoUnit(ioUnit) ! in sscmisc

    open(ioUnit, iostat=ioStatus, file=listOfSimSrcLists, status='old'&
      , action='read')
if (ioStatus > 0) call error('unableToOpenFile','')

!***** attempt to read 1st simSrcSetName from file
    read(ioUnit, '(a)', iostat=ioStatus) simSrcSetName

if (ioStatus .ne. 0) call error('unableToReadFirstRecord','')

    simSrcSet = dataSet(simSrcSetName, READ)
!******* NOTE! This should be filtered beforehand.
    simSrcTab = table(simSrcSet, 'SRCLIST')
    numSimSrc = numberOfRows(simSrcTab)
if (numSimSrc <= 0) call error('noSourcesInFirstSet','')
!*** handle this better. Ie just make an empty table.

    maskAreaDeg1st = real64Attribute(simSrcTab, 'SKY_AREA')
!*** should be templateTab?? Rest of this module also needs a spring clean.
call error('fixThis','')

    simRateCol = column(simSrcTab, rateColumnName)
    simRate => real32Data(simRateCol)
    allocate(log10SimRate(numSimSrc))
    log10SimRate = log10(simRate)
    call release(simSrcSet)

    call makeHistogram(log10SimRate, newHisto, binSize=real(binSize))
    deallocate(log10SimRate)

    masterHistoSet = dataSet(masterHistoSetName, CREATE)
    masterHistoTab = addTable(masterHistoSet, 'HISTO', size(newHisto))

    nCol   = addColumn(masterHistoTab, 'N_SIM',       INTEGER32)
    xLoCol = addColumn(masterHistoTab, 'LOG10_SR_LO', REAL32)
    xHiCol = addColumn(masterHistoTab, 'LOG10_SR_HI', REAL32)
    n   => int32Data(   nCol)
    xLo => real32Data(xLoCol)
    xHi => real32Data(xHiCol)
    n   = newHisto%n
    xLo = newHisto%xLo
    xHi = newHisto%xHi

    call setAttribute(masterHistoTab, 'SIM_SKAR', maskAreaDeg1st, 'deg^2'&
      , 'Sky area of sim sources.')
    call setAttribute(masterHistoTab, 'BIN_SIZE',    binSize)

    numHistoBins = size(newHisto)
    deallocate(newHisto)

    numFields = 1

    do
!print *, 'numFields: ', numFields
      read(ioUnit, '(a)', iostat=ioStatus) simSrcSetName
!print *, 'simSrcSetName: ', trim(simSrcSetName)

if (ioStatus > 0) call error('unableToReadRecord','')
    if (ioStatus < 0) exit

      simSrcSet = dataSet(simSrcSetName, READ)
!******* NOTE! This should be filtered beforehand.
      simSrcTab = table(simSrcSet, 'SRCLIST')
      numSimSrc = numberOfRows(simSrcTab)
    if (numSimSrc <= 0) cycle

      maskAreaDeg = real64Attribute(simSrcTab, 'SKY_AREA')
if (maskAreaDeg .ne. maskAreaDeg1st) call error('badMaskArea','')

      simRateCol = column(simSrcTab, rateColumnName)
      simRate => real32Data(simRateCol)
      allocate(log10SimRate(numSimSrc))
      log10SimRate = log10(simRate)
      call release(simSrcSet)

!print *, 'numHistoBins, maxval(log10SimRate), xHi(numHistoBins-1): ', numHistoBins, maxval(log10SimRate), xHi(numHistoBins-1)
      numHistoRowsToInsertAtStart = 0
!      if (minval(simRate) < xLo(0)) then
!        numHistoRowsToInsertAtStart = ceiling((xLo(0) - minval(simRate))&
if (minval(log10SimRate) < xLo(0)) then
  numHistoRowsToInsertAtStart = ceiling((xLo(0) - minval(log10SimRate))&
          / binSize)
      end if

      numHistoRowsToInsertAtFinis = 0
!      if (maxval(simRate) > xHi(numHistoBins-1)) then
!        numHistoRowsToInsertAtFinis = ceiling((maxval(simRate)&
if (maxval(log10SimRate) > xHi(numHistoBins-1)) then
  numHistoRowsToInsertAtFinis = ceiling((maxval(log10SimRate)&
          - xHi(numHistoBins-1)) / binSize)
      end if

!print *, 'numHistoRowsToInsertAtStart, numHistoRowsToInsertAtFinis:', numHistoRowsToInsertAtStart, numHistoRowsToInsertAtFinis

      if (numHistoRowsToInsertAtStart > 0 .or. numHistoRowsToInsertAtFinis > 0)&
      then
        call release(nCol)
        call release(xLoCol)
        call release(xHiCol)

        if (numHistoRowsToInsertAtFinis > 0) then
          call insertRows(masterHistoTab, numHistoBins&
            , numHistoRowsToInsertAtFinis)
        end if
        if (numHistoRowsToInsertAtStart > 0) then
print *, '############## INSERTING AT START!'
call error('','')
          call insertRows(masterHistoTab, 0, numHistoRowsToInsertAtStart)
        end if

        numHistoBins = numberOfRows(masterHistoTab)

        n   => int32Data( column(masterHistoTab, 'N_SIM'))
        xLo => real32Data(column(masterHistoTab, 'LOG10_SR_LO'))
        xHi => real32Data(column(masterHistoTab, 'LOG10_SR_HI'))

        do bin = numHistoRowsToInsertAtStart-1, 0, -1
          n(bin) = 0
          xHi(bin) = xLo(bin+1)
          xLo(bin) = xLo(numHistoRowsToInsertAtStart)&
            - (numHistoRowsToInsertAtStart - bin) * binSize
        end do

        do bin = numHistoBins-numHistoRowsToInsertAtFinis, numHistoBins-1
          n(bin) = 0
          xLo(bin) = xHi(bin-1)
          xHi(bin) = xLo(numHistoRowsToInsertAtStart)&
            + (bin + 1 - numHistoRowsToInsertAtStart) * binSize
        end do
      end if ! some rows to insert either at start or finish.

      allocate(binEdges(numHistoBins+1))
      binEdges(:numHistoBins) = xLo
      binEdges(numHistoBins+1) = xHi(numHistoBins-1)

      call makeHistogram(log10SimRate, newHisto, binEdges=binEdges)
      n = n + newHisto%n
      deallocate(newHisto, log10SimRate, binEdges)

      numFields = numFields + 1
    end do ! loop over entries in list of sim source lists.

    call setAttribute(masterHistoTab, 'N_SIM_FD', numFields&
      , comment='Number of sim fields included in the histogram.')

    call release(masterHistoSet)
    close(ioUnit)

  end subroutine makeMergedSimSrcHistogram

  !---------------------------------------------------------------------
  subroutine calcCompletenessHistogramsPart1
  !---------------------------------------------------------------------
  ! function=completeness1
  !
  ! This sub adds columns pertaining to detected sources to the histogram table giving the incidence of sim sources as a function of their rate. (The bins are of equal size in log10(SIM_RATE).) Various other useful columns are calculated. The final tally of columns should be as follows:
  !
  ! Pre-existing:
  !
  !	N_SIM		Number of simulated sources.
  !	LOG10_SR_LO	Lower edge of the bin in log10(SIM_RATE).
  !	LOG10_SR_HI	Upper edge of the bin in log10(SIM_RATE).
  !
  ! New:
  !
  !	N_SIM_ERR	Square root of N_SIM.
  !	N_SIM_INT	Integrated total of N_SIM.
  !	DENS_SIM	Average sky-density (in deg^-2) of sim sources.
  !	DENS_SIM_ERR	Error in DENS_SIM.
  !	DENS_SIM_INT	Integrated total of DENS_SIM.
  !	SIM_RATE_MID	SIM_RATE at the middle of the bin.
  !	SIM_RATE_INT	SIM_RATE at the lower edge of the bin.
  !	N_ALL_DET	All detections.
  !	N_ALL_DET_ERR	Square root of N_ALL_DET.
  !	N_ALL_DET_INT	Integrated total of N_ALL_DET.
  !	N_TRUE_DET	Reliable detections.
  !	N_TRUE_DET_ERR	Square root of N_TRUE_DET.
  !	N_TRUE_DET_INT	Integrated total of N_TRUE_DET.
  !	FLUX_MID	Calculated flux (in band 200-2000 eV) at bin centre.
  !	FLUX_INT	Calculated flux (in band 200-2000 eV) at the lower edge of the bin.

    implicit none

    character(255) ::&
      histoSetName, totalDetSrcSetName, trueDetSrcSetTabName,&
      trueDetSetName, trueDetTabName, logNLogSSetName
    character(20) :: filterName
    type(DataSetT) :: histoSet, detSrcSet, logNLogSSet
    type(TableT) :: histoTab, detSrcTab, logNLogSTab, histoTheoryTab
type(ColumnT) :: nSimErrCol, nSimIntCol, simSkyDensityCol,&
 simSkyDensityErrCol, simSkyDensityIntCol, simRateMidCol,&
 simRateIntCol, nTotalDetCol, nTotalDetErrCol, nTotalDetIntCol,&
 nTrueDetCol, nTrueDetErrCol, nTrueDetIntCol,&
 fluxSoftMidCol, fluxSoftIntCol
    integer(int32) :: numBins, bin, numSimFields, filtrId, numLogNLogSRows
    integer(int32), pointer ::&
      nSim(:), nSimInt(:), nTotalDet(:), nTotalDetInt(:),&
      nTrueDet(:), nTrueDetInt(:)
    real(single), pointer ::&
      nSimErr(:), simHistoXlo(:), simHistoXhi(:), simRateMid(:),&
      simRateInt(:), log10SimRate(:), nTotalDetErr(:), nTrueDetErr(:),&
      simSkyDensity(:), simSkyDensityErr(:), simSkyDensityInt(:),&
      fluxSoftMid(:), fluxSoftInt(:),&
      logNlogSFlux(:), logNlogSDens(:),&
      histoLogNLogSRate(:), histoLogNLogSFlux(:), histoLogNLogSDens(:)
    real(single), allocatable :: binEdges(:), amplitude(:), amplitudeMid(:), amplitudeInt(:)
    type(histoType), pointer :: detHisto(:)
    real(double) ::&
      simMaskAreaDeg, photonSpectralIndex, EeVloRate, EeVhiRate,&
      EeVloFlux, EeVhiFlux, logNlogSLoSlope, logNlogSHiSlope

    !-------------------------------------------------------------------
    ! Start

    histoSetName         = stringParameter('histoset')
    totalDetSrcSetName   = stringParameter('detsrclistset')
    trueDetSrcSetTabName = stringParameter('tab')
    logNLogSSetName      = stringParameter('lognlogsset')

    call splitSetTabName(trueDetSrcSetTabName, trueDetSetName, trueDetTabName)

    !...................................................................
    ! Open histo set and get some keywords:

    histoSet = dataSet(histoSetName, MODIFY)
    histoTab = table(histoSet, 'HISTO')
    numBins = numberOfRows(histoTab)

    simMaskAreaDeg      = real64Attribute(histoTab, 'SIM_SKAR')
    numSimFields         = int32Attribute(histoTab, 'N_SIM_FD')

    call getCalInfo(totalDetSrcSetName, instrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude) ! sets instrument in cal.

!    filterName = stringAttribute(detSrcSet, 'FILTER')
!    filtrId = filterId(filterName) ! in caloalutils/src/caloaldefs.f90

    !...................................................................
    ! N_SIM_ERR, N_SIM_INT:

    nSim    => int32Data(column(histoTab, 'N_SIM'))
    nSimErrCol = addOrOpenColumn(histoTab, 'N_SIM_ERR', REAL32)
    nSimErr => real32Data(nSimErrCol)
    nSimErr = sqrt(real(nSim))
    nSimIntCol = addOrOpenColumn(histoTab, 'N_SIM_INT', INTEGER32)
    nSimInt => int32Data(nSimIntCol)
    nSimInt(numBins-1) = nSim(numBins-1)
    do bin = numBins-2, 0, -1
      nSimInt(bin) = nSimInt(bin+1) + nSim(bin)
    end do

    !...................................................................
    ! DENS_SIM, DENS_SIM_ERR, DENS_SIM_INT:

    simSkyDensityCol = addOrOpenColumn(histoTab, 'DENS_SIM', REAL32&
      , 'sources/deg^2')
    simSkyDensity => real32Data(simSkyDensityCol)

    simSkyDensityErrCol = addOrOpenColumn(histoTab, 'DENS_SIM_ERR', REAL32&
      , 'sources/deg^2')
    simSkyDensityErr => real32Data(simSkyDensityErrCol)

    simSkyDensityIntCol = addOrOpenColumn(histoTab, 'DENS_SIM_INT', REAL32&
      , 'sources/deg^2')
    simSkyDensityInt => real32Data(simSkyDensityIntCol)

    simSkyDensity    = real(nSim)    / simMaskAreaDeg / real(numSimFields)
    simSkyDensityErr =      nSimErr  / simMaskAreaDeg / real(numSimFields)
    simSkyDensityInt = real(nSimInt) / simMaskAreaDeg / real(numSimFields)

    !...................................................................
    ! SIM_RATE_MID, SIM_RATE_INT:

    simHistoXlo => real32Data(column(histoTab, 'LOG10_SR_LO'))
    simHistoXhi => real32Data(column(histoTab, 'LOG10_SR_HI'))

    simRateMidCol = addOrOpenColumn(histoTab, 'SIM_RATE_MID', REAL32&
      , 'counts/s')
    simRateMid => real32Data(simRateMidCol)

    simRateMid = 10.0**((simHistoXlo + simHistoXhi)/2.0)
    simRateIntCol = addOrOpenColumn(histoTab, 'SIM_RATE_INT', REAL32&
      , 'counts/s')
    simRateInt => real32Data(simRateIntCol)

    simRateInt = 10.0**simHistoXlo

    !...................................................................
    ! Add det list keywords:

    allocate(binEdges(numBins+1))
    binEdges(:numBins) = simHistoXlo
    binEdges(numBins+1) = simHistoXhi(numBins-1)

    detSrcSet = dataSet(totalDetSrcSetName, READ)
    detSrcTab = table(detSrcSet, 'SRCLIST')

    call setAttribute(histoTab, 'DET_SKAR'&
      , real64Attribute(detSrcTab, 'SKY_AREA'), 'deg^2'&
      , 'Sky area of det sources.')
    call setAttribute(histoTab, 'N_DET_FD'&
      , int32Attribute(detSrcTab, 'N_FIELDS')&
      , comment='Number of det fields included in the histogram.')

    filterName = stringAttribute(detSrcSet, 'FILTER')
    filtrId = filterId(filterName) ! in caloalutils/src/caloaldefs.f90
    call setAttribute(histoTab, 'FILTER', filterName)
    call CAL_setState(filterId=filtrId)

    !...................................................................
    ! N_ALL_DET, N_ALL_DET_ERR:

    log10SimRate => real32Data(column(detSrcTab, 'LOG10_SR'))
    call makeHistogram(log10SimRate, detHisto, binEdges=binEdges)
    call release(detSrcSet)

    nTotalDetCol = addOrOpenColumn(histoTab, 'N_ALL_DET', INTEGER32)
    nTotalDet => int32Data(nTotalDetCol)

    nTotalDetErrCol = addOrOpenColumn(histoTab, 'N_ALL_DET_ERR', REAL32)
    nTotalDetErr => real32Data(nTotalDetErrCol)

    nTotalDet    = detHisto%n
    nTotalDetErr = sqrt(real(nTotalDet))
    deallocate(detHisto)

    !...................................................................
    ! N_ALL_DET_INT:

    nTotalDetIntCol = addOrOpenColumn(histoTab, 'N_ALL_DET_INT', INTEGER32)
    nTotalDetInt => int32Data(nTotalDetIntCol)

    nTotalDetInt(numBins-1) = nTotalDet(numBins-1)
    do bin = numBins-2, 0, -1
      nTotalDetInt(bin) = nTotalDetInt(bin+1) + nTotalDet(bin)
    end do

    !...................................................................
    ! N_TRUE_DET, N_TRUE_DET_ERR, N_TRUE_DET_INT:

    detSrcSet = dataSet(trueDetSetName, READ)
    detSrcTab = table(detSrcSet, 'SRCLIST')

    log10SimRate => real32Data(column(detSrcTab, 'LOG10_SR'))
    call makeHistogram(log10SimRate, detHisto, binEdges=binEdges)
    call release(detSrcSet)

    nTrueDetCol = addOrOpenColumn(histoTab, 'N_TRUE_DET', INTEGER32)
    nTrueDet => int32Data(nTrueDetCol)

    nTrueDet = detHisto%n
    deallocate(detHisto)

    nTrueDetErrCol = addOrOpenColumn(histoTab, 'N_TRUE_DET_ERR', REAL32)
    nTrueDetErr => real32Data(nTrueDetErrCol)

    nTrueDetErr = sqrt(real(nTrueDet))

    nTrueDetIntCol = addOrOpenColumn(histoTab, 'N_TRUE_DET_INT', INTEGER32)
    nTrueDetInt => int32Data(nTrueDetIntCol)

    nTrueDetInt(numBins-1) = nTrueDet(numBins-1)
    do bin = numBins-2, 0, -1
      nTrueDetInt(bin) = nTrueDetInt(bin+1) + nTrueDet(bin)
    end do

    !...................................................................
    ! Open the input logN-logS distribution and get energy limits:

    logNLogSSet = dataSet(logNLogSSetName, READ)
    logNLogSTab = table(logNLogSSet, 'LOGNLOGS')
    EeVloFlux = real64Attribute(logNLogSTab, 'E_MIN')
    EeVhiFlux = real64Attribute(logNLogSTab, 'E_MAX')
    photonSpectralIndex = real64Attribute(logNLogSTab, 'SPECINDX')

    !...................................................................
    ! FLUX_MID, FLUX_INT:

    allocate(amplitudeMid(numBins), amplitudeInt(numBins))
!    filtrId = filterId(filterName) ! in caloalutils/src/caloaldefs.f90

    EeVloRate = 200.0d0		! Appropriate for the RATE values,
    EeVhiRate = 12000.0d0	! which are those in band 8.

    call setAttribute(histoTab, 'ERATEMIN', EeVloRate, 'eV'&
      , 'Valid for the RATE values.')
    call setAttribute(histoTab, 'ERATEMAX', EeVhiRate, 'eV'&
      , 'Valid for the RATE values.')

    amplitudeMid = rateToFluxDensity0(photonSpectralIndex&
     , EeVloRate, EeVhiRate, simRateMid)!, filtrId)

    amplitudeInt = rateToFluxDensity0(photonSpectralIndex&
     , EeVloRate, EeVhiRate, simRateInt)!, filtrId)

    call setAttribute(histoTab, 'EFLUXMIN', EeVloFlux, 'eV'&
      , 'Valid for the FLUX and FLUX_INT values.')
    call setAttribute(histoTab, 'EFLUXMAX', EeVhiFlux, 'eV'&
      , 'Valid for the FLUX and FLUX_INT values.')

!    fluxSoftMid => addOrOpenColumnDataReal32(histoTab, 'FLUX_MID'&
!      , 'erg/cm^2/s', 'Valid for the E bounds and spectral index given.')
fluxSoftMidCol = addOrOpenColumn(histoTab, 'FLUX_MID', REAL32&
  , 'erg/cm^2/s', 'Valid for the E bounds and spectral index given.')
fluxSoftMid => real32Data(fluxSoftMidCol)

!    fluxSoftInt => addOrOpenColumnDataReal32(histoTab, 'FLUX_INT'&
!      , 'erg/cm^2/s', 'Valid for the E bounds and spectral index given.')
fluxSoftIntCol = addOrOpenColumn(histoTab, 'FLUX_INT', REAL32&
  , 'erg/cm^2/s', 'Valid for the E bounds and spectral index given.')
fluxSoftInt => real32Data(fluxSoftIntCol)

    fluxSoftMid = fluxDensity0ToFlux(photonSpectralIndex&
      , EeVloFlux, EeVhiFlux, amplitudeMid)

    fluxSoftInt = fluxDensity0ToFlux(photonSpectralIndex&
      , EeVloFlux, EeVhiFlux, amplitudeInt)

    deallocate(amplitudeMid, amplitudeInt)

    !...................................................................
    ! Add an extension containing the input (theoretical) logN-logS
    ! distribution:

    numLogNLogSRows = numberOfRows(logNlogSTab)
    logNlogSFlux => real32Data(column(logNLogSTab, 'FLUX'))
    logNlogSDens => real32Data(column(logNLogSTab, 'DENSITY'))

!    if (hasBlock(histoSet, 'LOGNLOGS')) call deleteBlock(histoSet, 'LOGNLOGS')
!    histoTheoryTab = addtable(histoSet, 'LOGNLOGS', numLogNLogSRows+2)
if (hasBlock(histoSet, 'THEORY')) call deleteBlock(histoSet, 'THEORY')
histoTheoryTab = addtable(histoSet, 'THEORY', numLogNLogSRows+2)

    call setAttributes(attributable(histoTheoryTab)&
      , attributable(logNlogSTab))

    histoLogNLogSRate => real32Data(addColumn(histoTheoryTab, 'RATE'&
      , REAL32, 'counts/s'))
    histoLogNLogSFlux => real32Data(addColumn(histoTheoryTab, 'FLUX'&
      , REAL32, 'erg/cm^2/s'))
    histoLogNLogSDens => real32Data(addColumn(histoTheoryTab, 'DENSITY'&
      , REAL32, 'sources/deg^2'&
      , comment='Sky density of sources brighter than the given flux/rate.'))

    histoLogNLogSFlux(0) = real64Attribute(logNLogSTab, 'MINFLUX')
    histoLogNLogSFlux(1:numLogNLogSRows) = logNlogSFlux(0:numLogNLogSRows-1)
    histoLogNLogSFlux(numLogNLogSRows+1) = 10.0 * maxval(fluxSoftMid)

    logNlogSLoSlope = real64Attribute(logNlogSTab, 'LO_SLOPE')
    logNlogSHiSlope = real64Attribute(logNlogSTab, 'HI_SLOPE')

    histoLogNLogSDens(0) = logNlogSDens(0)&
      * (histoLogNLogSFlux(0) / logNlogSFlux(0))**logNlogSLoSlope
    histoLogNLogSDens(1:numLogNLogSRows) = logNlogSDens(0:numLogNLogSRows-1)
    histoLogNLogSDens(numLogNLogSRows+1) = logNlogSDens(numLogNLogSRows-1)&
      * (histoLogNLogSFlux(numLogNLogSRows+1)&
      / logNlogSFlux(numLogNLogSRows-1))**logNlogSHiSlope

    call release(logNLogSSet)

    allocate(amplitude(numLogNLogSRows+2))
    amplitude = fluxToFluxDensity0(photonSpectralIndex, EeVloFlux, EeVhiFlux&
      , histoLogNLogSFlux)

    histoLogNLogSRate = fluxDensity0ToRate(photonSpectralIndex, EeVloRate&
     , EeVhiRate, amplitude)!, filtrId)

    deallocate(amplitude)

    call release(histoSet)

  end subroutine calcCompletenessHistogramsPart1

  !---------------------------------------------------------------------
  subroutine calcCompletenessHistogramsPart2
  !---------------------------------------------------------------------
  ! function=completeness2
  !
  ! This sub adds columns pertaining to detected sources to the histogram table giving the incidence of sim sources as a function of their rate. (The bins are of equal size in log10(SIM_RATE).) Various other useful columns are calculated. The final tally of columns should be as follows:
  !
  ! Pre-existing:
  !
  !	N_SIM		Number of simulated sources.
  !	LOG10_SR_LO	Lower edge of the bin in log10(SIM_RATE).
  !	LOG10_SR_HI	Upper edge of the bin in log10(SIM_RATE).
  !	N_SIM_ERR	Square root of N_SIM.
  !	N_SIM_INT	Integrated total of N_SIM.
  !	DENS_SIM	Average sky-density (in deg^-2) of sim sources.
  !	DENS_SIM_ERR	Error in DENS_SIM.
  !	DENS_SIM_INT	Integrated total of DENS_SIM.
  !	SIM_RATE_MID	SIM_RATE at the middle of the bin. Use this for the x coordinate when plotting against differential quantities.
  !	SIM_RATE_INT	SIM_RATE at the lower edge of the bin. Use this for the x coordinate when plotting against integrated quantities.
  !	N_ALL_DET	All detections.
  !	N_ALL_DET_ERR	Square root of N_ALL_DET.
  !	N_ALL_DET_INT	Integrated total of N_ALL_DET.
  !	N_TRUE_DET	Reliable detections.
  !	N_TRUE_DET_ERR	Square root of N_TRUE_DET.
  !	N_TRUE_DET_INT	Integrated total of N_TRUE_DET.
  !	DENS_DET	Average sky-density (in deg^-2) of reliable detections.
  !	DENS_DET_ERR	Error in DENS_DET.
  !	DENS_DET_INT	Integrated total of DENS_DET.
  !	FLUX_MID	Calculated flux (in band 200-2000 eV) at bin centre.
  !	FLUX_INT	Calculated flux (in band 200-2000 eV) at the lower edge of the bin.
  !
  ! New:
  !
  !	COMP_RATIO	DENS_DET/DENS_SIM.
  !	COMP_RATIO_ERR	Error in COMP_RATIO.
  !	COMP_RATIO_INT	DENS_DET_INT/DENS_SIM_INT.

    implicit none

    character(255) ::&
      histoSetName
    type(DataSetT) :: histoSet
    type(TableT) :: histoTab
    type(ColumnT) ::&
      completenessRatioCol, completenessRatioErrCol, completenessRatioIntCol
    integer(int32) :: numBins, bin
    real(single), pointer ::&
      simSkyDensity(:), simSkyDensityErr(:), simSkyDensityInt(:),&
      completenessRatio(:), completenessRatioErr(:), completenessRatioInt(:)
    real(double), pointer ::&
      detSkyDensity(:), detSkyDensityErr(:), detSkyDensityInt(:)

    !-------------------------------------------------------------------
    ! Start

    histoSetName = stringParameter('histoset')
    histoSet = dataSet(histoSetName, MODIFY)
    histoTab = table(histoSet, 'HISTO', MODIFY)
    numBins = numberOfRows(histoTab)

    detSkyDensity => real64Data(column(histoTab, 'DENS_DET'))
    simSkyDensity => real32Data(column(histoTab, 'DENS_SIM'))
    completenessRatioCol = addOrOpenColumn(histoTab, 'COMP_RATIO', REAL32)
    do bin = 0, numBins - 1
      if (simSkyDensity(bin) <= 0.0) call setToNull(completenessRatioCol, bin)
    end do
    completenessRatio => real32Data(completenessRatioCol)
    where(simSkyDensity > 0.0)&
      completenessRatio = min(1.0, max(0.0&
        , real(detSkyDensity) / simSkyDensity))

    detSkyDensityInt => real64Data(column(histoTab, 'DENS_DET_INT'))
    simSkyDensityInt => real32Data(column(histoTab, 'DENS_SIM_INT'))
    completenessRatioIntCol = addOrOpenColumn(histoTab, 'COMP_RATIO_INT'&
      , REAL32)
    do bin = 0, numBins - 1
      if (simSkyDensityInt(bin) <= 0.0)&
        call setToNull(completenessRatioIntCol, bin)
    end do
    completenessRatioInt => real32Data(completenessRatioIntCol)
    where(simSkyDensityInt > 0.0)&
      completenessRatioInt&
        = min(1.0, max(0.0, real(detSkyDensityInt) / simSkyDensityInt))

    detSkyDensityErr => real64Data(column(histoTab, 'DENS_DET_ERR'))
    simSkyDensityErr => real32Data(column(histoTab, 'DENS_SIM_ERR'))
    completenessRatioErrCol = addOrOpenColumn(histoTab, 'COMP_RATIO_ERR'&
      , REAL32)

    completenessRatioErr => real32Data(completenessRatioErrCol)
    where(simSkyDensityErr > 0.0 .and. detSkyDensityErr > 0.0)
      completenessRatioErr = completenessRatio&
      *sqrt(detSkyDensityErr * detSkyDensityErr / detSkyDensity / detSkyDensity&
      +     simSkyDensityErr * simSkyDensityErr / simSkyDensity / simSkyDensity)
    elsewhere
      completenessRatioErr = 0.0
    end where

    call release(histoSet)

  end subroutine calcCompletenessHistogramsPart2

  !---------------------------------------------------------------------
  subroutine calcReliabilityHistogramPart1
  !---------------------------------------------------------------------
  ! Makes a single fits file that contains some histograms (usually over
  ! DET_ML). The column names are as follows:
  !
  !	N_ALL		Number sources detected.
  !	N_ALL_ERR	Square root of N_ALL.
  !	N_ALL_INT	Integrated value of N_ALL.
  !	N_TRUE		Number of reliable detections.
  !	N_TRUE_ERR	Square root of N_TRUE.
  !	N_TRUE_INT	Integrated value of N_TRUE.
  !	DETML_LO	Lower edge of the bin.
  !	DETML_MID	Middle of the bin.
  !	DETML_HI	Upper edge of the bin.

    implicit none

    real(single) :: binSize, minDetml, maxDetml, overShoot
    character(255) ::&
      histoSetName, allDetSetName, truDetSetTabName, truDetSetName, truDetTabName
    type(DataSetT) :: allDetSet, truDetSet, histoSet
    type(TableT) :: allDetTab, truDetTab, histoTab
    integer(int32), pointer :: nAll(:), nAllInt(:), nTru(:), nTruInt(:)
    real(single), pointer ::&
      nAllErr(:), nTruErr(:),&
      allDetDetml(:), truDetDetml(:), detmlLo(:), detmlMid(:), detmlHi(:)
    integer(int32) :: numBins, bin
    real(single), allocatable :: binEdges(:)
    type(histoType), pointer :: allDetHisto(:), truDetHisto(:)

    !-------------------------------------------------------------------
    ! Start

    binSize = realParameter('histobinsize')
    histoSetName = stringParameter('histoset')
    allDetSetName = stringParameter('detsrclistset')
    truDetSetTabName = stringParameter('tab')
    call splitSetTabName(truDetSetTabName, truDetSetName, truDetTabName)

    allDetSet = dataSet(allDetSetName, READ)
    allDetTab = table(allDetSet, 'SRCLIST')
    allDetDetml => real32Data(column(allDetTab, 'DET_ML'))

    truDetSet = dataSet(truDetSetName, READ)
    truDetTab = table(truDetSet, 'SRCLIST')
    truDetDetml => real32Data(column(truDetTab, 'DET_ML'))

    minDetml = min(minval(allDetDetml), minval(truDetDetml))
    maxDetml = max(maxval(allDetDetml), maxval(truDetDetml))

    maxDetml = min(maxDetml, 100.0)
!**** note this truncation in the doco!!!

    numBins = max(1, ceiling((maxDetml - minDetml) / binSize)) ! there is
    ! at least 1 bin, even if all detml values are the same.

    overShoot = ((numBins * binSize) - (maxDetml - minDetml)) / 2.0

    allocate(binEdges(numBins+1))
    do bin = 1, numBins+1
      binEdges(bin) = fracval(minDetml - overShoot, maxDetml + overShoot&
        , (bin - 1) / real(numBins))
    end do

    call makeHistogram(allDetDetml, allDetHisto, binEdges=binEdges)
    call makeHistogram(truDetDetml, truDetHisto, binEdges=binEdges)

    histoSet = dataSet(histoSetName, CREATE)
    histoTab = addTable(histoSet, 'HISTO', numBins)

    call setAttribute(histoTab, 'SKY_AREA'&
!      , real32Attribute(allDetTab, 'SKY_AREA'), 'deg^2')
, real64Attribute(allDetTab, 'SKY_AREA'), 'deg^2')
    call setAttribute(histoTab, 'N_FIELDS'&
      , int32Attribute(allDetTab, 'N_FIELDS'))

    nAll     =>  int32Data(addColumn(histoTab, 'N_ALL',      INTEGER32))
    nAllErr  => real32Data(addColumn(histoTab, 'N_ALL_ERR',  REAL32))
    nAllInt  =>  int32Data(addColumn(histoTab, 'N_ALL_INT',  INTEGER32))
    nTru     =>  int32Data(addColumn(histoTab, 'N_TRUE',     INTEGER32))
    nTruErr  => real32Data(addColumn(histoTab, 'N_TRUE_ERR', REAL32))
    nTruInt  =>  int32Data(addColumn(histoTab, 'N_TRUE_INT', INTEGER32))
    detmlLo  => real32Data(addColumn(histoTab, 'DETML_LO',   REAL32))
    detmlMid => real32Data(addColumn(histoTab, 'DETML_MID',  REAL32))
    detmlHi  => real32Data(addColumn(histoTab, 'DETML_HI',   REAL32))

    nAll = allDetHisto%n
    nAllErr = sqrt(real(nAll))
    nAllInt(numBins-1) = nAll(numBins-1)
    do bin = numBins-2, 0, -1
      nAllInt(bin) = nAllInt(bin+1) + nAll(bin)
    end do

    nTru = truDetHisto%n
    nTruErr = sqrt(real(nTru))
    nTruInt(numBins-1) = nTru(numBins-1)
    do bin = numBins-2, 0, -1
      nTruInt(bin) = nTruInt(bin+1) + nTru(bin)
    end do

    detmlLo = allDetHisto%xLo
    detmlHi = allDetHisto%xHi
    detmlMid = (detmlLo + detmlHi) / 2.0

    call release(histoSet)
    deallocate(allDetHisto, truDetHisto)
    call release(truDetSet)
    call release(allDetSet)

  end subroutine calcReliabilityHistogramPart1

  !---------------------------------------------------------------------
  subroutine calcReliabilityHistogramPart2
  !---------------------------------------------------------------------
  ! Makes a single fits file that contains some histograms (usually over
  ! DET_ML). The column names are as follows:
  !
  ! Pre-existing:
  !
  !	N_ALL		Number sources detected.
  !	N_ALL_ERR	Square root of N_ALL.
  !	N_ALL_INT	Integrated value of N_ALL.
  !	N_TRUE		Number of reliable detections.
  !	N_TRUE_ERR	Square root of N_TRUE.
  !	N_TRUE_INT	Integrated value of N_TRUE.
  !	DETML_LO	Lower edge of the bin.
  !	DETML_MID	Middle of the bin.
  !	DETML_HI	Upper edge of the bin.
  !	DENS_ALL	Average sky-density (in deg^-2) of all detections.
  !	DENS_ALL_ERR	Error in DENS_ALL.
  !	DENS_ALL_INT	Integrated total of DENS_ALL.
  !	DENS_TRUE	Average sky-density (in deg^-2) of reliable detections.
  !	DENS_TRUE_ERR	Error in DENS_TRUE.
  !	DENS_TRUE_INT	Integrated total of DENS_TRUE.
  !
  ! New:
  !
  !	FIRMFRAC	DENS_TRUE/DENS_ALL.
  !	FIRMFRAC_ERR	Error in FIRMFRAC.
  !	FIRMFRAC_INT	DENS_TRUE_INT/DENS_ALL_INT.

    implicit none

    character(255) ::&
      histoSetName
    type(DataSetT) :: histoSet
    type(TableT) :: histoTab
    type(ColumnT) :: firmFractionCol, firmFractionIntCol, firmFractionErrCol
    real(double), pointer ::&
      skyDensityAll(:),    skyDensityTru(:),&
      skyDensityAllInt(:), skyDensityTruInt(:),&
      skyDensityAllErr(:), skyDensityTruErr(:)
    real(single), pointer ::&
      firmFraction(:), firmFractionInt(:), firmFractionErr(:)
    integer(int32) :: numBins, bin

    !-------------------------------------------------------------------
    ! Start

    histoSetName = stringParameter('histoset')
    histoSet = dataSet(histoSetName, MODIFY)
    histoTab = table(histoSet, 'HISTO')!, MODIFY)
    numBins = numberOfRows(histoTab)

    skyDensityAll => real64Data(column(histoTab, 'DENS_ALL'))
    skyDensityTru => real64Data(column(histoTab, 'DENS_TRUE'))
    firmFractionCol = addOrOpenColumn(histoTab, 'FIRMFRAC', REAL32)
    do bin = 0, numBins - 1
      if (skyDensityAll(bin) <= 0.0) call setToNull(firmFractionCol, bin)
    end do
    firmFraction => real32Data(firmFractionCol)
    where(skyDensityAll > 0.0)&
      firmFraction = min(1.0d0, max(0.0d0, skyDensityTru / skyDensityAll))

    skyDensityAllInt => real64Data(column(histoTab, 'DENS_ALL_INT'))
    skyDensityTruInt => real64Data(column(histoTab, 'DENS_TRUE_INT'))
    firmFractionIntCol = addOrOpenColumn(histoTab, 'FIRMFRAC_INT', REAL32)
    do bin = 0, numBins - 1
      if (skyDensityAllInt(bin) <= 0.0) call setToNull(firmFractionIntCol, bin)
    end do
    firmFractionInt => real32Data(firmFractionIntCol)
    where(skyDensityAllInt > 0.0)&
      firmFractionInt = min(1.0d0, max(0.0d0&
        , skyDensityTruInt / skyDensityAllInt))

    skyDensityAllErr => real64Data(column(histoTab, 'DENS_ALL_ERR'))
    skyDensityTruErr => real64Data(column(histoTab, 'DENS_TRUE_ERR'))
    firmFractionErrCol = addOrOpenColumn(histoTab, 'FIRMFRAC_ERR', REAL32)
    do bin = 0, numBins - 1
      if (skyDensityAllErr(bin) <= 0.0 .or. skyDensityTruErr(bin) <= 0.0)&
        call setToNull(firmFractionErrCol, bin)
    end do
    firmFractionErr => real32Data(firmFractionErrCol)
    where(skyDensityAllErr > 0.0 .and. skyDensityTruErr > 0.0)
      firmFractionErr = firmFraction&
      *sqrt(skyDensityAllErr * skyDensityAllErr / skyDensityAll / skyDensityAll&
      +     skyDensityTruErr * skyDensityTruErr / skyDensityTru / skyDensityTru)
    end where

    call release(histoSet)

  end subroutine calcReliabilityHistogramPart2

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module eimsim_reduction
