!
! Module: eimsim_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines to generate a simulated XMM EPIC image and to compare the list of simulated sources used to make that image with the list of sources detected in it.
! 
! Comments: TODO:
!
! $Id: eimsim_utils.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eimsim_utils

  use dal
  use cal
  use errorHandling
  use datafilter
  use dss		! for RangeT

  use array_utils	! for indicize(), getBinNumber(), smallestTrueIndex()
  use cal_aux
  use constants
  use coordinate
  use dal_aux		! for readArrayData(), addOrOpenColumn(),
			! readColDataToPtr().
  use detection_stats	! for inverseEddington().
  use dump
  use fftw_aux		! just used in calcClumpProbMap.
  use flux_transforms
  use geometric_types
  use math_utils
!  use polygon		! for regrid() in getCalPsfPatch.
  use psf_ims
  use random_aux
  use rebinners		! for rebinImage
  use regridders	! for regrid() in getCalPsfPatch.
  use save_image	! for saveImage()
  use sort_mod
  use ssc_misc		! for ucase()
use temp_ssclib	! for myAddColumn() + all the cone stuff.
use test_routines
  use wcs_aux

  implicit none 

  private 

  ! eimsim_prepare calls:
  public makeEcfFile
  public makeLocationTemplate
  public makePaddedMask
  public makeBkgMap
  public makeMapFaintSrcBkg

  ! eimsim_single calls:
  public makeSimSrcList
  public makeRateImage
  public makeAvCountsImage
  public poissonizeCountsImage
  public correctEddington
  public correctPosAndRates
  public detRateToFluxRandom
  public compareDetWithSim
  public addSimRates

  public eventizeBlankImage

! Called by eimsimtest:
  public makeUncorrPositions

! Was called by eimsimtest at one stage:
!  public makeUncorrSrcListSet
!  public recalcFluxRandoms

public calcNumSrcsInVolume ! just so as to allow testing of this routine.

  type(WcsT) :: wcs
  integer(int32) ::&
    instrumentId
  type(SpacecraftAttitudeType) ::&
    scAttitude, instrumAttitude
  real(double) ::&
    expStartTimeStamp

!  real(double) ::&
!    mmToArcsec,&
!    focalLengthMm

  type, public :: localSrcDataType !!! only used in makeSourceListClumped.
    real(single) ::&
      xPixels,&
      yPixels,&
      exposure,&
      rate,&
      peakRate,&
      extentWorld,&
      extentPixels
    real(double) ::&
      raDeg,&
      decDeg,&
      EeV,&
      thetaArcsec,&
      phi
    integer(int32) ::&
      flag
  end type localSrcDataType

  type :: rangesPtrType
    type(RangeT), pointer :: ranges(:)
  end type rangesPtrType

  integer(int32), public ::& ! flag bits
    NOT_IN_M1_FOV =  0,&
    NOT_IN_M2_FOV =  1,&
    NOT_IN_PN_FOV =  2,&
    NOT_IN_M1_EXP =  3,&
    NOT_IN_M2_EXP =  4,&
    NOT_IN_PN_EXP =  5,&
    DETECTED_M1   =  6,&
    DETECTED_M2   =  7,&
    DETECTED_PN   =  8,&
    COARSE_PSF(5)  =  (/9, 10, 11, 12, 13/)

  real(single), parameter :: detSrcXoffset = -0.5, detSrcYoffset = -0.5
  ! Both in image pixels.

contains 

  !---------------------------------------------------------------------
  subroutine makeEcfFile(ecfFileName, templateSetName)
  !---------------------------------------------------------------------
  ! This makes a text file (read by the script eimsim_single) which contains
  ! the energy limits of the five energy bands and the ecf values which are
  ! used to convert between source count rate R (in counts/s) and flux S
  ! (in erg/cm^2/s). An ecf value Y is defined such that
  !
  !	R = 10^11 Y S.
  !
  ! Both R and S are assumed to be measured within the same energy band.


    implicit none

    character(*), intent(in) :: ecfFileName, templateSetName
    integer(int32) :: eLo(5), eHi(5)
    integer :: ioUnit, instId, filterId
    character(1) :: tabChr
    real(single) :: ecf(5)
    character(2) :: inst(3)
    character(6) :: filter(4)
real(single) :: effArea
integer(int32) :: instrumentId

    !-------------------------------------------------------------------
    ! Start

do instrumentId = EMOS1, EPN
!instrumentId = EPN
  call CAL_setState(instrument=instrumentId)
  effArea = 0.0
  effArea = CAL_getEffectiveArea(900.0d0, 0.0d0, 0.0d0)
  call CAL_getState(instrument=instrumentId)
  print *, 'instrumentId, effArea: ', instrumentId, effArea
end do

call warning('test','This is a test routine! Real makeecf function has been suspended.')

return

    eLo = (/200,  500, 2000, 4500,  7500/)
    eHi = (/500, 2000, 4500, 7500, 12000/)



!!    call CAL_setState(instrument=EMOS1)
!    call CAL_setState(instrument=EPN)
!
!    call CAL_setState(filterId=THIN1)
!    ecf0 = calcEcf(-1.4d0, real(eLo(2), double), real(eHi(2), double)) ! in flux_transforms
!print *, 'ecf:', ecf0
!
!  return



    tabChr = achar(9)
    inst = (/'M1', 'M2', 'PN'/)
    filter = (/'OPEN  ', 'THIN  ', 'MEDIUM', 'THICK '/)




    call getFreeIoUnit(ioUnit)
    open(ioUnit, file=ecfFileName, status='REPLACE', access='SEQUENTIAL'&
      , form='FORMATTED')

    write(ioUnit, '(a)')&
      '# Bands'//tabChr//'1'//tabChr//'2'//tabChr//'3'//tabChr//'4'//tabChr//'5'
    write(ioUnit, '(a)') ''
    write(ioUnit, '(a)') '# Band start (eV):'
    write(ioUnit, '(a1, i5, a1, i5, a1, i5, a1, i5, a1, i5)')&
      tabChr,eLo(1),tabChr,eLo(2),tabChr,eLo(3),tabChr,eLo(4),tabChr,eLo(5)
    write(ioUnit, '(a)') '# Band stop (eV):'
    write(ioUnit, '(a1, i5, a1, i5, a1, i5, a1, i5, a1, i5)')&
      tabChr,eHi(1),tabChr,eHi(2),tabChr,eHi(3),tabChr,eHi(4),tabChr,eHi(5)
    write(ioUnit, '(a)') ''

    do instId = 1, 3
      write(ioUnit, '(a)') '# inst='//inst(instId)//':'
      do filterId = 1, 4
        write(ioUnit, '(a)') '# filter='//trim(filter(filterId))//':'
        call calcEcfs(eLo, eHi, inst(instId), filter(filterId), templateSetName&
          , ecf)

        if (ecf(1) >= 100.0 .or. any(ecf(2:) >= 10.0))&
call error('badEcfValue','eimsim_utils: Computed ECF value too big.')

        write(ioUnit, '(a1, f8.5, a1, f8.6, a1, f8.6, a1, f8.6, a1, f8.6)')&
          tabChr,ecf(1),tabChr,ecf(2),tabChr,ecf(3),tabChr,ecf(4),tabChr,ecf(5)
      end do
      write(ioUnit, '(a)') ''
    end do

    close(ioUnit)

  end subroutine makeEcfFile

  !---------------------------------------------------------------------
  subroutine calcEcfs(eLo, eHi, instStr, filterStr, templateSetName, ecfs)
  !---------------------------------------------------------------------
  ! This calculates (for a given combination of instrument and filter) the
  ! ecf values which are used to convert between source count rate R (in
  ! counts/s) and flux S (in erg/cm^2/s). An ecf value Y is defined such
  ! that
  !
  !	R = 10^11 Y S.
  !
  ! Both R and S are assumed to be measured within the same energy band.

    implicit none

    integer(int32), intent(in)  :: eLo(5), eHi(5)
    character(*),   intent(in)  :: instStr, filterStr, templateSetName
    real(single),   intent(out) :: ecfs(5)

    type(DataSetT) :: logNlogSset
    type(TableT) :: logNlogStab
    real(double) :: photonSpectralIndex
    integer :: band

    !-------------------------------------------------------------------
    ! Start

    logNlogSset = dataSet(templateSetName, READ)
    logNlogStab = table(logNlogSset, 'LOGNLOGS')
    photonSpectralIndex = real64Attribute(logNlogStab, 'SPECINDX')
    call release(logNlogSset)

    select case(instStr)
    case('M1')
      call CAL_setState(instrument=EMOS1)
    case('M2')
      call CAL_setState(instrument=EMOS2)
    case('PN')
      call CAL_setState(instrument=EPN)
    case default
call error('badInstrument',"eimsim_utils: Unrecognized instrument ID.")
    end select

    select case(filterStr)
    case('OPEN')
      call CAL_setState(filterId=OPEN)
    case('THIN')
      call CAL_setState(filterId=THIN1)
    case('MEDIUM')
      call CAL_setState(filterId=MEDIUM)
    case('THICK')
      call CAL_setState(filterId=THICK)
    case default
call error('badFilter',"eimsim_utils: Unrecognized filter ID.")
    end select

    do band = 1, 5
      ecfs(band) = calcEcf(photonSpectralIndex, real(eLo(band), double)&
        , real(eHi(band), double)) ! in flux_transforms

print *, 'instStr, filterStr, band, photonSpectralIndex, eLo(band), eHi(band), ecfs(band)'&
, instStr, filterStr, band, photonSpectralIndex, eLo(band), eHi(band), ecfs(band)
print *, ''
print *, ''

    end do

  end subroutine calcEcfs

  !---------------------------------------------------------------------
  subroutine makeLocationTemplate(expMapNames, templateSetName, padSizeArcsec)
  !---------------------------------------------------------------------
  ! This subroutine takes a list of exposure maps and calculates the
  ! direction and angle of the narrowest conical locus (or, circle on
  ! the celestial sphere) which includes all non-zero parts of each
  ! exposure map. The padSizeArcsec is then added to this angle and the
  ! resulting pair of values (ie the direction and the angle) are written
  ! to the specified template file. The intention is that the routine which
  ! constructs the list of sim sources should read the cone details from
  ! the template set (which should also contain the logN-logS details for
  ! the source brightness distribution) and restrict the spatial
  ! distribution of sources to within the cone.
  !
  ! In practice (at least at first writing) I think I will attempt a
  ! rather simpler problem. I'm going to represent the non-zero exposure
  ! of a single exposure map by a cone of 15 arcmin radius, centred on the
  ! optic axis. Such a cone ought to include all the non-zero exposure,
  ! and this will make construction of the 'union' cone for all maps much
  ! easier.
  !
  ! The intent of this procedure is to allow simulations with overlapping
  ! fields. There isn't much point in doing this for a list of exposure
  ! maps if they don't substantially overlap. However at present there is
  ! no protection against calling makeLocationTemplate with a list of
  ! widely-scattered exposure maps.

    implicit none

    character(*), intent(in) :: expMapNames(:)
    character(*), intent(in) :: templateSetName
    real(double), intent(in) :: padSizeArcsec

    integer(int8), parameter :: X=1, Y=2, Z=3
    integer :: numFiles, fileNum
    type(WcsT) :: wcs
    real(double) ::&
      centralRay(X:Z), coneAngle, centreRa, centreDec, coneAngleArcSec,&
      coneAreaDeg
    real(double), allocatable :: rays(:,:)
    type(DataSetT) :: set
    type(TableT) :: tab

    !-------------------------------------------------------------------
    ! Start

    numFiles = size(expMapNames)
    if (numFiles <= 0)&
call error('noInfiles','eimsim_utils: no exposure maps were supplied.')

    allocate(rays(numFiles, X:Z))

    do fileNum = 1, numFiles
      call readWcs(expMapNames(fileNum), wcs) ! in ssclib/wcs_aux
      rays(fileNum, X:Z) = skyToCartesian(wcs%x%refWorld * PiOn180&
        , wcs%y%refWorld * PiOn180) ! in ssclib/coordinate
    end do

    select case(numFiles)
    case(1)
      centralRay = rays(1,X:Z)
      coneAngle = 0.0d0

    case(2)
      call calcSmallestConePair(rays(1,X:Z), rays(2,X:Z), centralRay, coneAngle)
      ! at present this is in temp_ssclib.

    case(3)
      call calcSmallestConeTriplet(rays, centralRay, coneAngle)
      ! at present this is in temp_ssclib.

    case default
      call calcSmallestCone(rays, centralRay, coneAngle)
      ! at present this is in temp_ssclib.

!call test_doConePlot(rays, centralRay, coneAngle)
    end select

    coneAngleArcSec = (coneAngle * 3600.0d0 / PiOn180) + 900.0 + padSizeArcsec
    ! The 900 is the approximate radius in arcsec of EPIC fields.

    coneAngle = coneAngleArcSec / 3600.0 * PiOn180
    coneAreaDeg = (TwoPi * (1.0d0 - cos(coneAngle))) / PiOn180 / PiOn180

!print *, 'coneAngleArcSec, coneAngle, coneArea, coneAreaDeg:'&
!, coneAngleArcSec, coneAngle, TwoPi * (1.0d0 - cos(coneAngle)), coneAreaDeg

    call cartesianToSky(centralRay, centreRa, centreDec) ! in ssclib/coordinate
    set = dataSet(templateSetName, MODIFY)
    tab = table(set, 'LOGNLOGS')
    call setAttribute(tab, 'CONE_RA',  centreRa  / PiOn180, 'deg')
    call setAttribute(tab, 'CONE_DEC', centreDec / PiOn180, 'deg')
    call setAttribute(tab, 'CONE_RAD', coneAngleArcSec, 'arcsec')
    call setAttribute(tab, 'SKY_AREA', coneAreaDeg, 'deg^2')

    call release(set)

  end subroutine makeLocationTemplate

  !---------------------------------------------------------------------
  subroutine makePaddedMask(expMapSetName, paddedMaskSetName, padSizeArcsec&
  , tempSetName)
  !---------------------------------------------------------------------
  ! The output from this sub is used to select which sources from the list
  ! (which covers in general a larger area of the sky) will be added to the
  ! rate image. The padded mask is formed by first creating a mask from an
  ! exposure map and then adding a border of padSizeArcsec width. This
  ! ensures that sources which have centres which fall outside the exposure
  ! map but nevertheless would be expected to make some contribution to the
  ! image through their wings are included on the rate image.

    implicit none

    character(*), intent(in) :: expMapSetName, paddedMaskSetName, tempSetName
    real(double), intent(in) :: padSizeArcsec

    integer(int16) :: xSize, ySize, xiPad, yiPad,&
      xi,&
      yi, startXi, finisXi, startYi, finisYi
    type(WcsT) :: wcs
    logical(bool), pointer :: mask(:,:), edgeMask(:,:)
    logical(bool), allocatable :: padArray(:,:)
    real(single) :: x, y, pixelSizeXWorld, pixelSizeYWorld
    real(single), pointer :: expMapImage(:,:)
    type(DataSetT) :: maskSet
    type(ArrayT) :: maskArray

    !-------------------------------------------------------------------
    ! Start

    call readWcs(expMapSetName, wcs)
    call readArrayData(expMapSetName, expMapImage)
    xSize = size(expMapImage, 1)
    ySize = size(expMapImage, 2)
    allocate(mask(xSize, ySize))
    mask = (expMapImage > 0.0)
    deallocate(expMapImage)

    ! Now add the pad:

!    xiPad = ceiling(padSizeArcsec / 3600.0 / abs(wcs%x%pixelSizeWorld))
!    yiPad = ceiling(padSizeArcsec / 3600.0 / abs(wcs%y%pixelSizeWorld))
!vvvvv
! Have to replace the above by the following kludge because pixelSizeWorld is no longer a component of the WcsAxisT type.

    if (wcs%x%ctype .ne. 'RA---TAN' .or. wcs%y%ctype .ne. 'DEC--TAN')&
call error('',"")

    pixelSizeXWorld = sqrt(wcs%x%matrixOnAxis * wcs%x%matrixOnAxis&
     + wcs%x%matrixOffAxis * wcs%x%matrixOffAxis)
    pixelSizeYWorld = sqrt(wcs%y%matrixOnAxis * wcs%y%matrixOnAxis&
     + wcs%y%matrixOffAxis * wcs%y%matrixOffAxis)
!*** this is only valid if there is no skew.

    select case(wcs%x%worldUnit)
    case('deg')
      xiPad = ceiling(padSizeArcsec / 3600.0 / pixelSizeXWorld)
    case default
call error('notSupported',"")
    end select

    select case(wcs%y%worldUnit)
    case('deg')
      yiPad = ceiling(padSizeArcsec / 3600.0 / pixelSizeYWorld)
    case default
call error('notSupported',"")
    end select

!^^^^^

    ! First make elliptical pad array (elliptical rather than circular
    ! because the image pixels may not be square):

    allocate(padArray(-xiPad:xiPad,-yiPad:yiPad))
    padArray = .false. ! default
    do xi = -xiPad, xiPad
      x = xi / real(xiPad)
      do yi = -yiPad, yiPad
        y = yi / real(yiPad)
        if (x*x+y*y .lt. 1.0) padArray(xi, yi) = .true.
      end do
    end do

    allocate(edgeMask(xSize, ySize))
    call findEdges(mask, edgeMask)

    ! For each of the edge points of mask, add a copy of the pad array.
    ! (Another way to think of this is as a convolution of the edge mask
    ! with the pad array.) The result is added to mask.

    do xi = 1, xSize
    if (all(.not. edgeMask(xi, :))) cycle
      do yi = 1, ySize
      if (.not. edgeMask(xi, yi)) cycle
        startXi = max(1_int16, xi - xiPad)
        finisXi = min(xSize,   xi + xiPad)
        startYi = max(1_int16, yi - yiPad)
        finisYi = min(ySize,   yi + yiPad)

        mask(startXi:finisXi, startYi:finisYi)&
        = mask(startXi:finisXi, startYi:finisYi)&
        .or. padArray(startXi-xi:finisXi-xi,startYi-yi:finisYi-yi)
      end do
    end do
    deallocate(edgeMask, padArray)

    call saveImage(mask, expMapSetName, paddedMaskSetName, tempSetName)
    deallocate(mask)

    maskSet = dataSet(paddedMaskSetName, MODIFY)
    maskArray = array(maskSet, 0, MODIFY)

    call setAttribute(maskArray, 'PADSIZE', padSizeArcsec, 'arcsec')

    call release(maskSet)

  end subroutine makePaddedMask

  !---------------------------------------------------------------------
  subroutine makeBkgMap(expMapSetName, nonVigSetName, bkgMapSetName&
  , tempImageName, vigBkgFlux, nonVigBkgFlux)
  !---------------------------------------------------------------------

    implicit none

    character(*),   intent(in) :: expMapSetName,&
                                  nonVigSetName,&
                                  bkgMapSetName,&
                                  tempImageName
    real(double),   intent(in) :: vigBkgFlux, nonVigBkgFlux
    ! both in counts / arcsec^2 / s.

    real(single), pointer ::&
      expMapImage(:,:), nonVigImage(:,:)
    real(single) :: pixelAreaArcsec
    type(DataSetT) :: bkgSet
    integer(int16) :: xSize, ySize
    real(single), allocatable :: bkgCountsMapImage(:,:)
    type(WcsT) :: imageWcs

    !-------------------------------------------------------------------
    ! Start

    call readArrayData(expMapSetName, expMapImage)
    call readArrayData(nonVigSetName, nonVigImage)

    xSize = size(expMapImage, 1)
    ySize = size(expMapImage, 2)
    if ( size(nonVigImage, 1) .ne. xSize&
    .or. size(nonVigImage, 2) .ne. ySize)&
call error('imageSizesDifferent','eimsim_utils: the exposure map and &
&the nonvignetted map have different image sizes.')

    allocate(bkgCountsMapImage(xSize, ySize))

    call readWcs(expMapSetName, imageWcs) ! in ssclib/wcs_aux.
!    pixelAreaArcsec = 3600.0 * 3600.0 * abs(imageWcs%x%pixelSizeWorld&
!      * imageWcs%y%pixelSizeWorld)
pixelAreaArcsec = 3600.0 * 3600.0 * wcsPixelAreaWorld(imageWcs)
!*** strictly speaking should check that the world units are degrees.

    bkgCountsMapImage = ((nonVigImage / maxval(nonVigImage))&
      * nonVigBkgFlux * pixelAreaArcsec&
      + (expMapImage / maxval(expMapImage))&
      * vigBkgFlux * pixelAreaArcsec) * maxval(expMapImage)
    ! Units of the bkgCountsMapImage are counts/pixel.

    call saveImage(bkgCountsMapImage, expMapSetName, bkgMapSetName&
      , tempImageName)

    bkgSet = dataSet(bkgMapSetName, MODIFY)
    if (vigBkgFlux + nonVigBkgFlux > 0.0) then
      call setAttribute(bkgSet, 'VIG_FRAC', vigBkgFlux &
        / (vigBkgFlux + nonVigBkgFlux)&
        , comment='Fraction of the total background which is vignetted.')
    end if
    call release(bkgSet)

  end subroutine makeBkgMap

  !---------------------------------------------------------------------
  subroutine makeMapFaintSrcBkg(sensitivityMapName, templateSetName&
  , expMapSetName, EeVlo, EeVhi, ecf, faintSrcBkgMapName, tempImageName)
  !---------------------------------------------------------------------
  ! Takes the logN-logS template, exposure map and a sensitivity map as inputs. At each pixel, the subroutine estimates the average flux contribution due to undetectable sources and writes this to the output.

!imageWcs
    implicit none

    character(*), intent(in) :: sensitivityMapName,&
                                templateSetName,&
                                expMapSetName,&
                                faintSrcBkgMapName,&
                                tempImageName
    real(double), intent(in) :: EeVlo,&
                                EeVhi,&
                                ecf

    type(DataSetT) :: templateSet, sensMapSet
    type(TableT) :: templateTab
    real(single), pointer ::&
      modelFluxDensity0(:), modelCumSkyDensity(:), modelFlux(:),&
      expMapImage(:,:), sensMapImage(:,:), sensMapImageFlux(:,:)
    real(single) :: hiPower, convertToFlux, pixelAreaDeg
    real(double) :: photonSpectralIndex
    real(single), allocatable ::&
      modelPowers(:), modelAccumFlux(:), faintSrcBkgMapImage(:,:)
    type(WcsT) :: wcs
    integer(int32) :: xSize, ySize, xi, yi
    type(ArrayT) :: sensMapArray

    !-------------------------------------------------------------------
    ! Start

    templateSet = dataSet(templateSetName, READ)
    templateTab = table(templateSet, 'LOGNLOGS')
    call getLogNlogS(templateTab, modelFluxDensity0, modelCumSkyDensity&
      , hiPower, photonSpectralIndex)

    convertToFlux = fluxDensity0ToFlux(photonSpectralIndex&
      , EeVlo, EeVhi, 1.0)

    modelFlux => modelFluxDensity0
    modelFlux = modelFluxDensity0 * convertToFlux

    allocate(modelPowers(size(modelFlux)), modelAccumFlux(size(modelFlux)))
    call calcLogNLogSPowers(modelFluxDensity0, modelCumSkyDensity, hiPower&
      , modelPowers)

    call calcAccumFlux(modelFlux, modelCumSkyDensity, modelPowers&
      , modelAccumFlux)

    call readArrayData(expMapSetName, expMapImage)
    call readWcs(expMapSetName, wcs)
    xSize = size(expMapImage, 1)
    ySize = size(expMapImage, 2)

    sensMapSet = dataSet(sensitivityMapName, READ)
    sensMapArray = array(sensMapSet, 0)
    call readArrayData(sensMapArray, sensMapImage)
    if (size(sensMapImage,1) .ne. xSize .or. size(sensMapImage,2) .ne. ySize)&
call error('','')

    sensMapImageFlux => sensMapImage
!    sensMapImageFlux = sensMapImage&
!      * rateToFluxDensity0(photonSpectralIndex&
!      , EeVlo, EeVhi, 1.0) * multiplier ! convert to flux.
sensMapImageFlux = sensMapImage / ecf / 1.0e11

    allocate(faintSrcBkgMapImage(xSize, ySize))
    faintSrcBkgMapImage = 0.0
    do xi = 1, xSize
      do yi = 1, ySize
      if (isNullArray_ims(sensMapArray, xi, yi)) cycle
      if (sensMapImageFlux(xi, yi) <= modelFlux(1)) cycle

        call calcUndetectableFlux(modelFlux, modelCumSkyDensity, modelPowers&
          , modelAccumFlux, sensMapImageFlux(xi, yi)&
          , faintSrcBkgMapImage(xi, yi))
      end do
    end do

    ! Convert from flux /deg^2 to counts / s / pixel:

!    pixelAreaDeg = abs(wcs%x%pixelSizeWorld * wcs%y%pixelSizeWorld)
pixelAreaDeg = wcsPixelAreaWorld(wcs)
!*** strictly speaking should check that the world units are degrees.

!    faintSrcBkgMapImage = faintSrcBkgMapImage * pixelAreaDeg / convertToFlux&
!      * fluxDensity0ToRate(photonSpectralIndex, EeVlo, EeVhi, 1.0)
faintSrcBkgMapImage = faintSrcBkgMapImage * pixelAreaDeg * ecf * 1.0e11

    ! Finally, convert to counts / pixel:

    faintSrcBkgMapImage = faintSrcBkgMapImage * expMapImage

    call saveImage(faintSrcBkgMapImage, expMapSetName, faintSrcBkgMapName&
      , tempImageName)

    call release(sensMapSet)
    call release(templateSet)
    deallocate(modelFluxDensity0, modelCumSkyDensity, expMapImage, sensMapImage)

  end subroutine makeMapFaintSrcBkg

  !---------------------------------------------------------------------
  subroutine calcLogNLogSPowers(fluxDensity0, cumSkyDensity, hiPower&
  , powers)
  !---------------------------------------------------------------------
!*** not yet tested

    implicit none

    real(single), intent(in)  :: fluxDensity0(:),&
                                 cumSkyDensity(size(fluxDensity0)),&
                                 hiPower
    real(single), intent(out) :: powers(size(fluxDensity0))

    integer :: numSamples, sampleNum

    !-------------------------------------------------------------------
    ! Start

    numSamples = size(fluxDensity0)
    do sampleNum = 1, numSamples - 1
      powers(sampleNum)&
        = (log(cumSkyDensity(sampleNum+1)) - log(cumSkyDensity(sampleNum)))&
        / (log(fluxDensity0( sampleNum+1)) - log(fluxDensity0( sampleNum)))
    end do

    powers(numSamples) = hiPower

    powers = powers - 1 ! to return the differential rather than integral
    ! powers.

  end subroutine calcLogNLogSPowers

  !---------------------------------------------------------------------
  subroutine calcAccumFlux(flux, cumSkyDensity, powers, accumFlux)
  !---------------------------------------------------------------------
!*** not yet tested

    implicit none

    real(single), intent(in)  :: flux(:),&
                                 cumSkyDensity(size(flux)),&
                                 powers(size(flux))
    real(single), intent(out) :: accumFlux(size(flux))

    integer :: numSamples, sampleNum

    !-------------------------------------------------------------------
    ! Start

    numSamples = size(flux)

if (powers(numSamples) >= -2.0) call error('fluxIntegralUnbounded','')

    accumFlux(numSamples)&
      = ((1.0 + powers(numSamples)) / (2.0 + powers(numSamples)))&
      * cumSkyDensity(numSamples) * flux(numSamples)

    do sampleNum = numSamples - 1, 1, -1
      if (2.0 + powers(sampleNum) == 0.0) then
call error('notYetSupported','')

      else
        accumFlux(sampleNum) = accumFlux(sampleNum + 1)&
          + ((1.0 + powers(sampleNum)) / (2.0 + powers(sampleNum)))&
          * (cumSkyDensity(sampleNum))&
          / flux(sampleNum)**(1.0 + powers(sampleNum))&
          * (flux(sampleNum    )**(2.0 + powers(sampleNum))&
          -  flux(sampleNum + 1)**(2.0 + powers(sampleNum)))
      end if
    end do

  end subroutine calcAccumFlux

  !---------------------------------------------------------------------
  subroutine calcUndetectableFlux(flux, cumSkyDensity, powers&
  , accumFlux, fluxSample, undetectableTotalFlux)
  !---------------------------------------------------------------------
!*** not yet tested

    implicit none

    real(single), intent(in)  :: flux(:),&
                                 cumSkyDensity(size(flux)),&
                                 powers(size(flux)),&
                                 accumFlux(size(flux)),&
                                 fluxSample
    real(single), intent(out) :: undetectableTotalFlux

    integer(int32) :: numSamples, binNum
    real(single) :: detectableFluxAtSample

    !-------------------------------------------------------------------
    ! Start

    numSamples = size(flux)

    if (fluxSample <= flux(1)) then
      undetectableTotalFlux = 0.0
  return

    else if (fluxSample >= flux(numSamples)) then
      undetectableTotalFlux = accumFlux(numSamples)
  return
    end if

    binNum = getBinNumber(flux, fluxSample) ! in ssclib/array_utils

    if (2.0 + powers(binNum) == 0.0) then
call error('notYetSupported','')

    else
      detectableFluxAtSample = accumFlux(binNum + 1)&
        + ((1.0 + powers(binNum)) / (2.0 + powers(binNum)))&
        * (cumSkyDensity(binNum))&
        / flux(binNum)**(1.0 + powers(binNum))&
        * (      fluxSample**(2.0 + powers(binNum))&
        -  flux(binNum + 1)**(2.0 + powers(binNum)))
    end if

    undetectableTotalFlux = accumFlux(numSamples) - detectableFluxAtSample

  end subroutine calcUndetectableFlux






  !---------------------------------------------------------------------
  subroutine makeSimSrcList(templateSetName, sourceListName, obsIdList)
  !---------------------------------------------------------------------
  ! This subroutine creates a list of simulated sources. Each source has
  ! a position in the sky (which must fall within the area specified by
  ! the template set) and an amplitude. The spectrum of every source is
  ! assumed to be of the same form, namely given by the power law
  !
  !	F = F(1) * E^g							(1)
  !
  ! where F is the flux density in counts s^-1 cm^-2 eV^-1 (or cgs/eV),
  ! F(1) is the amplitude or reference flux density at E = 1 eV
  ! (calculated herein and stored in the output source list), E is the
  ! photon energy in eV and g is a spectral index (eg -1.4). Thus given
  ! the amplitude F(1) + energy boundaries + the spectral index g, all
  ! to be stored in the output source list, the flux S in ergs s^-1 cm^-2
  ! (or cgs) of each source can be calculated; given in addition the
  ! effective area and other exposure attenuations as a function of E,
  ! the count rates can also be deduced.
  !
  ! Positions are random with a uniform distribution over the celestial
  ! sphere, and the amplitudes are random with a distribution according
  ! to the logN-logS stored in the template set.
  !
  ! The format of templateSet is as follows:
  !
  !	- The circular area to contain the sources is specified by the
  ! header keywords CONE_RA, CONE_DEC and CONE_RAD, with units decimal
  ! degrees, decimal degrees and arcsec respectively. (The sky area of
  ! the cone is given by the keyword SKY_AREA in deg^2, but this is a
  ! convenience merely.)
  !
  !	- The logN-logS information is contained in a binary table called
  ! LOGNLOGS. This table should contain the following real32 columns:
  !
  !		- FLUX: the source flux in ergs cm^-2 s^-1.
  !
  !		- DENSITY: the total sky density (in sources deg^-2) below
  ! that flux.
  !
  !	- This table should contain the following keywords:
  !
  !		- EUNIT: either ergs, ev or kev are recognised. These
  ! units refer to the following pair of keywords:
  !
  !		- E_MIN: the lower energy band to use in converting flux
  ! to amplitude (the quantity F(1) in equation (1) above).
  !
  !		- E_MAX: the upper energy band to use in converting flux
  ! to amplitude (the quantity F(1) in equation (1) above).
  !
  !		- SPECINDX: the common spectral index, g as described
  ! in equation (1) above.
  !
  !		- HI_SLOPE. This real32 keyword gives the exponent
  ! (= slope on a log-log graph) of the high-flux power-law extension.
  !
  ! The table represents a graph of integrated source density in sources
  ! per square degree against the source flux in ergs cm^-2 s^-1. The
  ! FLUX/DENSITY pairs are taken to represent `knees' or vertices in a
  ! graph of integrated sky density N(>S) against flux S. The function
  ! is assumed to obey a power law in the gaps between vertices: thus
  ! in log-log space it is represented by a sequence of line segments.
  ! The pairs of values should occur in order of increasing flux, and
  ! non-physical situations (eg non-positive values, integrated densities
  ! which increase with flux) are naturally not permitted.
  !.....................................................................

    implicit none

    character(*), intent(in) :: templateSetName, sourceListName
    character(10), intent(in) :: obsIdList(:)

    integer(int32) ::&
      numSources, numModelPoints, obsIdNum
    real(single), allocatable ::&
      srcFluxDensity0(:), origRandoms(:), recalcRandoms(:),&
      theta(:), phi(:)
    real(single) ::&
      hiPower,&
      numSourcesReal
    real(single), pointer ::&
      modelFluxDensity0(:),&
      modelCumSkyDensity(:),&
      plotFluxDensity0(:),&
      plotCumSkyDensityDeg(:)
    real(double) ::&
!      photonSpectralIndex,&
!      fluxEeVLo, fluxEeVHi
      coneRaDeg, coneDecDeg, coneRadiusArcsec, coneRadius,&
      coneArea
    real(double), allocatable ::&
      srcRaDeg(:),&
      srcDecDeg(:),&
srcRa(:), srcDec(:)
    type(DataSetT) :: templateSet, srcSet
    type(TableT) :: templateTab, srcTab, plotTab
    integer :: foreignStatus
    type(ColumnT) :: flagCol
    integer(int16), pointer :: flag(:)
    integer(int32), pointer :: indices(:)
integer(int32) :: i

    !-------------------------------------------------------------------
    ! Start

    ! Read the logN-logS distribution model from the template file and
    ! transform the X values from units of flux (in the energy band also
    ! specified in the template) to flux density:
    templateSet = dataSet(templateSetName, READ)
    templateTab = table(templateSet, 'LOGNLOGS')
    call getLogNlogS(templateTab, modelFluxDensity0, modelCumSkyDensity&
      , hiPower)

    numModelPoints = size(modelFluxDensity0)

    !...................................................................
    ! Calculate numSources by multiplying maxDensityDeg (the total density
    ! of sources per square degree) as got from logNlogSSet by the area
    ! of the cone in square degrees:

    coneRaDeg        = real64Attribute(templateTab, 'CONE_RA')
    coneDecDeg       = real64Attribute(templateTab, 'CONE_DEC')
    coneRadiusArcsec = real64Attribute(templateTab, 'CONE_RAD')
    call release(templateTab)

    coneRadius = coneRadiusArcsec * PiOn180 / 3600.0d0
    coneArea = TwoPi * (1.0d0 - cos(coneRadius))

    numSourcesReal = (coneArea / PiOn180 / PiOn180) * modelCumSkyDensity(0)
    ! Add some noise to the number:
    numSources = nint(numSourcesReal + gaussianRandom() * sqrt(numSourcesReal))
    if (numSources <= 0)&
call error('noSourcesInSkyArea','eimsim_utils: The sky area for simulated &
&sources times the density of sources is probably too small.')

    !...................................................................
    ! Get random amplitudes and positions for the sources:

!    allocate(srcAmplitude(numSources))
    allocate(srcFluxDensity0(numSources))
allocate(srcRa(numSources), srcDec(numSources))
    allocate(srcRaDeg(numSources), srcDecDeg(numSources))
    allocate(theta(numSources), phi(numSources))
    allocate(origRandoms(numSources), recalcRandoms(numSources))

    call getSineRandoms(real(coneRadius), theta, foreignStatus)
    ! in ssclib/random

if (foreignStatus .ne. 0) then
print *, 'getSineRandoms status:', foreignStatus
call error('getSineRandomsError',"eimsim_utils: The function getSineRandoms &
&returned an error status.")
end if
    call getRandomNumbers(phi) ! in ssclib/random
    phi = TwoPi * phi

!    call polarsToRaDec(coneRaDeg, coneDecDeg, real(theta, double)&
!    , real(phi, double), srcRaDeg, srcDecDeg)
call polarsToRaDec(real(theta, double)&
 , real(phi, double), coneRaDeg*PiOn180, coneDecDeg*PiOn180, srcRa, srcDec)
    ! in ssclib/coordinate

srcRaDeg  = srcRa  / PiOn180
srcDecDeg = srcDec / PiOn180

    call getLogNlogSrandoms(modelFluxDensity0, modelCumSkyDensity, hiPower&
      , srcFluxDensity0, origRandoms, foreignStatus)

    if (foreignStatus < 0) then
print *, 'ssclib: Status return from getLogNlogSrandoms was', foreignStatus
call error('getLogNlogSrandoms',"eimsim_utils: The function getLogNlogSrandoms &
&returned an error status.")
    end if

    !...................................................................
    ! Recover the original (evenly-distributed between 0 and 1) random
    ! numbers used to generate the flux values:

    call invertLogNlogSrandoms(modelFluxDensity0, modelCumSkyDensity&
      , srcFluxDensity0, hiPower, recalcRandoms, foreignStatus)

!*** dump origRandoms and recalcRandoms and compare - they should be the same value in each case.

    if (foreignStatus < 0) then
print *, 'ssclib: Status return from invertLogNlogSrandoms was', foreignStatus
call error('invertLogNlogSrandoms','')
    end if

    !...................................................................
    ! Write source list to file:

    srcSet = dataSet(sourceListName, CREATE)
!    srcTab = addTable(srcSet, 'SRCLIST', size(srcFluxDensity0))
srcTab = addTable(srcSet, 'SRCLIST', numSources)

allocate(indices(numSources))
!    indices => indicize(size(srcFluxDensity0))
indices = (/(i,i=1,numSources)/)

    call myAddColumn(srcTab, 'INDEX',    indices)
    deallocate(indices)
    call myAddColumn(srcTab, 'S_DENS_0', srcFluxDensity0, 'cgs/eV'&
      , 'The flux density at E = 1 eV.')
    call myAddColumn(srcTab, 'RA',       srcRaDeg,  'deg')
    call myAddColumn(srcTab, 'DEC',      srcDecDeg, 'deg')
    call myAddColumn(srcTab, 'FLUXRAND', recalcRandoms, ''&
      , 'The original random numbers used to generate the S_DENS_0 values.')

    ! The flags are not used until the images are made:
    do obsIdNum = 1, size(obsIdList)
      flagCol = addColumn(srcTab, 'FLAG_'//obsIdList(obsIdNum), INTEGER16)
      flag => int16Data(flagCol)
      flag = 0
      call release(flagCol)
    end do
    call release(srcTab)

    !...................................................................
    ! Add the template table to the source list dataset:

    call copyBlock(srcSet, block(templateSet, 'LOGNLOGS'))

    !...................................................................
    ! Generate a further table to assist in plotting the logN-logS model
    ! alongside the actual distribution of sources:

    plotTab = addtable(srcSet, 'MODEL_PL', numModelPoints + 1)
    plotFluxDensity0 => real32Data(addColumn(plotTab, 'S_DENS_0', REAL32&
      , 'cgs/eV', comment='The flux density at E = 1 eV.'))
    plotCumSkyDensityDeg => real32Data(addColumn(plotTab, 'SKY_DENSY', REAL32&
      , 'deg^-2'))

    plotFluxDensity0(0:numModelPoints-1) = modelFluxDensity0
    plotFluxDensity0(numModelPoints) = 10.0 * maxval(srcFluxDensity0)
    plotCumSkyDensityDeg(0:numModelPoints-1) = modelCumSkyDensity

    plotCumSkyDensityDeg(numModelPoints)&
      = plotCumSkyDensityDeg(numModelPoints-1)&
      * (plotFluxDensity0(numModelPoints)&
      / plotFluxDensity0(numModelPoints-1))**(hiPower)

    !...................................................................
    ! Clean up:

    call release(srcSet)
    call release(templateSet)
    deallocate(modelFluxDensity0)
    deallocate(modelCumSkyDensity)

  end subroutine makeSimSrcList

  !---------------------------------------------------------------------
  subroutine getSrcPerSquareDeg(logNlogSSetName, srcPerSquareDeg)
  !---------------------------------------------------------------------
  ! Kind of a bodgy way to get the srcPerSquareDeg.

    implicit none

    character(*), intent(in)  :: logNlogSSetName
    real(single), intent(out) :: srcPerSquareDeg

    type(DataSetT) ::&
      logNlogSset
    type(TableT) ::&
      logNlogStab
    real(single) ::&
      hiPower
    real(single), pointer ::&
      fluxErgCmSec(:), modelCumSkyDensity(:)

    !-------------------------------------------------------------------
    ! Start

    logNlogSset = dataSet(logNlogSSetName, READ)
    logNlogStab = table(logNlogSset, 'LOGNLOGS')

    !...................................................................
    ! Get the integrated logN-logS values from the set:

    call loadActualLogNlogS(logNlogStab, fluxErgCmSec, modelCumSkyDensity&
      , hiPower)

    srcPerSquareDeg = modelCumSkyDensity(1)

    deallocate(fluxErgCmSec, modelCumSkyDensity)
    call release(logNlogSset)

  end subroutine getSrcPerSquareDeg

  !---------------------------------------------------------------------
  subroutine getLogNlogS(logNlogStab, modelFluxDensity0, modelCumSkyDensity&
  , hiPower, photonSpectralIndex, EeVLo, EeVHi)
  !---------------------------------------------------------------------
  ! Remember to deallocate pointers modelFluxDensity(:) and
  ! modelCumSkyDensity(:) after use.

    implicit none

    type(TableT), intent(in)            :: logNlogStab
    real(single), pointer               :: modelFluxDensity0(:),&
                                           modelCumSkyDensity(:)
    real(single), intent(out), optional :: hiPower
    real(double), intent(out), optional :: EeVLo,&
                                           photonSpectralIndex,&
                                           EeVHi

    integer(int32) ::&
      i
    character(20) ::&
      energyUnit
    real(single), pointer ::&
      modelFluxErgCmSec(:)
    real(double) :: local_photonSpectralIndex
    real(double) :: local_EeVLo, local_EeVHi

    !-------------------------------------------------------------------
    ! Start

    !...................................................................
    ! Get the integrated logN-logS values from the set:

    call loadActualLogNlogS(logNlogStab, modelFluxErgCmSec, modelCumSkyDensity&
      , hiPower)
!print *, 'size(modelCumSkyDensity):', size(modelCumSkyDensity)
!call error('','')

    !...................................................................
    ! Calculate the flux-to-amplitude conversion and shift the axis of
    ! the logNlogS accordingly:

    energyUnit = stringAttribute(logNlogStab, 'EUNIT')
    select case(ucase(energyUnit))
    case('EV')
      local_EeVLo = real64Attribute(logNlogStab, 'E_MIN')
      local_EeVHi = real64Attribute(logNlogStab, 'E_MAX')
    case('KEV')
      local_EeVLo = real64Attribute(logNlogStab, 'E_MIN') * 1000.0
      local_EeVHi = real64Attribute(logNlogStab, 'E_MAX') * 1000.0
    case('ERG')
      local_EeVLo = real64Attribute(logNlogStab, 'E_MIN') * 1.0e-7 / ElectronQ
      local_EeVHi = real64Attribute(logNlogStab, 'E_MAX') * 1.0e-7 / ElectronQ
    case default
call error('badEnergyUnit','')
    end select

    local_photonSpectralIndex = real64Attribute(logNlogStab, 'SPECINDX')

    allocate(modelFluxDensity0(0:size(modelFluxErgCmSec)-1))
    do i = 0, size(modelFluxErgCmSec) - 1
      modelFluxDensity0(i) = fluxToFluxDensity0(local_photonSpectralIndex&
        , local_EeVLo, local_EeVHi, modelFluxErgCmSec(i))
    end do

    if (present(photonSpectralIndex))&
      photonSpectralIndex = local_photonSpectralIndex
    if (present(EeVLo)) EeVLo = local_EeVLo
    if (present(EeVHi)) EeVHi = local_EeVHi

  end subroutine getLogNlogS

  !---------------------------------------------------------------------
  subroutine loadActualLogNlogS(logNlogStab, fluxErgCmSec, modelCumSkyDensity&
  , hiPower)
  !---------------------------------------------------------------------
  ! Note that the pointers fluxErgCmSec and modelCumSkyDensity should
  ! be deallocated after use.

    implicit none

    type(TableT), intent(in)  :: logNlogStab
    real(single), pointer     :: fluxErgCmSec(:), modelCumSkyDensity(:)
    real(single), intent(out) :: hiPower

!    real(single) :: minFluxErgCmSec
    integer(int32) :: numFluxes
!, startBin, numPoints
!    real(single), allocatable ::&
!      setFluxErgCmSec(:), setCumDensityDeg(:)

    !-------------------------------------------------------------------
    ! Start

    numFluxes = numberOfRows(logNlogStab)
if (numFluxes < 1) call error('noFluxValues','')
    call readColDataToPtr(logNlogStab, 'FLUX',    fluxErgCmSec)
    call readColDataToPtr(logNlogStab, 'DENSITY', modelCumSkyDensity)
!print *, 'size(modelCumSkyDensity):', size(modelCumSkyDensity)
!call error('','')

    if (numFluxes > 1) then
      if (any(modelCumSkyDensity(:numFluxes-2) < modelCumSkyDensity(1:)))&
call error('badDensityOrder','')
      if (any(fluxErgCmSec(:numFluxes-2) >= fluxErgCmSec(1:)))&
call error('badFluxOrder','')
    end if
if (fluxErgCmSec(0) <= 0.0) call error('nonPositiveFlux','')
if (modelCumSkyDensity(0) <= 0.0) call error('nonPositiveDensity','')

    hiPower = real32Attribute(logNlogStab, 'HI_SLOPE')

  end subroutine loadActualLogNlogS

  !---------------------------------------------------------------------
  subroutine invertLogNlogSrandoms(x, yInt, xSamples, hiPower, harvest&
  , status)
  !---------------------------------------------------------------------
  ! This sub was modified from the routine getLogNlogSrandoms in random.f90.
  !.....................................................................
  !
  ! The subroutine takes a vector of random values xSamples() which have a
  ! distribution P which, when integrated, follows a piecewise power law
  ! as follows. The ith segment of the integrated piecewise function is
  ! defined between x(i) and x(i+1) and is of the form
  !
  !	log(yInt) = log(yInt(i)) + slope * (log(x) - log(x(i)))
  !
  ! where
  !	         log(yInt(i+1)) - log(yInt(i))
  !	slope = -------------------------------.
  !	            log(x(i+1)) - log(x(i))
  !
  ! Above x(N) the integrated function is defined by
  !
  !	log(yInt) = log(yInt(N)) + hiPower * (log(x) - log(x(N))).
  !
  ! In the present subroutine the original random values, which are evenly
  ! distributed between 0 and 1, are returned.
  !.....................................................................
  !
  ! Notes: (i) The x values must be in increasing order, and they cannot be
  ! negative. (ii) The y values must all be >0, and must never increase with
  ! increasing i. (iii) N must be >0. (iv) hiPower must be <0.
  !
  ! In the event of an error, the subroutine returns values of -1 for all
  ! elements of numbers().
  !.....................................................................

    implicit none

    real,    intent(in)            :: x(:),&
                                      yInt(size(x)),&
                                      xSamples(:),&
                                      hiPower
    real,    intent(out)           :: harvest(size(xSamples))
    integer, intent(out), optional :: status

    integer ::&
      localStatus, foreignStatus, numBins, numXsamples, i, bin
    real :: slopes(size(x))

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    harvest = -1

    numBins = size(x)
    numXsamples = size(xSamples)

    call doLogNlogSchecks(x, yInt, hiPower, foreignStatus)

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus
  return
    end if

    ! Calculate slopes:

    do bin = 1, numBins - 1
      slopes(bin) = (log(yInt(bin+1)) - log(yInt(bin)))&
                  / (log(x(   bin+1)) - log(x(   bin)))
    end do
    slopes(numBins) = hiPower

    ! Main part of the routine:

    do i = 1, numXsamples
      bin = getBinNumber(x, xSamples(i)) ! in array_utils
if (bin < 1) bin = 1 ! to cope with zero- and non-(ID_BAND==0)-RATE values.
      harvest(i) = yInt(bin) * (xSamples(i) / x(bin))**slopes(bin)
    end do
    harvest = harvest / yInt(1)

    if (present(status)) status = localStatus

  end subroutine invertLogNlogSrandoms

  !---------------------------------------------------------------------
  subroutine doLogNlogSchecks(x, yInt, hiPower, status)
  !---------------------------------------------------------------------
  ! Taken straight from random.f90.

    implicit none

    real,    intent(in)  :: x(:), yInt(size(x)), hiPower
    integer, intent(out) :: status

    integer :: numPoints

    !-------------------------------------------------------------------
    ! Start:

    numPoints = size(x)

    if (numPoints < 1) then
      status = -1
  return
    end if

    if (numPoints > 1) then ! Check order of x values:
      if (any(x(2:) <= x(:numPoints-1))) then
        status = -2
  return
      end if

      if (any(yInt(2:) > yInt(:numPoints-1))) then
        status = -3
  return
      end if
    end if

    ! Check all x values >= 0:
    if (x(1) < 0.0) then
      status = -4
  return
    end if

    ! Check all yInt values > 0:
    if (any(yInt <= 0.0)) then
      status = -5
  return
    end if

    if (hiPower >= -1.0) then
      status = -6
  return
    end if

  end subroutine doLogNlogSchecks

  !---------------------------------------------------------------------
  subroutine makeRateImage(expMapSetName, sourceListSetName, paddedMaskSetName&
  , psfEnergyFraction, psfCutoffFluxDensity, ecf, rateEeVLo, rateEeVHi&
  , rateImageSetName, tempImageName, obsId, bandId)
  !---------------------------------------------------------------------
  ! This makes a simulated sky map from the source list, convolving each
  ! source position with the PSF appropriate to that position in the focal
  ! plane.The resulting sky map values are in counts/pixel/sec.
  !
  ! The sky map is created in the projection plane implicit in the wcs
  ! keywords of the exposure map. Sky map pixels are left set to zero
  ! where the exposure map is zero.
  !
  ! The exposure map also carries information about the instrument.
  !
  ! The medium-accuracy PSF from the cal is used for bright sources (ie
  ! those which have flux densities in erg cm^-2 s^-1 eV^-1 greater than
  ! psfCutoffFluxDensity) and a simpler, quicker PSF model is used for
  ! all fainter sources. The cal PSF requires instrument type and
  ! instrument pointing (obtained from the exposure map) as well as
  ! energy and energy fraction (supplied by the arguments psfEeV and
  ! psfEnergyFraction).
  !
  ! The flux densities p (in erg cm^-2 s^-1 eV^-1) for each source (used
  ! for comparison with psfCutoffFluxDensity) are calculated at E=psfEeV
  ! using the equation
  !
  !	p(E) = p(1) * E**gamma
!*** wrong, look at fluxTransforms stuff
  !
  ! where p(1) is the flux density at 1 eV, obtained from the source table.
  !.....................................................................

    implicit none

    character(*),  intent(in) :: expMapSetName,&
                                 sourceListSetName,&
                                 paddedMaskSetName,&
                                 rateImageSetName,&
                                 tempImageName
    character(10), intent(in) :: obsId
    integer(int8), intent(in) :: bandId
    real(double),  intent(in) :: ecf,&
                                 rateEeVLo,&
                                 rateEeVHi,&
                                 psfEnergyFraction,&
                                 psfCutoffFluxDensity

    type(DataSetT) :: sourceSet, imageSet
    type(TableT) :: sourceTab, templateTab
    type(ArrayT) :: imageArray
    integer(int32) ::&
      numSources, src
    type(ColumnT) ::&
      raDegCol, decDegCol, fluxDensity0Col, flagCol
    real(single), pointer :: fluxDensity0(:), patch(:,:)
    real(double), pointer :: raDeg(:), decDeg(:)
    integer(int16), pointer :: flag(:)
    integer(int16) :: xSize, ySize, xi, yi
    real(single), allocatable :: image(:,:), rate(:)
    type(WcsT) :: wcs
    real(double) ::&
      photonSpectralIndex,&
      psfEeV,&
sigmaTan
    real(double), allocatable ::&
      srcTanX(:), srcTanY(:), srcThetaArcsec(:), srcPhi(:)
    real(single), pointer :: srcXpixels(:), srcYpixels(:)
    real(single) ::&
      sigmaXpixels, sigmaYpixels,&
      fluxDensity, convertToFlux
    real(single), parameter :: sigmaArcsec = 5.0 ! for coarse (gaussian) PSF.
    integer :: foreignStatus
    real(single), pointer :: expMapImage(:,:)
    character(8) :: wcsType
    logical(bool), pointer :: paddedMask(:,:)

    !-------------------------------------------------------------------
    ! Start

    !...................................................................
    ! Load the information from the exposure map:

    call readArrayData(expMapSetName, expMapImage)

    xSize = size(expMapImage, 1)
    ySize = size(expMapImage, 2)

    call readArrayData(paddedMaskSetName, paddedMask)
    if (size(paddedMask,1) .ne. xSize .or. size(paddedMask,2) .ne. ySize)&
call error('','')

    call readWcs(expMapSetName, wcs)
    call easifyWcsType(wcs%x%ctype, wcs%y%ctype, wcsType)

    call getCalInfo(expMapSetName, instrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude) ! sets instrument in cal.


    !...................................................................
    ! Set up rest of cal (inst already set in call to getCalInfo())
    ! and calculate some miscellaneous quantities:

    call CAL_setState(accuracyLevel=ACCURACY_MEDIUM)

!    sigmaXpixels = sigmaArcsec / abs(wcs%x%pixelSizeWorld) / 3600.0
!    sigmaYpixels = sigmaArcsec / abs(wcs%y%pixelSizeWorld) / 3600.0
sigmaTan = sigmaArcsec * PiOn180 / 3600.0d0
call tanToPixels(wcs, sigmaTan, sigmaTan, sigmaXpixels, sigmaYpixels)
sigmaXpixels = abs(sigmaXpixels)
sigmaYpixels = abs(sigmaYpixels)

    psfEeV = (rateEeVLo + rateEeVHi) / 2.0

    !...................................................................
    ! Open source set and load the columns:

    sourceSet = dataSet(sourceListSetName, MODIFY) ! 'modify' because need
    ! to set flag col values.

    sourceTab = table(sourceSet, 'SRCLIST')
    numSources = numberOfRows(sourceTab)
if (numSources <= 0) call error('noSources','')

    raDegCol        = column(sourceTab, 'RA')
    decDegCol       = column(sourceTab, 'DEC')
    fluxDensity0Col = column(sourceTab, 'S_DENS_0')
    flagCol         = column(sourceTab, 'FLAG_'//obsId)

    raDeg        => real64Data(raDegCol)
    decDeg       => real64Data(decDegCol)
    fluxDensity0 => real32Data(fluxDensity0Col)
    flag         => int16Data(flagCol)

    !...................................................................
    ! Initialize flag column:

    select case(instrumentId)
    case(EPN)
      flag = ibclr(flag, NOT_IN_PN_EXP)
      flag = ibclr(flag, NOT_IN_PN_FOV)
    case(EMOS1)
      flag = ibclr(flag, NOT_IN_M1_EXP)
      flag = ibclr(flag, NOT_IN_M1_FOV)
    case(EMOS2)
      flag = ibclr(flag, NOT_IN_M2_EXP)
      flag = ibclr(flag, NOT_IN_M2_FOV)
    case default
call error('badInstrument','')
    end select

    !...................................................................
    ! Open the template table in the source set and get the source
    ! spectrum information:

    templateTab = table(sourceSet, 'LOGNLOGS')
    photonSpectralIndex = real64Attribute(templateTab, 'SPECINDX')
    call release(templateTab)

!coneRaDeg        = real64Attribute(templateTab, 'CONE_RA')
!coneDecDeg       = real64Attribute(templateTab, 'CONE_DEC')
!coneRadiusArcsec = real64Attribute(templateTab, 'CONE_RAD')

!call test_doSrcSkyPlot(fluxDensity0, )

!call release(sourceSet)
!call error('','')

    !...................................................................
    ! Now convert S_DENS_0 values to count rates within the supplied
    ! rate band. This is done in two steps: first a multiplier to convert
    ! from fluxDensity0 to flux is calculated; then the ecf value is used
    ! to convert from flux to rate.

    convertToFlux = fluxDensity0ToFlux(photonSpectralIndex&
      , rateEeVLo, rateEeVHi, 1.0)
    allocate(rate(numSources))
    rate = fluxDensity0 * convertToFlux * ecf * 1.0e11

    !...................................................................
    ! Convert source Ra/Dec to both image coordinates and TELCOORD coordinates:

    allocate(srcTanX(numSources), srcTanY(numSources))
!    call raDecToTan(wcs%x%refWorld, wcs%y%refWorld&
!      , raDeg, decDeg, srcTanX, srcTanY) ! in ssclib/coordinate
call raDecToTan(raDeg*PiOn180, decDeg*PiOn180&
 , wcs%x%refWorld*PiOn180, wcs%y%refWorld*PiOn180&
 , srcTanX, srcTanY) ! in ssclib/coordinate

    allocate(srcXpixels(numSources), srcYpixels(numSources))
    call tanToPixels(wcs, srcTanX, srcTanY, srcXpixels, srcYpixels)
    ! in ssclib/wcs_aux
    deallocate(srcTanX, srcTanY)

    allocate(srcThetaArcsec(numSources), srcPhi(numSources))
!    call raDecToInst(raDeg, decDeg, scAttitude&
!      , expStartTimeStamp, thetaArcsec=srcThetaArcsec, phi=srcPhi)
call raDecToTel(raDeg*PiOn180, decDeg*PiOn180, scAttitude&
 , expStartTimeStamp, srcThetaArcsec, srcPhi)
    ! in ssclib/coordinate

    call release(raDegCol)
    call release(decDegCol)

    !...................................................................
    ! Now work through the list of sources and add a correctly-normalized
    ! PSF image for each one that falls on a non-zero pixel of the
    ! exposure map:

    allocate(image(xSize, ySize))
    image = 0.0 ! default

    do src = 1, numSources
      call printEvery(src, numSources, fraction=0.05)

      xi = nint(srcXpixels(src))
      yi = nint(srcYpixels(src))
    if (xi < 1 .or. xi > xSize .or. yi < 1 .or. yi > ySize) cycle
    if (.not. paddedMask(xi, yi)) cycle

      if (expMapImage(xi, yi) <= 0.0) then
        select case(instrumentId)
        case(EPN)
          flag(src-1) = ibset(flag(src-1), NOT_IN_PN_EXP)
        case(EMOS1)
          flag(src-1) = ibset(flag(src-1), NOT_IN_M1_EXP)
        case(EMOS2)
          flag(src-1) = ibset(flag(src-1), NOT_IN_M2_EXP)
        end select
      end if

      ! Calculate the flux density at psfEeV:
      fluxDensity = fluxDensity0ToFluxDensity(photonSpectralIndex, psfEeV&
        , fluxDensity0(src-1))

!*** should rather calculate the psfCutoffFluxDensity at E = 1 eV. This way there is only 1 calculation involved in total instead of 1 for each source.

      if (fluxDensity > psfCutoffFluxDensity) then ! use high-precision model:
        flag(src-1) = ibclr(flag(src-1), COARSE_PSF(bandId)) ! ie, unset this
        ! flag.

!        call getCalPsfPatch(xSize, ySize, wcs, wcsType, psfEnergyFraction&
!          , psfEeV, srcThetaArcsec(src), srcPhi(src), patch, foreignStatus)
        call getCalPsfPatch(xSize, ySize, wcs, psfEnergyFraction&
          , psfEeV, srcThetaArcsec(src), srcPhi(src), patch, foreignStatus)

        if (foreignStatus < 0) then
          if (foreignStatus == -1) then ! theta was out of bounds. Used the
          ! boundary value of theta and we should flag the source to that
          ! effect:
            select case(instrumentId)
            case(EPN)
              flag(src-1) = ibset(flag(src-1), NOT_IN_PN_FOV)
            case(EMOS1)
              flag(src-1) = ibset(flag(src-1), NOT_IN_M1_FOV)
            case(EMOS2)
              flag(src-1) = ibset(flag(src-1), NOT_IN_M2_FOV)
            end select

          else
call error('','')
          end if
        end if

      else ! can use low-precision PSF model:
        flag(src-1) = ibset(flag(src-1), COARSE_PSF(bandId)) ! ie, set this
        ! flag.

        call getGaussianPsfPatch(srcXpixels(src), srcYpixels(src)&
          , sigmaXpixels, sigmaYpixels, patch)!, testFlag)
!call dumpImageToFits(patch, 'junkpsf.ds')
!call error('','')

      end if

      call addPatch(image, patch, rate(src))
      deallocate(patch)
    end do
    deallocate(expMapImage, paddedMask)
    deallocate(srcXpixels, srcYpixels)

    !...................................................................
    ! Save the output:

    call saveImage(image, expMapSetName, rateImageSetName, tempImageName)
    ! Copies over all keywords (eg wcs, INSTRUM) in the expMap.

    imageSet = dataSet(rateImageSetName, MODIFY)
    imageArray = array(imageSet, 0)
    call setAttribute(imageArray, 'PSFENERG', psfEeV, 'eV'&
      , 'The energy at which the cal PSF was obtained.')
    call setAttribute(imageArray, 'E_FRAC',  psfEnergyFraction, ''&
      , 'The energy fraction contained in the (cal) PSF array.')
    call setAttribute(imageArray, 'S_CUTOFF',  psfCutoffFluxDensity&
      , 'erg / sec / cm^2 / eV', 'Flux density boundary between use &
      &of coarse vs cal PSF.')
    call setAttribute(imageArray, 'ECF',  ecf,   '10^-11 counts cm^2 / erg')
    call setAttribute(imageArray, 'E_LO', rateEeVLo, 'eV')
    call setAttribute(imageArray, 'E_HI', rateEeVHi, 'eV')

    call copyBlock(imageSet, block(sourceSet, 'LOGNLOGS'))

    call release(imageSet)
    call release(sourceSet)

  end subroutine makeRateImage

  !---------------------------------------------------------------------
  subroutine getGaussianPsfPatch(srcXpixels, srcYpixels&
  , sigmaXpixels, sigmaYpixels, patch)
  !---------------------------------------------------------------------
  !
    implicit none

    real(single), intent(in) ::&
      srcXpixels, srcYpixels, sigmaXpixels, sigmaYpixels
    real(single), pointer :: patch(:,:)

    real(single), parameter :: patchWidthInSigmas = 7.0
    integer(int16) :: xStart, xFinis, yStart, yFinis, i
    real(single), allocatable :: x(:), y(:), xExp(:), yExp(:)
    integer(int16), pointer :: xi(:), yi(:)

    !-------------------------------------------------------------------
    ! Start

    xStart = floor(  srcXpixels - patchWidthInSigmas * sigmaXpixels)
    xFinis = ceiling(srcXpixels + patchWidthInSigmas * sigmaXpixels)
    yStart = floor(  srcYpixels - patchWidthInSigmas * sigmaYpixels)
    yFinis = ceiling(srcYpixels + patchWidthInSigmas * sigmaYpixels)

    allocate(patch(xStart:xFinis,yStart:yFinis))
    allocate(x(xStart:xFinis), xExp(xStart:xFinis))
    allocate(y(yStart:yFinis), yExp(yStart:yFinis))
allocate(xi(xStart:xFinis), yi(yStart:yFinis))

!    xi => indicize(xStart, xFinis)
xi = (/(i,i=xStart, xFinis)/)
    x = (real(xi, single) - srcXpixels) / RootTwo / sigmaXpixels

!print *, 'xStart, xi(xStart), x(xStart):', xStart, xi(xStart), x(xStart)

!    yi => indicize(yStart, yFinis)
yi = (/(i,i=yStart, yFinis)/)
    y = (real(yi, single) - srcYpixels) / RootTwo / sigmaYpixels
    xExp = exp(-(x*x))
    yExp = exp(-(y*y))

    patch = spread(xExp, 2, (1 + yFinis - yStart))&
          * spread(yExp, 1, (1 + xFinis - xStart))

!call dumpImageToFits(patch, 'junkpsf.ds')
!call error('','')
    call vignetPatch(patch)

    deallocate(xi, x, xExp, yi, y, yExp)

  end subroutine getGaussianPsfPatch

  !---------------------------------------------------------------------
  subroutine addPatch(image, patch, rate)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(inout) :: image(:,:)
    real(single), pointer :: patch(:,:)
    real(single), intent(in) :: rate

    integer :: xStart, xFinis, yStart, yFinis, xSize, ySize

    !-------------------------------------------------------------------
    ! Start

    xStart = lbound(patch, 1)
    xFinis = ubound(patch, 1)
    yStart = lbound(patch, 2)
    yFinis = ubound(patch, 2)
    xSize = size(image, 1)
    ySize = size(image, 2)

    if (xStart > xSize .or. xFinis < 1 .or. yStart > ySize .or. yFinis < 1) then
  return
    end if

    xStart = max(1, xStart)
    xFinis = min(xSize, xFinis)
    yStart = max(1, yStart)
    yFinis = min(ySize, yFinis)

    image(xStart:xFinis, yStart:yFinis) = image(xStart:xFinis, yStart:yFinis)&
      + patch(xStart:xFinis, yStart:yFinis) * rate / sum(patch)

  end subroutine addPatch

  !---------------------------------------------------------------------
!  subroutine getCalPsfPatch(xSize, ySize, wcs, wcsType, psfEnergyFraction&
!  , psfEeV, srcThetaArcsec, srcPhi, patch, status)
  subroutine getCalPsfPatch(xSize, ySize, wcs, psfEnergyFraction&
  , psfEeV, srcThetaArcsec, srcPhi, patch, status)
  !---------------------------------------------------------------------
  ! Requires cal to have been set up for instrumentId and accuracyLevel.

    implicit none

    integer(int16), intent(in) :: xSize, ySize
    type(WcsT), intent(in) :: wcs
!    character(*), intent(in) :: wcsType
    real(double), intent(in) ::&
      psfEnergyFraction, psfEeV, srcThetaArcsec, srcPhi
    real(single), pointer :: patch(:,:)
    integer, intent(out) :: status

    real(double) :: effectiveSrcThetaArcsec
    type(PsfValidityRangesT) :: validPsfRanges
    type(PsfT) :: psfHandle
    type(PsfWindowT) :: psfWindow
    type(PsfBinSizeT) :: psfPixelSizeMm
    real(single), pointer :: psfImage(:,:)
    integer(int16) :: startXi, finisXi, startYi, finisYi
    type(Point2dT), allocatable ::&
      psfImagePixelCorners(:,:)
    type(Point2dDbleT), allocatable ::&
      psfImgPixCrnrTel(:,:), psfImgPixCrnrTan(:,:), psfImgPixCrnrDet(:,:)
!    type(Point2dT) :: psfCentrePixels
      ! types defined in coordinate.f90.
    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start

    status = 0

    validPsfRanges = CAL_psfValidityRanges()

    if  (psfEeV < validPsfRanges%eMin&
    .or. psfEeV > validPsfRanges%eMax&
    .or. srcThetaArcsec < validPsfRanges%thetaMin&
    .or. srcPhi < validPsfRanges%phiMin&
    .or. srcPhi > validPsfRanges%phiMax)&
call error('','')

    if (srcThetaArcsec > validPsfRanges%thetaMax) then
      status = -1
      effectiveSrcThetaArcsec = validPsfRanges%thetaMax&
        * (1.0d0 - epsilon(srcThetaArcsec))
    else
      effectiveSrcThetaArcsec = srcThetaArcsec
    end if

    psfHandle = CAL_getPSF(psfEeV, effectiveSrcThetaArcsec, srcPhi)
    psfPixelSizeMm = CAL_psfBinSize(psfHandle)
    psfWindow = CAL_psfEnboxedEnergy(psfHandle, fraction=psfEnergyFraction)
    call CAL_psfImage(psfHandle, psfWindow, psfImage)

    ! Normalize, then vignet the psf image edges:

    psfImage = psfImage / (sum(psfImage) / psfEnergyFraction)
    call vignetPatch(psfImage)

    allocate(psfImgPixCrnrTel(size(psfImage,1)+1, size(psfImage,2)+1))
    allocate(psfImgPixCrnrDet(size(psfImage,1)+1, size(psfImage,2)+1))
    allocate(psfImgPixCrnrTan(size(psfImage,1)+1, size(psfImage,2)+1))
    allocate(psfImagePixelCorners(size(psfImage,1)+1, size(psfImage,2)+1))

!    call getPsfImagePixelCorners(srcThetaArcsec, srcPhi&
!      , psfPixelSizeMm, psfImage, wcs, wcsType&
!      , instrumentId, expStartTimeStamp, scAttitude&
!      , psfImagePixelCorners, psfCentrePixels) ! The pairs of numbers
!    ! (type(Point2dT)) in this array represent the coordinates of the
!    ! corners of the psfImage pixels, expressed in the exposure-map
!    ! image (ie, sky) coordinate system.

call getPsfImagePixelCorners(srcThetaArcsec, srcPhi&
  , psfPixelSizeMm, psfImage, psfImgPixCrnrTel)
! in ssclib/coordinate.
!
! The pairs of numbers (type(Point2dT)) in this array represent the
! coordinates of the corners of the psfImage pixels, expressed in
! the exposure-map image (ie, sky) coordinate system.

!call psfToDet(psfImgPixCrnrTel%x, psfImgPixCrnrTel%y&
!  , psfImgPixCrnrDet%x, psfImgPixCrnrDet%y)
call psfToDet(psfImgPixCrnrTel%x, psfImgPixCrnrTel%y&
  , psfImgPixCrnrDet%x, psfImgPixCrnrDet%y)
! in ssclib/coordinate.

!call instToTan(psfImgPixCrnrTan%x, psfImgPixCrnrTan%y, wcs%x%refWorld&
!  , wcs%y%refWorld, scAttitude, expStartTimeStamp&
!  , psfImgPixCrnrDet%x, psfImgPixCrnrDet%y)
call detToTan(psfImgPixCrnrDet%x&
  , psfImgPixCrnrDet%y&
  , wcs%x%refWorld*PiOn180, wcs%y%refWorld*PiOn180&
  , scAttitude, expStartTimeStamp&
  , psfImgPixCrnrTan%x, psfImgPixCrnrTan%y)
! in ssclib/coordinate.

call tanToPixels(wcs, psfImgPixCrnrTan%x, psfImgPixCrnrTan%y&
  , psfImagePixelCorners%x, psfImagePixelCorners%y)
! in ssclib/wcs_aux.

    startXi = max(1,     nint(minval(psfImagePixelCorners%x)))
    finisXi = min(xSize, nint(maxval(psfImagePixelCorners%x), int16))
    startYi = max(1,     nint(minval(psfImagePixelCorners%y)))
    finisYi = min(ySize, nint(maxval(psfImagePixelCorners%y), int16))
    allocate(patch(startXi:finisXi,startYi:finisYi))

    psfImagePixelCorners%x = psfImagePixelCorners%x - real(startXi - 1)
    psfImagePixelCorners%y = psfImagePixelCorners%y - real(startYi - 1)

    call regridCartesian(psfImagePixelCorners, psfImage&
      , patch, status=foreignStatus) ! in ssclib/regridders.
    ! This rebins the psfImage values from its original CAMCOORD2 system
    ! to the exposure-map image (ie, sky) coordinate system. The array
    ! returned in convolvers() is centred as near as possible on the
    ! centre of the PSF.

if (foreignStatus < 0) call error('','')
    if (foreignStatus > 0) then
      print *, 'regrid exit status:', foreignStatus
    end if

    call CAL_releaseMemory(psfImage)
!    deallocate(psfImagePixelCorners)

!call dumpImageToFits(patch, 'junkpsf.ds')
!call error('','')

  end subroutine getCalPsfPatch

  !---------------------------------------------------------------------
  subroutine vignetPatch(patch)
  !---------------------------------------------------------------------
  ! This is meant to take an array containing a source PSF image and
  ! multiply it by a local vignetting function, so as to avoid the otherwise
  ! sharp drop-off at the edge of the array.

    implicit none

    real(single), intent(inout) :: patch(:,:)

    integer(int16) :: xSize, ySize, i
!    integer(int16), pointer :: indices(:)
    real(single) ::&
      xHalfSize, yHalfSize,&
      xVector(size(patch, 1)), yVector(size(patch, 2))

    !-------------------------------------------------------------------
    ! Start

    xSize = size(patch, 1)
    xHalfSize = xSize / 2.0 - 0.5
    ySize = size(patch, 2)
    yHalfSize = ySize / 2.0 - 0.5

!    indices => indicize(xSize)
!    xVector = (indices - 1.0 - xHalfSize) / (xHalfSize + 1.0)
xVector = (/((i - 1.0 - xHalfSize) / (xHalfSize + 1.0), i=1,xSize)/)
    ! The +1 prevents patch being 0 all around the rim, which would be
    ! a waste of pixels.
    xVector = 1.0 - xVector * xVector
!    deallocate(indices)

!    indices => indicize(ySize)
!    yVector = (indices - 1.0 - yHalfSize) / (yHalfSize + 1.0)
yVector = (/((i - 1.0 - yHalfSize) / (yHalfSize + 1.0), i=1,ySize)/)
    ! The +1 prevents patch being 0 all around the rim, which would be
    ! a waste of pixels.
    yVector = 1.0 - yVector * yVector
!    deallocate(indices)

    patch = patch * spread(xVector, 2, ySize)&
                  * spread(yVector, 1, xSize)

  end subroutine vignetPatch

  !---------------------------------------------------------------------
  subroutine makeAvCountsImage(expMapSetName, bkgCountMapSetName&
  , rateImageSetName, avCountsImageSetName, tempImageName)
  !---------------------------------------------------------------------

    implicit none

    character(*),   intent(in) :: expMapSetName,&
                                  bkgCountMapSetName,&
                                  rateImageSetName,&
                                  avCountsImageSetName,&
                                  tempImageName

    real(single), pointer ::&
      rateImage(:,:), expImage(:,:), bkgCountsImage(:,:)
!    real(single) :: maxBkgRate
real(single) :: maxBkgCounts
    type(DataSetT) :: avCountsImageSet, bkgSet, rateImageSet
    integer(int16) :: xSize, ySize
    real(single), allocatable :: outImageReal(:,:)
!    integer(int32), allocatable :: outImageInt(:,:)

    !-------------------------------------------------------------------
    ! Start

    call readArrayData(rateImageSetName,  rateImage)
    call readArrayData(expMapSetName,     expImage)
    call readArrayData(bkgCountMapSetName, bkgCountsImage)

    xSize = size(expImage, 1)
    ySize = size(expImage, 2)
    if  (size(bkgCountsImage, 1) .ne. xSize&
    .or. size(bkgCountsImage, 2) .ne. ySize&
    .or. size(rateImage, 1) .ne. xSize&
    .or. size(rateImage, 2) .ne. ySize)&
call error('imageSizesDifferent','')

!*** also compare instrument, wcs between expImageSet and rateImageSet.

    allocate(outImageReal(xSize, ySize))
    outImageReal = rateImage * expImage + bkgCountsImage

    call saveImage(outImageReal, rateImageSetName, avCountsImageSetName&
      , tempImageName)
    ! Copies over all keywords (eg wcs, INSTRUM, ecf) from the rateImage.

!    maxBkgRate = maxval(bkgCountsImage / expImage, (expImage > 0.0))
maxBkgCounts = maxval(bkgCountsImage)
    deallocate(rateImage, bkgCountsImage, expImage)

    avCountsImageSet = dataSet(avCountsImageSetName, MODIFY)
!    call setAttribute(avCountsImageSet, 'MAX_BACK', maxBkgRate&
!      , 'counts/pixel/s', 'Max value of bkg rate added.')
call setAttribute(avCountsImageSet, 'MAX_BACK', maxBkgCounts&
  , 'counts/pixel', 'Max value of bkg counts added.')

!    if (maxBkgRate > 0.0) then
if (maxBkgCounts > 0.0) then
      bkgSet = dataSet(bkgCountMapSetName, READ)
      call setAttribute(avCountsImageSet, attribute(bkgSet, 'VIG_FRAC'))
      call release(bkgSet)
    end if

    rateImageSet = dataSet(rateImageSetName, READ)
    call copyBlock(avCountsImageSet, block(rateImageSet, 'LOGNLOGS'))
    call release(rateImageSet)
    call release(avCountsImageSet)

  end subroutine makeAvCountsImage

  !---------------------------------------------------------------------
  subroutine poissonizeCountsImage(avCountsImageName, countsImageName&
  , tempImageName)
  !---------------------------------------------------------------------

    implicit none

    character(*),   intent(in) :: avCountsImageName,&
                                  countsImageName,&
                                  tempImageName

    real(single), pointer ::&
      avCountsImage(:,:)
!    type(DataSetT) :: avCountsImageSet, countsImageSet
    integer(int16) :: xSize, ySize
    integer(int32), allocatable :: outImageInt(:,:)

    !-------------------------------------------------------------------
    ! Start

    call readArrayData(avCountsImageName, avCountsImage)
    xSize = size(avCountsImage, 1)
    ySize = size(avCountsImage, 2)
    allocate(outImageInt(xSize, ySize))

    call eventizeRateImage_slave(avCountsImage, outImageInt)
    call saveImage(outImageInt, avCountsImageName, countsImageName&
      , tempImageName)
    ! Copies over all keywords (eg wcs, INSTRUM, ecf) from the avCountsImage.

    deallocate(avCountsImage)

!    countsImageSet = dataSet(countsImageName, MODIFY)
!    avCountsImageSet = dataSet(avCountsImageName, READ)
!    call copyBlock(countsImageSet, block(avCountsImageSet, 'LOGNLOGS'))
!    call release(avCountsImageSet)
!    call release(countsImageSet)

  end subroutine poissonizeCountsImage

  !---------------------------------------------------------------------
  subroutine eventizeRateImage_slave(imageReal, image, status)
  !---------------------------------------------------------------------
  ! Adapted from asmooth-2.5 (temp_ssclib)

    implicit none

    real(single),   intent(in)            :: imageReal(:,:)
    integer(int32), intent(out)           :: image(:,:)
    integer,        intent(out), optional :: status

    integer :: xi, yi, xSize, ySize, localStatus

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    xSize = size(image, 1)
    ySize = size(image, 2)

    do xi = 1, xSize
      do yi = 1, ySize
        image(xi, yi) = poissonRandom(imageReal(xi, yi))
      end do
    end do

    if (present(status)) status = localStatus

  end subroutine eventizeRateImage_slave

  !---------------------------------------------------------------------
  subroutine correctEddington(detSrcSetName, minLikelihood)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: detSrcSetName
    real(single), intent(in) :: minLikelihood

    type(DataSetT) :: sourceSet
    type(TableT) :: sourceTab
    type(ColumnT) :: srcLCorrCol, srcLErrCorrCol, detBoxCountsErrCol
    real(single), pointer, dimension(:) ::&
      srcL,&
      srcLCorr, srcLErrCorr,&
      detBoxCounts, detBoxCountsErr, bkgBoxCounts,&
      likelihood
    integer(int32), pointer :: idBandData(:), mlIdSrcData(:)
    integer(int32) :: row, numRows, mlIdSrc
    real(single) :: avSrcBoxCounts, avSrcBoxCountsErr
    integer :: foreignStatus
    logical(bool), allocatable :: mask(:)

    !-------------------------------------------------------------------
    ! Start

    !...................................................................
    ! Open the source set:

    sourceSet = dataSet(detSrcSetName, MODIFY)
    sourceTab = table(sourceSet, 'SRCLIST')
    numRows = numberOfRows(sourceTab)
    allocate(mask(numRows))

    srcL    => real32Data(column(sourceTab, 'RATE'))

    srcLCorrCol    = addOrOpenColumn(sourceTab, 'RATE_CORR'&
      , REAL32, 'pixels')
    srcLErrCorrCol = addOrOpenColumn(sourceTab, 'RATE_ERR_CORR'&
      , REAL32, 'pixels')

    do row = 0, numRows - 1
      call setToNull(srcLCorrCol,    row)
      call setToNull(srcLErrCorrCol, row)
    end do

    srcLCorr    => real32Data(srcLCorrCol)
    srcLErrCorr => real32Data(srcLErrCorrCol)

    detBoxCountsErrCol = column(sourceTab, 'SRC_BOX_CTS_ERR')

    idBandData  => int32Data(column(sourceTab, 'ID_BAND'))
    mlIdSrcData => int32Data(column(sourceTab, 'ML_ID_SRC'))
    detBoxCounts    => real32Data(column(sourceTab, 'SRC_BOX_CTS'))
    detBoxCountsErr => real32Data(detBoxCountsErrCol)
    bkgBoxCounts    => real32Data(column(sourceTab, 'BKG_BOX_CTS'))
    likelihood      => real32Data(column(sourceTab, 'LIKE'))

    !...................................................................
    ! Now work through the sources and do the Eddington correction:

    do row = 0, numRows - 1
    if (likelihood(row) < minLikelihood) cycle ! leave these as nulls.
    if (isNull(detBoxCountsErrCol, row)) cycle

      call inverseEddington(bkgBoxCounts(row)&
        , detBoxCounts(row), detBoxCountsErr(row)&
        , minLikelihood, avSrcBoxCounts, avSrcBoxCountsErr, foreignStatus)

      if (foreignStatus < 0) then
        if (foreignStatus == -4) then
          call warning('cantDoEddington', "")
    cycle

        else
print *, 'ssclib: Status return from inverseEddington was', foreignStatus
call error('inverseEddington','')
        end if
      end if

      srcLCorr(row)    = srcL(row) * (avSrcBoxCounts / detBoxCounts(row))
      srcLErrCorr(row) = srcL(row) * (avSrcBoxCountsErr / avSrcBoxCounts)
    end do

    ! Now fix up the band0 numbers:

    do mlIdSrc = minval(mlIdSrcData), maxval(mlIdSrcData)
    if (.not. any(mlIdSrcData == mlIdSrc)) cycle

      if (count(mlIdSrcData == mlIdSrc .and. idBandData == 0) .ne. 1)&
call error('noBand0ThisSrc','')

      mask = (mlIdSrcData == mlIdSrc&
      .and. idBandData >= 1&
      .and. idBandData <= 9)

!if (count(mask) .ne. 5) call error('wrongNumBands','')

      row = smallestTrueIndex(logical(mlIdSrcData == mlIdSrc&
        .and. idBandData == 0, bool)) - 1 ! Note that we have already
      ! established that there is only 1 source which matches this criteria.
      !
      ! The function smallestTrueIndex() is in ssclib/array_utils.

      srcLCorr(row) = sum(srcLCorr, mask)
      srcLErrCorr(row) = sqrt(sum(srcLErrCorr * srcLErrCorr, mask))
    end do

    call release(sourceSet)

  end subroutine correctEddington

  !---------------------------------------------------------------------
  subroutine correctPosAndRates(detSrcSetName, ximaErrCorrSetName&
      , yimaErrCorrSetName, rateCorrSetName, readCorrRates)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: detSrcSetName, ximaErrCorrSetName&
      , yimaErrCorrSetName, rateCorrSetName
    logical, intent(in) :: readCorrRates

    type(DataSetT) :: sourceSet
    type(TableT) :: sourceTab
    real(single), pointer, dimension(:) ::&
      srcX, srcY, srcXErr, srcYErr,&
      srcXCorr, srcYCorr, srcXErrCorr, srcYErrCorr
!, srcLCorr, srcLErrCorr
!!!, srcL, srcLErr
character(10) :: dummy

    !-------------------------------------------------------------------
    ! Start

! This is just to stop compiler warnings while these arguments are otherwised unused!
if (readCorrRates) then
  dummy = ximaErrCorrSetName
  dummy = yimaErrCorrSetName
  dummy = rateCorrSetName
end if

    sourceSet = dataSet(detSrcSetName, READ)
    sourceTab = table(sourceSet, 'SRCLIST')

    srcX => real32Data(column(sourceTab, 'X_IMA'))
    srcY => real32Data(column(sourceTab, 'Y_IMA'))
!!!    srcL => real32Data(column(sourceTab, 'RATE'))
    srcXErr => real32Data(column(sourceTab, 'X_IMA_ERR'))
    srcYErr => real32Data(column(sourceTab, 'Y_IMA_ERR'))
!!!    srcLErr => real32Data(column(sourceTab, 'RATE_ERR'))

    srcXCorr => real32Data(addOrOpenColumn(sourceTab, 'X_IMA_CORR'&
      , REAL32, 'pixels'))
    srcYCorr => real32Data(addOrOpenColumn(sourceTab, 'Y_IMA_CORR'&
      , REAL32, 'pixels'))
!!!    srcLCorr => real32Data(addOrOpenColumn(sourceTab, 'RATE_CORR'&
!!!      , REAL32, 'pixels'))
    srcXErrCorr => real32Data(addOrOpenColumn(sourceTab, 'X_IMA_ERR_CORR'&
      , REAL32, 'pixels'))
    srcYErrCorr => real32Data(addOrOpenColumn(sourceTab, 'Y_IMA_ERR_CORR'&
      , REAL32, 'pixels'))
!!!    srcLErrCorr => real32Data(addOrOpenColumn(sourceTab, 'RATE_ERR_CORR'&
!!!      , REAL32, 'pixels'))

    !...................................................................
    ! Now correct for systematic errors:

!!  call correctY('xima_corrector.ds', srcL, srcX) ! purely additive
!  call correctY(ximaErrCorrSetName, srcL, srcXuncert)
!    ! purely additive
!
!!  call correctY('yima_corrector.ds', srcL, srcY) ! purely additive
!  call correctY(yimaErrCorrSetName, srcL, srcYuncert)
!    ! purely additive

srcXCorr = srcX
srcYCorr = srcY

srcXErrCorr = sqrt((2.5 * srcXErr)**2.0 + 0.25**2.0)
srcYErrCorr = sqrt((2.5 * srcYErr)**2.0 + 0.25**2.0)

!*** at the moment there is no change in rate or rate_err so there is no point in overwriting the Eddington:
!if (.not. readCorrRates) then
!  srcLCorr    = srcL
!  srcLErrCorr = srcLErr
!end if


!  call correctY(rateCorrSetName, srcC, srcLCorr) ! purely multiplicative
!  call correctY(rateCorrSetName, srcC, srcLErrCorr) ! purely multiplicative

    call release(sourceSet)

  end subroutine correctPosAndRates

  !---------------------------------------------------------------------
  subroutine correctY(correctorSetName, x, y)
  !---------------------------------------------------------------------
  ! This applies correction factors to y, giving
  !
  !	y_new = (y_old - sub) / div.
  !
  ! The two correction functions add and mul are defined via a series of
  ! samples, stored in two tables named 'SUB' and 'DIV'.
  !
  ! Naturally 'div' is not permitted to equal zero.

    implicit none

    character(*), intent(in)    :: correctorSetName
    real(single), intent(in)    :: x(:)
    real(single), intent(inout) :: y(size(x))

    type(DataSetT) :: correctorSet
    type(TableT) :: subtractTab, divisionTab
    real(single) :: sub(size(x)), div(size(x))

    !-------------------------------------------------------------------
    ! Start

    correctorSet = dataSet(correctorSetName, READ)

    if (hasBlock(correctorSet, 'SUB')) then
      subtractTab = table(correctorSet, 'SUB')
      call getCorrectorFunction(subtractTab, x, sub)

    else
      sub = 0.0
    end if

    if (hasBlock(correctorSet, 'DIV')) then
      divisionTab = table(correctorSet, 'DIV')
      call getCorrectorFunction(divisionTab, x, div)
if (any(div == 0.0)) call error('','')

    else
      div = 1.0
    end if

    call release(correctorSet)

    y = (y - sub) / div

  end subroutine correctY

  !---------------------------------------------------------------------
  subroutine getCorrectorFunction(tab, x, y)
  !---------------------------------------------------------------------
  ! Each correction function is defined via a series of (x,y) samples
  ! stored in the table given. The table must have at least 1 row.

    implicit none

    type(TableT), intent(in) :: tab
    real(single), intent(in) :: x(:)
    real(single), intent(out) :: y(size(x))

    integer(int32) :: numSamples, i
    real(single), pointer :: samplesX(:), samplesY(:)

    !-------------------------------------------------------------------
    ! Start

    numSamples = numberOfRows(tab)
if (numSamples <= 0) call error('bug','')

    samplesX   => real32Data(column(tab, 'X'))
    samplesY   => real32Data(column(tab, 'Y'))
    if (any(samplesX(:numSamples-2) >= samplesX(1:)))&
call error('badSampleOrder','')

    where(x < samplesX(0)) y = samplesY(0)
    do i = 0, numSamples-2
      where(x >= samplesX(i) .and. x <  samplesX(i+1))&
        y = fracVal(samplesY(i), samplesY(i+1)&
          , samplesX(i), samplesX(i+1), x)
    end do
    where(x >= samplesX(numSamples-1)) y = samplesY(numSamples-1)

  end subroutine getCorrectorFunction

  !---------------------------------------------------------------------
  subroutine detRateToFluxRandom(detSrcListSetName, useCorrected)
  !---------------------------------------------------------------------
  ! This subroutine reads in RATE values (in counts/sec) from a list of
  ! detected sources + information about the energy band, logN-logS and
  ! photon index and converts the RATE values to FLUXRAND. The latter is
  ! defined in terms of a list of simulated sources which have flux
  ! values that follow the specified logN-logS distribution. The flux
  ! values of the simulated source list are obtained via a process that
  ! takes a set of random numbers having an even distribution between 0
  ! and 1 and subjects them to a transform so that they become flux values
  ! which obey the specified logN-logS distribution. (It is intended that
  ! the original random numbers should be retained for each simulated
  ! source, however it is possible to recalculate them in the case in
  ! which this is not done.) These original random numbers are the
  ! FLUXRAND values. Errors in RATE (RATE_ERR) are also transformed to
  ! this system to become FLXRNERR.

    implicit none

    character(*), intent(in) :: detSrcListSetName
    logical, intent(in) :: useCorrected

    real(single), pointer :: modelFluxDensity0(:), modelCumSkyDensity(:)
    real(single) ::&
      hiPower, multiplier(3, 0:9)
    real(double) ::&
      photonSpectralIndex,&
      fluxEeVLo(3, 0:9), fluxEeVHi(3, 0:9), ecf(3, 9)
    type(DataSetT) :: srcSet
    type(TableT) :: templateTab, srcTab
    type(ColumnT) :: ranCol, ranErrCol
    real(single), dimension(:), pointer ::&
      ran, ranErr, rate, rateErr
    integer(int32) :: numRows, idInst, idBand, mlIdSrc
    real(single), allocatable ::&
      srcFlux(:), srcFluxErr(:), srcFluxDensity0(:), srcFluxDensity0Err(:)
    integer :: foreignStatus, row
    integer(int32), pointer :: idBandData(:), idInstData(:), mlIdSrcData(:)
    logical(bool), allocatable :: mask(:)
    logical(bool) :: kwdsPresent(3,0:9)
type(rangesPtrType) :: fluxEeVranges(3)
!type(RangeT), pointer :: fluxEeVranges_temp(:)

    !-------------------------------------------------------------------
    ! Start

    !...................................................................
    ! Read the logN-logS distribution model from the template extension and
    ! transform the X values from units of flux (in the energy band also
    ! specified in the template) to flux density:
    
    srcSet = dataSet(detSrcListSetName, MODIFY)
    templateTab = table(srcSet, 'LOGNLOGS')
    call getLogNlogS(templateTab, modelFluxDensity0, modelCumSkyDensity&
      , hiPower, photonSpectralIndex)!, fluxEeVLo, fluxEeVHi)

    !...................................................................
    ! Now open the source table and read the columns of interest:

    srcTab = table(srcSet, 'SRCLIST')
    numRows = numberOfRows(srcTab)
    allocate(srcFlux(numRows), srcFluxErr(numRows))
    allocate(srcFluxDensity0(numRows), srcFluxDensity0Err(numRows))
    allocate(mask(numRows))

    ranCol    = addOrOpenColumn(srcTab, 'FLUXRAND', REAL32&
      , comment='Should be evenly distributed between 0 and 1.')
    ranErrCol = addOrOpenColumn(srcTab, 'FLXRNERR', REAL32)

    do row = 0, numberOfRows(srcTab) - 1
      call setToNull(ranCol, row)
      call setToNull(ranErrCol, row)
    end do

    ran     => real32Data(ranCol)
    ranErr  => real32Data(ranErrCol)

    if (useCorrected) then
      rate    => real32Data(column(srcTab, 'RATE_CORR'))
      rateErr => real32Data(column(srcTab, 'RATE_ERR_CORR'))
    else
      rate    => real32Data(column(srcTab, 'RATE'))
      rateErr => real32Data(column(srcTab, 'RATE_ERR'))
    end if

    idBandData  => int32Data(column(srcTab, 'ID_BAND'))
    idInstData  => int32Data(column(srcTab, 'ID_INST'))
    mlIdSrcData => int32Data(column(srcTab, 'ML_ID_SRC'))

    !...................................................................
    ! Load the band edges and ecfs and other inst/band dependent factors:

    call loadEcfStuff(srcSet, ecf, fluxEeVLo, fluxEeVHi, fluxEeVranges&
      , kwdsPresent)

    do idInst = 1, 3
    if (.not. kwdsPresent(idInst, 0)) cycle

!fluxEeVranges_temp => fluxEeVranges(idInst)%ranges
      multiplier(idInst, 0)&
!        = fluxToFluxDensity0(photonSpectralIndex, fluxEeVranges_temp, 1.0_single)
        = fluxToFluxDensity0(photonSpectralIndex, fluxEeVranges(idInst)%ranges, 1.0_single)

      do idBand = 1, 9
      if (.not. kwdsPresent(idInst, idBand)) cycle

        multiplier(idInst, idBand)&
          = fluxToFluxDensity0(photonSpectralIndex&
          , fluxEeVLo(idInst, idBand), fluxEeVHi(idInst, idBand), 1.0)
      end do
    end do

    !...................................................................
    ! Calculate the flux values:

    do idInst = 1, 3
      do idBand = 1, 9
        mask = (idInstData == idInst .and. idBandData == idBand)
if (any(mask) .and. .not. kwdsPresent(idInst, idBand)) call error('','')

        where(mask)&
          srcFlux    = rate    * 1.0e-11 / ecf(idInst, idBand)
          srcFluxErr = rateErr * 1.0e-11 / ecf(idInst, idBand)
      end do
    end do

    ! Now fix up the band0 numbers:

    do mlIdSrc = minval(mlIdSrcData), maxval(mlIdSrcData)
    if (.not. any(mlIdSrcData == mlIdSrc)) cycle

      if (count(mlIdSrcData == mlIdSrc .and. idBandData == 0) < 1)&
call error('noBand0','')
      if (count(mlIdSrcData == mlIdSrc .and. idBandData == 0) > 1)&
call error('tooManyBand0','')

      mask = (mlIdSrcData == mlIdSrc&
      .and. idBandData >= 1&
      .and. idBandData <= 9)

if (count(mask) <= 0) call error('wrongNumBands','')

      row = smallestTrueIndex(logical(mlIdSrcData == mlIdSrc&
        .and. idBandData == 0, bool)) - 1

      srcFlux(row+1) = sum(srcFlux, mask)
      srcFluxErr(row+1) = sqrt(sum(srcFluxErr * srcFluxErr, mask))
    end do

    !...................................................................
    ! Now calculate the fluxDensity0 values:

    do idInst = 1, 3
      do idBand = 0, 9 ! Note no special treatment now for band 0.
        where(idInstData == idInst .and. idBandData == idBand)&
          srcFluxDensity0    = srcFlux    * multiplier(idInst, idBand)
          srcFluxDensity0Err = srcFluxErr * multiplier(idInst, idBand)
      end do
    end do

    !...................................................................
    ! Now at last we can invert the fluxDensity0 values to get fluxRand:

    call invertLogNlogSrandoms(modelFluxDensity0, modelCumSkyDensity&
      , srcFluxDensity0, hiPower, ran, foreignStatus)

    if (foreignStatus < 0) then
print *, 'ssclib: Status return from invertLogNlogSrandoms was', foreignStatus
call error('invertLogNlogSrandoms',"eimsim_utils: The function &
&invertLogNlogSrandoms returned an error status.")
    end if

    call ampErrToFluxRandErr(modelFluxDensity0, modelCumSkyDensity&
      , srcFluxDensity0, srcFluxDensity0Err, hiPower, ranErr&
      , foreignStatus)

    if (foreignStatus < 0) then
print *, 'ssclib: Status return from ampErrToFluxRandErr was', foreignStatus
call error('ampErrToFluxRandErr',"eimsim_utils: The function &
&ampErrToFluxRandErr returned an error status.")
    end if

    call release(srcSet)
    deallocate(modelFluxDensity0, modelCumSkyDensity)
    do idInst = 1, 3
      deallocate(fluxEeVranges(idInst)%ranges)
    end do

  end subroutine detRateToFluxRandom

  !---------------------------------------------------------------------
  subroutine loadEcfStuff(set, ecf, EeVLo, EeVHi, EeVranges, kwdsPresent)
  !---------------------------------------------------------------------
!*** note this is called in many places and I haven't added the rangesPtrType call argument everywhere yet.

!*** also this is all a bit new and I am as yet unsure about whether to change kwdsPresent(idInst, 0) conditions etc. Check check check!

    implicit none

    type(DataSetT), intent(in) :: set
    real(double), intent(out) :: ecf(3,9), EeVLo(3,0:9), EeVHi(3,0:9)
    type(rangesPtrType) :: EeVranges(3)
    logical(bool), intent(out) :: kwdsPresent(3,0:9)

    character(2) :: instName(3)
    integer :: idInst, idBand, rangeNum
    character(8) :: ecfKwdName, eLoKwdName, eHiKwdName

    !-------------------------------------------------------------------
    ! Start

    instName = (/'PN', 'M1', 'M2'/)
    kwdsPresent = .false. ! default
    ecf       = 0.0 ! default
    EeVLo = 0.0 ! default
    EeVHi = 0.0 ! default

    do idInst = 1, 3
      do idBand = 1, 9
        write(ecfKwdName, '(a2, a1, i1, a4)') instName(idInst), '_'&
          , idBand, '_ECF'

        write(eLoKwdName, '(a2, a1, i1, a4)') instName(idInst), '_'&
          , idBand, '_ELO'

        write(eHiKwdName, '(a2, a1, i1, a4)') instName(idInst), '_'&
          , idBand, '_EHI'

        if   (hasAttribute(set, ecfKwdName)&
        .and. hasAttribute(set, eLoKwdName)&
        .and. hasAttribute(set, eHiKwdName)) then
          kwdsPresent(idInst, idBand) = .true.

          ecf(  idInst, idBand) = real32Attribute(set, ecfKwdName)
          EeVLo(idInst, idBand) = real32Attribute(set, eLoKwdName)
          EeVHi(idInst, idBand) = real32Attribute(set, eHiKwdName)
        end if
      end do

      if (any(kwdsPresent(idInst, 1:9))) kwdsPresent(idInst, 0) = .true.

      EeVLo(idInst, 0) = EeVLo(idInst, 1)
      EeVHi(idInst, 0) = EeVHi(idInst, 9)
    end do

    do idInst = 1, 3
      allocate(EeVranges(idInst)%ranges(count(kwdsPresent(idInst,:))))
      rangeNum = 0
      do idBand = 1, 9
      if (.not. kwdsPresent(idInst, idBand)) cycle
        rangeNum = rangeNum + 1
        EeVranges(idInst)%ranges(rangeNum)%lower%type = INCLUSIVE
        EeVranges(idInst)%ranges(rangeNum)%lower%value = EeVLo(idInst, idBand)
        EeVranges(idInst)%ranges(rangeNum)%upper%type = INCLUSIVE
        EeVranges(idInst)%ranges(rangeNum)%upper%value = EeVHi(idInst, idBand)
      end do
    end do

  end subroutine loadEcfStuff

  !---------------------------------------------------------------------
  subroutine loadEcfStuff_20030912(set, ecf, EeVLo, EeVHi, kwdsPresent)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set
    real(double), intent(out) :: ecf(3,5), EeVLo(3,0:5), EeVHi(3,0:5)
    logical(bool), intent(out) :: kwdsPresent(3,0:5)

    character(2) :: instName(3)
    integer :: idInst, idBand
    character(8) :: ecfKwdName, eLoKwdName, eHiKwdName

    !-------------------------------------------------------------------
    ! Start

    instName = (/'PN', 'M1', 'M2'/)
    kwdsPresent = .false. ! default
    ecf       = 0.0 ! default
    EeVLo = 0.0 ! default
    EeVHi = 0.0 ! default

    do idInst = 1, 3
      do idBand = 1, 5
        write(ecfKwdName, '(a2, a1, i1, a4)') instName(idInst), '_'&
          , idBand, '_ECF'

        write(eLoKwdName, '(a2, a1, i1, a4)') instName(idInst), '_'&
          , idBand, '_ELO'

        write(eHiKwdName, '(a2, a1, i1, a4)') instName(idInst), '_'&
          , idBand, '_EHI'

        if   (hasAttribute(set, ecfKwdName)&
        .and. hasAttribute(set, eLoKwdName)&
        .and. hasAttribute(set, eHiKwdName)) then
          kwdsPresent(idInst, idBand) = .true.

          ecf(  idInst, idBand) = real32Attribute(set, ecfKwdName)
          EeVLo(idInst, idBand) = real32Attribute(set, eLoKwdName)
          EeVHi(idInst, idBand) = real32Attribute(set, eHiKwdName)
        end if
      end do

      if (all(kwdsPresent(idInst, 1:5))) kwdsPresent(idInst, 0) = .true.

      EeVLo(idInst, 0) = EeVLo(idInst, 1)
      EeVHi(idInst, 0) = EeVHi(idInst, 5)
    end do

  end subroutine loadEcfStuff_20030912

  !---------------------------------------------------------------------
  subroutine ampErrToFluxRandErr(x, yInt, xSamples, xErrs, hiPower, randErrs&
  , status)
  !---------------------------------------------------------------------
  ! This sub was modified from invertLogNlogSrandoms.
  !.....................................................................
  !
  ! The subroutine takes a vector of random values xSamples() which have a
  ! distribution P which, when integrated, follows a piecewise power law
  ! as follows. The ith segment of the integrated piecewise function is
  ! defined between x(i) and x(i+1) and is of the form
  !
  !	log(yInt) = log(yInt(i)) + slope * (log(x) - log(x(i)))
  !
  ! where
  !	         log(yInt(i+1)) - log(yInt(i))
  !	slope = -------------------------------.
  !	            log(x(i+1)) - log(x(i))
  !
  ! Above x(N) the integrated function is defined by
  !
  !	log(yInt) = log(yInt(N)) + hiPower * (log(x) - log(x(N))).
  !
  ! In the present subroutine the original random values, which are evenly
  ! distributed between 0 and 1, are returned.
  !.....................................................................
  !
  ! Notes: (i) The x values must be in increasing order, and they cannot be
  ! negative. (ii) The y values must all be >0, and must never increase with
  ! increasing i. (iii) N must be >0. (iv) hiPower must be <0.
  !
  ! In the event of an error, the subroutine returns values of -1 for all
  ! elements of numbers().
  !.....................................................................

    implicit none

    real,    intent(in)            :: x(:),&
                                      yInt(size(x)),&
                                      xSamples(:),&
                                      xErrs(size(xSamples)),&
                                      hiPower
    real,    intent(out)           :: randErrs(size(xSamples))
    integer, intent(out), optional :: status

    integer ::&
      localStatus, foreignStatus, numBins, numXsamples, i, bin
    real :: slopes(size(x))

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    randErrs = -1

    numBins = size(x)
    numXsamples = size(xSamples)

    call doLogNlogSchecks(x, yInt, hiPower, foreignStatus)

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus
  return
    end if

    ! Calculate slopes:

    do bin = 1, numBins - 1
      slopes(bin) = (log(yInt(bin+1)) - log(yInt(bin)))&
                  / (log(x(   bin+1)) - log(x(   bin)))
    end do
    slopes(numBins) = hiPower

    ! Main part of the routine:

    do i = 1, numXsamples
      bin = getBinNumber(x, xSamples(i)) ! in array_utils
if (bin < 1) bin = 1 ! to cope with zero- and non-(ID_BAND==0)-RATE values.
      randErrs(i) = xErrs(i) * slopes(bin) * yInt(bin) / x(bin)&
        * (xSamples(i) / x(bin))**(slopes(bin) - 1.0)
    end do
    randErrs = abs(randErrs / yInt(1))

    if (present(status)) status = localStatus

  end subroutine ampErrToFluxRandErr

  !---------------------------------------------------------------------
  subroutine compareDetWithSim(detSrcSetName, simSrcSetName, tempSetName&
  , obsId, idBand)
  !---------------------------------------------------------------------
  ! Given RATE, X_IMA and Y_IMA values for two source lists, one simulated
  ! and the other detected, this subroutine performs two functions: (i) it
  ! attempts to deduce which of the simulated sources corresponds with each
  ! detected source; (ii) it calculates the probability of the chance
  ! occurrence of a source 'closer' than this to the detected source.
  !
  ! The first part of this task, namely the search for the 'best' match
  ! between each detection and the simulated source list, is carried out by
  ! finding the nearest simulated neighbour to each detection in an
  ! abstract space. This space is defined by the three axes
  ! (X_IMA/X_IMA_ERR), (Y_IMA/Y_IMA_ERR) and (RATE/RATE_ERR).
  !.....................................................................
  !
  ! The present subroutine has been modified from the original
  ! compareDetWithSim. In this version the distribution of fluxes (or rates,
  ! or brightnesses, or whatever linear scaling you prefer to think of in)
  ! is subjected to a non-linear scaling in order to recover the original
  ! evenly-distributed random numbers. The nearest neighbour search is thus
  ! performed in an abstract space defined here by the same X and Y axes
  ! as previously but the third axis is replaced by (FLUXRAND/FLXRNERR).
  ! Hopefully this will avoid the problem of too many 'false' matches
  ! between a detection and a much fainter (but spatially near by) sim source.
  !.....................................................................

    implicit none

    character(*),   intent(in) :: detSrcSetName, simSrcSetName, tempSetName
    character(10),  intent(in) :: obsId
    integer(int8),  intent(in) :: idBand

    real(single), dimension(:), pointer ::&
      detSrcXima, detSrcYima, detSrcFluxRand,&
      detSrcXimaErr, detSrcYimaErr, detSrcFluxRandErr,&
      simSrcFluxRand, simFluxDensity0
    integer(int16), pointer :: flag(:), simSrcFlag(:)
    integer(int32), pointer ::&
      detSrcMlId(:), simIndices(:)
!, detSrcIdBand(:)
    real(single), dimension(:), allocatable ::&
      simSrcXima, simSrcYima, devOverStdDev, nullProb
    integer(int32) ::&
      numDetSrcs, numSimSrcs, detSrc, bitToFlagDetection, mlIdVal
    integer(int32), allocatable :: indexOfNearestSim(:)
    logical, allocatable :: alsoRan(:)
    type(WcsT) :: imageWcs
    real(single) :: srcPerPixel, srcPerSquareDeg, pixelAreaDeg
    type(DataSetT) :: simSourceSet
    type(TableT) :: simSourceTab
    real(double), pointer :: simSrcRaDeg(:), simSrcDecDeg(:)
    real(double), allocatable :: tanX(:), tanY(:)

character(12) :: style = 'ellipse' !*** don't try halfellipse just yet. Need to think about the correct volume shape.

    !-------------------------------------------------------------------
    ! Start

    !...................................................................
    ! First, compare the template tables in the det and sim lists and exit
    ! if they are incompatible:

    if (templatesUnequal(detSrcSetName, simSrcSetName))&
call error('templatesUnequal',"eimsim_utils: The LOGNLOGS extension in the &
&list of simulated sources does not match that in the list of detections.")

    !...................................................................
    ! Load the relevant det and sim source columns:

    call readDetSourceList(detSrcSetName, detSrcXima, detSrcYima&
      , detSrcFluxRand, detSrcXimaErr, detSrcYimaErr, detSrcFluxRandErr&
      , detSrcMlId, idBand, tempSetName)

    do mlIdVal = minval(detSrcMlId), maxval(detSrcMlId)
      if (count(detSrcMlId == mlIdVal) > 1)&
call error('tooManyRowsPerBand',"eimsim_utils: In the list of detected sources, &
&there is more than 1 row per source with the specified value of ID_BAND.")
    end do

    numDetSrcs = size(detSrcXima)
    allocate(indexOfNearestSim(numDetSrcs), devOverStdDev(numDetSrcs))
    allocate(nullProb(numDetSrcs), alsoRan(numDetSrcs))

    call readSimSourceList(simSrcSetName, obsId, simSrcRaDeg, simSrcDecDeg&
      , simSrcFluxRand, simSrcFlag, simIndices, simFluxDensity0)

    !...................................................................
    ! Set the cal state for instrument and translate sim src ra/dec into
    ! image coordinates:

    call getCalInfo(detSrcSetName, instrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude)

    call readWcsSourceList(detSrcSetName, imageWcs)

    numSimSrcs = size(simSrcRaDeg)
    allocate(tanX(numSimSrcs), tanY(numSimSrcs))
!    call raDecToTan(imageWcs%X%refWorld, imageWcs%Y%refWorld&
!      , simSrcRaDeg, simSrcDecDeg, tanX, tanY) ! in ssclib/coordinate.
call raDecToTan(simSrcRaDeg*PiOn180, simSrcDecDeg*PiOn180&
 , imageWcs%X%refWorld*PiOn180, imageWcs%Y%refWorld*PiOn180&
 , tanX, tanY) ! in ssclib/coordinate.
    deallocate(simSrcRaDeg, simSrcDecDeg)

    allocate(simSrcXima(numSimSrcs), simSrcYima(numSimSrcs))
    call tanToPixels(imageWcs, tanX, tanY, simSrcXima, simSrcYima)
    deallocate(tanX, tanY)

    !...................................................................
    ! Misc setups:

    call getSrcPerSquareDeg(detSrcSetName, srcPerSquareDeg)

!    pixelAreaDeg = abs(imageWcs%X%pixelSizeWorld)&
!      * abs(imageWcs%Y%pixelSizeWorld)
pixelAreaDeg = wcsPixelAreaWorld(imageWcs)
!*** strictly speaking should check that the world units are degrees.

    srcPerPixel = srcPerSquareDeg * pixelAreaDeg

    select case(instrumentId)
    case(EPN)
      bitToFlagDetection = DETECTED_PN
    case(EMOS1)
      bitToFlagDetection = DETECTED_M1
    case(EMOS2)
      bitToFlagDetection = DETECTED_M2
    case default
call error('badInstrument',"eimsim_utils: Unrecognized instrument ID.")
    end select

    simSrcFlag = ibclr(simSrcFlag, bitToFlagDetection) ! default

    !...................................................................
    ! Main loop over detected sources:

    do detSrc = 1, numDetSrcs
      call findNearestSimSource(detSrcXima(detSrc), detSrcYima(detSrc)&
        , detSrcFluxRand(detSrc), detSrcXimaErr(detSrc), detSrcYimaErr(detSrc)&
        , detSrcFluxRandErr(detSrc), simSrcXima, simSrcYima, simSrcFluxRand&
        , simSrcFlag, simIndices, style, bitToFlagDetection&
        , indexOfNearestSim(detSrc), devOverStdDev(detSrc))

      call probOfOneOrMoreByChance(detSrcFluxRand(detSrc)&
        , detSrcXimaErr(detSrc), detSrcYimaErr(detSrc)&
        , detSrcFluxRandErr(detSrc), devOverStdDev(detSrc), srcPerPixel&
        , nullProb(detSrc))
    end do

    deallocate(simIndices)

!!!    call flagAlsoRans(indexOfNearestSim, devOverStdDev, alsoRan, status)
!!!    if (status .lt. 0) then
!!!      print *, 'det src', detSrc, ': flagAlsoRans status return'&
!!!      , status
!!!  return
!!!    end if

    !...................................................................
    ! Wrap up:

    call saveFinalDetSrcList(detSrcSetName, simSrcXima, simSrcYima&
      , simSrcFluxRand, simSrcFlag, detSrcMlId, indexOfNearestSim&
      , nullProb, simFluxDensity0)
      !!!, alsoRan)!, devOverStdDev)

    simSourceSet = dataSet(simSrcSetName, MODIFY)
    simSourceTab = table(simSourceSet, 'SRCLIST')
    flag => int16Data(column(simSourceTab, 'FLAG_'//obsId))
    flag = simSrcFlag
    call release(simSourceSet)

    deallocate(detSrcXima, detSrcYima, detSrcFluxRand)
    deallocate(detSrcXimaErr, detSrcYimaErr, detSrcFluxRandErr)
    deallocate(simSrcXima, simSrcYima, simSrcFluxRand, simFluxDensity0)
    deallocate(simSrcFlag, detSrcMlId)

  end subroutine compareDetWithSim

  !---------------------------------------------------------------------
  logical function templatesUnequal(setNameA, setNameB)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: setNameA, setNameB

    type(DataSetT) :: setA, setB
    type(TableT) :: tabA, tabB
!    character(20) :: instrumentNameA, instrumentNameB
!    integer(int32) :: instrumentIdA, instrumentIdB
    real(double) ::&
      ecfA(3,9), fluxEeVLoA(3,0:9), fluxEeVHiA(3,0:9),&
      ecfB(3,9), fluxEeVLoB(3,0:9), fluxEeVHiB(3,0:9)
    logical(bool) :: kwdsPresentA(3,0:9), kwdsPresentB(3,0:9)
    real(single), pointer :: fd0A(:), fd0B(:), csdA(:), csdB(:)
    real(single) :: hiPowerA, hiPowerB
    real(double) ::&
      modelEeVLoA, psiA, modelEeVHiA, modelEeVLoB, psiB, modelEeVHiB,&
      coneRaA, coneDecA, coneAngleA, coneAreaA,&
      coneRaB, coneDecB, coneAngleB, coneAreaB

    !-------------------------------------------------------------------
    ! Start

    templatesUnequal = .false. ! default

    setA = dataSet(setNameA, READ)
    setB = dataSet(setNameB, READ)

    ! Compare inst, ecf, elo, ehi:

!    instrumentNameA = stringAttribute(setA, "INSTRUME")
!    instrumentIdA = instId(instrumentNameA) ! in caloalutils/src/caloaldefs.f90

!    instrumentNameB = stringAttribute(setB, "INSTRUME")
!    instrumentIdB = instId(instrumentNameB) ! in caloalutils/src/caloaldefs.f90

!  if (instrumentIdA .ne. instrumentIdB) return

    call loadEcfStuff_20030912(setA, ecfA, fluxEeVLoA, fluxEeVHiA, kwdsPresentA)
    call loadEcfStuff_20030912(setB, ecfB, fluxEeVLoB, fluxEeVHiB, kwdsPresentB)

    if (any(ecfA .ne. ecfB&
    .and. kwdsPresentA(:,1:9) .and. kwdsPresentB(:,1:9)))&
  return

    if (any(fluxEeVLoA .ne. fluxEeVLoB&
    .and. kwdsPresentA .and. kwdsPresentB))&
  return

    if (any(fluxEeVHiA .ne. fluxEeVHiB&
    .and. kwdsPresentA .and. kwdsPresentB))&
  return

    tabA = table(setA, 'LOGNLOGS')
    tabB = table(setB, 'LOGNLOGS')

    ! Compare sim source template:

    call getLogNlogS(tabA, fd0A, csdA&
      , hiPowerA, psiA, modelEeVLoA, modelEeVHiA)

    call getLogNlogS(tabB, fd0B, csdB&
      , hiPowerB, psiB, modelEeVLoB, modelEeVHiB)

    if (size(fd0A) .ne. size(fd0B)) then
      deallocate(fd0A, fd0B, csdA, csdB)
  return
    end if

    if (any(fd0A .ne. fd0B) .or. any(csdA .ne. csdB)) then
      deallocate(fd0A, fd0B, csdA, csdB)
  return
    end if

    deallocate(fd0A, fd0B, csdA, csdB)

  if (hiPowerA    .ne. hiPowerB)    return
  if (psiA        .ne. psiB)        return
  if (modelEeVLoA .ne. modelEeVLoB) return
  if (modelEeVHiA .ne. modelEeVHiB) return

    coneRaA    = real64Attribute(tabA, 'CONE_RA')
    coneDecA   = real64Attribute(tabA, 'CONE_DEC')
    coneAngleA = real64Attribute(tabA, 'CONE_RAD')
    coneAreaA  = real64Attribute(tabA, 'SKY_AREA')
    coneRaB    = real64Attribute(tabB, 'CONE_RA')
    coneDecB   = real64Attribute(tabB, 'CONE_DEC')
    coneAngleB = real64Attribute(tabB, 'CONE_RAD')
    coneAreaB  = real64Attribute(tabB, 'SKY_AREA')

  if (coneRaA    .ne. coneRaB)    return
  if (coneDecA   .ne. coneDecB)   return
  if (coneAngleA .ne. coneAngleB) return
  if (coneAreaA  .ne. coneAreaB) return

    call release(setB)
    call release(setA)

  end function templatesUnequal

  !---------------------------------------------------------------------
  subroutine readDetSourceList(detSrcSetName, srcX, srcY&
  , srcR, srcXuncert, srcYuncert, srcRuncert, srcMlId, idBand&
  , tempSetName)
  !---------------------------------------------------------------------
  ! Designed to work with compareDetWithSim.
  !
  ! NOTE! You should deallocate all the returned pointers after using them.

    implicit none

    character(*), intent(in) :: detSrcSetName, tempSetName
    real(single), dimension(:), pointer ::&
      srcX, srcY, srcR,&
      srcXuncert, srcYuncert, srcRuncert
    integer(int32), pointer :: srcMlId(:)!, srcIdBand(:)
    integer(int8), intent(in) :: idBand

    type(DataSetT) :: sourceSet
    type(TableT)   :: sourceTab
    integer :: numSrcs
    real(single), dimension(:), pointer ::&
      x,    y,    fluxRand,&
      xUnc, yUnc, fluxRandUnc
    integer(int32), pointer :: index(:)!, idBand(:)
    character(2) :: bandStr

    !-------------------------------------------------------------------
    ! Start

!    sourceSet = dataSet(detSrcSetName, READ)
sourceSet = clone(detSrcSetName, tempSetName, TEMP)
sourceTab = table(sourceSet, 'SRCLIST')
bandStr = intToStr(int(idBand), leftJustify=.true._bool) ! in ssclib/ssc_misc
call initFilter('ID_BAND=='//trim(bandStr))
numSrcs = doFilter(sourceTab)
call release(sourceTab)

    sourceTab = table(sourceSet, 'SRCLIST')
    numSrcs = numberOfRows(sourceTab)

if (numSrcs .le. 0) call error('noDetSources',"eimsim_utils: No sources of &
&specified ID_BAND were found in the list of detected sources.")

    allocate(srcMlId(numSrcs))
!, srcIdBand(numSrcs))
    allocate(srcX(numSrcs), srcY(numSrcs), srcR(numSrcs))
    allocate(srcXuncert(numSrcs), srcYuncert(numSrcs), srcRuncert(numSrcs))

    index       => int32Data( column(sourceTab, 'ML_ID_SRC'))
!    idBand      => int32Data( column(sourceTab, 'ID_BAND'))
    x           => real32Data(column(sourceTab, 'X_IMA'))
    xUnc        => real32Data(column(sourceTab, 'X_IMA_ERR'))
    y           => real32Data(column(sourceTab, 'Y_IMA'))
    yUnc        => real32Data(column(sourceTab, 'Y_IMA_ERR'))
    fluxRand    => real32Data(column(sourceTab, 'FLUXRAND'))
    fluxRandUnc => real32Data(column(sourceTab, 'FLXRNERR'))

    srcMlId    = index
!    srcIdBand  = idBand
    srcX       = x
    srcXuncert = xUnc
    srcY       = y
    srcYuncert = yUnc
    srcR       = fluxRand
    srcRuncert = fluxRandUnc

    call release(sourceSet)

  end subroutine readDetSourceList

  !---------------------------------------------------------------------
  subroutine readSimSourceList(simSrcSetName, obsId, simSrcRaDeg, simSrcDecDeg&
  , simSrcFluxRand, simSrcFlag, simIndices, simFluxDensity0)
  !---------------------------------------------------------------------
  ! Designed to work with compareDetWithSim.
  !
  ! NOTE! You should deallocate all the returned pointers after using them.

    implicit none

    character(255), intent(in) :: simSrcSetName
    character(10),  intent(in) :: obsId
    real(double),   pointer    :: simSrcRaDeg(:), simSrcDecDeg(:)
    real(single),   pointer    :: simSrcFluxRand(:), simFluxDensity0(:)
    integer(int16), pointer    :: simSrcFlag(:)
    integer(int32), pointer    :: simIndices(:)

    type(DataSetT) :: sourceSet
    type(TableT)   :: sourceTab
    integer(int32) :: numSources
    real(single), pointer :: fluxRand(:), fluxDensity0(:)
    real(double), pointer :: raDeg(:), decDeg(:)
    integer(int16), pointer :: flag(:)
    integer(int32), pointer :: indices(:)

    !-------------------------------------------------------------------
    ! Start

    sourceSet = dataSet(simSrcSetName, READ)
    sourceTab = table(sourceSet, 'SRCLIST')
    numSources = numberOfRows(sourceTab)
if (numSources <= 0) call error('noSimSources',"eimsim_utils: No sources were &
&found in the list of simulated sources.")

    indices      => int32Data( column(sourceTab, 'INDEX'))
    fluxDensity0 => real32Data(column(sourceTab, 'S_DENS_0'))
    raDeg        => real64Data(column(sourceTab, 'RA'))
    decDeg       => real64Data(column(sourceTab, 'DEC'))
    fluxRand     => real32Data(column(sourceTab, 'FLUXRAND'))
    flag         => int16Data( column(sourceTab, 'FLAG_'//obsId))

    allocate(simSrcRaDeg(numSources), simSrcDecDeg(numSources)&
      , simSrcFluxRand(numSources), simFluxDensity0(numSources)&
      , simSrcFlag(numSources), simIndices(numSources))

    simIndices      = indices
    simFluxDensity0 = fluxDensity0
    simSrcRaDeg     = raDeg
    simSrcDecDeg    = decDeg
    simSrcFluxRand  = fluxRand
    simSrcFlag      = flag

    call release(sourceSet)

  end subroutine readSimSourceList

  !---------------------------------------------------------------------
  subroutine findNearestSimSource(srcX, srcY, srcL, srcXuncert, srcYuncert&
  , srcLuncert, simSrcX, simSrcY, simSrcL, simSrcFlag, simIndex, style&
  , bitToFlagDetection, indexOfNearestSim, devOverStdDev)
  !---------------------------------------------------------------------
  ! Style = 'ellipse':
  ! ------------------
  !
  ! Given the position (X and Y) and brightness (L) of a detected source,
  ! plus the uncertainties in these values, and given a list of the simulated
  ! sources, determine that simulated source which is 'closest', that is,
  ! which gives the smallest value of the quantity
  !
  !	      (simSrcX - srcX)^2     (simSrcY - srcY)^2     (simSrcL - srcL)^2
  !	r2 = -------------------- + -------------------- + --------------------
  !	         srcXuncert^2           srcYuncert^2           srcLuncert^2
  !
  ! Style = 'halfellipse':
  ! ------------------
  !
  ! For this style, the 'closest' simulated source is defined as that one
  ! which gives the smallest value of the quantity
  !
  !	      (simSrcX - srcX)^2     (simSrcY - srcY)^2     (simSrcL - srcL)^2
  !	r2 = -------------------- + -------------------- + --------------------
  !	         srcXuncert^2           srcYuncert^2           srcLuncert^2
  !
  ! (same as above) for simSrcL <= srcL, and
  !
  !	      (simSrcX - srcX)^2     (simSrcY - srcY)^2
  !	z2 = -------------------- + --------------------
  !	         srcXuncert^2           srcYuncert^2
  !
  ! for simSrcL > srcL.
  !.....................................................................

    implicit none

    real(single),   intent(in)    :: srcX,&
                                     srcY,&
                                     srcL,&
                                     srcXuncert,&
                                     srcYuncert,&
                                     srcLuncert,&
                                     simSrcX(:),&
                                     simSrcY(size(simSrcX)),&
                                     simSrcL(size(simSrcX))
    integer(int32), intent(in)    :: simIndex(size(simSrcX)),&
                                     bitToFlagDetection
    character(*),   intent(in)    :: style
    integer(int16), intent(inout) :: simSrcFlag(size(simSrcX))
    integer(int32), intent(out)   :: indexOfNearestSim
    real(single),   intent(out)   :: devOverStdDev

    integer(int32) :: numSrcs, iSpatiallyNearestBrighterSim
    real(double), dimension(size(simSrcL)) :: dX, dY, dL, r2, z2
    integer :: rLocVec(1), zLocVec(1)

    !-------------------------------------------------------------------
    ! Start

    indexOfNearestSim = 0 ! nonsensical default
    devOverStdDev = 0.0
    numSrcs = size(simSrcL)
    if (numSrcs .le. 0)&
call error('noSimSources',"eimsim_utils: No sources were found in the list &
&of simulated sources.")

    if (srcXuncert .le. 0.0&
    .or. srcYuncert .le. 0.0&
    .or. srcLuncert .le. 0.0)&
call error('negativeUncerts',"eimsim_utils: Non-positive values of position &
&or rate uncertainties were found in the list of detected sources.")

    if (numSrcs .eq. 1) then ! trivial
      indexOfNearestSim = simIndex(1)
      simSrcFlag(1) = ibset(simSrcFlag(1), bitToFlagDetection)
      call warning('onlyOneSimSource',"eimsim_utils: There is only one &
        &simulated source to match to.")
  return
    end if

    dX = simSrcX - srcX
    dY = simSrcY - srcY
    dL = simSrcL - srcL

    z2 = dX * dX / srcXuncert / srcXuncert&
    + dY * dY / srcYuncert / srcYuncert

    r2 = z2 + dL * dL / srcLuncert / srcLuncert

    select case(style)
    case('ellipse')
      rLocVec = minloc(r2)
      indexOfNearestSim = simIndex(rLocVec(1))
      simSrcFlag(rLocVec(1)) = ibset(simSrcFlag(rLocVec(1)), bitToFlagDetection)
      devOverStdDev = sqrt(r2(indexOfNearestSim))

    case('halfellipse')
      if (any(dL > 0.0)) then
        zLocVec = minloc(z2, (dL > 0.0)) ! closest source (in a spatial
        ! sense) which is brighter than the detection.

        iSpatiallyNearestBrighterSim = zLocVec(1)

        if (any(dL <= 0.0)) then
          rLocVec = minloc(r2, (dL <= 0.0)) ! closest source (in the absolute
          ! sense) which is fainter than the detection.

          indexOfNearestSim = simIndex(rLocVec(1))
          simSrcFlag(rLocVec(1)) = ibset(simSrcFlag(rLocVec(1))&
            , bitToFlagDetection)

          if (r2(indexOfNearestSim) < z2(iSpatiallyNearestBrighterSim))&
          then
            devOverStdDev = sqrt(r2(indexOfNearestSim))

          else
            indexOfNearestSim = simIndex(iSpatiallyNearestBrighterSim)
            simSrcFlag(iSpatiallyNearestBrighterSim)&
              = ibset(simSrcFlag(iSpatiallyNearestBrighterSim)&
              , bitToFlagDetection)

            devOverStdDev = sqrt(z2(indexOfNearestSim))
!*** flag the detection?
          end if

        else ! there are no sim sources fainter than detection.
          indexOfNearestSim = simIndex(iSpatiallyNearestBrighterSim)
          simSrcFlag(iSpatiallyNearestBrighterSim)&
            = ibset(simSrcFlag(iSpatiallyNearestBrighterSim)&
            , bitToFlagDetection)

          devOverStdDev = sqrt(z2(indexOfNearestSim))
!*** flag the detection?
        end if

      else ! there are no sim sources brighter than the detection.
        rLocVec = minloc(r2)
        indexOfNearestSim = simIndex(rLocVec(1))
         simSrcFlag(rLocVec(1)) = ibset(simSrcFlag(rLocVec(1))&
           , bitToFlagDetection)

        devOverStdDev = sqrt(r2(indexOfNearestSim))
      end if

    case default
call error('badStyle',"eimsim_utils: Source matching style not recognized.")
    end select

  end subroutine findNearestSimSource

  !---------------------------------------------------------------------
  subroutine probOfOneOrMoreByChance(srcL, srcXuncert, srcYuncert&
  , srcLuncert, devOverStdDev, srcPerPixel, probability)
  !---------------------------------------------------------------------
  ! This has been developed to work in conjunction with compareDetWithSim.
  !.....................................................................
  !
  ! Where sources are randomly distributed in 3 dimensions r_ = (X,Y,L)
  ! with a probability density q(r_), the probability P that 1 or more
  ! sources will lie within a given volume V is given by
  !
  !	P = 1 - exp(-N_V)						(1)
  !
  ! where N_V is the number of sources that lies (on average) inside V; or,
  ! in other words,
  !
  !	               V
  !	           (  /           )
  !	P = 1 - exp( -| dr_ q[r_] ).
  !	           (  /           )
  !
  ! In cases (such as the present) where q is a function of L only, N_V
  ! can be expressed as
  !
  !	     L_V_max
  !	       /
  !	N_V =  | dL A_V(L) q(L).
  !	       /
  !	    L_V_min
  !
  ! where L_V_min and L_V_max are respectively the smallest and largest
  ! values of L enclosed within V, and A_V(L) is the area of the cross-
  ! section through V normal to the L axis at the specified value of L.
  !
  ! In the present subroutine, V is assumed to be an ellipsoid centred on
  ! srcX, srcY and srcL, with axes dX, dY and dL obtained by multiplying
  ! devOverStdDev times respectively srcXuncert, srcYuncert and srcLuncert.
  ! The cross-section function A_V(L) is therefore given by
  !
  !	A_V(L) = Pi dX dY (1 - [L - srcL]^2 / dL^2).
  !
  ! In the present case q(L) is constant at
  !
  !	q(L) = srcPerPixel for 0.0 <= L < 1.0
  !
  !	     = 0 elsewhere.
  !
  !.....................................................................

    implicit none

    real(single), intent(in)  :: srcL,&
                                 srcLuncert,&
                                 devOverStdDev,&
                                 srcPerPixel,&
                                 srcXuncert,&
                                 srcYuncert
    real(single), intent(out) :: probability

    real(single) ::&
      deltaX, deltaY, deltaL
    real(single) :: numV

    !-------------------------------------------------------------------
    ! Start

    probability = 0.0 ! default

    if (srcXuncert .le. 0.0&
    .or. srcYuncert .le. 0.0&
    .or. srcLuncert .le. 0.0&
!    .or. devOverStdDev .le. 0.0&
    )then
call error('negativeUncerts',"eimsim_utils: Non-positive values of position &
&or rate uncertainties were found in the list of detected sources.")
    end if

    deltaX = srcXuncert * devOverStdDev
    deltaY = srcYuncert * devOverStdDev
    deltaL = srcLuncert * devOverStdDev

    numV = calcNumSrcsInVolume(srcL, deltaX, deltaY, deltaL, srcPerPixel)

    if (numV <= 0.0) then
      probability = 0.0
      call warning('negativeProbability', 'eimsim_utils: This ought not to &
        &happen. Best contact the code developer.')

    else if (numV .ge. -log(tiny(numV))) then
      probability = 1.0 ! saves the possibility of an underflow on the exp().

    else if (log(numV * numV / 2.0) .lt. (1 - precision(numV))) then
      probability = numV
!      call warning('probabilityUnderflow', 'No cause for concern. This &
!      &is just to inform you that the probability has been calculated &
!      &by use of the approximation prob = N rather than the exact &
!      &prob = 1 - exp(-N).')

    else
      probability = 1.0 - exp(-numV)
    end if

  end subroutine probOfOneOrMoreByChance

  !---------------------------------------------------------------------
  function calcNumSrcsInVolume(srcL, deltaX, deltaY, deltaL, srcPerPixel)&
  result(numV)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: srcL,&
                                deltaX, deltaY, deltaL, srcPerPixel
    real(single)             :: numV

    real(double) ::&
      L_V_max, L_V_min, deltaLsquared,&
      L_start, L_finis

    !-------------------------------------------------------------------
    ! Start

    numV = 0.0 ! default

    L_V_max = real(srcL, double) + real(deltaL, double)
    L_V_min = real(srcL, double) - real(deltaL, double)

    if (L_V_max .le. 0.0d0) then
      call warning('srcTooFaint',"eimsim_utils: The brightness of the &
        &detected source, plus its brightness uncertainty, was fainter &
        &than any of the simulated sources.")
  return
    end if

    if (L_V_min .ge. 1.0d0) then
call error('bug0',"eimsim_utils: If this occurs, please contact the code &
&developer.")
    end if

    deltaLsquared = deltaL * deltaL
    L_start = max(0.0d0, L_V_min)
    L_finis = min(1.0d0, L_V_max)
    numV = ellipsoidIntegral(L_start, L_finis, deltaLsquared&
          , 0.0d0, real(srcPerPixel, double), real(srcL, double))
    
    numV = numV * Pi * deltaX * deltaY

  end function calcNumSrcsInVolume

  !---------------------------------------------------------------------
  function ellipsoidIntegral(L_start, L_finis, deltaLsquared, power, amp&
  , srcL) result(term)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: L_start,&
                                L_finis,&
                                power,&
                                srcL,&
                                deltaLsquared,&
                                amp
    real(double)             :: term

    real(double) ::&
      intgrlTerm1, intgrlTerm2, intgrlTerm3,&
      pPlus1, pPlus2, pPlus3

    !-------------------------------------------------------------------
    ! Start

    pPlus1 = 1.0d0 + power
    pPlus2 = 2.0d0 + power
    pPlus3 = 3.0d0 + power

    if (pPlus1 == 0.0 .or. pPlus2 == 0.0 .or. pPlus3 == 0.0)&
call error('noIntegralPossible',"eimsim_utils: This integral cannot be done at &
&the value of power supplied.")

    intgrlTerm1 =  (L_finis**pPlus1 - L_start**pPlus1) / pPlus1&
      * (1.0d0 - srcL * srcL / deltaLsquared)

    intgrlTerm2 =  (L_finis**pPlus2 - L_start**pPlus2) / pPlus2&
      * 2.0d0 * srcL / deltaLsquared

    intgrlTerm3 = -(L_finis**pPlus3 - L_start**pPlus3) / pPlus3&
      / deltaLsquared

    term = amp * (intgrlTerm1 + intgrlTerm2 + intgrlTerm3)

  end function ellipsoidIntegral

  !---------------------------------------------------------------------
  subroutine saveFinalDetSrcList(detSrcSetName, simSrcX, simSrcY, simSrcR&
  , simSrcFlag, mlId, indexOfNearestSim&
  , nullProb, simSrcFluxDens0)!!!, alsoRan)!, devOverStdDev)
  !---------------------------------------------------------------------

    implicit none

    character(255), intent(in) :: detSrcSetName
    real(single),   intent(in) :: simSrcX(:),&
                                  simSrcY(size(simSrcX)),&
                                  simSrcR(size(simSrcX)),&
                                  simSrcFluxDens0(size(simSrcX))
    integer(int16), intent(in) :: simSrcFlag(size(simSrcX))
    integer(int32), intent(in) :: mlId(:),&
                                  indexOfNearestSim(size(mlId))
    real(single),   intent(in) :: nullProb(size(mlId))
!, devOverStdDev(size(mlId))
!!!    logical,        intent(in) :: alsoRan(size(mlId))

    type(DataSetT) :: sourceSet
    type(TableT)   :: sourceTab
    integer :: numRows, numDetSources, detSrc, simRow
    real(single), dimension(:), pointer ::&
      simX, simY, simR, diffX, diffY, nulP, xima, yima,&
      simFluxDens0
    integer(int16), pointer :: simFlag(:)
    integer(int32), pointer :: simIndex(:), mlIdData(:)

    !-------------------------------------------------------------------
    ! Start

    sourceSet = dataSet(detSrcSetName, MODIFY)
    sourceTab = table(sourceSet, 'SRCLIST')
    numRows = numberOfRows(sourceTab)
    numDetSources = size(mlId)

    mlIdData => int32Data( column(sourceTab, 'ML_ID_SRC'))
    xima     => real32Data(column(sourceTab, 'X_IMA'))
    yima     => real32Data(column(sourceTab, 'Y_IMA'))

    simX         => real32Data(addOrOpenColumn(sourceTab, 'SIM_XIMA',REAL32))
    simY         => real32Data(addOrOpenColumn(sourceTab, 'SIM_YIMA',REAL32))
    simR         => real32Data(addOrOpenColumn(sourceTab, 'SIM_RAND',REAL32))
    simFluxDens0 => real32Data(addOrOpenColumn(sourceTab, 'SIM_SDY0',REAL32&
      , 'cgs/eV', 'Sim source flux density at E = 1 eV.'))
    diffX        => real32Data(addOrOpenColumn(sourceTab, 'XIMADIFF',REAL32&
      , 'pixels', 'SIM_XIMA - X_IMA'))
    diffY        => real32Data(addOrOpenColumn(sourceTab, 'YIMADIFF',REAL32&
      , 'pixels', 'SIM_YIMA - Y_IMA'))

    nulP         => real32Data(addOrOpenColumn(sourceTab, 'NULLPROB',REAL32))
    simFlag      => int16Data( addOrOpenColumn(sourceTab, 'SIM_FLAG',INTEGER16))
    simIndex     => int32Data( addOrOpenColumn(sourceTab, 'SIM_INDX',INTEGER32))
    ! addOrOpenColumn is in ssclib/dal_aux.

simIndex = -1
!*** use this (temporarily?) to mark those rows where RATE was a null or RATE_ERR <= 0.

    do detSrc = 1, numDetSources
      simRow = indexOfNearestSim(detSrc)

      where(mlIdData == mlId(detSrc))
        simX         = simSrcX(simRow)
        simY         = simSrcY(simRow)
        simR         = simSrcR(simRow)
        simFluxDens0 = simSrcFluxDens0(simRow)
        diffX        = simSrcX(simRow) - xima
        diffY        = simSrcY(simRow) - yima
        simFlag      = simSrcFlag(simRow)
      
        simIndex = indexOfNearestSim(detSrc)
        nulP     = nullProb(detSrc)
      end where
    end do

    call release(sourceSet)

  end subroutine saveFinalDetSrcList

  !---------------------------------------------------------------------
  subroutine flagAlsoRans(indexOfNearestSim, devOverStdDev, alsoRan, status)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in)            :: indexOfNearestSim(:)
    real(single),   intent(in)            :: devOverStdDev(:)
    logical,        intent(out)           :: alsoRan(size(indexOfNearestSim))
    integer,        intent(out), optional :: status

    integer(int32) :: numSrcs, minIndex, maxIndex, index
    real(single)   :: minR
    logical(bool)  :: mask(size(indexOfNearestSim))
    integer        :: locVec(1)

    !-------------------------------------------------------------------
    ! Start

    status = 0

    alsoRan = .true. ! nonsensical default
    numSrcs = size(indexOfNearestSim)
    if (numSrcs .le. 0) then
      if (present(status)) status = -1
  return
    end if
    if (size(devOverStdDev) .ne. numSrcs) then
      if (present(status)) status = -2
  return
    end if

    minIndex = minval(indexOfNearestSim)
    maxIndex = maxval(indexOfNearestSim)

    if (minIndex .le. 0) then
      if (present(status)) status = -3
  return
    end if

    alsoRan = .false. ! sensible default
    do index = minIndex, maxIndex
      mask = (indexOfNearestSim .eq. index)
    if (count(mask) .le. 1) cycle
      minR = minval(devOverStdDev, mask)
      if (count(mask .and. devOverStdDev .eq. minR) .gt. 1) then ! there is
      ! more than 1 det source at the minimum 'distance' from the sim source
      ! of this index. Choose one of them to be the offical det source,
      ! setting alsoRan = .true. for all the others:
        where(mask) alsoRan = .true. ! covers everything.
        locVec = minloc(devOverStdDev, mask) ! just picks up 1 best location.
        alsoRan(locVec(1)) = .false.

      else ! there is only 1 det source at the minimum 'distance' from the
      ! sim source of this index. Leave alsoRan at .false. for this source,
      ! setting alsoRan = .true. for all the others at greater r:
        where(mask .and. devOverStdDev .gt. minR) alsoRan = .true.
      end if
    end do

  end subroutine flagAlsoRans

  !---------------------------------------------------------------------
  subroutine addSimRates(detSrcListSetName)
  !---------------------------------------------------------------------
  ! This subroutine should be called after the list of detected sources
  ! has been matched with simulation sources from the starting list.
  !
  ! The subroutine adds to the list of detections several quantities from
  ! the list of simulated sources, including calculating the count rates
  ! of the simulated sources in the various bands of interest.

    implicit none

    character(*), intent(in) :: detSrcListSetName

    real(single) :: multiplier(3, 0:9)
    real(double) ::&
      photonSpectralIndex,&
      fluxEeVLo(3, 0:9), fluxEeVHi(3, 0:9), ecf(3, 9)
    type(DataSetT) :: srcSet
    type(TableT) :: templateTab, srcTab
    real(single), dimension(:), pointer ::&
      simSrcFluxDensity0, simRate
    integer(int32) :: numRows, idInst, idBand, mlIdSrc, row
    real(single), allocatable ::&
      simSrcFlux(:)
    integer(int32), pointer :: idBandData(:), idInstData(:), mlIdSrcData(:)
    logical(bool), allocatable :: mask(:)
    logical(bool) :: kwdsPresent(3,0:9)

    !-------------------------------------------------------------------
    ! Start

    srcSet = dataSet(detSrcListSetName, MODIFY)
    templateTab = table(srcSet, 'LOGNLOGS')
    photonSpectralIndex = real64Attribute(templateTab, 'SPECINDX')
    call release(templateTab)
    srcTab = table(srcSet, 'SRCLIST')
    numRows = numberOfRows(srcTab)
    allocate(simSrcFlux(numRows), mask(numRows))

    simSrcFluxDensity0    => real32Data(column(srcTab, 'SIM_SDY0'))
    simRate    => real32Data(addOrOpenColumn(srcTab, 'SIM_RATE', REAL32&
      , 'counts/s'))
    mlIdSrcData => int32Data(column(srcTab, 'ML_ID_SRC'))
    idInstData  => int32Data(column(srcTab, 'ID_INST'))
    idBandData  => int32Data(column(srcTab, 'ID_BAND'))

    !...................................................................
    ! Load the band edges and ecfs and other inst/band dependent factors:

    call loadEcfStuff_20030912(srcSet, ecf, fluxEeVLo, fluxEeVHi, kwdsPresent)

    do idInst = 1, 3
      do idBand = 0, 5
      if (.not. kwdsPresent(idInst, idBand)) cycle

        multiplier(idInst, idBand)&
          = fluxToFluxDensity0(photonSpectralIndex&
          , fluxEeVLo(idInst, idBand), fluxEeVHi(idInst, idBand), 1.0)
      end do
    end do

    !...................................................................
    ! Calculate the flux values from fluxDensity0:

    do idInst = 1, 3
      do idBand = 0, 5 ! Note no special treatment now for band 0.
        mask = (idInstData == idInst .and. idBandData == idBand)
        if (any(mask) .and. .not. kwdsPresent(idInst, idBand))&
call error('noEcfKwdsThisInstBand',"eimsim_utils: I have no ECF for the &
&specified instrument and band!")

        where(mask)&
          simSrcFlux    = simSrcFluxDensity0    / multiplier(idInst, idBand)
      end do
    end do

    !...................................................................
    ! Calculate the rate values from flux:

    do idInst = 1, 3
      do idBand = 1, 5
        where(idInstData == idInst .and. idBandData == idBand)&
          simRate    = simSrcFlux    * 1.0e11 * ecf(idInst, idBand)
      end do
    end do

    ! Now fix up the band0 numbers:

    do mlIdSrc = minval(mlIdSrcData), maxval(mlIdSrcData)
    if (.not. any(mlIdSrcData == mlIdSrc)) cycle

      if (count(mlIdSrcData == mlIdSrc .and. idBandData == 0) .ne. 1)&
call error('noBand0',"eimsim_utils: There is no ID_BAND=0 row for this &
&value of ML_ID_SRC.")

      mask = (mlIdSrcData == mlIdSrc&
      .and. idBandData >= 1&
      .and. idBandData <= 5)

if (count(mask) .ne. 5) call error('wrongNumBands',"eimsim_utils: One (or more) &
&of the ID_BAND rows in the range 1 through 5 is missing for this value of &
&ML_ID_SRC.")

      row = smallestTrueIndex(logical(mlIdSrcData == mlIdSrc&
        .and. idBandData == 0, bool)) - 1

      simRate(row) = sum(simRate, mask)
    end do

    call release(srcSet)

  end subroutine addSimRates







































  !---------------------------------------------------------------------
  subroutine eventizeBlankImage(expMapSetName, imageName&
  , vigBkgFlux, nonVigBkgFlux, tempImageName)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: expMapSetName, imageName, tempImageName
    real(double), intent(in) :: vigBkgFlux, nonVigBkgFlux
    ! These values are background counts in counts / pixel (= rates / pixel multiplied by exposure time).

    real(single), pointer :: expMapImage(:,:)
    real(single), allocatable :: imageReal(:,:)
    integer(int32), allocatable :: image(:,:)
    type(DataSetT) :: set
    integer(int16) :: xSize, ySize

    !-------------------------------------------------------------------
    ! Start

!    call loadExposureMap(expMapSetName, expMapImage) ! sets cal->instrument

call getCalInfo(expMapSetName, instrumentId, expStartTimeStamp, scAttitude&
, instrumAttitude) ! also sets cal instrument state.

call readArrayData(expMapSetName, expMapImage)

    xSize = size(expMapImage, 1)
    ySize = size(expMapImage, 2)
    allocate(imageReal(xSize, ySize), image(xSize, ySize))
    imageReal = 0.0
    where(expMapImage > 0.0)&
      imageReal = nonVigBkgFlux + vigBkgFlux * (expMapImage&
      / maxval(expMapImage))

    call eventizeRateImage_slave(imageReal, image)
    call saveImage(image, expMapSetName, imageName, tempImageName)

print *, 'check on bkg flux: '&
, sum(image, (expMapImage > 0.0)) / real(count(expMapImage > 0.0))

    deallocate(expMapImage)

    set = dataSet(imageName, MODIFY)
    call setAttribute(set, 'VIGBFLUX', vigBkgFlux, 'counts/pixel'&
      , 'Vignetted bkg flux component at maximum exposure.')
    call setAttribute(set, 'NONVIGBF', nonVigBkgFlux, 'counts/pixel'&
      , 'Non-vignetted bkg flux component.')
    call release(set)

  end subroutine eventizeBlankImage

  !---------------------------------------------------------------------
  subroutine makeSourceListClumped(mask, maskWcs, logNlogSslope, minRate&
  , autoCorrPower, maxExtentWorld, srcData)
  !---------------------------------------------------------------------
  ! This makes a source list which has some spatial autocorrelation,
  ! the probability P(r) of another source being at radius r being given
  ! by
  !
  !	P(r) = k * r^autoCorrPower	r > r_min,
  !
  ! where k is a normalisation constant. The cutoff at r_min is necessary
  ! for convergence. autoCorrPower can of course be supplied to the
  ! subroutine directly; the same facility is not yet available for r_min.
  ! However r_min can be chosen approximately by varying the value of
  ! numDivsPerPixel. This integer specifies the fineness of the
  ! probability map compared to the exposure map. Within
  ! calcClumpProbMap() the autocorrelation function of the (1,1) pixel
  ! is set to zero, which reflects in effect a choice of r_min equal to
  ! the size of the clumpProbMap() pixels.
  !---------------------------------------------------------------------

    implicit none

    logical(bool), intent(in) :: mask(:,:)
    type(WcsT), intent(in) :: maskWcs
    real(single), intent(in) ::&
      logNlogSslope, autoCorrPower, minRate, maxExtentWorld
    type(localSrcDataType), intent(out) :: srcData(:)

    integer(int16) ::&
      minXi, maxXi, minYi, maxYi
    integer ::&
      numSources, srcNum, trialSrc, xSize, ySize,&
      maskXi, maskYi, clumpMapXsize, clumpMapYsize, clumpXi, clumpYi,&
      numDivsPerPixel = 3, status
    integer, parameter :: batchSize = 512
    real :: theta(batchSize), phi(batchSize), clumpTest(batchSize),&
      relativeFlux(size(srcData)), extent(size(srcData)),&
      coneHalfAngle
!, x, y
real(double) :: x, y, maxRsquared
    logical(bool) :: edgeImage(size(mask, 1), size(mask, 2))
    real(single) ::&
      avePixelSizeWorld, xPixels, yPixels
!, xPixelsToRadians, yPixelsToRadians
    real(single), allocatable :: clumpProbMap(:,:)

    !-------------------------------------------------------------------
    ! Start

if (count(mask) .le. 0) call error('emptyMask',"eimsim_utils: There are no &
&true pixels in the supplied mask.")

    xSize = size(mask, 1)
    ySize = size(mask, 2)

    call findEdges(mask, edgeImage)!, status=status)
!if (status .ne. 0) call error('findEdgesStatus','')

!    xPixelsToRadians = abs(maskWcs%x%pixelSizeWorld) * PiOn180
!    yPixelsToRadians = abs(maskWcs%y%pixelSizeWorld) * PiOn180
    maxRsquared = 0.0 ! default
    do maskXi = 1, xSize
    if (all(.not. edgeImage(maskXi, :))) cycle
!      x = (maskXi - maskWcs%x%refPixels) * xPixelsToRadians
      do maskYi = 1, ySize
      if (.not. edgeImage(maskXi, maskYi)) cycle
!        y = (maskYi - maskWcs%y%refPixels) * yPixelsToRadians
call pixelsToTan(real(maskXi, single), real(maskYi, single), maskWcs, x, y)
! in ssclib/wcs_aux
        maxRsquared = max(maxRsquared, x * x + y * y)
      end do
    end do
    coneHalfAngle = atan(sqrt(maxRsquared))

!    avePixelSizeWorld = (abs(maskWcs%X%pixelSizeWorld)&
!    + abs(maskWcs%Y%pixelSizeWorld)) / 2.0
avePixelSizeWorld = sqrt(wcsPixelAreaWorld(maskWcs))

    call maskArrayLimits(mask, minXi, maxXi, minYi, maxYi, status)
if (status .ne. 0) call error('maskArrayLimitsStatus',"eimsim_utils: &
&subroutine maskArrayLimits returned a nonzero error status.")

    clumpMapXsize = (1 + maxXi - minXi) * numDivsPerPixel
    clumpMapYsize = (1 + maxYi - minYi) * numDivsPerPixel
    allocate(clumpProbMap(clumpMapXsize,clumpMapYsize))

    call calcClumpProbMap(autoCorrPower, clumpProbMap)

    numSources = size(srcData)

!    call getMinus5on2Randoms(relativeFlux)
call getPowerRandoms(logNlogSslope, relativeFlux)
    call getRandomNumbers(extent)

    srcNum = 0
    BATCH: do
      call getSineRandoms(coneHalfAngle, theta, status)
if (status .ne. 0) then
print *, 'getSineRandoms status:', status
call error('getSineRandomsStatus','')
end if
      call getRandomNumbers(phi)
      phi = TwoPi * phi

      call getRandomNumbers(clumpTest)

      do trialSrc = 1, batchSize
        x = cos(phi(trialSrc)) * tan(theta(trialSrc))
        y = sin(phi(trialSrc)) * tan(theta(trialSrc))
!        xPixels = maskWcs%x%refPixels + x / xPixelsToRadians
!        yPixels = maskWcs%y%refPixels + y / yPixelsToRadians
call tanToPixels(wcs, x, y, xPixels, yPixels)
        maskXi = nint(xPixels)
        maskYi = nint(yPixels)
!print *, 'srcNum, trialSrc, phi(trialSrc), theta(trialSrc), x, y, xPixels, yPixels:'&
!, srcNum, trialSrc, phi(trialSrc), theta(trialSrc), x, y, xPixels, yPixels
      if (.not. mask(maskXi, maskYi)) cycle

        clumpXi = nint((xPixels - minXi + 0.5) * numDivsPerPixel + 0.5)
        clumpYi = nint((yPixels - minYi + 0.5) * numDivsPerPixel + 0.5)
      if (clumpProbMap(clumpXi, clumpYi) < clumpTest(trialSrc)) cycle

        srcNum = srcNum + 1
    if (srcNum .gt. numSources) exit BATCH

        srcData(srcNum)%xPixels = xPixels 
        srcData(srcNum)%yPixels = yPixels
        srcData(srcNum)%rate = relativeFlux(srcNum) * minRate
        srcData(srcNum)%extentWorld = extent(srcNum) * maxExtentWorld
        srcData(srcNum)%extentPixels = srcData(srcNum)%extentWorld&
          / avePixelSizeWorld
      end do ! trialSrc
    end do BATCH

!call test_dumpColumn(srcData%rate / minRate)

  end subroutine makeSourceListClumped

  !---------------------------------------------------------------------
  subroutine test_dumpColumn(vector)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: vector(:)

    type(DataSetT) :: set
    type(TableT) :: tab
    type(ColumnT) :: col
    real(single), pointer :: myData(:)
    integer(int32) :: numRows

    !-------------------------------------------------------------------
    ! Start

    numRows = size(vector)
    set = dataSet('junk.ds', CREATE)
    tab = addTable(set, 'SRCLIST', numRows)
    col = addColumn(tab, 'RATE', REAL32)
    myData => real32Data(col)
    myData = vector
    call release(set)

  end subroutine test_dumpColumn

  !---------------------------------------------------------------------
  subroutine calcClumpProbMap(autoCorrPower, clumpProbMap)
  !---------------------------------------------------------------------
  ! The purpose of this subroutine is to create a probability map for the
  ! occurrence of a source in the field. The source positions are assumed
  ! to have some degree of autocorrelation or 'clumping', this probability
  ! P(r) given by a power-law rule
  !
  !	P(r) = k * r^autoCorrPower	r > r_min
  !
  ! where r is the separation between sources and k is a normalisation
  ! constant. Note that autoCorrPower must be more negative than -1
  ! and also that there must be some low-end cutoff in r. In the present
  ! subroutine r_min is set approximately equal to 1 pixel width, and
  ! is thus controllable by the user via the fineness of the pixel
  ! grid chosen for clumpProbMap(). This is described in more detail
  ! below.
  !
  ! The probability map is obtained via a Fourier method. The power-law
  ! autocorrelation function is first loaded into the complex array
  ! mapFFT(). This will eventually (after Fourier manipulations) hold
  ! the prototype clumpProbMap(), however it is made a little bigger
  ! than clumpProbMap() in order to force its dimensions to be multiples
  ! of 2, 3, 5 or 7, as is necessary for most efficient FFT.
  !
  ! The autocorrelation function is first loaded into the real values of
  ! the low-x, low-y corner of mapFFT(), then mirrored into the other
  ! corners. This is standard stuff, however to recap briefly, consider
  ! a slice at y=0; let us consider 2 examples, 1 with an odd x-dimension
  ! and the other with an even x-dimension:
  !
  ! Odd:
  !
  !	0.0000	1.0000	0.3299	0.1724	0.1088	0.1088	0.1724	0.3299	1.0000
  !	   1	   2	   3	   4	   5	   6	   7	   8	   9
  !
  ! Even:
  !
  !	0.0000	1.0000	0.3299	0.1724	0.1088	0.1724	0.3299	1.0000
  !	   1	   2	   3	   4	   5	   6	   7	   8
  !
  ! In the first case the corner with x indices from 1 to 5 should be first
  ! filled, then mirrored into the rest; in the second case the corner to
  ! be first filled again runs from x = 1 to 5.
  !
  ! Note that the central value (at x = 1) is set to zero. This reflects
  ! (in an approximate fashion) the low-r cutoff mentioned earlier.
  !
  ! After the real part of mapFFT() is filled with the autocorrelation
  ! function (the imaginary part is set to zero), is is Fourier transformed,
  ! and the complex square root taken. Random phases are then applied to
  ! these pixels. Finally, this array is back-transformed, and the real
  ! values are taken for clumpProbMap().
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: autoCorrPower
    real(single), intent(out) :: clumpProbMap(:,:)

    complex(single), allocatable :: mapFFT(:,:)
    real(single) :: autoCorrPowerOn2, x, y, autoCorrValue
    integer :: paddedXsize, paddedYsize, cornerLimitX, cornerLimitY, xi, yi
    real(single), allocatable :: phase(:)
    type(FftwPlanType) :: fftForwardPlanNum, fftReversePlanNum
    ! type def is in fftw_aux.

    !-------------------------------------------------------------------
    ! Start

if (autoCorrPower >= 0.0) call error('powerTooBig','')
autoCorrPowerOn2 = autoCorrPower / 2.0

    paddedXsize = findNextHighest2357multiple(size(clumpProbMap, 1))
    paddedYsize = findNextHighest2357multiple(size(clumpProbMap, 2))
    ! this function is in fftw_aux.

    allocate(mapFFT(paddedXsize, paddedYsize))
    mapFFT = (0.0, 0.0) ! array op, default

    ! Now want to fill in the autocorrelation function:

    cornerLimitX = 1 + paddedXsize / 2 ! this should produce the desired
    ! result whether paddedXsize is odd or even. Eg paddedXsize = 16
    ! produces cornerLimitX = 9; paddedXsize = 35 produces
    ! cornerLimitX = 18; both correct.

    cornerLimitY = 1 + paddedYsize / 2

    xi = 1
    do yi = 2, cornerLimitY
      y = yi - 1.0
      autoCorrValue = y ** autoCorrPower
!if (y < 5) then
!  autoCorrValue = 1.0
!else
!  autoCorrValue = 0.0
!end if
      mapFFT(xi, yi) = cmplx(autoCorrValue, 0.0, single)
      mapFFT(xi, paddedYsize + 2 - yi) = mapFFT(xi, yi)
    end do

    do xi = 2, cornerLimitX
      x = xi - 1.0
      yi = 1
      autoCorrValue = x ** autoCorrPower
!if (x < 5) then
!  autoCorrValue = 1.0
!else
!  autoCorrValue = 0.0
!end if
      mapFFT(xi, yi) = cmplx(autoCorrValue, 0.0, single)
      mapFFT(paddedXsize + 2 - xi, yi) = mapFFT(xi, yi)

      do yi = 2, cornerLimitY
        y = yi - 1.0
        autoCorrValue = (x * x + y * y) ** autoCorrPowerOn2
!if ((x * x + y * y) < 25) then
!  autoCorrValue = 1.0
!else
!  autoCorrValue = 0.0
!end if
        mapFFT(xi, yi) = cmplx(autoCorrValue, 0.0, single)
        mapFFT(xi, paddedYsize + 2 - yi) = mapFFT(xi, yi)
        mapFFT(paddedXsize + 2 - xi, yi) = mapFFT(xi, yi)
        mapFFT(paddedXsize + 2 - xi, paddedYsize + 2 - yi) = mapFFT(xi, yi)
      end do
    end do

    ! Now FFT:

!allocate(phase(paddedXsize))
!do yi = 1, paddedYsize
!  call getRandomNumbers(phase)
!  mapFFT(:,yi) = cmplx(phase, 0.0, single)
!end do
!deallocate(phase)

    call fftw2d_f77_create_plan(fftForwardPlanNum, paddedXsize&
      , paddedYsize, FFTW_FORWARD, FFTW_ESTIMATE + FFTW_IN_PLACE)

    call fftwnd_f77_one(fftForwardPlanNum, mapFFT, 0)

!call dumpImageToFits(real(mapFFT), 'auto_corr_real.ds')
!call dumpImageToFits(aimag(mapFFT), 'auto_corr_imag.ds')
!call error('','')

    ! Take the square root, and add random phase:

    mapFFT = sqrt(mapFFT)

    allocate(phase(paddedXsize))
    do yi = 1, paddedYsize
      call getRandomNumbers(phase)
      phase = phase * TwoPi
      mapFFT(:, yi) = mapFFT(:, yi) * cmplx(cos(phase), sin(phase), single)
    end do

!    allocate(phase(cornerLimitY))
!    xi = 1
!    call getRandomNumbers(phase)
!    phase = phase * TwoPi
!    do yi = 2, cornerLimitY
!      mapFFT(xi, yi) = mapFFT(xi, yi) * cmplx(cos(phase(yi)), sin(phase(yi))&
!        , single)
!      mapFFT(xi, paddedYsize + 2 - yi) = conjg(mapFFT(xi, yi))
!    end do
!
!    do xi = 2, cornerLimitX
!      call getRandomNumbers(phase)
!      phase = phase * TwoPi
!      yi = 1
!      mapFFT(xi, yi) = mapFFT(xi, yi) * cmplx(cos(phase(yi)), sin(phase(yi))&
!        , single)
!      mapFFT(paddedXsize + 2 - xi, yi) = conjg(mapFFT(xi, yi))
!
!      do yi = 2, cornerLimitY
!        mapFFT(xi, yi) = mapFFT(xi, yi) * cmplx(cos(phase(yi)), sin(phase(yi))&
!          , single)
!        mapFFT(xi, paddedYsize + 2 - yi) = conjg(mapFFT(xi, yi))
!        mapFFT(paddedXsize + 2 - xi, yi) = conjg(mapFFT(xi, yi))
!        mapFFT(paddedXsize + 2 - xi, paddedYsize + 2 - yi) = mapFFT(xi, yi)
!        ! In the opposite corner, the two conjugations cancel.
!      end do
!    end do
!
!    ! Note that, if a dimension is even, the cornerLimit pixel must be
!    ! strictly real-valued - otherwise the required antisymmetry could
!    ! not hold. This is now done:
!
!    if (.not. isOdd(paddedXsize))&
!      mapFFT(cornerLimitX, :) = cmplx(abs(mapFFT(cornerLimitX, :))&
!      , 0.0 , single)
!
!    if (.not. isOdd(paddedYsize))&
!      mapFFT(:, cornerLimitY) = cmplx(abs(mapFFT(:, cornerLimitY))&
!      , 0.0 , single)
!
!    ! The function isOdd() is in math_utils (should be in ssc_misc).

    ! Now, back-transform, and take the absolute value. The result should be
    ! a probability
    ! map whose autocorrelation function is rotationally symmetrical
    ! and described by the original power-law + cutoff radial function
    ! P(r) = k * r^autoCorrPower.

    call fftwnd_f77_destroy_plan(fftForwardPlanNum)
    call fftw2d_f77_create_plan(fftReversePlanNum, paddedXsize&
    , paddedYsize, FFTW_BACKWARD, FFTW_ESTIMATE + FFTW_IN_PLACE)

    call fftwnd_f77_one(fftReversePlanNum, mapFFT, 0)
    call fftwnd_f77_destroy_plan(fftReversePlanNum)

    clumpProbMap = abs(mapFFT(:size(clumpProbMap, 1)&
      , :size(clumpProbMap, 2)))

    clumpProbMap = clumpProbMap / maxval(clumpProbMap)

!call dumpImageToFits(clumpProbMap, 'prob_map.ds')
!call error('','')

  end subroutine calcClumpProbMap

  !---------------------------------------------------------------------
  subroutine makeUncorrSrcListSet(srcListSetName, numSources, maskSetName&
  , antiCorrLength, avoidEdges)
  !---------------------------------------------------------------------
  ! Makes a table of source RAs and decs. The source positions are evenly distributed in the plane of the mask image and are restricted to the true-valued mask pixels. In addition, the positions are anti-correlated and also optionally avoid the mask edges.

    implicit none

    character(*),   intent(in) :: srcListSetName, maskSetName
    integer(int32), intent(in) :: numSources
    real(single),   intent(in) :: antiCorrLength
    logical,        intent(in) :: avoidEdges

    type(DataSetT) :: set
    type(TableT) :: tab
!    real(single) :: srcX(numSources), srcY(numSources)
real(single), pointer :: srcX(:), srcY(:), rate(:)
!    real(double) :: tanX(numSources), tanY(numSources), srcRaDeg(numSources), srcDecDeg(numSources)
real(double) :: tanX(numSources), tanY(numSources), srcRa(numSources), srcDec(numSources)
real(double), pointer :: srcRaDeg(:), srcDecDeg(:)
    logical(bool), pointer :: mask(:,:)
    type(WcsT) :: wcs

    !-------------------------------------------------------------------
    ! Start

    call readArrayData(maskSetName, mask) ! in fits_utils

    set = dataSet(srcListSetName, CREATE)
    tab = addTable(set, 'SRCLIST', numSources)
    srcX      => real32Data(addColumn(tab, 'X_IMA', REAL32))
    srcY      => real32Data(addColumn(tab, 'Y_IMA', REAL32))
    rate      => real32Data(addColumn(tab, 'RATE',  REAL32))
rate = 1.0
    srcRaDeg  => real64Data(addColumn(tab, 'RA',    REAL64))
    srcDecDeg => real64Data(addColumn(tab, 'DEC',   REAL64))

    call makeUncorrPositions(mask, antiCorrLength, avoidEdges, srcX, srcY)
!&
!, maskSetName)
    call readWcs(maskSetName, wcs)
!    tanX = (srcX - wcs%x%refPixels) * abs(wcs%x%pixelSizeWorld) * PiOn180
!    tanY = (srcY - wcs%y%refPixels) * abs(wcs%y%pixelSizeWorld) * PiOn180
call pixelsToTan(srcX, srcY, wcs, tanX, tanY) ! in ssclib/wcs_aux
!    call tanToRaDec(wcs%x%refWorld, wcs%y%refWorld&
!      , tanX, tanY, srcRaDeg, srcDecDeg) ! in ssclib/coordinate.
call tanToRaDec(tanX, tanY, wcs%x%refWorld*PiOn180, wcs%y%refWorld*PiOn180&
 , srcRa, srcDec) ! in ssclib/coordinate.

srcRaDeg  = srcRa  / PiOn180
srcDecDeg = srcDec / PiOn180

    call release(set)

  end subroutine makeUncorrSrcListSet

  !---------------------------------------------------------------------
  subroutine makeUncorrPositions(mask, antiCorrLength, avoidEdges, srcX, srcY)
!&
!, oldImageSetName)
  !---------------------------------------------------------------------
  ! This subroutine returns a list of positions. These are random and evenly distributed over all pixels which are TRUE in the input mask image, except firstly that, if avoidEdges is set TRUE, the probability function decreases smoothly to zero at the mask edges, and secondly that each added source modifies the probability function further such that subsequent sources avoid the vicinity of present ones.
  !
  ! Each modification of the probability function P(x,y) is carried out in a similar fashion. The previous sources are avoided by multiplying the previous P(x,y) by 1 - exp(-(dx_s^2 + dy_s^2) / 2 / ACL^2), where dx_s, dy_s are the displacements from the source position and ACL is antiCorrLength. The edge is avoided (if avoidEdges is set) by the fairly simplistic procedure of, at the start of the algorithm, treating all the edge pixels as sources and multiplying the initially flat P by the appropriate functions.
  !
  ! NOTE! I'm not bothering here with getting the sources evenly distributed on the celestial sphere. They're even on the tangent plane, that's good enough.

    implicit none

    logical(bool), intent(in) :: mask(:,:)
    logical, intent(in) :: avoidEdges
    real, intent(in) :: antiCorrLength
    real, intent(out) :: srcX(:), srcY(size(srcX))
!character(*), intent(in) :: oldImageSetName
! Just so as to be able to dump out the prob map with the right wcs kwds

    integer :: xSize, ySize, numSources, xi, yi, srcNum, trialSrc
!, foreignStatus
    real :: prob(size(mask, 1), size(mask, 2))
    logical(bool) :: edgeImage(size(mask, 1), size(mask, 2))
    integer, parameter :: batchSize = 512
    real :: x(batchSize), y(batchSize), z(batchSize), maxProb

    !-------------------------------------------------------------------
    ! Start

!print *, 'lbound(srcX), ubound(srcX): ', lbound(srcX), ubound(srcX)

if (count(mask) .le. 0) call error('emptyMask','')

    xSize = size(mask, 1)
    ySize = size(mask, 2)

    numSources = size(srcX)
  if (numSources <= 0) return

    prob = 0.0
    where(mask) prob = 1.0

!call saveOutImage(mask, oldImageSetName, 'mask.ds')
    if (avoidEdges) then
      call findEdges(mask, edgeImage)!, status=foreignStatus)
!if (foreignStatus .ne. 0) call error('findEdgesStatus','')

!call saveOutImage(edgeImage, oldImageSetName, 'edge.ds')

      ! Depresses prob() at the edges by making a gaussian-section dimple
      ! at each edge pixel. (Yes it's a slow way to do things - but simple!)
      do xi = 1, xSize
        do yi = 1, ySize
          if (edgeImage(xi, yi)) then
!            call depressProb(prob, real(xi), real(yi), antiCorrLength)
!            call takeBiteOutOfProb(prob, real(xi), real(yi), antiCorrLength)
            call cutWellInProb(prob, real(xi), real(yi), antiCorrLength)
          end if
        end do
      end do
    end if

!call saveOutImage(prob, oldImageSetName, 'pppp.ds')

!print *, 'nint(0.49999): ', nint(0.49999)
!print *, 'nint(0.5): ', nint(0.5)
!print *, 'nint(0.5-epsilon(1.0)): ', nint(0.5-epsilon(1.0))

maxProb = maxVal(prob)
    srcNum = 0
    BATCH: do
      call getRandomNumbers(x)
      x = fracVal(0.5, real(xSize) + 0.5, x)
      call getRandomNumbers(y)
      y = fracVal(0.5, real(ySize) + 0.5 - epsilon(1.0), y)
      call getRandomNumbers(z)

      do trialSrc = 1, batchSize
        xi = max(1, min(xSize, nint(x(trialSrc))))
        yi = max(1, min(ySize, nint(y(trialSrc))))
!xi = nint(x(trialSrc))
!yi = nint(y(trialSrc))
!print *, 'xi/xSize, yi/ySize: ', xi,'/',xSize, yi,'/',ySize

!        if (z(trialSrc) < prob(xi, yi)) then
if (z(trialSrc) < prob(xi, yi) / maxProb) then
        ! success!
          srcNum = srcNum + 1
    if (srcNum .gt. numSources) exit BATCH
!print *, srcNum, '/', numSources

          srcX(srcNum) = x(trialSrc)
          srcY(srcNum) = y(trialSrc)

!          call depressProb(prob, x(trialSrc), y(trialSrc), antiCorrLength)
!          call takeBiteOutOfProb(prob, x(trialSrc), y(trialSrc), antiCorrLength)
          call cutWellInProb(prob, x(trialSrc), y(trialSrc), antiCorrLength)
maxProb = maxVal(prob)
        end if
      end do
    end do BATCH

!call dumpImageToFits(prob, 'prob.ds')
!call saveOutImage(prob, oldImageSetName, 'prob.ds')

  end subroutine makeUncorrPositions

  !---------------------------------------------------------------------
  subroutine depressProb(prob, x, y, antiCorrLength)
  !---------------------------------------------------------------------

    implicit none

    real, intent(inout) :: prob(:,:)
    real, intent(in)    :: x, y, antiCorrLength

    real, parameter :: numSigmas = 6.0
    real :: sigmaSquaredTimes2, dx, dy, rSquared
    real, allocatable :: mat(:,:)
    integer :: xSize, ySize, startXi, finisXi, startYi, finisYi, xi, yi

    !-------------------------------------------------------------------
    ! Start

    sigmaSquaredTimes2 = 2.0 * antiCorrLength * antiCorrLength

    xSize = size(prob, 1)
    ySize = size(prob, 2)
!    xi = nint(x)
!    yi = nint(yi)
    startXi = max(1,     floor(  x - numSigmas * antiCorrLength))
    finisXi = min(xSize, ceiling(x + numSigmas * antiCorrLength))
    startYi = max(1,     floor(  y - numSigmas * antiCorrLength))
    finisYi = min(ySize, ceiling(y + numSigmas * antiCorrLength))
    allocate(mat(startXi:finisXi,startYi:finisYi))
    do xi = startXi, finisXi
      dx = real(xi) - x
      do yi = startYi, finisYi
        dy = real(yi) - y
        rSquared = dx * dx + dy * dy
        mat(xi, yi) = 1.0 - exp(-rSquared / sigmaSquaredTimes2)
      end do
    end do

    prob(startXi:finisXi,startYi:finisYi)&
      = prob(startXi:finisXi,startYi:finisYi) * mat

  end subroutine depressProb

  !---------------------------------------------------------------------
  subroutine takeBiteOutOfProb(prob, x, y, antiCorrLength)
  !---------------------------------------------------------------------

    implicit none

    real, intent(inout) :: prob(:,:)
    real, intent(in)    :: x, y, antiCorrLength

    real, parameter :: numSigmas = 6.0
    real :: sigmaSquaredTimes2, dx, dy, rSquared
    real, allocatable :: mat(:,:)
    integer :: xSize, ySize, startXi, finisXi, startYi, finisYi, xi, yi

    !-------------------------------------------------------------------
    ! Start

    sigmaSquaredTimes2 = 2.0 * antiCorrLength * antiCorrLength

    xSize = size(prob, 1)
    ySize = size(prob, 2)
!    xi = nint(x)
!    yi = nint(yi)
    startXi = max(1,     floor(  x - numSigmas * antiCorrLength))
    finisXi = min(xSize, ceiling(x + numSigmas * antiCorrLength))
    startYi = max(1,     floor(  y - numSigmas * antiCorrLength))
    finisYi = min(ySize, ceiling(y + numSigmas * antiCorrLength))
    allocate(mat(startXi:finisXi,startYi:finisYi))
    do xi = startXi, finisXi
      dx = real(xi) - x
      do yi = startYi, finisYi
        dy = real(yi) - y
        rSquared = dx * dx + dy * dy
        mat(xi, yi) = 1.0 - exp(-rSquared / sigmaSquaredTimes2)
      end do
    end do

    where(prob(startXi:finisXi,startYi:finisYi) > mat)&
      prob(startXi:finisXi,startYi:finisYi) = mat

  end subroutine takeBiteOutOfProb

  !---------------------------------------------------------------------
  subroutine cutWellInProb(prob, x, y, antiCorrLength)
  !---------------------------------------------------------------------

    implicit none

    real, intent(inout) :: prob(:,:)
    real, intent(in)    :: x, y, antiCorrLength

    real, parameter :: numSigmas = 6.0
    real :: sigmaSquaredTimes2, dx, dy, rSquared
    real, allocatable :: mat(:,:)
    integer :: xSize, ySize, startXi, finisXi, startYi, finisYi, xi, yi

    !-------------------------------------------------------------------
    ! Start

!print *, 'lbound(prob, 1), ubound(prob, 1): ', lbound(prob, 1), ubound(prob, 1)
!print *, 'lbound(prob, 2), ubound(prob, 2): ', lbound(prob, 2), ubound(prob, 2)

    sigmaSquaredTimes2 = 2.0 * antiCorrLength * antiCorrLength

    xSize = size(prob, 1)
    ySize = size(prob, 2)
!    xi = nint(x)
!    yi = nint(yi)
    startXi = max(1,     floor(  x - numSigmas * antiCorrLength))
    finisXi = min(xSize, ceiling(x + numSigmas * antiCorrLength))
    startYi = max(1,     floor(  y - numSigmas * antiCorrLength))
    finisYi = min(ySize, ceiling(y + numSigmas * antiCorrLength))
!print *, 'startXi, finisXi: ', startXi, finisXi
!print *, 'startYi, finisYi: ', startYi, finisYi
    allocate(mat(startXi:finisXi,startYi:finisYi))
    mat = 1.0
    do xi = startXi, finisXi
      dx = real(xi) - x
      do yi = startYi, finisYi
        dy = real(yi) - y
        rSquared = dx * dx + dy * dy
        if (rSquared <= sigmaSquaredTimes2) mat(xi, yi) = 0.0
      end do
    end do

    where(prob(startXi:finisXi,startYi:finisYi) > mat)&
      prob(startXi:finisXi,startYi:finisYi) = mat

  end subroutine cutWellInProb

  !---------------------------------------------------------------------
  subroutine recalcFluxRandoms(simSrcListSetName, logNlogSSetName)
  !---------------------------------------------------------------------
  ! This subroutine reads in S_DENS_0 values (in counts/sec) from a (simulated) source list + information about the logN-logS and photon index and converts the S_DENS_0 values to FLUXRAND. The latter is defined as follows. The flux values of the simulated source list are obtained via a process that takes a set of random numbers having an even distribution between 0 and 1 and subjects them to a transform so that they become flux values which obey the specified logN-logS distribution. These original random numbers are recalculated here and named the FLUXRAND values.

    implicit none

    character(*), intent(in) :: simSrcListSetName, logNlogSSetName

    real(single) ::&
!      photonSpectralIndex,&
      hiPower
    real(single), pointer ::&
      modelFluxDensity0(:),&
      modelCumSkyDensity(:)
!    real(double) ::&
!photonSpectralIndex,&
!      EeVLo, EeVHi
    type(DataSetT) :: set, templateSet
    type(TableT) :: tab, templateTab
    type(ColumnT) :: ranCol
    real(single), pointer :: ran(:), srcFluxDensity0(:)
    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start

    templateSet = dataSet(logNlogSSetName, READ)
    templateTab = table(templateSet, 'LOGNLOGS')
    call getLogNlogS(templateTab, modelFluxDensity0, modelCumSkyDensity&
      , hiPower)
    call release(templateSet)

!    simSrcListSetName = stringParameter('simsrclistset')
    set = dataSet(simSrcListSetName, MODIFY)
    tab = table(set, 'SRCLIST')
    if (hasColumn(tab, 'FLUXRAND')) then
      ranCol = column(tab, 'FLUXRAND')
    else
      ranCol = addColumn(tab, 'FLUXRAND', REAL32&
        , comment='Should be evenly distributed between 0 and 1.')
    end if
    ran => real32Data(ranCol)
    srcFluxDensity0 => real32Data(column(tab, 'S_DENS_0'))

    call invertLogNlogSrandoms(modelFluxDensity0, modelCumSkyDensity&
      , srcFluxDensity0, hiPower, ran, foreignStatus)

    if (foreignStatus < 0) then
print *, 'ssclib: Status return from invertLogNlogSrandoms was', foreignStatus
call error('invertLogNlogSrandoms','')
    end if

    call release(set)
    deallocate(modelFluxDensity0, modelCumSkyDensity)

  end subroutine recalcFluxRandoms













! Possibly obsolete:

  !---------------------------------------------------------------------
  subroutine loadExposureMap(expMapSetName, expMapImage)
  !---------------------------------------------------------------------
  ! Adapted from asmooth-2.5
    
    implicit none

    character(*), intent(in) :: expMapSetName
    real(single), pointer ::&
      expMapImage(:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(expMapSetName, READ)

    call getCalInfo(set, instrumentId, expStartTimeStamp, scAttitude&
    , instrumAttitude) ! also sets cal instrument state.

    inArray = array(set, 0, READ)
    call readWcs(set, inArray, wcs)
!    call loadImage(inArray, expMapImage)
call readArrayData(inArray, expMapImage)

    call release(set)

  end subroutine loadExposureMap

  !---------------------------------------------------------------------
  subroutine loadImage(inArray, image)
  !---------------------------------------------------------------------
  ! Adapted from asmooth-2.5
  !
  ! Note that (up til 23 Apr 2002 anyway) the DAL call boolArray2Data(array)
  ! is not yet implemented.

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(single), dimension(:,:), pointer :: image

    integer(int32), dimension(:), pointer ::&
      dimVector
    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:), pointer ::&
      arrayDataReal64
!    logical(bool), dimension(:,:), pointer ::&
!      arrayDataBoolean
    integer(int16) :: xSize, ySize

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)
    xSize = dimVector(0)
    ySize = dimVector(1)

    allocate(image(xSize, ySize))

    if (dataType(inArray) .eq. INTEGER8) then
      arrayDataInt8 => int8Array2Data(inArray)
      image = real(arrayDataInt8, single)
    else if (dataType(inArray) .eq. INTEGER16) then
      arrayDataInt16 => int16Array2Data(inArray)
      image = real(arrayDataInt16, single)
    else if (dataType(inArray) .eq. INTEGER32) then
      arrayDataInt32 => int32Array2Data(inArray)
      image = real(arrayDataInt32, single)
    else if (dataType(inArray) .eq. REAL32) then
      arrayDataReal32 => real32Array2Data(inArray)
      image = arrayDataReal32
!    else if (dataType(inArray) .eq. BOOLEAN) then
!      arrayDataBoolean => boolArray2Data(inArray)
!      image = 0.0
!      where(arrayDataBoolean) image = 1.0
    else ! this gives a DAL error if data type is none of the above.
      arrayDataReal64 => real64Array2Data(inArray)
      image = real(arrayDataReal64, single)
    end if

  end subroutine loadImage

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module eimsim_utils
