#!/usr/local/bin/perl -w

# 

my ($testflag);

sub eimsimdetect {

  use SAS;

  my $taskname = "eimsimdetect";

  my ($i, $band, $j, $evlist, $imageset, $status, $mode, $command);
  my ($pix_delta_x, $pix_delta_y, $n_mask_pixels, $sky_area);
  my (%expmaps, %expmaps, %expmaps, %idbandlo, %idbandhi, %ecf, %sim_images, %splinemaps);

  # Prints all output to stderr:
  select(STDERR);

  #......................................................................
  # Read and check parameters:

  my $first_stage = stringParameter('entrystage');
  my $last_stage  = stringParameter('finalstage');
  my $prod_dir    = stringParameter('prdsdir');
  my $simop_dir   = stringParameter('simopdir');
  my $instexpid   = stringParameter('instexpid');
  my $obsid       = stringParameter('obsid');
  my $stream_num  = intParameter('streamnumber');
  my $id_number   = intParameter('idnumber');

  my $num_entries = parameterCount('bandlist');
SAS::error('noBands', "There must be at least one member in the "
."list --bandlist")
  if ($num_entries <= 0);

  my @used_bands = ();
  foreach $i (0 .. $num_entries-1) {
    $used_bands[$i] = intParameter('bandlist', $i);
  }

  my $likemin = realParameter('likemin');
  my $mlmin   = realParameter('mlmin');
  $testflag   = booleanParameter('astest');

  my $inst_uc = substr($instexpid, 0, 2);

  foreach $i (0 .. $#used_bands) {
    $band = $used_bands[$i];
SAS::error('badBandNumber', "Band must be between 1 and 9!")
    if ($band < 1 || $band > 8);

    foreach $j (0 .. $#used_bands) {
    next if ($i == $j);
SAS::error('repeatedBandNumber', "Repeat bands not allowed.")
      if ($band == $used_bands[$j]);
    }
  }

  #......................................................................
  # Set up all file names and other variables:

  $stream_num =~ s/^0+//; # chop off any leading zeros
  $id_number  =~ s/^0+//; # chop off any leading zeros
  my $stream_num_str = '0'x(2-length($stream_num))."$stream_num";
  my $id_number_str  = '0'x(4-length($id_number)) ."$id_number";
  my $id_str = "S$stream_num_str"."F$id_number_str";

  my $detmaskset = "$simop_dir/Sim_$instexpid"
    ."_edetectMask.fits";

  my $boxlocalset = "$simop_dir/P$id_str$instexpid"
    ."EBLSLI0000.FIT";

  my $boxmapset = "$simop_dir/P$id_str$instexpid"
    ."EBMSLI0000.FIT";

  my $mlset = "$simop_dir/P$id_str$instexpid"
    ."EMSRLI0000.FIT";

  if ($inst_uc eq 'PN') {
    $evlist = "$prod_dir/P$obsid$instexpid"."PIEVLI0000.FIT";
  } else {
    $evlist = "$prod_dir/P$obsid$instexpid"."MIEVLI0000.FIT";
  }

  foreach $band (@used_bands) {
    $expmaps{$band} = "$prod_dir/P$obsid$instexpid"
      ."EXPMAP$band"."000.FIT";

    $sim_images{$band} = "$simop_dir/P$id_str$instexpid"
      ."IMAGE_$band"."000.FIT";

    $splinemaps{$band} = "$simop_dir/Sim_$id_str$instexpid"
      ."_splineMap_$band.fits";
  } # end foreach $band (@used_bands)

  #......................................................................
  # Read ECF info from the sim images and setup ecf and band lists and arrays:

  foreach $band (@used_bands) {
    $imageset = $sim_images{$band};

    if ($testflag) {
      $idbandlo{$band} = 999;
      $idbandhi{$band} = 999;
      $ecf{$band} = 999;

    } else {
SAS::error('noSimImage', "Couldn't find $imageset.") if (!-e "$imageset");

      ($status, $idbandlo{$band}) = &getKeywordValue($imageset, 0, 'E_LO');
SAS::error('badEloKwd', "E_LO kwd: getKeywordValue status = $status.")
      if ($status);

      $idbandlo{$band} = int($idbandlo{$band});

      ($status, $idbandhi{$band}) = &getKeywordValue($imageset, 0, 'E_HI');
SAS::error('badEhiKwd', "E_HI kwd: getKeywordValue status = $status.")
      if ($status);

      $idbandhi{$band} = int($idbandhi{$band});

      ($status, $ecf{$band})      = &getKeywordValue($imageset, 0, 'ECF');
SAS::error('badEcfKwd', "ECF kwd: getKeywordValue status = $status.")
      if ($status);

      $ecf{$band} = sprintf "%9.6f", $ecf{$band};
      $ecf{$band} =~ s/^\s+//;
    }
  }

  my $band_lo_string = join(' ', @idbandlo{@used_bands});
  my $band_hi_string = join(' ', @idbandhi{@used_bands});
  my $ecfstring      = join(' ', @ecf{@used_bands});

SAS::error('noEventList', "Couldn't find $evlist") if (!-e "$evlist");

  ($status, $mode) = &getKeywordValue($evlist, 0, 'DATAMODE');
SAS::error('badDataModeKwd', "DATAMODE kwd: getKeywordValue status = $status.", 1)
  if ($status);

  if($inst_uc eq 'PN' && uc($mode) =~ /^SMALL/ ){
SAS::error('badDataMode', "No source detection for PN mode $mode.");}

  #......................................................................
  # Decide where to go:

#  if ($first_stage eq 'emask') {
#    goto EMASK;
#  } elsif($first_stage eq 'boxlocal') {
if($first_stage eq 'boxlocal') {
    goto BOXLOCAL;
  } elsif($first_stage eq 'splinemap') {
    goto SPLINEMAP;
  } elsif($first_stage eq 'boxmap') {
    goto BOXMAP;
  } elsif($first_stage eq 'mldetect') {
    goto MLDETECT;
#  } elsif($first_stage eq 'finalize') {
#    goto FINALIZE;
#  } elsif($first_stage eq '') {
  } else {
#&quit("Unrecognized entry stage $first_stage.", 9);
#    goto EMASK;
  }

  #......................................................................
  # Start main routine:

#  EMASK:
#
#  $command = "emask "
#  ."expimageset=$expmap_list[1] "
#  ."detmaskset=$detmask "
#  ."threshold1=0.05 "
#  ;
#
#&quit("Couldn't run emask.", 10) if (&run("$command"));
#
#exit 0 if ($last_stage eq 'emask');
#
# Now done in eimsim_prepare.

  # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  BOXLOCAL:

  foreach $band (@used_bands) {
SAS::error('noExpMap', "Couldn't find $expmaps{$band}")
    if (!-e "$expmaps{$band}");
  }

SAS::error('noDetMaskSet', "Couldn't find $detmaskset")  if (!$testflag && !-e "$detmaskset");

  # Already checked image sets before branching to functions.

  $command = "eboxdetect "
  ."imagesets='".join(' ', @sim_images{@used_bands})."' "
  ."expimagesets='".join(' ', @expmaps{@used_bands})."' "
  ."detmasksets=$detmaskset "
  ."boxlistset=$boxlocalset "
  ."usemap=false "
  ."likemin=$likemin "
  ."boxsize=5 "
  ."withdetmask=true "
  ."withexpimage=true "
  ."ecf='$ecfstring' "
#  ."hrdef='$ebox_hrdef' "
  ."pimin='$band_lo_string' "
  ."pimax='$band_hi_string' "
  ;

SAS::error('eboxLocalFailed', "Couldn't run eboxdetect in local mode.")
  if (&run("$command"));

  goto ENDD if ($last_stage eq 'boxlocal');

  # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  SPLINEMAP:

SAS::error('noBoxLocalSet', "Couldn't find $boxlocalset")
  if (!$testflag && !-e "$boxlocalset");

SAS::error('noDetMaskSet', "Couldn't find $detmaskset")  if (!$testflag && !-e "$detmaskset");

  # Already checked image sets before branching to functions.

  foreach $band (sort(@used_bands)) {
    $command = "esplinemap "
    ."boxlistset=$boxlocalset "
    ."scut=0.001 "
    ."idband=$band "
    ."imageset=$sim_images{$band} "
    ."detmaskset=$detmaskset "
    ."withexpimage=no "
    ."withdetmask=yes "
    ."nsplinenodes=16 "
    ."bkgimageset=$splinemaps{$band} "
    ."pimin='$idbandlo{$band}' "
    ."pimax='$idbandhi{$band}' "
    ;

SAS::error('esplinemapFailed', "Couldn't run esplinemap on band $band.")
    if (&run("$command"));
  }

  goto ENDD if ($last_stage eq 'splinemap');

  # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  BOXMAP:

  foreach $band (@used_bands) {
SAS::error('noExpMapSet', "Couldn't find $expmaps{$band}")
    if (!-e "$expmaps{$band}");
SAS::error('noSplineMapSet', "Couldn't find $splinemaps{$band}")
    if (!$testflag && !-e "$splinemaps{$band}");
  }

SAS::error('noDetMaskSet', "Couldn't find $detmaskset")
  if (!$testflag && !-e "$detmaskset");

  # Already checked image sets before branching to functions.

  $command = "eboxdetect "
  ."imagesets='".join(' ', @sim_images{@used_bands})."' "
  ."expimagesets='".join(' ', @expmaps{@used_bands})."' "
  ."bkgimagesets='".join(' ', @splinemaps{@used_bands})."' "
  ."detmasksets=$detmaskset "
  ."boxlistset=$boxmapset "
  ."usemap=true "
  ."likemin=$likemin "
  ."boxsize=5 "
  ."withdetmask=true "
  ."withexpimage=true "
  ."ecf='$ecfstring' "
#  ."hrdef='$ebox_hrdef' "
  ."pimin='$band_lo_string' "
  ."pimax='$band_hi_string' "
  ;

SAS::error('eboxMapFailed', "Couldn't run eboxdetect in map mode.")
  if (&run("$command"));

  goto ENDD if ($last_stage eq 'boxmap');

  # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  MLDETECT:

  foreach $band (@used_bands) {
SAS::error('noExpMapSet', "Couldn't find $expmaps{$band}")
    if (!-e "$expmaps{$band}");
SAS::error('noSplineMapSet', "Couldn't find $splinemaps{$band}")
    if (!$testflag && !-e "$splinemaps{$band}");
  }

SAS::error('noDetMaskSet', "Couldn't find $detmaskset")
  if (!$testflag && !-e "$detmaskset");
SAS::error('noBoxMapSet', "Couldn't find $boxmapset")
  if (!$testflag && !-e "$boxmapset");

  # Already checked image sets before branching to functions.

  $command = "emldetect "
  ."imagesets='".join(' ', @sim_images{@used_bands})."' "
  ."expimagesets='".join(' ', @expmaps{@used_bands})."' "
  ."bkgimagesets='".join(' ', @splinemaps{@used_bands})."' "
  ."boxlistset=$boxmapset "
  ."mllistset=$mlset "
  ."ecf='$ecfstring' "
#  ."hrm1def='$hrm1def' "
#  ."hrm2def='$hrm2def' "
#  ."hrpndef='$hrpndef' "
#  ."withxidband=Y "
."withxidband=no "
#  ."xidm1def='$xid_band_list_str' "
#  ."xidm2def='$xid_band_list_str' "
#  ."xidpndef='$xid_band_list_str' "
#  ."xidecf=$xidecf{$filter} "
  ."determineerrors=Y "
  ."mlmin=$mlmin "
  ."pimin='$band_lo_string' "
  ."pimax='$band_hi_string' "

##### withdetmask, detmasksets???????? check pcms module
  ;

SAS::error('emldetectFailed', "Couldn't run emldetect.") if (&run("$command"));

  # Copy over some useful columns from the eboxdetect output:

SAS::error('noMlSrcListSet', "Couldn't find $mlset")
  if (!$testflag && !-e "$mlset");

  $command = "esrclistpatch "
  ."boxset=$boxmapset "
  ."mlset=$mlset "
  ;

SAS::error('addBoxListColumnsFailed', "Couldn't add columns from boxset.")
  if (&run("$command"));

  #......................................................................
  # Now add extension LOGNLOGS; columns STREAM_N, FIELD_N; keywords *_ECF,
  # SKY_AREA.

  $command = "eimsim "
  ."function=addcols "
  ."tab=$mlset".":SRCLIST "
  ."newcolumnlist='STREAM_N FIELD_N' "
  ."newcolvaluelist='$stream_num $id_number' "
  ;

SAS::error('addColsFailed', "Couldn't add stream & field numbers to "
."list of detected sources.") if (&run("$command"));

  foreach $band (sort(@used_bands)) {
    $command = "addattribute "
    ."attributename='$inst_uc"."_$band"."_ECF' "
    ."attributetype='real' "
###    ."attributeunit=''10^-11 counts cm^2 erg^-1'' "
# Can't yet do this. DJF is going to modify addattribute to accept a unit parameter.
    ."withattributecomment='yes' "
    ."attributecomment='\"Flux to rate for ID_BAND $band\"' "
    ."set=$mlset "
    ."realvalue='$ecf{$band}' "
    ;
 
SAS::error('addEcfKwdFailed', "Couldn't run addattribute.")
    if (&run("$command"));
  }

  if (!$testflag) {
    ($status, $pix_delta_x) = &getKeywordValue($mlset, 0, 'TCDELT1');
SAS::error('badTcdelt1Kwd', "TCDELT1 kwd: getKeywordValue status = $status.")
    if ($status);

    ($status, $pix_delta_y) = &getKeywordValue($mlset, 0, 'TCDELT2');
SAS::error('badTcdelt2Kwd', "TCDELT2 kwd: getKeywordValue status = $status.")
    if ($status);

SAS::error('getPixelDimsFailed', "Couldn't retrieve pixel dimensions.")
    if ($pix_delta_x == 0 || $pix_delta_y == 0);

    $command = "fimgstat "
    ."infile=$detmaskset "
    ."threshlo=INDEF "
    ."threshup=INDEF "
    ;

    my @lines = `$command | grep sum`;

SAS::error('getAreaDetMaskFailed', "Couldn't get area of detmask.")
    if (!defined($lines[0]));

    if ($lines[0] =~ /=\s*(\d+)/) {
      $n_mask_pixels = $1;
    } else {
SAS::error('getAreaDetMaskFailed', "Couldn't get area of detmask.");
    }

    $sky_area = abs($n_mask_pixels * $pix_delta_x * $pix_delta_y);
  } else {
    $sky_area = 999;
  }

#  $command = "fparkey "
#  ."value=$sky_area "
#  ."fitsfile=$mlset+1 "
#  ."keyword='SKY_AREA' "
#  ."add=yes "
#  ;
#
#&quit("Couldn't add SKY_AREA keyword.", 20)
#if (&run("$command"));

  $command = "addattribute "
  ."attributename='SKY_AREA' "
  ."attributetype='real' "
###  ."attributeunit=''deg^2'' "
# Can't yet do this. DJF is going to modify addattribute to accept a unit parameter.
  ."withattributecomment='yes' "
###  ."attributecomment=''Area of detection mask.'' "
."attributecomment='\"Area in deg^2 of detection mask.\"' "
  ."set=$mlset "
  ."realvalue='$sky_area' "
  ;
 
SAS::error('addSkyAreaFailed', "Couldn't add SKY_AREA keyword.")
  if (&run("$command"));

  $imageset = $sim_images{$used_bands[0]}; # any band ought to do.

  $command = "fappend "
  ."infile=$imageset'[LOGNLOGS]' "
  ."outfile=$mlset "
  ;
 
SAS::error('addLogNLogSExtFailed', "Couldn't add LOGNLOGS extension.")
  if (&run("$command"));

  goto ENDD if ($last_stage eq 'mldetect');

  # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
#  FINALIZE:

#### esky2det stuff here

#### source plot stuff here

#### extended source stuff here


#  $command = "esensmap "
#  ."bkgimagesets='$xid_splinemap_list_str' "
#  ."detmasksets=$detmask "
#  ."expimagesets='$xid_expmap_list_str' "
#  ."sensimageset=$sensemapset "
#  ;
#
#&quit("Couldn't run esensmap.", 13)
#if (&run("$command"));

  ENDD:

}

  #---------------------------------------------------------------------
  sub run {
  #---------------------------------------------------------------------
  # $testflag should be defined globally.

    my ($command) = @_;
    my $status = 0;

    print "\ninvoke $command\n\n";
    if (!$testflag) {
      $status = system($command);
    }

  return $status;
  }

  #----------------------------------------------------------------------
  sub getKeywordValue {
  #----------------------------------------------------------------------

    my ($filename, $extname, $kwd) = @_;

    my ($rump, $command);
    my $value = 0; # default

  return (-1, 0) if (!defined($filename) || !-e "$filename");
  return (-2, 0) if (!defined($extname));
  return (-3, 0) if (!defined($kwd) || $kwd eq '');

    if ($extname =~ /^\d+$/) {
      $command = "fkeyprint $filename+$extname $kwd | grep =";
    } else {
      $command = "fkeyprint $filename'[$extname]' $kwd | grep =";
    }

    my @lines = `$command`;
    if (defined($lines[0])){
      if ($lines[0] =~ /^${kwd}\s*=\s*/) {
        $rump = $';
        if ($rump =~ /^'(.+)'/ || $rump =~ /(\S+)/) {
          $value = $1;
        } else {
  return (-4, 0);
        }
      } else {
  return (-5, 0);
      }
    } else {
  return (-6, 0);
    }

  return (0, $value);
  }
