!
! Module: eimsimloadcorr_mod
!
! Authors:  Ian Stewart - SSC-LUX
!          
! Description: This is an interactive tool to help load tables to correct systematic errors arising in the source detection.
! 
! Comments:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eimsimloadcorr_mod

  use dal
  use param  
  use errorHandling

  use dal_aux
  use array_utils
  use ssc_misc

  implicit none

  private 

  ! Command-line parameters:
  character(255) ::&
    sysErrorSetTabName,&
    correctorSetTabName
  character(20) ::&
    xColName,&
    yColName,&
    yErrColName

  ! Other globals:
  type :: pinT
    real :: x, y
  end type pinT
  type :: pinPosT
    type(pinT), pointer :: pos(:)
  end type pinPosT

!  logical :: withInSet
!  character(len=255) :: inSetName, outSetName
  real, parameter ::&
    stripMinX = 0.0,&
    stripMaxX = 0.05,&
    stripMinY = 0.02,&
    stripMaxY = 0.98
  integer, parameter ::&
    numBoxes = 1
  real, parameter ::&
    boxRowMinXVp = 0.08,&
    boxRowMaxXVp = 0.84,&
    boxRowMinYVp = stripMinY,&
    boxRowMaxYVp = stripMaxY,&
    boxPadFraction = 0.1,&
    delVpFrac = 0.005
  real, dimension(numBoxes) ::&
    boxMinXVp, boxMaxXVp,&
    boxMinXWd, boxMaxXWd, boxMinYWd, boxMaxYWd,&
    delXWd, delYWd
  integer, parameter ::&
    numBtns = 4
  real, parameter ::&
    btnColMinX = 0.87,&
    btnColMaxX = 1.0,&
    btnColMinY = stripMinY,&
    btnColMaxY = stripMaxY,&
    btnPadFraction = 0.15
  real, dimension(numBtns) :: btnMinY, btnMaxY
!  integer :: minClrIndex, maxClrIndex, numClrIndices
  character(len=20), dimension(numBtns) ::&
    btnText = (/'Delete     ', 'Save       ', 'EXIT       ', 'Save & exit'/)
  integer :: selectedPinNum, selectedBoxNum
  logical :: somePinIsSelected
  integer, parameter :: pinSymbol = 0, selectionSymbol = 16, bkgSymbol = 5

  public eimsimloadcorr

contains 
  
  !---------------------------------------------------------------------
  subroutine eimsimloadcorr
  !---------------------------------------------------------------------
    
    implicit none

    character(255) :: sysErrorSetName, correctorSetName
    character(20) :: sysErrorTabName, correctorTabName
    type(DataSetT) :: sysErrorSet, correctorSet
    type(TableT) :: sysErrorTab, correctorTab
    integer :: numPoints, numCorrPoints
    real(single), allocatable :: xValues(:), yValues(:), yErrValues(:)
    logical :: fileExists, fileOpened, canLoadCorrector
!    real(single) :: loX, loY, loSlope, hiY, hiSlope
    type(pinPosT) :: pins(numBoxes)
    real, parameter :: pad = 0.05
    integer :: status, pgopen, btnNum, pgcurs
    real :: x, y, newX, newY
    character(len=1) :: char
    real(single), pointer :: xData(:), yData(:)

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    ! Load the syserror fits table:
    call splitSetTabName(sysErrorSetTabName, sysErrorSetName, sysErrorTabName)
    sysErrorSet = dataSet(sysErrorSetName, READ)
    sysErrorTab = table(sysErrorSet, sysErrorTabName)
    numPoints = numberOfRows(sysErrorTab)
if (numPoints <= 1) call error('','')
    allocate(xValues(numPoints), yValues(numPoints), yErrValues(numPoints))
    call readColDataToFixed(sysErrorTab, xColName, xValues)
    call readColDataToFixed(sysErrorTab, yColName, yValues)
    call readColDataToFixed(sysErrorTab, yErrColName, yErrValues)
    call release(sysErrorSet)

if (any(xValues <= 0.0)) call error('','')
    xValues = log10(xValues)

    ! Set up world-coord limits:
    boxMinXWd(1) = minval(xValues)
    boxMaxXWd(1) = maxval(xValues)
    boxMinYWd(1) = minval(yValues + yErrValues, (yValues + yErrValues > 0.0))
    boxMaxYWd(1) = maxval(yValues + yErrValues)
if (boxMinXWd(1) >= boxMaxXWd(1)) call error('','')
if (boxMinYWd(1) >= boxMaxYWd(1)) call error('','')

    ! Load the syserror corrector:
    call splitSetTabName(correctorSetTabName, correctorSetName&
      , correctorTabName)

    canLoadCorrector = .false.
    inquire(file=correctorSetName, exist=fileExists, opened=fileOpened)
    if (fileExists) then
      correctorSet = dataSet(correctorSetName, READ)
      if (hasBlock(correctorSet, correctorTabName)) canLoadCorrector = .true.
      call release(correctorSet)
    end if

    if (canLoadCorrector) then
      correctorSet = dataSet(correctorSetName, READ)
      correctorTab = table(correctorSet, correctorTabName)
      numCorrPoints = numberOfRows(correctorTab)
if (numCorrPoints <= 0) call error('','')
      allocate(pins(1)%pos(numCorrPoints))
      call readColDataToFixed(correctorTab, 'X', pins(1)%pos%x)
      call readColDataToFixed(correctorTab, 'Y', pins(1)%pos%y)
      call release(correctorSet)

if (any(pins(1)%pos%x <= 0.0)) call error('','')
      pins(1)%pos%x=log10(pins(1)%pos%x)

      boxMinXWd(1) = min(boxMinXWd(1), minval(pins(1)%pos%x))
      boxMaxXWd(1) = max(boxMaxXWd(1), maxval(pins(1)%pos%x))
      boxMinYWd(1) = min(boxMinYWd(1), minval(pins(1)%pos%y))
      boxMaxYWd(1) = max(boxMaxYWd(1), maxval(pins(1)%pos%y))

    else
      allocate(pins(1)%pos(0))
      numCorrPoints = 0
    end if

    boxMinXWd(1) = boxMinXWd(1) - (boxMaxXWd(1) - boxMinXWd(1))&
      * pad / (1.0 - 2.0 * pad)
    boxMaxXWd(1) = boxMaxXWd(1) + (boxMaxXWd(1) - boxMinXWd(1))&
      * pad / (1.0 - pad)
    boxMinYWd(1) = boxMinYWd(1) - (boxMaxYWd(1) - boxMinYWd(1))&
      * pad / (1.0 - 2.0 * pad)
    boxMaxYWd(1) = boxMaxYWd(1) + (boxMaxYWd(1) - boxMinYWd(1))&
      * pad / (1.0 - pad)

    ! Setup for interactive plotting routine:

    status = pgopen('/xs')
    call pgsvp(0.0, 1.0, 0.0, 1.0)
    call pgswin(0.0, 1.0, 0.0, 1.0)
    call plotButtons
    call setUpBoxSizes

    ! Now enter the interactive plotting loop:

    x = 0.0
    y = 0.0
    do
      call pgbbuf
      call drawBkg(boxMinXVp(1), boxMaxXVp(1), boxRowMinYVp, boxRowMaxYVp&
        , boxMinXWd(1), boxMaxXWd(1), boxMinYWd(1), boxMaxYWd(1)&
        , xValues, yValues, yErrValues)

      call drawPinGraph(boxMinXVp(1), boxMaxXVp(1), boxRowMinYVp, boxRowMaxYVp&
        , boxMinXWd(1), boxMaxXWd(1), boxMinYWd(1), boxMaxYWd(1)&
        , pins(1)%pos, 1)

      call pgebuf

      status = pgcurs(x, y, char)
      btnNum = readButton(x, y)

      select case (btnNum)
      case (1) ! delete selected point
print *, 'delete point'
        call deletePin(pins(1)%pos, selectedPinNum, somePinIsSelected)

      case (2, 4)
print *, 'save file'
        if (fileExists) then
          correctorSet = dataSet(correctorSetName, MODIFY)
        else
          correctorSet = dataSet(correctorSetName, CREATE)
        end if

        if (hasBlock(correctorSet, correctorTabName)) then
          call deleteBlock(correctorSet, correctorTabName)
        end if

        if (size(pins(1)%pos) > 0) then
          correctorTab = addTable(correctorSet, correctorTabName&
            , size(pins(1)%pos))
          xData => real32Data(addColumn(correctorTab, 'X',     REAL32))
          yData => real32Data(addColumn(correctorTab, 'Y',     REAL32))

          xData = 10.0**pins(1)%pos%x
          yData = pins(1)%pos%y
        end if

        call release(correctorSet)

    if (btnNum == 4) exit

      case (3) ! just exit
print *, 'just exit'
    exit

      case default ! no button, drag the relevant pinning diagram:
        call doPinningMod(x, y, newX, newY, pins)
        x = newX
        y = newY
      end select
    end do

    call pgclos

    deallocate(pins(1)%pos)

  end subroutine eimsimloadcorr 

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    sysErrorSetTabName = stringParameter('syserrset')
    xColName = stringParameter('xcol')
    yColName = stringParameter('ycol')
    yErrColName = stringParameter('yerrcol')
    correctorSetTabName = stringParameter('corrset')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine plotButtons
  !---------------------------------------------------------------------

    implicit none

    real ::&
      btnHeight,&
      textX,&
      textY
    integer ::&
       btnNum

    !-------------------------------------------------------------------
    ! Start

    call pgsci(1)
    call pgsfs(2)

    btnHeight = (btnColMaxY - btnColMinY) / (numBtns + (numBtns - 1)&
    * (btnPadFraction) / (1.0 - btnPadFraction))
    do btnNum = 1, numBtns
      call calcFracWithPad(btnColMinY, btnColMaxY, btnPadFraction&
        , numBtns, btnNum, btnMinY(btnNum), btnMaxY(btnNum))

      call pgrect(btnColMinX, btnColMaxX, btnMinY(btnNum), btnMaxY(btnNum))
      textX = btnColMinX + 0.02
      textY = (btnMaxY(btnNum) + btnMinY(btnNum)) / 2.0
      call pgtext(textX, textY, btnText(btnNum))
    end do

  end subroutine plotButtons

  !---------------------------------------------------------------------
  subroutine calcFracWithPad(minn, maxx, padFraction&
  , numBins, binNum, binMinn, binMaxx, status)
  !---------------------------------------------------------------------
!** Should make this a vector calculation.

!** note that padFraction is the fraction of the bin width that the pad should equal. This may be any number >= 0.

    implicit none

    real, intent(in) :: minn, maxx, padFraction
    integer, intent(in) :: numBins, binNum
    real, intent(out) :: binMinn, binMaxx
    integer, intent(out), optional :: status

    integer :: localStatus
    real :: maxBinMin, minBinMax

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0

    if (numBins <= 0) then
      if (present(status)) status = -1
  return
    end if
    if (binNum <= 0 .or. binNum > numBins) then
      if (present(status)) status = -2
  return
    end if
    if (minn >= maxx) then
      if (present(status)) status = -3
  return
    end if

    if (numBins == 1) then
      binMinn = minn
      binMaxx = maxx

    else
      maxBinMin = fracVal(minn, maxx&
        , ((numBins - 1) * (1.0 + padFraction))&
        / (numBins + (numBins - 1) * padFraction))

      binMinn = fracVal(minn, maxBinMin, (binNum - 1) / real(numBins - 1))

      minBinMax = fracVal(minn, maxx&
        , 1.0 / (numBins + (numBins - 1) * padFraction))

      binMaxx = fracVal(minBinMax, maxx, (binNum - 1) / real(numBins - 1))

    end if

    if (present(status)) status = localStatus
    
  end subroutine calcFracWithPad

  !---------------------------------------------------------------------
  subroutine setUpBoxSizes
  !---------------------------------------------------------------------

    implicit none

    integer ::&
       boxNum

    !-------------------------------------------------------------------
    ! Start

    do boxNum = 1, numBoxes
      call calcFracWithPad(boxRowMinXVp, boxRowMaxXVp, boxPadFraction&
        , numBoxes, boxNum, boxMinXVp(boxNum), boxMaxXVp(boxNum))

      delXWd(boxNum) = fracVal(0.0, (boxMaxXWd(boxNum) - boxMinXWd(boxNum))&
        , 0.0, 1.0, delVpFrac)
      delYWd(boxNum) = fracVal(0.0, (boxMaxYWd(boxNum) - boxMinYWd(boxNum))&
        , 0.0, 1.0, delVpFrac)
    end do

  end subroutine setUpBoxSizes

  !---------------------------------------------------------------------
  subroutine drawBkg(boxMinXVp, boxMaxXVp, boxMinYVp, boxMaxYVp&
  , boxMinXWd, boxMaxXWd, boxMinYWd, boxMaxYWd&
  , bkgX, bkgY, bkgYErr)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: boxMinXVp, boxMaxXVp, boxMinYVp, boxMaxYVp,&
                        boxMinXWd, boxMaxXWd, boxMinYWd, boxMaxYWd
    real, intent(in) :: bkgX(:), bkgY(size(bkgX)), bkgYErr(size(bkgX))

    real ::&
       oldMinXVp, oldMaxXVp, oldMinYVp, oldMaxYVp,&
       oldMinXWd, oldMaxXWd, oldMinYWd, oldMaxYWd
    integer ::&
       j, oldCi

    !-------------------------------------------------------------------
    ! Start

    call pgqci(oldCi)

    call pgqvp(0, oldMinXVp, oldMaxXVp, oldMinYVp, oldMaxYVp)
    call pgsvp(boxMinXVp, boxMaxXVp, boxMinYVp, boxMaxYVp)
    call pgqwin(oldMinXWd, oldMaxXWd, oldMinYWd, oldMaxYWd)
    call pgswin(boxMinXWd, boxMaxXWd, boxMinYWd, boxMaxYWd)
!    call pgeras
    call pgsci(0)
    call pgsfs(1)
    call pgrect(boxMinXWd, boxMaxXWd, boxMinYWd, boxMaxYWd)
    call pgsci(2)
    call pgsfs(2)

    do j = 1, size(bkgX)
      call pgmove(bkgX(j), bkgY(j) - bkgYErr(j))
      call pgdraw(bkgX(j), bkgY(j) + bkgYErr(j))
      call pgpt1(bkgX(j), bkgY(j), bkgSymbol)
    end do

    call pgswin(oldMinXWd, oldMaxXWd, oldMinYWd, oldMaxYWd)
    call pgsvp(oldMinXVp, oldMaxXVp, oldMinYVp, oldMaxYVp)
    call pgsci(oldCi)

  end subroutine drawBkg

  !---------------------------------------------------------------------
  subroutine drawPinGraph(boxMinXVp, boxMaxXVp, boxMinYVp, boxMaxYVp&
  , boxMinXWd, boxMaxXWd, boxMinYWd, boxMaxYWd&
  , pins, boxNum)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: boxMinXVp, boxMaxXVp, boxMinYVp, boxMaxYVp,&
                        boxMinXWd, boxMaxXWd, boxMinYWd, boxMaxYWd
    type(pinT), intent(in) :: pins(:)
    integer, intent(in) :: boxNum

    integer :: pinNum
    real ::&
       oldMinXVp, oldMaxXVp, oldMinYVp, oldMaxYVp,&
       oldMinXWd, oldMaxXWd, oldMinYWd, oldMaxYWd

    !-------------------------------------------------------------------
    ! Start

    call pgqvp(0, oldMinXVp, oldMaxXVp, oldMinYVp, oldMaxYVp)
    call pgsvp(boxMinXVp, boxMaxXVp, boxMinYVp, boxMaxYVp)
    call pgqwin(oldMinXWd, oldMaxXWd, oldMinYWd, oldMaxYWd)
    call pgswin(boxMinXWd, boxMaxXWd, boxMinYWd, boxMaxYWd)
!    call pgrect(boxMinXWd, boxMaxXWd, boxMinYWd, boxMaxYWd)
call pgbox('BC', 0.0, 0, 'BC', 0.0, 0)

    if (size(pins) > 0) then
      call pgmove(pins(1)%x, pins(1)%y)
      call pgpt1(pins(1)%x, pins(1)%y, pinSymbol)
      do pinNum = 2, size(pins)
        call pgdraw(pins(pinNum)%x, pins(pinNum)%y)
        call pgpt1(pins(pinNum)%x, pins(pinNum)%y, pinSymbol)
      end do

      if (somePinIsSelected .and. boxNum .eq. selectedBoxNum) then
        pinNum = selectedPinNum
        call pgpt1(pins(pinNum)%x, pins(pinNum)%y, selectionSymbol)
      end if
    end if

    call pgswin(oldMinXWd, oldMaxXWd, oldMinYWd, oldMaxYWd)
    call pgsvp(oldMinXVp, oldMaxXVp, oldMinYVp, oldMaxYVp)

  end subroutine drawPinGraph 

  !---------------------------------------------------------------------
  function readButton(csrX, csrY)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: csrX, csrY
    integer :: btnNum, readButton

    !-------------------------------------------------------------------
    ! Start

    readButton = -1 ! default

    if (csrX .ge. btnColMinX .and. csrX .le. btnColMaxX) then
      do btnNum = 1, numBtns
        if (csrY >= btnMinY(btnNum) .and. csrY < btnMaxY(btnNum)) then
          readButton = btnNum
      exit
        end if
      end do
    end if

  end function readButton

  !---------------------------------------------------------------------
  subroutine doPinningMod(oldCsrXVp, oldCsrYVp, newCsrXVp, newCsrYVp, pins)
  !---------------------------------------------------------------------

    implicit none

    real, intent(inout) ::&
      oldCsrXVp,&
      oldCsrYVp
    real, intent(out) ::&
      newCsrXVp,&
      newCsrYVp
    type(pinPosT), intent(inout) :: pins(:)

    integer ::&
      boxNum,&
      pinNum,&
      status,&
      pgband
    real ::&
      csrXWd,&
      csrYWd
    character(len=1) ::&
      char
    logical ::&
      isNearPin

    !-------------------------------------------------------------------
    ! Start

    newCsrXVp = oldCsrXVp ! Defaults: so the cursor doesn't get shifted
    newCsrYVp = oldCsrYVp ! when the click is not within a box.

!    if (oldCsrY .ge. boxRowMinYVp .and. oldCsrY .lt. boxRowMaxYVp) then
    if (oldCsrYVp >= 0.0 .and. oldCsrYVp < 1.0) then ! A tiny bit of a
    ! bodge so that the first and last pins can be moved without trouble.
      do boxNum = 1, numBoxes
        if (oldCsrXVp >= boxMinXVp(boxNum)&
        .and. oldCsrXVp <= boxMaxXVp(boxNum)) then

          csrXWd = fracVal(boxMinXWd(boxNum), boxMaxXWd(boxNum)&
            , boxMinXVp(boxNum), boxMaxXVp(boxNum), oldCsrXVp)
          csrYWd = fracVal(boxMinYWd(boxNum), boxMaxYWd(boxNum)&
            , boxRowMinYVp, boxRowMaxYVp, oldCsrYVp)

          call checkIsNearPin(csrXWd, csrYWd, delXWd(boxNum), delYWd(boxNum)&
            , pins(boxNum)%pos, isNearPin, pinNum)

print *, 'Is near pin?', isNearPin

          if (isNearPin) then
            selectedPinNum = pinNum
            somePinIsSelected = .true.
            call pgpt1(pins(boxNum)%pos(pinNum)%x&
              , pins(boxNum)%pos(pinNum)%y, selectionSymbol)

          else
            call addPin(pins(boxNum)%pos, csrXWd, csrYWd, selectedPinNum&
              , somePinIsSelected)
          end if
          selectedBoxNum = boxNum

          ! Relocate cursor position to the centre of the old pin position:
          oldCsrXVp = fracVal(boxMinXVp(boxNum), boxMaxXVp(boxNum)&
            , boxMinXWd(boxNum), boxMaxXWd(boxNum)&
            , pins(boxNum)%pos(selectedPinNum)%x)
          oldCsrYVp = fracVal(boxRowMinYVp, boxRowMaxYVp&
            , boxMinYWd(boxNum), boxMaxYWd(boxNum)&
            , pins(boxNum)%pos(selectedPinNum)%y)
print *, 'old pin position:'&
, pins(boxNum)%pos(selectedPinNum)%x, pins(boxNum)%pos(selectedPinNum)%y

          ! Drag the point:
          status = pgband(2, 1, oldCsrXVp, oldCsrYVp, newCsrXVp, newCsrYVp&
            , char)

          csrXWd = fracVal(boxMinXWd(boxNum), boxMaxXWd(boxNum)&
            , boxMinXVp(boxNum), boxMaxXVp(boxNum), newCsrXVp)
          csrYWd = fracVal(boxMinYWd(boxNum), boxMaxYWd(boxNum)&
            , boxRowMinYVp, boxRowMaxYVp, newCsrYVp)

          ! Limit the vertical travel to the box boundaries:
          csrYWd = min(boxMaxYWd(boxNum), max(boxMinYWd(boxNum), csrYWd))

          ! Limit the horizontal travel to the X positions of the pins
          ! on either side, or the box boundaries, whichever comes first:
          if (selectedPinNum == 1) then
            csrXWd = max(boxMinXWd(boxNum), csrXWd)
          else
            csrXWd = max(pins(boxNum)%pos(selectedPinNum-1)%x, csrXWd)
          end if
          if (selectedPinNum == size(pins(boxNum)%pos)) then
            csrXWd = min(boxMaxXWd(boxNum), csrXWd)
          else
            csrXWd = min(pins(boxNum)%pos(selectedPinNum+1)%x, csrXWd)
          end if

!          if (selectedPinNum > 1&
!          .and. selectedPinNum < size(pins(boxNum)%pos)) then
            pins(boxNum)%pos(selectedPinNum)%x = csrXWd
!          end if
          pins(boxNum)%pos(selectedPinNum)%y = csrYWd
print *, 'new pin position:'&
, pins(boxNum)%pos(selectedPinNum)%x, pins(boxNum)%pos(selectedPinNum)%y
print *, ''

      exit
        end if
      end do
    end if

  end subroutine doPinningMod

  !---------------------------------------------------------------------
  subroutine checkIsNearPin(csrX, csrY, delX, delY, pins, isNearPin, pinNum)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: csrX, csrY, delX, delY
    type(pinT), intent(in) :: pins(:)
    logical, intent(out) :: isNearPin
    integer, intent(out) :: pinNum

    logical(bool) :: mask(size(pins))
    integer :: tempVec(1)

    !-------------------------------------------------------------------
    ! Start

    pinNum = 0 ! meaningless default. Don't like to leave the value floating.
    isNearPin = .true. ! default

    mask = (abs(pins%x - csrX) < delX .and. abs(pins%y - csrY) < delY)
    if (count(mask) > 1) then
      tempVec = minloc(abs(pins%x - csrX) + abs(pins%y - csrY), mask)
      pinNum = tempVec(1)

    else if (count(mask) == 1) then
      pinNum = largestTrueIndex(mask)

    else
      isNearPin = .false.
    end if

  end subroutine checkIsNearPin

  !---------------------------------------------------------------------
  subroutine deletePin(pins, selectedPinNum, somePinIsSelected)
  !---------------------------------------------------------------------

    implicit none

    type(pinT), pointer :: pins(:)
    integer, intent(inout) :: selectedPinNum
    logical, intent(inout) :: somePinIsSelected

    integer :: numPins
    type(pinT), allocatable :: tempPins(:)

    !-------------------------------------------------------------------
    ! Start

    if (somePinIsSelected) then
      numPins = size(pins)
!if (numPins <= 2) call error('BUG','')

!      if (numPins > 1) then
        allocate(tempPins(numPins - 1))
        tempPins(:selectedPinNum-1)%x = pins(:selectedPinNum-1)%x
        tempPins(:selectedPinNum-1)%y = pins(:selectedPinNum-1)%y
        tempPins(selectedPinNum:)%x = pins(selectedPinNum+1:)%x
        tempPins(selectedPinNum:)%y = pins(selectedPinNum+1:)%y
        deallocate(pins)
        allocate(pins(numPins - 1))
        pins%x = tempPins%x
        pins%y = tempPins%y
        deallocate(tempPins)

!      else
!        deallocate(pins)
!        allocate(pins(0))
!      end if

      somePinIsSelected = .false.
    end if

  end subroutine deletePin

  !---------------------------------------------------------------------
  subroutine addPin(pins, csrX, csrY, selectedPinNum, somePinIsSelected)
  !---------------------------------------------------------------------

    implicit none

    type(pinT), pointer :: pins(:)
    real, intent(in) :: csrX, csrY
    integer, intent(inout) :: selectedPinNum
    logical, intent(inout) :: somePinIsSelected

    integer :: numPins
    type(pinT), allocatable :: tempPins(:)

    !-------------------------------------------------------------------
    ! Start

!if (csrX <= minval(pins%x) .or. csrX >= maxval(pins%x)) call error('','')

    selectedPinNum = 1 + largestTrueIndex(logical(pins%x < csrX, bool))
print *, 'selectedPinNum',selectedPinNum

    numPins = size(pins)
    allocate(tempPins(numPins + 1))
    tempPins(:selectedPinNum-1)%x = pins(:selectedPinNum-1)%x
    tempPins(:selectedPinNum-1)%y = pins(:selectedPinNum-1)%y
    tempPins(selectedPinNum)%x = csrX

    if (selectedPinNum <= 1 .or. selectedPinNum > numPins) then
      tempPins(selectedPinNum)%y = csrY

    else
      tempPins(selectedPinNum)%y = fracVal(pins(selectedPinNum-1)%y&
        , pins(selectedPinNum)%y, pins(selectedPinNum-1)%x&
        , pins(selectedPinNum)%x, csrX)
    end if

    tempPins(selectedPinNum+1:)%x = pins(selectedPinNum:)%x
    tempPins(selectedPinNum+1:)%y = pins(selectedPinNum:)%y
    deallocate(pins)
    allocate(pins(numPins + 1))
    pins%x = tempPins%x
    pins%y = tempPins%y
    deallocate(tempPins)

    somePinIsSelected = .true.

  end subroutine addPin

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module eimsimloadcorr_mod
