#!/usr/local/bin/perl -w

# ENDD
my ($testflag, $i, @used_insts_uc, @idbandlo, @idbandhi);
my (%ecf_info, $key, $band, $bkg_rate, $vig_frac, $obsid_root);
my ($prod_dir, @prod_files, $file, $obsid, $inst_uc);
my (@files, @chunks, $instexpid, $simop_dir, $evlist, $cif);
my ($ref_obsid, $expmap, $command, $expmapset, $padmaskset, $ahf);
my ($eventset, $band_id, $imageset, $nonvigset, $bkgmapset);
my ($detmaskset, $ecf_file);#, @band_ids);
my (@used_bands);

my ($already_found_obsid, %cifs);

sub eimsimprep {

  use SAS;

  my $taskname = "eimsimprep";

  # Prints all output to stderr:
  select(STDERR);

  #......................................................................
  # Read and check parameters:

  my $num_entries = parameterCount('obsidroots');
SAS::error('noObsIds', "There must be at least one member in the "
."list --obsidroots") if ($num_entries <= 0);

  my @obsid_roots = ();
  foreach $i (0 .. $num_entries-1) {
    $obsid_roots[$i] = stringParameter('obsidroots', $i);
  }

  my $first_stage      = stringParameter('entrystage');
  my $last_stage       = stringParameter('finalstage');
  my $ref_band         = intParameter('refband'); # for making sky mask
  my $with_ref_obsid   = booleanParameter('withrefobsidroot');
  my $ref_obsid_root   = stringParameter('refobsidroot'); # for making ecfs
  my $ecf_file         = stringParameter('ecffile');

  $num_entries = parameterCount('bkgrates');
SAS::error('noBkgRates', "There must be at least one member in the "
."list --bkgrates") if ($num_entries <= 0);

  my @bkgrates = ();
  foreach $i (0 .. $num_entries-1) {
    $bkgrates[$i] = realParameter('bkgrates', $i);
  }

  $num_entries = parameterCount('vigfractions');

  my @vigfractions = ();
  foreach $i (0 .. $num_entries-1) {
    $vigfractions[$i] = realParameter('vigfractions', $i);
  }

  my $prod_subdir      = stringParameter('prdssubdir');
  my $simop_subdir     = stringParameter('simopsubdir');
  my $padsizearcsec    = realParameter('padsizearcsec');
  my $template_set     = stringParameter('templateset');

  $testflag       = booleanParameter('astest');

  if (@bkgrates != @vigfractions) {
SAS::error('badNumVigFractions', "Numbers of --bkgrates and --vigfractions are different.");}

  #......................................................................
  # Misc setups:

  if ($first_stage eq 'makeecfs') {
    @used_bands = (1, 2, 3, 4, 5);
#@band_ids = (1, 2, 3, 4, 5);
    @used_insts_uc = ('M1', 'M2', 'PN');
    @idbandlo = (200,  500, 2000, 4500,  7500);
    @idbandhi = (500, 2000, 4500, 7500, 12000);

  } else {
    %ecf_info = &getEcfs($ecf_file); # contains band edge info too.
    @used_insts_uc = ();
    foreach $key (keys(%ecf_info)) {
      if ($key =~ /^(M1|M2|PN)/) {
        push @used_insts_uc, $key;
      }
    }

SAS::error('noInstruments', "There are no instrument lines in the ECF file.") if (!defined($used_insts_uc[0]));

    @idbandlo = @{$ecf_info{'lo'}};
    @idbandhi = @{$ecf_info{'hi'}};
#    @used_bands = (1 .. @idbandlo);
#@band_ids = @{$ecf_info{'id'}};
@used_bands = @{$ecf_info{'id'}};
  }

SAS::error('badNumBkgRates', "There should be the same number of --bkgrates and --vigfractions as the number of bands.") if (@used_bands != @bkgrates);

  my %vigbkgrates = ();
  my %nonvigbkgrates = ();
  foreach $band (@used_bands) {
#foreach $band (@band_ids) {
    $bkg_rate = shift @bkgrates;
    $vig_frac = shift @vigfractions;
    $vigbkgrates{$band} = $bkg_rate * $vig_frac;
    $nonvigbkgrates{$band}  = $bkg_rate * (1-$vig_frac);
  }

  #......................................................................
  # Set up obsid hash:

  my @expmap_vector = ();
  my %obs_ids = ();
  foreach $obsid_root (@obsid_roots) {
    $prod_dir  = "$obsid_root/$prod_subdir";
    @prod_files = `ls -1 $prod_dir`;

SAS::error('noProducts', "No product files in $prod_dir.") if (@prod_files <= 0);

    $already_found_obsid = 0;
    foreach $file (@prod_files) {
      chomp($file);

      if ($file =~ /^P(\d{10})/) {
        if ($already_found_obsid) {
          if ($1 != $obsid) {
SAS::error('tooManyPathsThisObsid', "Only 1 --obsidroots entry allowed per obsid.");
          }
        } else {
          $obsid = $1;
          $already_found_obsid = 1;
        }

      } # end if $file =~ /^P(\d{10})/
    } # end foreach $file (@prod_files)

SAS::error('noProducts', "No product files in $prod_dir.") if (!$already_found_obsid);

    $obs_ids{$obsid_root} = $obsid;
  } # end foreach $obsid_root (@obsid_roots)

  #......................................................................
  # Setup instexpid arrays from existence of any files for that obsid and
  # instrument:

  my %temp_hash = ();
  my %instexpids = ();
  foreach $obsid_root (@obsid_roots) {
    $prod_dir  = "$obsid_root/$prod_subdir";
    $obsid = $obs_ids{$obsid_root};

    %temp_hash = ();
    foreach $inst_uc (@used_insts_uc) {
      @files = `ls -1 $prod_dir/P$obsid$inst_uc*`;
      foreach $file (@files) {
        @chunks = split('\/', $file);
        if ($chunks[$#chunks] =~ /^P${obsid}${inst_uc}((S|U)\d{3})\S{6}(\d)/) {
          $instexpid = "$inst_uc$1";
          $temp_hash{$instexpid} = 1;
        }
      }
    }
    push @{$instexpids{$obsid}}, keys(%temp_hash)
  }

  #......................................................................
  # Setup file names:

  my %attitude_sets = ();
  my %expmap_refs = ();
  my %padmasksets = ();
  my %detmasksets = ();
  my %evlists = ();
  my %expmaps = ();
  my %images = ();
  my %nonvigs = ();
  my %bkgmaps = ();
  foreach $obsid_root (@obsid_roots) {
    $prod_dir  = "$obsid_root/$prod_subdir";
    $simop_dir = "$obsid_root/$simop_subdir";

    $obsid = $obs_ids{$obsid_root};

    $cifs{$obsid}          = "$prod_dir/P$obsid"."OBX000CALIND0000.FIT";
    $attitude_sets{$obsid} = "$prod_dir/P$obsid"."OBX000ATTTSR0000.FIT";

    foreach $instexpid (@{$instexpids{$obsid}}) {
      $expmap_refs{"$obsid$instexpid"} = "$prod_dir/P$obsid$instexpid"
        ."EXPMAP$ref_band"."000.FIT";

      $padmasksets{$obsid}{$instexpid} = "$simop_dir/Sim_$instexpid"
        ."_paddedMask.fits";

      $detmasksets{$obsid}{$instexpid} = "$simop_dir/Sim_$instexpid"
        ."_edetectMask.fits";

      if (substr($instexpid, 0, 2) eq 'PN') {
        $evlist = "$prod_dir/P$obsid$instexpid"."PIEVLI0000.FIT";
      } else {
        $evlist = "$prod_dir/P$obsid$instexpid"."MIEVLI0000.FIT";
      }
      $evlists{$obsid}{$instexpid} = $evlist;

      foreach $band (@used_bands) {
        $expmaps{$obsid}{$instexpid}{$band} = "$prod_dir/P$obsid$instexpid"
          ."EXPMAP$band"."000.FIT";

        $images{$obsid}{$instexpid}{$band} = "$prod_dir/P$obsid$instexpid"
          ."IMAGE_$band"."000.FIT";

        $nonvigs{$obsid}{$instexpid}{$band} = "$simop_dir/Sim_$instexpid"
          ."_NONVIG_$band.fits";

        $bkgmaps{$obsid}{$instexpid}{$band} = "$simop_dir/Sim_$instexpid"
          ."_BKGMAP_$band.fits";
      } # end foreach $band (@used_bands)
    } # end foreach $instexpid (@{$instexpids{$obsid}})
  } # end foreach $obsid_root (@obsid_roots)

  #......................................................................
  # Decide where to go:

  if ($first_stage eq 'makeecfs') {
    goto MAKEECFS;

  } elsif ($first_stage eq 'makeskymasks') {
    goto MAKESKYMASKS;

  } elsif ($first_stage eq 'makenonvig') {
    goto MAKENONVIG;

  } elsif ($first_stage eq 'makebkgmap') {
    goto MAKEBKGMAP;

  } elsif ($first_stage eq 'emask') {
    goto EMASK;

  } elsif ($first_stage eq 'sensmaps') {
    goto SENSMAPS;

  } else {
SAS::error('badEntryStage', "Unrecognized entry stage $first_stage.");
  }

  #......................................................................
  # Start main routine:

  MAKEECFS: # Makes a text file which contains band-edges for the five standard energy bands, and ecf values for these bands for four filter settings for each of the three EPIC instruments. Note that a CIf is necessary, which is obtained from the observation in the location supplied via the --refobsidroot parameter above.

  # Params accessed:
  #	- ecffile
  #	- refobsidroot

  #     - prdssubdir
  #	- entrystage
  #	- finalstage
  #	- templateset
  #	- astest

  if ($with_ref_obsid) {
    $prod_dir  = "$ref_obsid_root/$prod_subdir";
  } else {
    $prod_dir  = "$obsid_roots[0]/$prod_subdir";
  }

  $cif = "$prod_dir/P$ref_obsid"."OBX000CALIND0000.FIT";
SAS::error('noCif', "Can't find CIF $cif.") if (!-e "$cif");
  $ENV{'SAS_CCF'} = "$cif";

SAS::error('noTemplate', "Can't find $template_set.") if (!-e "$template_set");

SAS::error('makeEcfsFailed', "Couldn't make ecfs.") if (&makeEcfs($ecf_file, $template_set));

  goto ENDD if ($last_stage eq 'makeecfs');

  MAKESKYMASKS: # This routine calculates the location and size of a great circle which encloses all the fields of view supplied via the parameter --insets. Exposure maps are used for this purpose. Note that it is not necessary to supply this algorithm with exposure maps from all bands of a particular observation/exposure since the instrument pointing information is all that is required.
  #
  # Individual sky masks are also calculated for each obs/instexpid combination.

  # Params accessed:
  #	- padsizearcsec
  #	- refband

  #	- obsidroots
  #	- entrystage
  #	- finalstage
  #	- prdssubdir
  #	- templateset
  #	- astest

  foreach $expmap (values %expmap_refs) {
SAS::error('noExpMap', "Couldn't find $expmap.") if (!-e "$expmap");
  }

SAS::error('noTemplate', "Can't find $template_set.") if (!-e "$template_set");

  $command = "eimsim "
  ."function=skytemplate "
  ."insets='".join(' ', values(%expmap_refs))."' "
  ."templateset=$template_set "
  ."padsizearcsec=$padsizearcsec "
  ;

SAS::error('makeSkyTemplateFailed', "Couldn't make overall sky mask.") if (&run("$command"));

  foreach $obsid_root (@obsid_roots) {
    $obsid = $obs_ids{$obsid_root};

    foreach $instexpid (@{$instexpids{$obsid}}) {
      $expmapset = $expmap_refs{"$obsid$instexpid"};
      $padmaskset = $padmasksets{$obsid}{$instexpid};

      $command = "eimsim "
      ."function=mask "
      ."expmapset=$expmapset "
      ."paddedmaskset=$padmaskset "
      ."padsizearcsec=$padsizearcsec "
      ;

SAS::error('makeSkyMaskFailed', "Couldn't make sky mask for $obsid, $instexpid.") if (&run("$command"));
    }
  }

  goto ENDD if ($last_stage eq 'makeskymasks');

  MAKENONVIG:

  # Params accessed:
  #	- ecffile

  #	- obsidroots
  #	- entrystage
  #	- finalstage
  #	- prdssubdir
  #	- intersubdir
  #	- astest

  foreach $obsid_root (@obsid_roots) {
    $obsid = $obs_ids{$obsid_root};

    $cif = $cifs{$obsid};
SAS::error('noCif', "Can't find CIF $cif.") if (!-e "$cif");
    $ENV{'SAS_CCF'} = "$cif";

    $ahf = $attitude_sets{$obsid};
SAS::error('noAttFile', "Can't find $ahf.") if (!-e "$ahf");

    foreach $instexpid (@{$instexpids{$obsid}}) {
      $eventset  = $evlists{$obsid}{$instexpid};
SAS::error('noEventList', "Can't find $eventset.") if (!-e "$eventset");

      foreach $band_id (0 .. $#used_bands) {
        $band = $used_bands[$band_id];
        $imageset  = $images{$obsid}{$instexpid}{$band};
SAS::error('noImage', "Can't find $imageset.") if (!-e "$imageset");
        $nonvigset = $nonvigs{$obsid}{$instexpid}{$band};

        $command = "eexpmap "
        ."imageset=$imageset "
        # just need it for wcs, dss and image size I guess.
        ."attitudeset=$ahf "
        ."eventset=$eventset "
        ."expimageset=$nonvigset "
        ."withvignetting=no "
        ."attrebin=2 "
        ."pimin=$idbandlo[$band_id] "
        ."pimax=$idbandhi[$band_id] "
        ;

SAS::error('makeNonVigMapFailed', "Couldn't make nonvig exposure map.") if (&run("$command"));
      }
    }
  }

  goto ENDD if ($last_stage eq 'makenonvig');

  MAKEBKGMAP:

  # Params accessed:
  #	- bkgrates
  #	- vigfractions

  #	- obsidroots
  #	- entrystage
  #	- finalstage
  #	- prdssubdir
  #	- intersubdir
  #	- astest

  foreach $obsid_root (@obsid_roots) {
    $obsid = $obs_ids{$obsid_root};

    $cif = $cifs{$obsid};
SAS::error('noCif', "Can't find CIF $cif.") if (!-e "$cif");
    $ENV{'SAS_CCF'} = "$cif";

    foreach $instexpid (@{$instexpids{$obsid}}) {
      foreach $band (@used_bands) {
        $nonvigset = $nonvigs{$obsid}{$instexpid}{$band};
SAS::error('noNonVigMap', "Can't find $nonvigset.") if (!$testflag && !-e "$nonvigset");
        $expmapset = $expmaps{$obsid}{$instexpid}{$band};
SAS::error('noExpMap', "Can't find $expmapset.") if (!-e "$expmapset");
        $bkgmapset = $bkgmaps{$obsid}{$instexpid}{$band};

        $command = "eimsim "
        ."function=makebkgmap "
        ."expmapset=$expmapset "
        ."nonvigset=$nonvigset "
        ."bkgmapset=$bkgmapset "
        ."vigbkgrate=$vigbkgrates{$band} "
        ."nonvigrate=$nonvigbkgrates{$band} "
        ;

SAS::error('makeBkgMapFailed', "Couldn't make background map.") if (&run("$command"));
      }
    }
  }

  goto ENDD if ($last_stage eq 'makebkgmap');

  EMASK:

  foreach $obsid_root (@obsid_roots) {
    $obsid = $obs_ids{$obsid_root};
    foreach $instexpid (@{$instexpids{$obsid}}) {
      $expmapset = $expmap_refs{"$obsid$instexpid"};
SAS::error('noExpMap', "Can't find ref-band $expmapset.") if (!-e "$expmapset");
      $detmaskset = $detmasksets{$obsid}{$instexpid};

      $command = "emask "
      ."expimageset=$expmapset "
      ."detmaskset=$detmaskset "
      ."threshold1=0.05 "
      ;

SAS::error('emaskFailed', "Couldn't run emask.") if (&run("$command"));
    }
  }

  goto ENDD if ($last_stage eq 'emask');

#  SENSMAPS:
#
#  # Params accessed:
#
#  foreach $obsid_root (@obsid_roots) {
#    $obsid = $obs_ids{$obsid_root};
#    $prod_dir = "$obsid_root/$prod_subdir";
#    $inter_dir = "$obsid_root/$inter_subdir";
#    @instexpids = keys(%{$expmaps{$obsid}});
#
#    $command = "esensmap_single "
#    ."prdsdir=$prod_dir "
#    ."interdir=$inter_dir "
#    ."instexpids='".join(' ', @instexpids)."' "
#    ."obsid=$obsid "
#    ."astest=$params{'astest'} "
#    ."cleanup=yes "
#    ;
#
#    print "\ninvoke $command\n\n";
#
#&quit("Couldn't make sensitivity maps.", 26) if (system("$command"));
## Not using &run() here because we want esensmap_single to be run whether
## $params{'astest'} is set or not. $params{'astest'} is passed to
## esensmap_single and so is also effective therein.
#  }

  ENDD:

}

  #---------------------------------------------------------------------
  sub run {
  #---------------------------------------------------------------------
  # $testflag should be defined globally.

    my ($command) = @_;
    my $status = 0;

    print "\ninvoke $command\n\n";
    if (!$testflag) {
      $status = system($command);
    }

  return $status;
  }

  #---------------------------------------------------------------------
  sub getEcfs {
  #---------------------------------------------------------------------
  # Loading default parameter values:

    my ($ecf_file) = @_;
    my ($line, %bands, $filter, @ecf_line, $num_bands, $band, $inst_uc, $band_index);

SAS::error('noEcfFile', "Can't find $ecf_file") if (!-e "$ecf_file");

    open (ECFS, "$ecf_file");

    while(){
      $line=<ECFS>;
SAS::error('ecfTooFewDataLines', "Too few data lines in $ecf_file") if (!defined($line));
    next if ($line =~ /^\s*$/);
    next if ($line =~ /^#/);
    last;
    }
    if ($line =~ /^numBands\s*=\s*(\d+)/) {
      $num_bands = $1;
    } else {
SAS::error('ecfBadStructureFile', "First non-comment line must contain the number of bands.");
    }

SAS::error('ecfTooFewBands', "Too few bands in $ecf_file") if ($num_bands <= 0);

    while(){
      $line=<ECFS>;
SAS::error('ecfTooFewDataLines', "Too few data lines in $ecf_file") if (!defined($line));
    next if ($line =~ /^\s*$/);
    next if ($line =~ /^#/);
    last;
    }
    $line =~ s/^\s+//;
    chomp($line);
    my @band_line = split('\s+', $line);
SAS::error('ecfWrongNumBandIds', "There should be $num_bands entries in the band-id line in $ecf_file") if ($num_bands != @band_line);

    @{$bands{'id'}} = ();
    foreach $band (1 .. $num_bands) {
      push @{$bands{'id'}}, shift @band_line;
    }

    while(){
      $line=<ECFS>;
SAS::error('ecfTooFewDataLines', "Too few data lines in $ecf_file") if (!defined($line));
    next if ($line =~ /^\s*$/);
    next if ($line =~ /^#/);
    last;
    }
    $line =~ s/^\s+//;
    chomp($line);
    @band_line = split('\s+', $line);
SAS::error('ecfWrongNumBandLos', "There should be $num_bands entries in the band-lo-E line in $ecf_file") if ($num_bands != @band_line);

    @{$bands{'lo'}} = ();
    foreach $band (1 .. $num_bands) {
      push @{$bands{'lo'}}, shift @band_line;
    }

    while(){
      $line=<ECFS>;
SAS::error('ecfTooFewDataLines', "Too few data lines in $ecf_file") if (!defined($line));
    next if ($line =~ /^\s*$/);
    next if ($line =~ /^#/);
    last;
    }
    $line =~ s/^\s+//;
    chomp($line);
    @band_line = split('\s+', $line);
SAS::error('ecfWrongNumBandHis', "There should be $num_bands entries in the band-hi-E line in $ecf_file") if ($num_bands != @band_line);

    @{$bands{'hi'}} = ();
    foreach $band_index (1 .. $num_bands) {
      push @{$bands{'hi'}}, shift @band_line;
    }

    #....................................................................
    # Now look for ecfs:

    my %ecf_info = ();
    my $num_instruments = 0;
    while(defined($line=<ECFS>)){
      if ($line =~ /^instrument\s*=\s*(\w\w)/) {
        ++$num_instruments;
        $inst_uc = $1;
        foreach $filter ('OPEN', 'THIN', 'MEDIUM', 'THICK') {
          while(){
            $line=<ECFS>;
SAS::error('ecfNoFilterData', "There was no data for filter=$filter, instrument=$inst_uc in $ecf_file") if (!defined($line));
          next if ($line =~ /^\s*$/);
          next if ($line =~ /^#/);
          last;
          }

          $line =~ s/^\s+//;
          chomp($line);

          @ecf_line = split('\s+', $line);

SAS::error('ecfWrongNumEcfs', "There should be $num_bands entries in the line for filter=$filter of instrument=$inst_uc in $ecf_file") if ($num_bands != @ecf_line);

#          foreach $band (1 .. $num_bands) {
#          foreach $band (@{$bands{'id'}}) {
          foreach $band_index (0 .. @{$bands{'id'}} - 1) {
$band = $bands{'id'}[$band_index];
            $ecf_info{$inst_uc}{$filter}{$band} = $ecf_line[$band_index];
          }
        }
      }
    }

    close(ECFS);

SAS::error('ecfNoInstrumentData', "There was no data for instrument=$inst_uc in $ecf_file") if ($num_instruments==0);

  return (%ecf_info, %bands);
  }

  #---------------------------------------------------------------------
  sub getEcfs_old {
  #---------------------------------------------------------------------
  # Loading default parameter values:

    my ($ecf_file) = @_;
    my ($line, %bands, $filter, @ecf_line);

SAS::error('noEcfFile', "Can't find $ecf_file") if (!-e "$ecf_file");

    open (ECFS, "$ecf_file");

    while(){
      $line=<ECFS>;
SAS::error('tooFewEcfDataLines', "Too few data lines in $ecf_file") if (!defined($line));
    next if ($line =~ /^\s*$/);
    next if ($line =~ /^#/);
    last;
    }
    $line =~ s/^\s+//;
    chomp($line);
    my @band_line = split('\s+', $line);
    my $num_bands = @band_line;

SAS::error('tooFewEcfBands', "Too few bands in $ecf_file") if ($num_bands <= 0);

    @{$bands{'lo'}} = ();
    foreach $band (1 .. $num_bands) {
      push @{$bands{'lo'}}, shift @band_line;
    }

    while(){
      $line=<ECFS>;
SAS::error('tooFewEcfDataLines', "Too few data lines in $ecf_file") if (!defined($line));
    next if ($line =~ /^\s*$/);
    next if ($line =~ /^#/);
    last;
    }
    $line =~ s/^\s+//;
    chomp($line);
    @band_line = split('\s+', $line);
SAS::error('badNumEcfBands', "The number of band-hi-edge entries in $ecf_file doesn't match the number of band-lo-edges") if (@band_line != $num_bands);

    @{$bands{'hi'}} = ();
    foreach $band (1 .. $num_bands) {
      push @{$bands{'hi'}}, shift @band_line;
    }

    #....................................................................
    # Now look for ecfs:

    while(defined($line=<ECFS>)){
print "00 $line 00\n";
      if ($line =~ /^instrument\s*=\s*(\w\w)/) {
        $inst_uc = $1;
        foreach $filter ('OPEN', 'THIN', 'MEDIUM', 'THICK') {
          while(){
            $line=<ECFS>;
SAS::error('ecfNoInstrumentData', "There was no data for instrument $inst_uc in $ecf_file") if (!defined($line));
          next if ($line =~ /^\s*$/);
          next if ($line =~ /^#/);
          last;
          }

          $line =~ s/^\s+//;
          chomp($line);

          @ecf_line = split('\s+', $line);

SAS::error('badNumEcfs', "The number of ecfs for filter $filter of instrument $inst_uc in $ecf_file doesn't match the number of band-lo-edges") if (@ecf_line != $num_bands);

          foreach $band (1 .. $num_bands) {
            $ecf_info{$inst_uc}{$filter}{$band} = $ecf_line[$band-1];
          }
        }
      }
    }

    close(ECFS);

  return (%ecf_info, %bands);
  }

  #----------------------------------------------------------------------
  sub makeEcfs {
  #----------------------------------------------------------------------
  # This writes ecfs to a text file. The ecfs, in the standard 5 pcms bands,
  # are calculated according to the source spectrum contained in $template_set.

    my ($ecf_file, $template_set) = @_;
    my ($inst, $filter, $band, $ecf, $spec_index);

    my $status = 0;

    ($status, $spec_index) = &getKeywordValue($template_set, 'LOGNLOGS'
      , 'SPECINDX');

  return $status if ($status < 0);

    my %band_lo = (1=>200, 2=> 500, 3=>2000, 4=>4500, 5=> 7500);
    my %band_hi = (1=>500, 2=>2000, 3=>4500, 4=>7500, 5=>12000);

    open(ECFS, ">$ecf_file");
    print ECFS "numBands=5\n";
    print ECFS "\n";
    print ECFS "# Band IDs:\n";
    print ECFS "\t1\t2\t3\t4\t5\n";
    print ECFS "# Band start (eV):\n";
    print ECFS "\t".join('\t', sort(values(%band_lo)))."\n";
    print ECFS "# Band stop (eV):\n";
    print ECFS "\t".join('\t', sort(values(%band_hi)))."\n";
    print ECFS "\n";

    foreach $inst ('M1', 'M2', 'PN') {
      print ECFS "instrument=$inst\n";
      foreach $filter ('OPEN', 'THIN1', 'MEDIUM', 'THICK') {
        print ECFS "# filter=$filter\n";
        foreach $band (1 .. 5) {
          $ecf = &calc_ecf($spec_index, $inst, $filter
            , $band_lo{$band}, $band_hi{$band});

SAS::error('ecfTooBig', "Computed ECF value too big.") if ($ecf >= 100.0);

          printf ECFS "\t%8.6f";
        }
        print ECFS "\n";
      }
      print ECFS "\n";
    }

    close(ECFS);

  return $status;
  }

  #----------------------------------------------------------------------
  sub makeEcfs_old {
  #----------------------------------------------------------------------
  # This writes ecfs to a text file. The ecfs, in the standard 5 pcms bands,
  # are calculated according to the source spectrum contained in $template_set.

    my ($ecf_file, $template_set) = @_;
    my ($inst, $filter, $band, $ecf, $spec_index);

    my $status = 0;

    ($status, $spec_index) = &getKeywordValue($template_set, 'LOGNLOGS'
      , 'SPECINDX');

  return $status if ($status < 0);

    my %band_lo = (1=>200, 2=> 500, 3=>2000, 4=>4500, 5=> 7500);
    my %band_hi = (1=>500, 2=>2000, 3=>4500, 4=>7500, 5=>12000);

    open(ECFS, ">$ecf_file");
    print ECFS "# Bands\t1\t2\t3\t4\t5\n";
    print ECFS "\n";
    print ECFS "# Band start (eV):\n";
    print ECFS "\t".join('\t', sort(values(%band_lo)))."\n";
    print ECFS "# Band stop (eV):\n";
    print ECFS "\t".join('\t', sort(values(%band_hi)))."\n";
    print ECFS "\n";

    foreach $inst ('EMOS1', 'EMOS2', 'EPN') {
      print ECFS "instrument=$inst\n";
      foreach $filter ('OPEN', 'THIN1', 'MEDIUM', 'THICK') {
        print ECFS "# filter=$filter\n";
        foreach $band (1 .. 5) {
          $ecf = &calc_ecf($spec_index, $inst, $filter
            , $band_lo{$band}, $band_hi{$band});

SAS::error('ecfTooBig', "Computed ECF value too big.") if ($ecf >= 100.0);

          printf ECFS "\t%8.6f";
        }
        print ECFS "\n";
      }
      print ECFS "\n";
    }

    close(ECFS);

  return $status;
  }

  #----------------------------------------------------------------------
  sub getKeywordValue {
  #----------------------------------------------------------------------

    my ($filename, $extname, $kwd) = @_;
    my ($rump, $command);

    my $value = 0; # default

  return (-1, 0) if (!defined($filename) || !-e "$filename");
  return (-2, 0) if (!defined($extname));
  return (-3, 0) if (!defined($kwd) || $kwd eq '');

    if ($extname =~ /^\d+$/) {
      $command = "fkeyprint $filename+$extname $kwd | grep =";
    } else {
      $command = "fkeyprint $filename'[$extname]' $kwd | grep =";
    }

    my @lines = `$command`;
    if (defined($lines[0])){
      if ($lines[0] =~ /^${kwd}\s*=\s*/) {
        $rump = $';
        if ($rump =~ /^'(.+)'/ || $rump =~ /(\S+)/) {
          $value = $1;
        } else {
  return (-4, 0);
        }
      } else {
  return (-5, 0);
      }
    } else {
  return (-6, 0);
    }

  return (0, $value);
  }

  #----------------------------------------------------------------------
  sub calc_ecf {
  #----------------------------------------------------------------------

    my ($spec_index, $inst, $filter, $band_lo, $band_hi) = @_;
    my ($junk);

    my $command = "eimsimtest "
    ."function=printecf "
    ."specindex=$spec_index "
    ."inst=$inst "
    ."filter=$filter "
    ."elo=$band_lo "
    ."ehi=$band_hi "
    ;

    my $ecf = 0; # default

    print "\ninvoke $command\n\n";
    if (!$testflag) {
      my @lines = `$command`;
      foreach(@lines){
        chomp;
        if (/^result:/) {
          ($junk, $ecf) = split('=', $');
        }
      }
    }

    if (!$testflag) {
      my @lines = `$command`;
      chomp($lines[0]);
      if ($lines[0] =~ /^\s*result: ecf =\s*/) {
        if ($' =~ /\s*s\^-1/) {
          $ecf = $`;
        }
      }
    }

SAS::error('ecfCalcFailed', "Couldn't calc ecf for inst $inst, energy $band_lo - $band_hi.") if (!$testflag && $ecf<=0);

  return $ecf;
  }

