!
! Module: flux_transforms
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines to convert to and from flux to other scales.
! 
! Comments: TODO:
! 
! $Id: flux_transforms.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
!.......................................................................
! These transforms offer ways from transforming x-ray source strengths
! between various units. The energy density of the radiation dS/dE(1) at
! a photon energy of 1 eV is here employed as the fundamental unit of
! source strength. Routines are provided to allow the user to transform
! between this unit and both flux S (in ergs s^-1 cm^-2) and count rate
! R (in counts/sec).
!
! The theory behind these transforms is as follows. Let us assume that a
! source has a spectrum of the form
!
!	p(E) = p(1) * E**gamma
!
! where p(E) is the count flux spectral density in counts s^-1 cm^-2
! eV^-1 impinging on the camera (ie, no losses due to exposure are
! factored in yet), p(1) is of course p at E = 1 eV, E is energy in eV,
! and gamma is the spectral index. The flux density dS/dE in ergs s^-1
! cm^-2 eV^-1 is therefore
!
!	dS/dE = p * E * ergs/eV
!
!	dS/dE = p(1) * E**gamma * E * ergs/eV
!
!	dS/dE = p(1) * ergs/eV * E**(gamma+1)
!
!	dS/dE = dS/dE(1) * E**(gamma+1)
!
! where
!
!	dS/dE(1) = p(1) * 10**7 * qe
!
! and qe is the electronic charge in coulombs. S in ergs s^-1 cm^-2 is
! thus given by
!
!	            EeVhi
!	             /
!	S = dS/dE(1) | dE E**(gamma+1)
!	             /
!	           EeVlo
!
!	     dS/dE(1)
!	  = ------------ (EeVhi**(gamma + 2) - EeVlo**(gamma + 2))
!	     gamma + 2
!
! Hence the inverse equation giving dS/dE(1) can be written in terms
! either of dS/dE as
!
!	               dS/dE
!	dS/dE(1) = --------------
!	            E**(gamma+1)
!
! or in terms of the flux S in a finite energy band as
!
!
!	                      S * (gamma + 2)
!	dS/dE(1) = -------------------------------------------.
!	            (EeVhi**(gamma + 2) - EeVlo**(gamma + 2))
!
!
! Another way in which we might want to specify the source strength is by
! its count rate as detected by the appropriate EPIC camera. The count
! rate density dR/dE in counts s^-1 eV^-1 is given by
!
!	         dS/dE(1)
!	dR/dE = ---------- E**gamma EA(E) F(E)
!	         10**7 qe
!
! where EA(E) is the effective area in cm^2 and F(E) is the filter
! transmission fraction (+ any other energy-dependent attenuation); hence
! the total count rate R in the band EeVlo to EeVhi is
!
!	               EeVhi
!	     dS/dE(1)   /
!	R = ----------  | dE E**gamma EA(E) F(E)
!	     10**7 qe   /
!	              EeVlo
!
!.......................................................................

module flux_transforms

  use dal
  use cal
  use errorHandling

  use constants
  use array_utils ! for trapezoidSum().
  use exposure	! for sampleEnergyRange(), getEffectiveArea() etc.
  use dss	! for RangeT

  implicit none 

  private 

  public fluxToFluxDensity0
  public fluxDensityToFluxDensity0
  public fluxDensity0ToFlux
  public fluxDensity0ToFluxDensity
  public fluxDensity0ToRate
  public fluxDensity0ToRateDensity
  public rateToFluxDensity0
  public rateDensityToFluxDensity0
  public calcEcf

  interface fluxDensity0ToFlux
    module procedure fluxDensity0ToFluxScalar
    module procedure fluxDensity0ToFluxVector, fluxDensity0ToFluxArray
  end interface

  interface fluxDensity0ToRate
    module procedure fluxDensity0ToRateScalar
    module procedure fluxDensity0ToRateVector, fluxDensity0ToRateArray
  end interface

  interface rateToFluxDensity0
    module procedure rateToFluxDensity0Scalar
    module procedure rateToFluxDensity0Vector
  end interface

  interface fluxToFluxDensity0
    module procedure fluxToFluxDensity0Scalar, fluxToFluxDensity0Range
    module procedure fluxToFluxDensity0Vector, fluxToFluxDensity0Array
  end interface

  interface calcEcf
    module procedure calcEcfCore
    module procedure calcEcfWrapper
  end interface

  integer(int16), parameter :: numEsamples = 100

contains 

  !---------------------------------------------------------------------
  function fluxToFluxDensity0Range(eventSpectrumPowerIndex, EeVranges&
  , flux) result(fluxDensity0)
  !---------------------------------------------------------------------
  ! Uses the equation
  !
  !	                     S
  !	dS/dE(1) = ---------------------.
  !
  !	              /
  !	              | dE E**(gamma+1)
  !	              /
  !	           EeVrange
  !
  !.....................................................................

    implicit none

    real(single),   intent(in) :: flux
    real(double),   intent(in) :: eventSpectrumPowerIndex
    type(RangeT),   intent(in) :: EeVranges(:)
    real(single)               :: fluxDensity0

    real(single) :: gammaPlus2, fluxIntegral
    integer :: numRanges, rangeNum
    real(double) :: EeVlo, EeVhi

    !-------------------------------------------------------------------
    ! Start

    numRanges = size(EeVranges)
    gammaPlus2 = eventSpectrumPowerIndex + 2.0
    fluxIntegral = 0.0
    do rangeNum = 1, numRanges
      EeVlo = EeVranges(rangeNum)%lower%value
      EeVhi = EeVranges(rangeNum)%upper%value
      fluxIntegral = fluxIntegral + EeVhi**gammaPlus2 - EeVlo**gammaPlus2
    end do
    fluxDensity0 = flux * (gammaPlus2 / fluxIntegral)

  end function fluxToFluxDensity0Range

  !---------------------------------------------------------------------
  function fluxToFluxDensity0Scalar(eventSpectrumPowerIndex, EeVlo, EeVhi&
  , flux) result(fluxDensity0)
  !---------------------------------------------------------------------
  ! Uses the equation
  !
  !	                      S * (gamma + 2)
  !	dS/dE(1) = -------------------------------------------.
  !	            (EeVhi**(gamma + 2) - EeVlo**(gamma + 2))
  !
  !.....................................................................

    implicit none

    real(single),   intent(in) :: flux
    real(double),   intent(in) :: eventSpectrumPowerIndex, EeVlo, EeVhi
    real(single)               :: fluxDensity0

    real(single) :: gammaPlus2, fluxIntegral

    !-------------------------------------------------------------------
    ! Start

    gammaPlus2 = eventSpectrumPowerIndex + 2.0
    fluxIntegral = EeVhi**gammaPlus2 - EeVlo**gammaPlus2
    fluxDensity0 = flux * (gammaPlus2 / fluxIntegral)

  end function fluxToFluxDensity0Scalar

  !---------------------------------------------------------------------
  function fluxToFluxDensity0Vector(eventSpectrumPowerIndex, EeVlo, EeVhi&
  , flux) result(fluxDensity0)
  !---------------------------------------------------------------------
  ! Uses the equation
  !
  !	                      S * (gamma + 2)
  !	dS/dE(1) = -------------------------------------------.
  !	            (EeVhi**(gamma + 2) - EeVlo**(gamma + 2))
  !
  !.....................................................................

    implicit none

    real(single),   intent(in) :: flux(:)
    real(double),   intent(in) :: eventSpectrumPowerIndex, EeVlo, EeVhi
    real(single)               :: fluxDensity0(size(flux))

    real(single) :: gammaPlus2, fluxIntegral

    !-------------------------------------------------------------------
    ! Start

    gammaPlus2 = eventSpectrumPowerIndex + 2.0
    fluxIntegral = EeVhi**gammaPlus2 - EeVlo**gammaPlus2
    fluxDensity0 = flux * (gammaPlus2 / fluxIntegral)

  end function fluxToFluxDensity0Vector

  !---------------------------------------------------------------------
  function fluxToFluxDensity0Array(eventSpectrumPowerIndex, EeVlo, EeVhi, flux)&
  result(fluxDensity0)
  !---------------------------------------------------------------------
  ! Uses the equation
  !
  !	                      S * (gamma + 2)
  !	dS/dE(1) = -------------------------------------------.
  !	            (EeVhi**(gamma + 2) - EeVlo**(gamma + 2))
  !
  !.....................................................................

    implicit none

    real(single),   intent(in) :: flux(:,:)
    real(double),   intent(in) :: eventSpectrumPowerIndex, EeVlo, EeVhi
    real(single)               :: fluxDensity0(size(flux, 1), size(flux, 2))

    real(single) :: gammaPlus2, fluxIntegral

    !-------------------------------------------------------------------
    ! Start

    gammaPlus2 = eventSpectrumPowerIndex + 2.0
    fluxIntegral = EeVhi**gammaPlus2 - EeVlo**gammaPlus2
    fluxDensity0 = flux * (gammaPlus2 / fluxIntegral)

  end function fluxToFluxDensity0Array

  !---------------------------------------------------------------------
  function fluxDensityToFluxDensity0(eventSpectrumPowerIndex, EeV, fluxDensity)&
  result(fluxDensity0)
  !---------------------------------------------------------------------
  ! Uses the equation
  !
  !	               dS/dE
  !	dS/dE(1) = --------------.
  !	            E**(gamma+1)
  !
  !.....................................................................

    implicit none

    real(single),   intent(in) :: fluxDensity
    real(double),   intent(in) :: eventSpectrumPowerIndex, EeV
    real(single)               :: fluxDensity0

    !-------------------------------------------------------------------
    ! Start

    fluxDensity0 = fluxDensity / (EeV**(eventSpectrumPowerIndex + 1))

  end function fluxDensityToFluxDensity0

  !---------------------------------------------------------------------
  function fluxDensity0ToFluxScalar(eventSpectrumPowerIndex, EeVlo, EeVhi&
  , fluxDensity0) result(flux)
  !---------------------------------------------------------------------
  ! Uses the equation
  !
  !	     dS/dE(1)
  !	S = ----------- (EeVhi**(gamma + 2) - EeVlo**(gamma + 2)).
  !	     gamma + 2
  !
  !.....................................................................

    implicit none

    real(single),   intent(in) :: fluxDensity0
    real(double),   intent(in) :: eventSpectrumPowerIndex, EeVlo, EeVhi
    real(single)               :: flux

    real(single) :: gammaPlus2, fluxIntegral

    !-------------------------------------------------------------------
    ! Start

    gammaPlus2 = eventSpectrumPowerIndex + 2.0
    fluxIntegral = EeVhi**gammaPlus2 - EeVlo**gammaPlus2
    flux = fluxDensity0 * fluxIntegral / gammaPlus2

  end function fluxDensity0ToFluxScalar

  !---------------------------------------------------------------------
  function fluxDensity0ToFluxVector(eventSpectrumPowerIndex, EeVlo, EeVhi&
  , fluxDensity0) result(flux)
  !---------------------------------------------------------------------
  ! Uses the equation
  !
  !	     dS/dE(1)
  !	S = ----------- (EeVhi**(gamma + 2) - EeVlo**(gamma + 2)).
  !	     gamma + 2
  !
  !.....................................................................

    implicit none

    real(single),   intent(in) :: fluxDensity0(:)
    real(double),   intent(in) :: eventSpectrumPowerIndex, EeVlo, EeVhi
    real(single)               :: flux(size(fluxDensity0))

    real(single) :: gammaPlus2, fluxIntegral

    !-------------------------------------------------------------------
    ! Start

    gammaPlus2 = eventSpectrumPowerIndex + 2.0
    fluxIntegral = EeVhi**gammaPlus2 - EeVlo**gammaPlus2
    flux = fluxDensity0 * fluxIntegral / gammaPlus2

  end function fluxDensity0ToFluxVector

  !---------------------------------------------------------------------
  function fluxDensity0ToFluxArray(eventSpectrumPowerIndex, EeVlo, EeVhi&
  , fluxDensity0) result(flux)
  !---------------------------------------------------------------------
  ! Uses the equation
  !
  !	     dS/dE(1)
  !	S = ----------- (EeVhi**(gamma + 2) - EeVlo**(gamma + 2)).
  !	     gamma + 2
  !
  !.....................................................................

    implicit none

    real(single),   intent(in) :: fluxDensity0(:,:)
    real(double),   intent(in) :: eventSpectrumPowerIndex, EeVlo, EeVhi
    real(single)               :: flux(size(fluxDensity0, 1)&
                                     , size(fluxDensity0, 2))

    real(single) :: gammaPlus2, fluxIntegral

    !-------------------------------------------------------------------
    ! Start

    gammaPlus2 = eventSpectrumPowerIndex + 2.0
    fluxIntegral = EeVhi**gammaPlus2 - EeVlo**gammaPlus2
    flux = fluxDensity0 * fluxIntegral / gammaPlus2

  end function fluxDensity0ToFluxArray

  !---------------------------------------------------------------------
  function fluxDensity0ToFluxDensity(eventSpectrumPowerIndex, EeV&
  , fluxDensity0) result(fluxDensity)
  !---------------------------------------------------------------------
  ! Uses the equation
  !
  !	dS/dE = dS/dE(1) * E**(gamma+1).
  !
  !.....................................................................

    implicit none

    real(single),   intent(in) :: fluxDensity0
    real(double),   intent(in) :: eventSpectrumPowerIndex, EeV
    real(single)               :: fluxDensity

    !-------------------------------------------------------------------
    ! Start

    fluxDensity = fluxDensity0 * EeV**(eventSpectrumPowerIndex + 1)

  end function fluxDensity0ToFluxDensity

  !---------------------------------------------------------------------
  function fluxDensity0ToRateArray(eventSpectrumPowerIndex, EeVlo, EeVhi&
  , fluxDensity0) result(rate)
  !---------------------------------------------------------------------
  ! Uses the equation
  !
  !	               EeVhi
  !	     dS/dE(1)   /
  !	R = ----------  | dE E**gamma EA(E) F(E).
  !	     10**7 qe   /
  !	              EeVlo
  !.....................................................................
  ! NOTE! The cal MUST be set up for the correct instrument AND filter
  ! before this function is called.
  !.....................................................................

    implicit none

    real(single),   intent(in)  :: fluxDensity0(:,:)
    real(double),   intent(in)  :: eventSpectrumPowerIndex, EeVlo, EeVhi
    real(single)                :: rate(size(fluxDensity0, 1), size(fluxDensity0, 2))

    real(double), dimension(numEsamples) :: EeVsamples
    real(single), dimension(numEsamples) ::&
      effArea, filterTrans, spectrum, qe
    integer(int16) :: Ei
    real(single) :: multiplier
    logical(bool) :: usePattern(0:0)

    !-------------------------------------------------------------------
    ! Start

    call sampleEnergyRange(EeVlo, EeVhi, EeVsamples) ! in ssclib/exposure
    call sampleEnergyRange(EeVlo, EeVhi, EeVsamples) ! in ssclib/exposure
    call getEffectiveArea(EeVsamples, effArea) ! in ssclib/exposure
    call getFilterTransmission(EeVsamples, filterTrans) ! in ssclib/exposure
    usePattern(0) = .true.
    call getQe(EeVsamples, usePattern, qe) ! in ssclib/exposure

    do Ei = 1, numEsamples
      spectrum(Ei) = EeVsamples(Ei)**eventSpectrumPowerIndex
    end do

    multiplier = trapezoidSum(real(EeVsamples, single), spectrum * effArea * filterTrans * qe)
    ! in ssclib/array_utils

    rate = (fluxDensity0 * 1.0e-7 / ElectronQ) * multiplier

  end function fluxDensity0ToRateArray

  !---------------------------------------------------------------------
  function fluxDensity0ToRateVector(eventSpectrumPowerIndex, EeVlo, EeVhi&
  , fluxDensity0) result(rate)
  !---------------------------------------------------------------------
  ! Uses the equation
  !
  !	               EeVhi
  !	     dS/dE(1)   /
  !	R = ----------  | dE E**gamma EA(E) F(E).
  !	     10**7 qe   /
  !	              EeVlo
  !.....................................................................
  ! NOTE! The cal MUST be set up for the correct instrument AND filter
  ! before this function is called.
  !.....................................................................

    implicit none

    real(single),   intent(in)  :: fluxDensity0(:)
    real(double),   intent(in)  :: eventSpectrumPowerIndex, EeVlo, EeVhi
    real(single)                :: rate(size(fluxDensity0))

    real(double), dimension(numEsamples) :: EeVsamples
    real(single), dimension(numEsamples) ::&
      effArea, filterTrans, spectrum, qe
    integer(int16) :: Ei
    real(single) :: multiplier
    logical(bool) :: usePattern(0:0)

    !-------------------------------------------------------------------
    ! Start

    call sampleEnergyRange(EeVlo, EeVhi, EeVsamples) ! in ssclib/exposure
    call getEffectiveArea(EeVsamples, effArea) ! in ssclib/exposure
    call getFilterTransmission(EeVsamples, filterTrans) ! in ssclib/exposure
    usePattern(0) = .true.
    call getQe(EeVsamples, usePattern, qe) ! in ssclib/exposure

    do Ei = 1, numEsamples
      spectrum(Ei) = EeVsamples(Ei)**eventSpectrumPowerIndex
    end do

    multiplier = trapezoidSum(real(EeVsamples, single), spectrum * effArea * filterTrans * qe)
    ! in ssclib/array_utils

    rate = (fluxDensity0 * 1.0e-7 / ElectronQ) * multiplier

  end function fluxDensity0ToRateVector

  !---------------------------------------------------------------------
  function fluxDensity0ToRateScalar(eventSpectrumPowerIndex, EeVlo, EeVhi&
  , fluxDensity0) result(rate)
  !---------------------------------------------------------------------
  ! Uses the equation
  !
  !	               EeVhi
  !	     dS/dE(1)   /
  !	R = ----------  | dE E**gamma EA(E) F(E).
  !	     10**7 qe   /
  !	              EeVlo
  !.....................................................................
  ! NOTE! The cal MUST be set up for the correct instrument AND filter
  ! before this function is called.
  !.....................................................................

    implicit none

    real(single),   intent(in)  :: fluxDensity0
    real(double),   intent(in)  :: eventSpectrumPowerIndex, EeVlo, EeVhi
    real(single)                :: rate

    real(double), dimension(numEsamples) :: EeVsamples
    real(single), dimension(numEsamples) ::&
      effArea, filterTrans, spectrum, qe
    integer(int16) :: Ei
    real(single) :: multiplier
    logical(bool) :: usePattern(0:0)

!real(single), pointer ::&
!  EeVsamplesData(:), effAreaData(:), filterTransData(:), qeData(:),&
!  spectrumData(:)
!type(DataSetT) :: set
!type(TableT) :: tab

!integer(int32) :: instrumentId, filterId

    !-------------------------------------------------------------------
    ! Start

    call sampleEnergyRange(EeVlo, EeVhi, EeVsamples) ! in ssclib/exposure
    call getEffectiveArea(EeVsamples, effArea) ! in ssclib/exposure
    call getFilterTransmission(EeVsamples, filterTrans) ! in ssclib/exposure
    usePattern(0) = .true.
    call getQe(EeVsamples, usePattern, qe) ! in ssclib/exposure

    do Ei = 1, numEsamples
      spectrum(Ei) = EeVsamples(Ei)**eventSpectrumPowerIndex
    end do

!set = dataset('fe_dump.ds', CREATE)
!tab = addTable(set, 'FUN_E', int(numEsamples))
!EeVsamplesData  => real32Data(addColumn(tab, 'E_EV',     REAL32))
!effAreaData     => real32Data(addColumn(tab, 'EFF_AREA', REAL32))
!filterTransData => real32Data(addColumn(tab, 'FILTRANS', REAL32))
!qeData          => real32Data(addColumn(tab, 'QE',       REAL32))
!spectrumData    => real32Data(addColumn(tab, 'SPECTRUM', REAL32))
!
!EeVsamplesData  = EeVsamples
!effAreaData     = effArea
!filterTransData = filterTrans
!qeData          = qe
!spectrumData    = spectrum
!
!call release(set)

    multiplier = trapezoidSum(real(EeVsamples, single), spectrum * effArea * filterTrans * qe)
    ! in ssclib/array_utils

    rate = (fluxDensity0 * 1.0e-7 / ElectronQ) * multiplier

!call CAL_getState(instrument=instrumentId, filterId=filterId)
!print *, 'eventSpectrumPowerIndex, fluxDensity0, EeVlo, EeVhi: '&
!,         eventSpectrumPowerIndex, fluxDensity0, EeVlo, EeVhi
!print *, 'instrumentId, filterId, EeVsamples(50), effArea(50): '&
!,         instrumentId, filterId, EeVsamples(50), effArea(50)
!print *, 'filterTrans(50), qe(50), spectrum(50), multiplier, rate: '&
!,         filterTrans(50), qe(50), spectrum(50), multiplier, rate
!print *, ''

  end function fluxDensity0ToRateScalar

  !---------------------------------------------------------------------
  function fluxDensity0ToRateDensity(eventSpectrumPowerIndex, EeV&
  , fluxDensity0) result(rateDensity)
  !---------------------------------------------------------------------
  ! Uses the equation
  !
  !	         dS/dE(1)
  !	dR/dE = ---------- E**gamma EA(E) F(E).
  !	         10**7 qe
  !.....................................................................
  ! NOTE! The cal MUST be set up for the correct instrument AND filter
  ! before this function is called.
  !.....................................................................

    implicit none

    real(single),   intent(in)  :: fluxDensity0
    real(double),   intent(in)  :: eventSpectrumPowerIndex, EeV
    real(single)                :: rateDensity

    real(double), dimension(1) :: EeVsamples
    real(single), dimension(1) ::&
      effArea, filterTrans, qe
    logical(bool) :: usePattern(0:0)

    !-------------------------------------------------------------------
    ! Start

    EeVsamples(1) = EeV
    call getEffectiveArea(EeVsamples, effArea) ! in ssclib/exposure
    call getFilterTransmission(EeVsamples, filterTrans) ! in ssclib/exposure
    usePattern(0) = .true.
    call getQe(EeVsamples, usePattern, qe) ! in ssclib/exposure

    rateDensity = (fluxDensity0 * 1.0e-7 / ElectronQ)&
      * EeV**eventSpectrumPowerIndex * effArea(1)&
      * filterTrans(1) * qe(1)

  end function fluxDensity0ToRateDensity

  !---------------------------------------------------------------------
  function rateToFluxDensity0Scalar(eventSpectrumPowerIndex, EeVlo, EeVhi&
  , rate) result(fluxDensity0)
  !---------------------------------------------------------------------
  ! Inverts the equation
  !
  !	               EeVhi
  !	     dS/dE(1)   /
  !	R = ----------  | dE E**gamma EA(E) F(E).
  !	     10**7 qe   /
  !	              EeVlo
  !
  !.....................................................................
  ! NOTE! The cal MUST be set up for the correct instrument AND filter
  ! before this function is called.
  !.....................................................................

    implicit none

    real(single),   intent(in)  :: rate
    real(double),   intent(in)  :: eventSpectrumPowerIndex, EeVlo, EeVhi
    real(single)                :: fluxDensity0

    real(double), dimension(numEsamples) :: EeVsamples
    real(single), dimension(numEsamples) ::&
      effArea, filterTrans, spectrum, qe
    integer(int16) :: Ei
    real(single) :: multiplier
    logical(bool) :: usePattern(0:0)

    !-------------------------------------------------------------------
    ! Start

    call sampleEnergyRange(EeVlo, EeVhi, EeVsamples) ! in ssclib/exposure
    call getEffectiveArea(EeVsamples, effArea) ! in ssclib/exposure
    call getFilterTransmission(EeVsamples, filterTrans) ! in ssclib/exposure
    usePattern(0) = .true.
    call getQe(EeVsamples, usePattern, qe) ! in ssclib/exposure

    do Ei = 1, numEsamples
      spectrum(Ei) = EeVsamples(Ei)**eventSpectrumPowerIndex
    end do

    multiplier = trapezoidSum(real(EeVsamples, single), spectrum * effArea * filterTrans * qe)
    ! in ssclib/array_utils

    fluxDensity0 = rate * 1.0e7 * ElectronQ / multiplier

  end function rateToFluxDensity0Scalar

  !---------------------------------------------------------------------
  function rateToFluxDensity0Vector(eventSpectrumPowerIndex, EeVlo, EeVhi&
  , rate) result(fluxDensity0)
  !---------------------------------------------------------------------
  ! Inverts the equation
  !
  !	               EeVhi
  !	     dS/dE(1)   /
  !	R = ----------  | dE E**gamma EA(E) F(E).
  !	     10**7 qe   /
  !	              EeVlo
  !
  !.....................................................................
  ! NOTE! The cal MUST be set up for the correct instrument AND filter
  ! before this function is called.
  !.....................................................................

    implicit none

    real(single),   intent(in)  :: rate(:)
    real(double),   intent(in)  :: eventSpectrumPowerIndex, EeVlo, EeVhi
    real(single)                :: fluxDensity0(size(rate))

    real(double), dimension(numEsamples) :: EeVsamples
    real(single), dimension(numEsamples) ::&
      effArea, filterTrans, spectrum, qe
    integer(int16) :: Ei
    real(single) :: multiplier
    logical(bool) :: usePattern(0:0)

    !-------------------------------------------------------------------
    ! Start

    call sampleEnergyRange(EeVlo, EeVhi, EeVsamples) ! in ssclib/exposure
    call getEffectiveArea(EeVsamples, effArea) ! in ssclib/exposure
    call getFilterTransmission(EeVsamples, filterTrans) ! in ssclib/exposure
    usePattern(0) = .true.
    call getQe(EeVsamples, usePattern, qe) ! in ssclib/exposure

    do Ei = 1, numEsamples
      spectrum(Ei) = EeVsamples(Ei)**eventSpectrumPowerIndex
    end do

    multiplier = trapezoidSum(real(EeVsamples, single), spectrum * effArea * filterTrans * qe)
    ! in ssclib/array_utils

    fluxDensity0 = rate * 1.0e7 * ElectronQ / multiplier

  end function rateToFluxDensity0Vector

  !---------------------------------------------------------------------
  function rateDensityToFluxDensity0(eventSpectrumPowerIndex, EeV&
  , rateDensity) result(fluxDensity0)
  !---------------------------------------------------------------------
  ! Inverts the equation
  !
  !	         dS/dE(1)
  !	dR/dE = ---------- E**gamma EA(E) F(E).
  !	         10**7 qe
  !.....................................................................
  ! NOTE! The cal MUST be set up for the correct instrument AND filter
  ! before this function is called.
  !.....................................................................

    implicit none

    real(single),   intent(in)  :: rateDensity
    real(double),   intent(in)  :: eventSpectrumPowerIndex, EeV
    real(single)                :: fluxDensity0

    real(double), dimension(1) :: EeVsamples
    real(single), dimension(1) ::&
      effArea, filterTrans, qe
    logical(bool) :: usePattern(0:0)

    !-------------------------------------------------------------------
    ! Start

    EeVsamples(1) = EeV
    call getEffectiveArea(EeVsamples, effArea) ! in ssclib/exposure
    call getFilterTransmission(EeVsamples, filterTrans) ! in ssclib/exposure
    usePattern(0) = .true.
    call getQe(EeVsamples, usePattern, qe) ! in ssclib/exposure

    fluxDensity0 = rateDensity * 1.0e7 * ElectronQ&
      / (EeV**eventSpectrumPowerIndex * effArea(1)&
      * filterTrans(1) * qe(1))

  end function rateDensityToFluxDensity0

  !---------------------------------------------------------------------
  function calcEcfCore(eventSpectrumPowerIndex, EeVlo, EeVhi) result(ecf)
  !---------------------------------------------------------------------
  ! This calculates (for a given combination of instrument and filter)
  ! the ecf values which are used to convert between source count rate R
  ! (in counts/s) and flux S (in erg/cm^2/s). An ecf value is defined
  ! such that
  !
  !	R = 10^11 * ECF * S.
  !
  ! Both R and S are assumed to be measured within the same energy band.
  !.....................................................................
  ! NOTE! The cal MUST be set up for the correct instrument AND filter
  ! before this function is called.
  !.....................................................................

    implicit none

    real(double), intent(in)  :: eventSpectrumPowerIndex, EeVlo, EeVhi
    real(single)              :: ecf, fluxDensity0

    !-------------------------------------------------------------------
    ! Start

    fluxDensity0 = fluxToFluxDensity0(eventSpectrumPowerIndex&
      , EeVlo, EeVhi, 1.0e-11)
    ecf = fluxDensity0ToRate(eventSpectrumPowerIndex&
      , EeVlo, EeVhi, fluxDensity0)

  end function calcEcfCore

  !---------------------------------------------------------------------
  function calcEcfWrapper(eventSpectrumPowerIndex, EeVlo, EeVhi&
  , instrumentStr, filterStr) result(ecf)
  !---------------------------------------------------------------------
  ! This calculates (for a given combination of instrument and filter)
  ! the ecf values which are used to convert between source count rate R
  ! (in counts/s) and flux S (in erg/cm^2/s). An ecf value is defined
  ! such that
  !
  !	R = 10^11 * ECF * S.
  !
  ! Both R and S are assumed to be measured within the same energy band.
  !.....................................................................

    implicit none

    real(double), intent(in)  :: eventSpectrumPowerIndex, EeVlo, EeVhi
    character(*), intent(in)  :: instrumentStr, filterStr
    real(single)              :: ecf

    !-------------------------------------------------------------------
    ! Start

    call CAL_setState(instrument=instId(instrumentStr)&
      , filterId=filterId(filterStr))

    ecf = calcEcfCore(eventSpectrumPowerIndex, EeVlo, EeVhi)

  end function calcEcfWrapper

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module flux_transforms
