!+
! Module: simutils_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This is a collection of routines used to make simulated xmm images.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module simutils_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.

  use dal
  use cal	! contains 'use caloaldefs' and 'use caltypes'
  use param  
  use errorHandling
  use mathutils_mod
  use constants
  use fitsutils_mod
  use temp_ssclib
  use temp_simulations
  use sscmisc

  implicit none 

  private 

  type skyPositionType
    real(double) :: ra, dec  ! ra dec of position in sky. (radians epoch 2000)
  end type skyPositionType

  type srcTypeA
    integer(int32) :: ml_id, box_id
    real(single) :: scts, scts_err, rate, rate_err, xima, yima, xima_err, yima_err
  end type srcTypeA

  public simutils

contains 

  !---------------------------------------------------------------------
  subroutine simutils
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    select case(stringParameter('function'))
    case('mask')
      call addPadToMask
    case('srclist')
      call makeSimSrcList
    case('image')
      call makeSimImg
    case('eventize')
      call eventizeImg
    case('fluxcorrect')
      call correctDetectedFluxes
    case('compare')
      call compareDetWithSim
    case('syserror')
      call calcSysErrors
    case('merge')
      call makeMasterList
    case('extract')
      call extractMatches
    case('completeness')
      call calcCompletenessHistograms
    case('reliability')
      call calcReliabilityHistogram
    case default
call error('badStringPar','')
    end select

  end subroutine simutils
  
  !---------------------------------------------------------------------
  subroutine addPadToMask
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    call padOutMask(trim(stringParameter('intables', 0))&
    , paddedMaskName='padded_mask.ds', padSizeArcsec_arg=100.0)

  end subroutine addPadToMask

  !---------------------------------------------------------------------
  subroutine makeSimSrcList
  !---------------------------------------------------------------------

    implicit none

    character(255) :: maskSetName
    character(255) :: sourceListName
    real(single) :: brightestRate, faintestRate, sourceDensityArcsec
    integer(int32) :: numSources

    !-------------------------------------------------------------------
    ! Start


    maskSetName          = stringParameter('maskset')
    sourceListName       = stringParameter('simsrclistset')

    faintestRate  = realParameter('faintlimit') ! 5e-5 is a good value.

!    sourceDensityArcsec = realParameter('real', 1)
!    call makeSimulatedSourceList(sourceListName, maskSetName, faintestRate&
!    , sourceDensityArcsec=sourceDensityArcsec)

    numSources = intParameter('nsources')
    call makeSimulatedSourceList(sourceListName, maskSetName, faintestRate&
    , numSources=numSources)!!!, maxExtentArcsec=30.0)

!    brightestRate = realParameter('real', 1)
!    call makeSimulatedSourceList(sourceListName, maskSetName, faintestRate&
!    , brightestExpectedRate=brightestRate)


  end subroutine makeSimSrcList

  !---------------------------------------------------------------------
  subroutine makeSimImg
  !---------------------------------------------------------------------

    implicit none

    character(255) :: expMapSetName, sourceListSetName, rawImageName

    !-------------------------------------------------------------------
    ! Start

    expMapSetName     = stringParameter('expmapset')
    sourceListSetName = stringParameter('simsrclistset')
    rawImageName      = stringParameter('rawimageset')

    call makeSimulatedImage(trim(expMapSetName)&
      , trim(sourceListSetName), 'SRCLIST'&
      , trim(rawImageName), psfModel='cal_acc_2')

  end subroutine makeSimImg

  !---------------------------------------------------------------------
  subroutine eventizeImg
  !---------------------------------------------------------------------

    implicit none

    character(255) :: expMapSetName, rawImageName, imageName
    real(single)   :: exposureMultiplier, nonVigBkgRate

    !-------------------------------------------------------------------
    ! Start

    expMapSetName = stringParameter('expmapset')
    rawImageName  = stringParameter('rawimageset')
    imageName     = stringParameter('imageset')

    exposureMultiplier = realParameter('expmul')
    nonVigBkgRate      = realParameter('nonvigbkgrate')

    call eventizeImage(expMapSetName, rawImageName, imageName&
      , exposureMultiplier, nonVigBkgRate)

  end subroutine eventizeImg

  !---------------------------------------------------------------------
  subroutine correctDetectedFluxes
  !---------------------------------------------------------------------
  ! Explanation: the time-consuming step in this source simulation
  ! business is the creation of the raw image. Multiplying this by an
  ! exposure map, adding background and eventizing are relatively quick
  ! in contrast. Hence, to save time, different exposure conditions are
  ! often simulated by using the same raw image, list of sim sources
  ! and exposure map, but multiplying the exposure map * raw image
  ! by an additional factor before eventizing. However since the detect
  ! chain uses the same exposure map without this multiplication, the
  ! values of RATE etc that it finds are too large by a factor of
  ! exposureMultiplier. Therefore the RATE etc values of the simulated
  ! sources are here divided by this factor.

    implicit none

    real(single) :: exposureMultiplier
    type(DataSetT) :: set
    type(TableT) :: tab
    real(single), dimension(:), pointer ::&
      scts, sctsErr, bkgMap, flux, fluxErr, rate, rateErr

    !-------------------------------------------------------------------
    ! Start

    exposureMultiplier = realParameter('expmul')

    set = dataSet(stringParameter('detsrclistset'), MODIFY)
    tab = table(set, 'SRCLIST')

    scts    => real32Data(column(tab, 'SCTS'))
    sctsErr => real32Data(column(tab, 'SCTS_ERR'))
    bkgMap  => real32Data(column(tab, 'BG_MAP'))
    flux    => real32Data(column(tab, 'FLUX'))
    fluxErr => real32Data(column(tab, 'FLUX_ERR'))
    rate    => real32Data(column(tab, 'RATE'))
    rateErr => real32Data(column(tab, 'RATE_ERR'))

    scts    = scts / exposureMultiplier
    sctsErr = sctsErr / exposureMultiplier
    bkgMap  = bkgMap / exposureMultiplier
    flux    = flux / exposureMultiplier
    fluxErr = fluxErr / exposureMultiplier
    rate    = rate / exposureMultiplier
    rateErr = rateErr / exposureMultiplier

    call release(set)

  end subroutine correctDetectedFluxes

  !---------------------------------------------------------------------
  subroutine compareDetWithSim
  !---------------------------------------------------------------------
  ! Given RATE, X_IMA and Y_IMA values for two source lists, one simulated and the other detected, this subroutine performs two functions: (i) it attempts to deduce which of the simulated sources corresponds with each detected source; (ii) it calculates the probability of the chance occurrence of a source 'closer' than this to the detected source.

    implicit none

    real(double), dimension(:), pointer :: srcX, srcY, simSrcX, simSrcY
    real(single), dimension(:), pointer :: srcL, srcXuncert, srcYuncert, srcLuncert, simSrcL
    real(single), dimension(:), allocatable :: devOverStdDev, probability
    real(single) :: simDensity, simFaintestL
    integer(int32) :: numDetSrcs, detSrc
    integer(int32), allocatable :: indexOfNearestSim(:)
    integer :: status
    logical, allocatable :: alsoRan(:)
integer(int32), pointer :: simSrcFlag(:)

    !-------------------------------------------------------------------
    ! Start

! Need to call simutils with --intables='orig_src_list in_det_src_list out_det_src_list', --string=compare and real=<value of the exposure multiplier>.

    call readDetSourceList(srcX, srcY, srcL, srcXuncert, srcYuncert&
      , srcLuncert)

    numDetSrcs = size(srcX)
    allocate(indexOfNearestSim(numDetSrcs), devOverStdDev(numDetSrcs))
    allocate(probability(numDetSrcs), alsoRan(numDetSrcs))

    call readSimSourceList(simSrcX, simSrcY, simSrcL, simSrcFlag&
      , simDensity, simFaintestL)

    do detSrc = 1, numDetSrcs
      call findNearestSimSource(srcX(detSrc), srcY(detSrc)&, srcL(detSrc)&
        , srcXuncert(detSrc), srcYuncert(detSrc), srcLuncert(detSrc)&
        , simSrcX, simSrcY, simSrcL&
        , indexOfNearestSim(detSrc), devOverStdDev(detSrc), status)

      if (status .lt. 0) then
        print *, 'det src', detSrc, ': findNearestSimSource status return'&
        , status
  return
      end if

      call probOfOneOrMoreByChance(srcL(detSrc), srcXuncert(detSrc)&
        , srcYuncert(detSrc), srcLuncert(detSrc), devOverStdDev(detSrc)&
        , simDensity, simFaintestL, probability(detSrc), status)

      if (status .lt. 0) then
        print *, 'det src', detSrc, ': probOfOneOrMoreByChance status return'&
        , status
  return
      end if
    end do

    call flagAlsoRans(indexOfNearestSim, devOverStdDev, alsoRan, status)
    if (status .lt. 0) then
      print *, 'det src', detSrc, ': flagAlsoRans status return'&
      , status
  return
    end if

    call saveFinalDetSrcList(simSrcX, simSrcY, simSrcL, simSrcFlag, simDensity&
    , simFaintestL, indexOfNearestSim, devOverStdDev, probability)!, alsoRan)

    deallocate(srcX, srcY, simSrcX, simSrcY, srcL&
    , srcXuncert, srcYuncert, srcLuncert, simSrcL)

  end subroutine compareDetWithSim

  !---------------------------------------------------------------------
  subroutine readDetSourceList(srcX, srcY, srcL, srcXuncert, srcYuncert&
  , srcLuncert)
  !---------------------------------------------------------------------

    implicit none

    real(double), dimension(:), pointer :: srcX, srcY
    real(single), dimension(:), pointer ::&
      srcL, srcXuncert, srcYuncert, srcLuncert

    type(DataSetT) :: set
    type(TableT)   :: tab
    integer :: numSrcs
    real(single), dimension(:), pointer :: x, y, rate, xUnc, yUnc, rateUnc
character(255) :: inDetSrcSetName

    !-------------------------------------------------------------------
    ! Start

inDetSrcSetName = stringParameter('detsrclistset')

    set = dataSet(inDetSrcSetName, READ)
    tab = table(set, 'SRCLIST')
    numSrcs = numberOfRows(tab)
    allocate(srcX(numSrcs), srcY(numSrcs), srcL(numSrcs))
    allocate(srcXuncert(numSrcs), srcYuncert(numSrcs), srcLuncert(numSrcs))

    x => real32Data(column(tab, 'X_IMA'))
    srcX = x

    xUnc => real32Data(column(tab, 'X_IMA_ERR'))
    srcXuncert = xUnc

    y => real32Data(column(tab, 'Y_IMA'))
    srcY = y

    yUnc => real32Data(column(tab, 'Y_IMA_ERR'))
    srcYuncert = yUnc

    rate => real32Data(column(tab, 'RATE'))
    srcL = rate

    rateUnc => real32Data(column(tab, 'RATE_ERR'))
    srcLuncert = rateUnc

    call release(set)

  end subroutine readDetSourceList

  !---------------------------------------------------------------------
  subroutine readSimSourceList(simSrcX, simSrcY, simSrcL, simSrcFlag, simDensity&
  , simFaintestL)
  !---------------------------------------------------------------------

    implicit none

    real(double), dimension(:), pointer :: simSrcX, simSrcY
    real(single), dimension(:), pointer :: simSrcL
integer(int32), dimension(:), pointer :: simSrcFlag
    real(single), intent(out) :: simDensity, simFaintestL

    type(DataSetT) :: set
    type(TableT)   :: tab
    integer :: numSrcs
    real(single), dimension(:), pointer :: x, y, rate
integer(int32), dimension(:), pointer :: flag
character(255) :: inSimSrcSetName

    !-------------------------------------------------------------------
    ! Start

inSimSrcSetName = stringParameter('simsrclistset')

    set = dataSet(inSimSrcSetName, READ)
    tab = table(set, 'SRCLIST')
    simDensity   = real32Attribute(tab, 'NDENSITY')
    simFaintestL = real32Attribute(tab, 'L_FAINT')
    numSrcs = numberOfRows(tab)
    allocate(simSrcX(numSrcs), simSrcY(numSrcs), simSrcL(numSrcs), simSrcFlag(numSrcs))

    x => real32Data(column(tab, 'X_IMA'))
    simSrcX = x

    y => real32Data(column(tab, 'Y_IMA'))
    simSrcY = y

    rate => real32Data(column(tab, 'RATE'))
    simSrcL = rate

    flag => int32Data(column(tab, 'FLAG'))
    simSrcFlag = flag

    call release(set)

  end subroutine readSimSourceList

  !---------------------------------------------------------------------
  subroutine saveFinalDetSrcList(simSrcX, simSrcY, simSrcL, simsrcFlag&
  , simDensity&
  , simFaintestL, indexOfNearestSim, devOverStdDev, probability)!, alsoRan)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: simSrcL(:)
    real(double), intent(in) :: simSrcX(size(simSrcL)), simSrcY(size(simSrcL))
    real(single), intent(in) :: simDensity, simFaintestL
    integer(int32), intent(in) :: indexOfNearestSim(:), simSrcFlag(:)
    real(single), dimension(size(indexOfNearestSim)), intent(in) ::&
      devOverStdDev, probability
!    logical, intent(in) :: alsoRan(size(indexOfNearestSim))

    type(DataSetT) :: set
    type(TableT)   :: tab
    type(ColumnT)  :: logLCol, logLerrCol
    integer :: numSrcs, row, simRow
    real(single), dimension(:), pointer ::&
      simX, simY, simL, devR, nulP, lgSL, rate, logL, rErr, logLerr, logNulP
integer(int32), pointer :: simFlag(:), simIndex(:)
character(255) :: inDetSrcSetName, outDetSrcSetName

    !-------------------------------------------------------------------
    ! Start

inDetSrcSetName = stringParameter('detsrclistset')
outDetSrcSetName = stringParameter('outset')

!    set = dataSet(stringParameter('intables', 0), MODIFY)
set = clone(inDetSrcSetName, outDetSrcSetName, MODIFY)
    tab = table(set, 'SRCLIST')
    numSrcs = numberOfRows(tab)
!    allocate(srcX(numSrcs), srcY(numSrcs), srcL(numSrcs))
!    allocate(srcXuncert(numSrcs), srcYuncert(numSrcs), srcLuncert(numSrcs))

if (numSrcs .ne. size(indexOfNearestSim)) call error('','')

    call setAttribute(tab, 'NDENSITY', simDensity, 'sources / pixel')
    call setAttribute(tab, 'L_FAINT',  simFaintestL, 'counts / sec'&
      , 'Faint-end cutoff in the distribution of sources.')

!    simXcol     = addColumn(tab, 'SIM_XIMA', REAL32)
!    simYcol     = addColumn(tab, 'SIM_YIMA', REAL32)
!    simLcol     = addColumn(tab, 'SIM_RATE', REAL32)
!    devRatiocol = addColumn(tab, 'DEVRATIO', REAL32)
!    nullProbcol = addColumn(tab, 'NULLPROB', REAL32)

logLCol    = addColumn(tab, 'LOG10_RA', REAL32)
logLerrCol = addColumn(tab, 'LOG10_RE', REAL32)

    simX => real32Data(addColumn(tab, 'SIM_XIMA', REAL32))
    simY => real32Data(addColumn(tab, 'SIM_YIMA', REAL32))
    simL => real32Data(addColumn(tab, 'SIM_RATE', REAL32))
    devR => real32Data(addColumn(tab, 'DEVRATIO', REAL32))
    nulP => real32Data(addColumn(tab, 'NULLPROB', REAL32))
    logNulP => real32Data(addColumn(tab, 'LOG10_NP', REAL32))
    lgSL => real32Data(addColumn(tab, 'LOG10_SR', REAL32))
    rate => real32Data(   column(tab, 'RATE'))
    logL => real32Data(logLCol)
    rErr => real32Data(   column(tab, 'RATE_ERR'))
    logLerr => real32Data(logLerrCol)
simFlag  => int32Data(addColumn(tab, 'SIM_FLAG', INTEGER32))
simIndex => int32Data(addColumn(tab, 'SIM_INDX', INTEGER32))

    do row = 0, numSrcs - 1
      simRow = indexOfNearestSim(row + 1)
      simX(row) = simSrcX(simRow)
      simY(row) = simSrcY(simRow)
      simL(row) = simSrcL(simRow)
      lgSL(row) = log10(simSrcL(simRow))
      if (rate(row) .gt. 0.0) then
        logL(row) = log10(rate(row))
        logLerr(row) = 1.0 * rErr(row) / rate(row) / lnTen
      else
        call setToNull(logLCol, row)
        call setToNull(logLerrCol, row)
      end if
simFlag(row)  = simSrcFlag(simRow)
    end do

simIndex = indexOfNearestSim
    devR = devOverStdDev
    nulP = probability
    logNulP = log10(probability)

    call release(set)

  end subroutine saveFinalDetSrcList

  !---------------------------------------------------------------------
  subroutine calcSysErrors
  !---------------------------------------------------------------------

    implicit none

    character(255) :: srcSetName
    type(DataSetT) :: set
    type(TableT) :: tab
    integer :: numRows, status
    real, allocatable :: rate(:), rateErr(:), simRate(:), vector(:),&
      uncerts(:), xima(:), ximaErr(:), simXima(:), yima(:),&
      yimaErr(:), simYima(:)
    logical, allocatable :: mask(:)
    real :: maskedMean, uncertInMean, redChiSqu

    !-------------------------------------------------------------------
    ! Start

    srcSetName = stringParameter('detsrclistset')
    set = dataSet(srcSetName, MODIFY)
    tab = table(set, 'SRCLIST')
    numRows = numberOfRows(tab)
    allocate(rate(numRows), rateErr(numRows), simRate(numRows), vector(numRows), uncerts(numRows), mask(numRows))



    print *, ''
    print *, 'RATE / SIM_RATE:'

    call readColData(tab, 'RATE', rate)
    call readColData(tab, 'RATE_ERR', rateErr)
    call readColData(tab, 'SIM_RATE', simRate)

    vector  = rate    / simRate
    uncerts = rateErr / simRate

    call temp_robustMean(vector, uncerts, 5.0&
      , maskedMean, mask, uncertInMean, redChiSqu, status)
    if (status < 0) then
print *, 'robustMean error status', status
call error('','')
    end if

    print *, '    robust mean: ', maskedMean, '+-', uncertInMean
    print *, '  reduced chi^2: ', redChiSqu
    print *, '     num values: ', numRows
    print *, '  fraction used: ', count(mask) / real(numRows)

    deallocate(rate, simRate, rateErr)

    call setAttribute(tab, 'MRATERAT', maskedMean, ''&
      , '(Robust) mean of RATE / SIM_RATE.')
    call setAttribute(tab, 'ERATERAT', redChiSqu, ''&
      , 'Reduced chi^2 of MRATERAT.')


    print *, ''
    print *, 'X_IMA - SIM_XIMA:'

    allocate(xima(numRows), ximaErr(numRows), simXima(numRows))
    call readColData(tab, 'X_IMA',     xima)
    call readColData(tab, 'X_IMA_ERR', ximaErr)
    call readColData(tab, 'SIM_XIMA',  simXima)

    vector  = xima - simXima
    uncerts = ximaErr

    call temp_robustMean(vector, uncerts, 5.0&
      , maskedMean, mask, uncertInMean, redChiSqu, status)
    if (status < 0) then
print *, 'robustMean error status', status
call error('','')
    end if

    print *, '    robust mean: ', maskedMean, '+-', uncertInMean
    print *, '  reduced chi^2: ', redChiSqu
    print *, '     num values: ', numRows
    print *, '  fraction used: ', count(mask) / real(numRows)

    deallocate(xima, simXima, ximaErr)

    call setAttribute(tab, 'MXIMADEL', maskedMean, 'image pixels'&
      , '(Robust) mean of X_IMA - SIM_XIMA.')
    call setAttribute(tab, 'EXIMADEL', redChiSqu, ''&
      , 'Reduced chi^2 of MXIMADEL.')


    print *, ''
    print *, 'Y_IMA - SIM_YIMA:'

    allocate(yima(numRows), yimaErr(numRows), simYima(numRows))
    call readColData(tab, 'Y_IMA',     yima)
    call readColData(tab, 'Y_IMA_ERR', yimaErr)
    call readColData(tab, 'SIM_YIMA',  simYima)

    vector  = yima - simYima
    uncerts = yimaErr

    call temp_robustMean(vector, uncerts, 5.0&
      , maskedMean, mask, uncertInMean, redChiSqu, status)
    if (status < 0) then
print *, 'robustMean error status', status
call error('','')
    end if

    print *, '    robust mean: ', maskedMean, '+-', uncertInMean
    print *, '  reduced chi^2: ', redChiSqu
    print *, '     num values: ', numRows
    print *, '  fraction used: ', count(mask) / real(numRows)

    call release(set)

    call setAttribute(tab, 'MYIMADEL', maskedMean, 'image pixels'&
      , '(Robust) mean of Y_IMA - SIM_YIMA.')
    call setAttribute(tab, 'EYIMADEL', redChiSqu, ''&
      , 'Reduced chi^2 of MYIMADEL.')

  end subroutine calcSysErrors

  !---------------------------------------------------------------------
  subroutine makeMasterList
  !---------------------------------------------------------------------
  ! Adds or merges together separate source lists.

    implicit none

    character(255) :: masterSetName, setName
    integer :: numSets, setNum, numKeywords
    character(20), allocatable :: keywordNames(:)
    logical, allocatable :: keywordIsUniform(:)
    type(DataSetT) :: masterSet
    type(DataSetT), allocatable :: set(:)
    type(TableT) :: masterTab
    type(TableT), allocatable :: tab(:)

    !-------------------------------------------------------------------
    ! Start

    masterSetName = stringParameter('outset')
    numSets = parameterCount('intables')

if (numSets <= 1) call error('tooFewSetsProvided','')

    masterSet = dataSet(masterSetName, CREATE)
    masterTab = addTable(masterSet, 'SRCLIST', 0)

    ! Check for presence and consistency of NDENSITY, MYIMADEL etc kwds
    ! and copy them over:

    allocate(set(numSets), tab(numSets))
    do setNum = 0, numSets - 1
      setName = stringParameter('intables', setNum)
      set(setNum) = dataSet(setName, READ)
      tab(setNum) = table(set(setNum), 'SRCLIST')
    end do

    numKeywords = 8
    allocate(keywordNames(numKeywords), keywordIsUniform(numKeywords))
    keywordNames(1) = 'NDENSITY'
    keywordNames(2) = 'L_FAINT'
    keywordNames(3) = 'MRATERAT'
    keywordNames(4) = 'ERATERAT'
    keywordNames(5) = 'MXIMADEL'
    keywordNames(6) = 'EXIMADEL'
    keywordNames(7) = 'MYIMADEL'
    keywordNames(8) = 'EYIMADEL'

    call checkAndCopyKeywords(tab, masterTab, keywordNames&
      , keywordIsUniform)

    if (any(.not. keywordIsUniform)) call warning('nonUniformKwds','')

    ! Now the actual merging:

    do setNum = 0, numSets - 1
      call addTables(masterTab, tab(setNum), 'ID_FIELD', int(setNum, int16))
      call release(set(setNum))
    end do

    call release(masterSet)

  end subroutine makeMasterList

  !---------------------------------------------------------------------
  subroutine extractMatches
  !---------------------------------------------------------------------
  ! Inputs to this subroutine are three source lists comprising
  ! detections and (for each detected source) a best-guess associated
  ! sim source, 1 list for each of the three epic cameras. (Multi-field
  ! lists should not be used.) The subroutine outputs a single list
  ! comprising those detections which are associated with the same
  ! sim source for all three cameras. The assumption is that these
  ! are likely to be pretty firm detections and can then be used to
  ! assess correctly the systematic errors in RATE, X_IMA and Y_IMA.

    implicit none

    type :: vectorOfPointers
      integer(int32), pointer :: simIndex(:)
    end type vectorOfPointers

    integer(int32) ::&
      numSets, setNum, locVec(1), numOfSmallestSet, minNumRows,&
      numCopiableRows, row, simIndx, toRow, fromRow, rowPlus1, rowToCopy
    type(vectorOfPointers), allocatable :: vop(:)
    type(DataSetT), allocatable :: set(:)
    type(TableT),   allocatable :: tab(:)
    integer(int32), allocatable :: numRows(:), rowsToCopy(:,:)
    type(DataSetT) :: masterSet
    type(TableT)   :: masterTab
    character(255) :: masterSetRoot = 'extracted_sli', masterSetName, setName, idStr=''
    integer :: posLastUnderscore, posSuffix
logical :: testflag = .false.

    !-------------------------------------------------------------------
    ! Start

    numSets = parameterCount('intables')

if (numSets < 2 .or. numSets > 3) call error('wrongNumSets','')

    do setNum = 0, numSets - 1
      setName = stringParameter('intables', setNum)

      posLastUnderscore = index(setName, '_', .true.)
if (posLastUnderscore == 0) call error('badSetName0','')
      posSuffix = index(setName, '.ds')
if (posSuffix == 0) call error('badSetName1','')
if (posSuffix < posLastUnderscore + 2) call error('badSetName2','')

      if (setNum > 0) then
        if (setName(posLastUnderscore+1:posSuffix-1) .ne. trim(idStr))&
then
print *, 'setName(posLastUnderscore+1:posSuffix-1), trim(idStr): '&
, setName(posLastUnderscore+1:posSuffix-1), '*', trim(idStr), '*'
call error('inconsistentIdNumbers','')
end if
    cycle
      end if

      idStr = setName(posLastUnderscore+1:posSuffix-1)
    end do

    masterSetName = trim(masterSetRoot)//'_'//trim(idStr)//'.ds'

    allocate(set(numSets), tab(numSets), vop(numSets), numRows(numSets))
    set(1) = dataSet(stringParameter('intables', 0), READ)
    tab(1) = table(set(1), 'SRCLIST')

    ! Open all the SIM_INDX columns:

    vop(1)%simIndex => int32Data(column(tab(1), 'SIM_INDX'))
    numRows(1) = numberOfRows(tab(1))
    do setNum = 2, numSets
      set(setNum) = dataSet(stringParameter('intables', setNum - 1)&
        , READ)
      tab(setNum) = table(set(setNum), 'SRCLIST')
      vop(setNum)%simIndex => int32Data(column(tab(setNum), 'SIM_INDX'))
      numRows(setNum) = numberOfRows(tab(setNum))
    end do

    locVec = minloc(numRows)
    numOfSmallestSet = locVec(1)

    minNumRows = minval(numRows)
if (minNumRows < 1) call error('smallestSetIsEmpty','')
if (testflag) print *, 'numOfSmallestSet, minNumRows: '&
, numOfSmallestSet, minNumRows

    allocate(rowsToCopy(minNumRows, numSets))
    numCopiableRows = 0
    ROWS: do row = 0, minNumRows - 1
      simIndx = vop(numOfSmallestSet)%simIndex(row)
if (testflag) print *, '  row, simIndx: '&
, row, simIndx

      ! Look in all the sets apart from numOfSmallestSet:
      do setNum = 1, numSets
      if (setNum == numOfSmallestSet) cycle
    if (.not. any(vop(setNum)%simIndex == simIndx)) cycle ROWS
      end do

      ! At this point, we have established that the same value of SIM_INDX is found in all the sets. However now we have to check that it has not been duplicated in any one set:

      ! Look in all the sets, numOfSmallestSet included:
      do setNum = 1, numSets
    if (count(vop(setNum)%simIndex == simIndx) > 1) cycle ROWS
      end do

      ! The value of SIM_INDX has passed all the tests. Mark the relevant rows for inclusion in the output file:

      numCopiableRows = numCopiableRows + 1
if (testflag) print *, '    numCopiableRows: '&
, numCopiableRows
      do setNum = 1, numSets
        rowToCopy = smallestIndex(vop(setNum)%simIndex == simIndx) - 1
        rowsToCopy(numCopiableRows, setNum) = rowToCopy
if (testflag) print *, '      setNum, rowToCopy, matchingIndex: '&
, setNum, rowToCopy, vop(setNum)%simIndex(rowToCopy)
      end do
    end do ROWS

if (numCopiableRows <= 0) call error('noOutputRows','')

    ! Now construct the output set:

    masterSet = dataSet(masterSetName, CREATE)
    masterTab = addTable(masterSet, 'SRCLIST', numCopiableRows * numSets)

    toRow = 0
    do setNum = 1, numSets
      do rowPlus1 = 1, numCopiableRows
        fromRow = rowsToCopy(rowPlus1, setNum)
        call copyRowsBetweenTables(tab(setNum), fromRow, 1&
          , masterTab, toRow)
        toRow = toRow + 1
      end do
      call release(set(setNum))
    end do

    call release(masterSet)

  end subroutine extractMatches

  !---------------------------------------------------------------------
  subroutine calcCompletenessHistograms
  !---------------------------------------------------------------------
  ! Makes a single fits file that contains several histograms having
  ! the same x-axis bin arrangement, with column names as follows:
  !
  !	NUM_SIM		simulated sources that fall within the inst exposure map
  !	NUM_ALL		all sources detected by the inst.
  !	NUM_FIRM	detected sources with NULLPROB < --nullproblimit.
  !	DETFRAC		fraction of simulated sources (firmly) detected.
  !	FIRMFRAC	fraction OF firm sources.
  !	DETFRAC_ERR	obvious.
  !	FIRMFRAC_ERR	obvious.
  	

    implicit none

    integer(int16) :: numBins, bin
    real(single) :: xLo, xHi, nullProbLimit, binXlo, binXhi, exposureMultiplier
    character(255) :: histoXcolName
    type(DataSetT) :: outSet, simSet, detSet
    type(TableT) :: outTab, simTab, detTab
    real(single), allocatable :: simX(:), detX(:)
    real(single), pointer, dimension(:) :: outX, detNullProb,&
        outLogSim,&
       outDetFrac,&
      outFirmFrac,&
       outDetFracErr,&
      outFirmFracErr
    real(double), pointer, dimension(:) ::&
           outSim,&
           outAll,&
          outFirm
    type(ColumnT) ::&
        outLogSimCol,&
       outDetFracCol,&
      outFirmFracCol,&
       outDetFracErrCol,&
      outFirmFracErrCol
    integer(int32), pointer :: simFlag(:)
    logical, allocatable :: mask(:)
    character(2) :: instStr
    character(24) :: outSetName
    integer(int32) :: notOnExpMapBit, noPsfBit

    !-------------------------------------------------------------------
    ! Start

    numBins = intParameter('histonbins')
    xLo = realParameter('histoxlo')
    xHi = realParameter('histoxhi')
!    histoXcolName = stringParameter('histoxcol') ! expected to be LOG10_SR
histoXcolName = 'LOG10_SR'
    nullProbLimit = realParameter('nullproblimit')
    exposureMultiplier = realParameter('expmul')
    instStr = ucase(stringParameter('instrument'))

    select case(instStr)
    case('PN')
      notOnExpMapBit = NOT_IN_PN_EXP
      noPsfBit       = NO_PN_PSF
    case('M1')
      notOnExpMapBit = NOT_IN_M1_EXP
      noPsfBit       = NO_M1_PSF
    case('M2')
      notOnExpMapBit = NOT_IN_M2_EXP
      noPsfBit       = NO_M2_PSF
    case default
call error('badInstrument', '')
    end select

    outSetName = 'histo_completeness_'//instStr//'.ds'

    outSet = dataSet(outSetName, CREATE)
    outTab = addTable(outSet, 'HISTO', int(numBins, int32))

    outLogSimCol   = addColumn(outTab, 'LOG10_NS', REAL32)
    outDetFracCol  = addColumn(outTab, 'DETFRAC',  REAL32)
    outFirmFracCol = addColumn(outTab, 'FIRMFRAC', REAL32)
    outDetFracErrCol  = addColumn(outTab, 'DETFRAC_ERR',  REAL32)
    outFirmFracErrCol = addColumn(outTab, 'FIRMFRAC_ERR', REAL32)

    outX        => real32Data(addColumn(outTab, histoXcolName, REAL32))
    outSim      => real64Data(addColumn(outTab, 'NUM_SIM',  REAL64))
    outAll      => real64Data(addColumn(outTab, 'NUM_ALL',  REAL64))
    outFirm     => real64Data(addColumn(outTab, 'NUM_FIRM', REAL64))
    outLogSim   => real32Data(outLogSimCol)
    outDetFrac  => real32Data(outDetFracCol)
    outFirmFrac => real32Data(outFirmFracCol)
    outDetFracErr  => real32Data(outDetFracErrCol)
    outFirmFracErr => real32Data(outFirmFracErrCol)

    simSet = dataSet(stringParameter('mergedsimlist'), READ)
    simTab = table(simSet, 'SRCLIST')
    simFlag => int32Data(column(simTab, 'FLAG'))
    allocate(simX(numberOfRows(simTab)))
    call readColData(simTab, histoXcolName, simX)
simX = simX + log10(exposureMultiplier)
!*** assumes that histoXcolName = 'LOG10_SR'!
    allocate(mask(numberOfRows(simTab)))

    do bin = 1, numBins
      binXlo = fracVal(xLo, xHi, (bin - 1) / real(numBins))
      binXhi = fracVal(xLo, xHi,  bin      / real(numBins))
      outX(bin - 1) = (binXlo + binXhi) / 2.0

      outSim(bin - 1) = count(simX >= binXlo .and. simX < binXhi .and.&
      .not. (btest(simFlag, notOnExpMapBit) .or. btest(simFlag, noPsfBit)))
    end do

    call release(simSet)
    deallocate(simX, mask)

    detSet = dataSet(stringParameter('mergeddetlist'), READ)
    detTab = table(detSet, 'SRCLIST')
!*** check presence of NDENSITY and L_FAINT kwds, compare to those in sim_src_list, and copy over.
!*** check presence of EXP_MULT and NONVIGBR kwds and copy over.
    detNullProb => real32Data(column(detTab, 'NULLPROB'))
    allocate(detX(numberOfRows(detTab)))
    call readColData(detTab, histoXcolName, detX)

    do bin = 1, numBins
      binXlo = fracVal(xLo, xHi, (bin - 1) / real(numBins))
      binXhi = fracVal(xLo, xHi,  bin      / real(numBins))

      outAll(bin - 1)  = count(detX >= binXlo .and. detX < binXhi)
      outFirm(bin - 1) = count(detX >= binXlo .and. detX < binXhi&
        .and. detNullProb < nullProbLimit)
    end do

    call release(detSet)
    deallocate(detX)

    ! Now do arithmetic:

    do bin = 1, numBins
      if (outSim(bin - 1) > 0.0d0) then
        outLogSim(bin - 1)  = log10(outSim(bin - 1))
      else
        call setToNull(outLogSimCol, bin - 1)
      end if

      if (outSim(bin - 1) > 0.0d0) then
        outDetFrac(bin - 1)  = outFirm(bin - 1) / outSim(bin - 1)
        outDetFracErr(bin - 1)  = sqrt(outFirm(bin - 1)) / outSim(bin - 1)
      else
        call setToNull(outDetFracCol, bin - 1)
        call setToNull(outDetFracErrCol, bin - 1)
      end if

      if (outAll(bin - 1) > 0.0d0) then
        outFirmFrac(bin - 1) = outFirm(bin - 1) / outAll(bin - 1)
        outFirmFracErr(bin - 1) = sqrt(outFirm(bin - 1)) / outAll(bin - 1)
      else
        call setToNull(outFirmFracCol, bin - 1)
        call setToNull(outFirmFracErrCol, bin - 1)
      end if
    end do

    call release(outSet)

  end subroutine calcCompletenessHistograms

  !---------------------------------------------------------------------
  subroutine calcReliabilityHistogram
  !---------------------------------------------------------------------
  ! Makes a single fits file that contains some histograms (usually over
  ! DET_ML). The column names are as follows:
  !
  !	NUM_ALL		total number sources detected by this inst.
  !	NUM_FIRM	'firm' sources, ie detected sources with
  !                          NULLPROB < --nullproblimit.
  !	FIRMFRAC	fraction of firm sources.
  !	FIRMFRAC_ERR	obvious.
  	
    implicit none

    integer(int16) :: numBins, bin
    real(single) :: xLo, xHi, nullProbLimit, binXlo, binXhi
    character(255) :: histoXcolName
    type(DataSetT) :: outSet, detSet
    type(TableT) :: outTab, detTab
    real(single), allocatable :: detX(:)
    real(single), pointer, dimension(:) :: outX, detNullProb,&
      outFirmFrac, outFirmFracErr
    real(double), pointer, dimension(:) ::&
           outAll,&
          outFirm
    type(ColumnT) ::&
      outFirmFracCol, outFirmFracErrCol
    character(2) :: instStr
    character(24) :: outSetName

    !-------------------------------------------------------------------
    ! Start

    numBins = intParameter('histonbins')
    xLo = realParameter('histoxlo')
    xHi = realParameter('histoxhi')
    histoXcolName = stringParameter('histoxcol') ! expected to be DET_ML
    nullProbLimit = realParameter('nullproblimit')
    instStr = ucase(stringParameter('instrument'))

    outSetName = 'histo_reliability_'//instStr//'.ds'

    outSet = dataSet(outSetName, CREATE)
    outTab = addTable(outSet, 'HISTO', int(numBins, int32))

    outFirmFracCol = addColumn(outTab, 'FIRMFRAC', REAL32)
    outFirmFracErrCol = addColumn(outTab, 'FIRMFRAC_ERR', REAL32)

    outX        => real32Data(addColumn(outTab, histoXcolName, REAL32))
    outAll      => real64Data(addColumn(outTab, 'NUM_ALL',  REAL64))
    outFirm     => real64Data(addColumn(outTab, 'NUM_FIRM', REAL64))
    outFirmFrac => real32Data(outFirmFracCol)
    outFirmFracErr => real32Data(outFirmFracErrCol)

    detSet = dataSet(stringParameter('mergeddetlist'), READ)
    detTab = table(detSet, 'SRCLIST')
!*** check presence of NDENSITY and L_FAINT kwds, compare to those in sim_src_list, and copy over.
!*** check presence of EXP_MULT and NONVIGBR kwds and copy over.
    detNullProb => real32Data(column(detTab, 'NULLPROB'))
    allocate(detX(numberOfRows(detTab)))
    call readColData(detTab, histoXcolName, detX)

    do bin = 1, numBins
      binXlo = fracVal(xLo, xHi, (bin - 1) / real(numBins))
      binXhi = fracVal(xLo, xHi,  bin      / real(numBins))

      outX(bin - 1) = (binXlo + binXhi) / 2.0

      outAll(bin - 1)  = count(detX >= binXlo .and. detX < binXhi)
      outFirm(bin - 1) = count(detX >= binXlo .and. detX < binXhi&
        .and. detNullProb < nullProbLimit)
    end do

    call release(detSet)
    deallocate(detX)

    ! Now do arithmetic:

    do bin = 1, numBins
      if (outAll(bin - 1) > 0.0d0) then
        outFirmFrac(bin - 1) = outFirm(bin - 1) / outAll(bin - 1)
        outFirmFracErr(bin - 1) = sqrt(outFirm(bin - 1)) / outAll(bin - 1)
      else
        call setToNull(outFirmFracCol, bin - 1)
        call setToNull(outFirmFracErrCol, bin - 1)
      end if
    end do

    call release(outSet)

  end subroutine calcReliabilityHistogram

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module simutils_mod
