!+
! Module: temp_array_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines to process arrays in various ways.
! 
! Comments: TODO:
! 
! $Id: temp_array_utils.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module temp_array_utils

  implicit none 

  private 

  public rebinImage
  public rebin

  interface rebinImage
    module procedure rebinImageBasicReal, rebinImageWithEdgesReal
    module procedure rebinImageBasicDble, rebinImageWithEdgesDble
  end interface

  interface rebin
    module procedure rebinBasicReal, rebinWithWidthsReal, rebinWithEdgesReal
    module procedure rebinBasicDble, rebinWithWidthsDble, rebinWithEdgesDble
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine rebinImageBasicReal(oldImage, newImage, status)
  !---------------------------------------------------------------------
  ! The rebinning is done in two passes, x first then y, using the math_utils
  ! procedure rebin (cut and pasted in here until math_utils is accepted).

    implicit none

    real, intent(in) ::&
      oldImage(:,:)
    real, intent(out) ::&
      newImage(:,:)
    integer, intent(out), optional :: status

    real ::&
      tempImage(size(oldImage, 1), size(newImage, 2))
    integer :: localStatus, foreignStatus, xi, yi

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    tempImage = 0.0 ! array op, default
    do xi = 1, size(oldImage, 1)
      call rebin(oldImage(xi, :), tempImage(xi, :), status=foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 0)
!***
      end if
    end do

    newImage = 0.0 ! array op, default
    do yi = 1, size(newImage, 2)
      call rebin(tempImage(:, yi), newImage(:, yi), status=foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 1)
!***
      end if
    end do

    if (present(status)) status = localStatus

  end subroutine rebinImageBasicReal

  !---------------------------------------------------------------------
  subroutine rebinImageBasicDble(oldImage, newImage, status)
  !---------------------------------------------------------------------
  ! The rebinning is done in two passes, x first then y, using the math_utils
  ! procedure rebin (cut and pasted in here until math_utils is accepted).

    implicit none

    real(kind(1d0)), intent(in) ::&
      oldImage(:,:)
    real(kind(1d0)), intent(out) ::&
      newImage(:,:)
    integer, intent(out), optional :: status

    real(kind(1d0)) ::&
      tempImage(size(oldImage, 1), size(newImage, 2))
    integer :: localStatus, foreignStatus, xi, yi

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    tempImage = 0.0 ! array op, default
    do xi = 1, size(oldImage, 1)
      call rebin(oldImage(xi, :), tempImage(xi, :), status=foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 0)
!***
      end if
    end do

    newImage = 0.0 ! array op, default
    do yi = 1, size(newImage, 2)
      call rebin(tempImage(:, yi), newImage(:, yi), status=foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 1)
!***
      end if
    end do

    if (present(status)) status = localStatus

  end subroutine rebinImageBasicDble

  !---------------------------------------------------------------------
  subroutine rebinImageWithEdgesReal(oldImage, oldXedges, oldYedges&
  , newImage, newXedges, newYedges, status)
  !---------------------------------------------------------------------
  ! The rebinning is done in two passes, x first then y, using the math_utils
  ! procedure rebin (cut and pasted in here until math_utils is accepted).

    implicit none

    real, intent(in) ::&
      oldImage(:,:),&
      oldXedges(size(oldImage, 1) + 1), oldYedges(size(oldImage, 2) + 1),&
      newXedges(:), newYedges(:)
    real, intent(out) ::&
      newImage(size(newXedges) - 1, size(newYedges) - 1)
    integer, intent(out), optional :: status

    real ::&
      tempImage(size(oldImage, 1), size(newImage, 2))
    integer :: localStatus, foreignStatus, xi, yi

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    tempImage = 0.0 ! array op, default
    do xi = 1, size(oldImage, 1)
      call rebin(oldImage(xi, :), tempImage(xi, :), oldYedges, newYedges&
        , foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 0)
!***
      end if
    end do

    newImage = 0.0 ! array op, default
    do yi = 1, size(newImage, 2)
      call rebin(tempImage(:, yi), newImage(:, yi), oldXedges, newXedges&
        , status=foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 1)
!***
      end if
    end do

    if (present(status)) status = localStatus

  end subroutine rebinImageWithEdgesReal

  !---------------------------------------------------------------------
  subroutine rebinImageWithEdgesDble(oldImage, oldXedges, oldYedges&
  , newImage, newXedges, newYedges, status)
  !---------------------------------------------------------------------
  ! The rebinning is done in two passes, x first then y, using the math_utils
  ! procedure rebin (cut and pasted in here until math_utils is accepted).

    implicit none

    real(kind(1d0)), intent(in) ::&
      oldImage(:,:),&
      oldXedges(size(oldImage, 1) + 1), oldYedges(size(oldImage, 2) + 1),&
      newXedges(:), newYedges(:)
    real(kind(1d0)), intent(out) ::&
      newImage(size(newXedges) - 1, size(newYedges) - 1)
    integer, intent(out), optional :: status

    real(kind(1d0)) ::&
      tempImage(size(oldImage, 1), size(newImage, 2))
    integer :: localStatus, foreignStatus, xi, yi

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    tempImage = 0.0 ! array op, default
    do xi = 1, size(oldImage, 1)
      call rebin(oldImage(xi, :), tempImage(xi, :), oldYedges, newYedges&
        , foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 0)
!***
      end if
    end do

    newImage = 0.0 ! array op, default
    do yi = 1, size(newImage, 2)
      call rebin(tempImage(:, yi), newImage(:, yi), oldXedges, newXedges&
        , status=foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 1)
!***
      end if
    end do

    if (present(status)) status = localStatus

  end subroutine rebinImageWithEdgesDble

  !---------------------------------------------------------------------
  subroutine rebinBasicReal(oldYvalues, newYvalues, status)
  !---------------------------------------------------------------------

    implicit none 

    real, intent(in) ::&
      oldYvalues(:)
    real, intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins
    integer ::&
      oldBin, newBin, localStatus

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    do oldBin = 1, numOldBins + 1
      oldXedges(oldBin) = (oldBin - 1) / real(numOldBins)
    end do
    do newBin = 1, numNewBins + 1
      newXedges(newBin) = (newBin - 1) / real(numNewBins)
    end do

    call rebinWithEdgesReal(oldYvalues, newYvalues, oldXedges, newXedges&
    , localStatus)

    if (present(status)) status = localStatus

  end subroutine rebinBasicReal

  !---------------------------------------------------------------------
  subroutine rebinBasicDble(oldYvalues, newYvalues, status)
  !---------------------------------------------------------------------

    implicit none 

    real(kind(1d0)), intent(in) ::&
      oldYvalues(:)
    real(kind(1d0)), intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real(kind(1d0)) ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins
    integer ::&
      oldBin, newBin, localStatus

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    do oldBin = 1, numOldBins + 1
      oldXedges(oldBin) = (oldBin - 1) / real(numOldBins)
    end do
    do newBin = 1, numNewBins + 1
      newXedges(newBin) = (newBin - 1) / real(numNewBins)
    end do

    call rebinWithEdgesDble(oldYvalues, newYvalues, oldXedges, newXedges&
    , localStatus)

    if (present(status)) status = localStatus

  end subroutine rebinBasicDble

  !---------------------------------------------------------------------
  subroutine rebinWithWidthsReal(oldYvalues, newYvalues, oldXwidths&
  , newXwidths, oldXcentres, newXcentres, status)
  !---------------------------------------------------------------------

    implicit none 

    real, intent(in), dimension(:) ::&
      oldXwidths, oldXcentres, oldYvalues, newXwidths, newXcentres
    real, intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins,&
      localStatus, foreignStatus
    integer ::&
      oldBin, newBin

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    if (size(oldXwidths) .lt. numOldBins) then
      if(present(status)) status = -1
  return
    else if (size(oldXwidths) .gt. numOldBins) then
      localStatus = localStatus + 1
    end if
    if (size(newXwidths) .lt. numNewBins) then
      if(present(status)) status = -2
  return
    else if (size(newXwidths) .gt. numNewBins) then
      localStatus = localStatus + 2
    end if

    if (size(oldXcentres) .lt. numOldBins) then
      if(present(status)) status = -3
  return
    else if (size(oldXcentres) .gt. numOldBins) then
      localStatus = localStatus + 4
    end if
    if (size(newXcentres) .lt. numNewBins) then
      if(present(status)) status = -4
  return
    else if (size(newXcentres) .gt. numNewBins) then
      localStatus = localStatus + 8
    end if

    do oldBin = 1, numOldBins
      oldXedges(oldBin) = oldXcentres(oldBin) - 0.5 * oldXwidths(oldBin)
    end do
    oldXedges(numOldBins + 1) = oldXcentres(numOldBins) + 0.5&
    * oldXwidths(numOldBins)

    do newBin = 1, numNewBins
      newXedges(newBin) = newXcentres(newBin) - 0.5 * newXwidths(newBin)
    end do
    newXedges(numNewBins + 1) = newXcentres(numNewBins) + 0.5&
    * newXwidths(numNewBins)

    call rebinWithEdgesReal(oldYvalues, newYvalues, oldXedges, newXedges&
    , foreignStatus)
    if (foreignStatus .lt. 0) then
      if(present(status)) status = foreignStatus - 4
  return
    end if

    localStatus = localStatus + foreignStatus * 16
    if(present(status)) status = localStatus

  end subroutine rebinWithWidthsReal

  !---------------------------------------------------------------------
  subroutine rebinWithWidthsDble(oldYvalues, newYvalues, oldXwidths&
  , newXwidths, oldXcentres, newXcentres, status)
  !---------------------------------------------------------------------

    implicit none 

    real(kind(1d0)), intent(in), dimension(:) ::&
      oldXwidths, oldXcentres, oldYvalues, newXwidths, newXcentres
    real(kind(1d0)), intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real(kind(1d0)) ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins,&
      localStatus, foreignStatus
    integer ::&
      oldBin, newBin

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    if (size(oldXwidths) .lt. numOldBins) then
      if(present(status)) status = -1
  return
    else if (size(oldXwidths) .gt. numOldBins) then
      localStatus = localStatus + 1
    end if
    if (size(newXwidths) .lt. numNewBins) then
      if(present(status)) status = -2
  return
    else if (size(newXwidths) .gt. numNewBins) then
      localStatus = localStatus + 2
    end if

    if (size(oldXcentres) .lt. numOldBins) then
      if(present(status)) status = -3
  return
    else if (size(oldXcentres) .gt. numOldBins) then
      localStatus = localStatus + 4
    end if
    if (size(newXcentres) .lt. numNewBins) then
      if(present(status)) status = -4
  return
    else if (size(newXcentres) .gt. numNewBins) then
      localStatus = localStatus + 8
    end if

    do oldBin = 1, numOldBins
      oldXedges(oldBin) = oldXcentres(oldBin) - 0.5d0 * oldXwidths(oldBin)
    end do
    oldXedges(numOldBins + 1) = oldXcentres(numOldBins) + 0.5d0&
    * oldXwidths(numOldBins)

    do newBin = 1, numNewBins
      newXedges(newBin) = newXcentres(newBin) - 0.5d0 * newXwidths(newBin)
    end do
    newXedges(numNewBins + 1) = newXcentres(numNewBins) + 0.5d0&
    * newXwidths(numNewBins)

    call rebinWithEdgesDble(oldYvalues, newYvalues, oldXedges, newXedges&
    , foreignStatus)
    if (foreignStatus .lt. 0) then
      if(present(status)) status = foreignStatus - 4
  return
    end if

    localStatus = localStatus + foreignStatus * 16
    if(present(status)) status = localStatus

  end subroutine rebinWithWidthsDble

  !---------------------------------------------------------------------
  subroutine rebinWithEdgesReal(oldYvalues, newYvalues, oldXedges, newXedges&
  , status)
  !---------------------------------------------------------------------

    implicit none 

    real, intent(in) ::&
      oldXedges(:), oldYvalues(size(oldXedges)-1), newXedges(:)
    real, intent(out) ::&
      newYvalues(size(newXedges)-1)
    integer, intent(out), optional :: status

    integer :: foreignStatus
    real(kind(1d0)) :: newYvaluesDble(size(newYvalues))

    !-------------------------------------------------------------------
    ! Start:

    call rebinKernel(dble(oldYvalues), newYvaluesDble, dble(oldXedges)&
    , dble(newXedges), foreignStatus)

    if (present(status)) status = foreignStatus
    newYvalues = real(newYvaluesDble)

  end subroutine rebinWithEdgesReal

  !---------------------------------------------------------------------
  subroutine rebinWithEdgesDble(oldYvalues, newYvalues, oldXedges, newXedges&
  , status)
  !---------------------------------------------------------------------

    implicit none 

    real(kind(1d0)), intent(in) ::&
      oldXedges(:), oldYvalues(size(oldXedges)-1), newXedges(:)
    real(kind(1d0)), intent(out) ::&
      newYvalues(size(newXedges)-1)
    integer, intent(out), optional :: status

    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start:

    call rebinKernel(oldYvalues, newYvalues, oldXedges&
    , newXedges, foreignStatus)

    if (present(status)) status = foreignStatus

  end subroutine rebinWithEdgesDble

  !---------------------------------------------------------------------
  subroutine rebinKernel(oldYvalues, newYvalues, oldXedges, newXedges&
  , status)
  !---------------------------------------------------------------------
  ! This is designed to be a pretty general and robust rebinning routine.
  ! Neither the old or new series is constrained to have bins of equal width.

    implicit none 

    real(kind(1d0)), intent(in) ::&
      oldXedges(:), oldYvalues(size(oldXedges)-1), newXedges(:)
    real(kind(1d0)), intent(out) ::&
      newYvalues(size(newXedges)-1)
    integer, intent(out) :: status

    integer ::&
      numOldBins,&
      numNewBins
    integer ::&
      oldBin, oldStartBin, oldFinisBin, newBin
    logical ::&
      foundOverlappingBin
    real(kind(1d0)) ::&
      overlap, binWidth

    !-------------------------------------------------------------------
    ! Start:

    status = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    if (size(oldXedges) .lt. numOldBins + 1) then
      status = -1
  return
    else if (size(oldXedges) .gt. numOldBins + 1) then
      status = status + 1
    end if
    if (size(newXedges) .lt. numNewBins + 1) then
      status = -2
  return
    else if (size(newXedges) .gt. numNewBins + 1) then
      status = status + 2
    end if

    do oldBin = 1, numOldBins
      binWidth = oldXedges(oldBin + 1) - oldXedges(oldBin)
      if (binWidth .le. 0.0) then
        status = -3
  return
      end if
    end do

    do newBin = 1, numNewBins
      binWidth = newXedges(newBin + 1) - newXedges(newBin)
      if (binWidth .le. 0.0) then
        status = -4
  return
      end if
    end do

    newYvalues = 0.0 ! vector op, default.

    oldStartBin = 1
    NEW: do newBin = 1, numNewBins
      ! Find first old bin that overlaps newBin:
      foundOverlappingBin = .false. ! default
      OLD1: do oldBin = oldStartBin, numOldBins
        if (oldXedges(oldBin + 1) .gt. newXedges(newBin)&
        .and. oldXedges(oldBin) .lt. newXedges(newBin + 1)) then
          foundOverlappingBin = .true.
          oldStartBin = oldBin
      exit OLD1
        end if
      end do OLD1

      if (.not. foundOverlappingBin) then ! have either reached the end of the
      ! old series, or not reached the start. Leave new values at default 0.0.
    exit NEW
      end if

      ! Find the last old bin that overlaps newBin:
      foundOverlappingBin = .true. ! default
      oldFinisBin = numOldBins ! default
      OLD2: do oldBin = oldStartBin + 1, numOldBins
        if (oldXedges(oldBin) .ge. newXedges(newBin + 1)) then
          foundOverlappingBin = .false.
          oldFinisBin = oldBin - 1
      exit OLD2
        end if
      end do OLD2

      if (foundOverlappingBin) then ! have reached the end of the
      ! old series without leaving the new bin.
        oldFinisBin = numOldBins
      end if

      ! Now work though all the old bins that overlap the new, and pour in
      ! the correct proportional amounts from them into newBin:
      OLD3: do oldBin = oldStartBin, oldFinisBin
        if (oldXedges(oldBin + 1) .lt. newXedges(newBin + 1)) then
          if (oldXedges(oldBin) .gt. newXedges(newBin)) then !B - all
          ! of old bin goes into new:
            overlap = oldXedges(oldBin + 1) - oldXedges(oldBin)

          else !A - upper part of old goes into new:
            overlap = oldXedges(oldBin + 1) - newXedges(newBin)
          end if

        else
          if (oldXedges(oldBin) .gt. newXedges(newBin)) then !D - lower
          ! part of old goes into new:
            overlap = newXedges(newBin + 1) - oldXedges(oldBin)

          else !C - middle part of old goes into new:
            overlap = newXedges(newBin + 1) - newXedges(newBin)
          end if
        end if
        newYvalues(newBin) = newYvalues(newBin) + oldYvalues(oldBin)&
        * overlap
      end do OLD3

      oldStartBin = oldFinisBin
    end do NEW

    ! Divide by the new bin width:
    do newBin = 1, numNewBins
      binWidth = newXedges(newBin + 1) - newXedges(newBin)
      newYvalues(newBin) = newYvalues(newBin) / binWidth
    end do

  end subroutine rebinKernel

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module temp_array_utils
