!
! Module: temp_simulations
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines to generate a simulated XMM EPIC image.
! 
! Comments: TODO:
! 
! $Id: temp_simulations.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module temp_simulations

  use random_mod
  use my_psf_mod
  use fitsutils_mod ! indicize(), loadFitsArray(), copyAttribute()
  use dal
  use cal
  use constants
  use coordinate_mod
  use mathutils_mod
  use errorHandling

  implicit none 

  private 

  public padOutMask
  public makeSimulatedSourceList
  public makeSimulatedImage
  public findNearestSimSource
  public probOfOneOrMoreByChance
  public flagAlsoRans
  public eventizeImage

  interface makeSimulatedImage
    module procedure makeSimulatedImage_withSet
    module procedure makeSimulatedImage_withList
  end interface

  type(wcsDataType) :: wcs
  integer(int32) ::&
    instrumentId
  type(SpacecraftAttitudeType) ::&
    scAttitude, instrumAttitude
  real(double) ::&
    expStartTimeStamp

  real(double) ::&
    mmToArcsec,&
    focalLengthMm

  type, public :: localSrcDataType
    real(single) ::&
      xPixels,&
      yPixels,&
      exposure,&
      rate,&
      peakRate,&
      extentWorld,&
      extentPixels
    real(double) ::&
      raDeg,&
      decDeg,&
      EeV,&
      thetaArcsec,&
      phi
    integer(int32) ::&
      flag
  end type localSrcDataType

  integer(int32), public ::&
    NO_M1_PSF     = 0,&
    NO_M2_PSF     = 1,&
    NO_PN_PSF     = 2,&
    NOT_IN_M1_EXP = 3,&
    NOT_IN_M2_EXP = 4,&
    NOT_IN_PN_EXP = 5

contains 

  !---------------------------------------------------------------------
  subroutine padOutMask(inMaskFileName, paddedMaskName, padSizeArcsec_arg)
  !---------------------------------------------------------------------
  ! Two masks are used in compiling the source list, described as follows.
  ! Source positions are chosen with random sky coverage, but only those
  ! positions that fall within the first mask are included in the
  ! simulation. The second mask is derived from the exposure map, being
  ! true where this map is greater than zero. This mask is used in
  ! conjunction with the requested source sky density to derive the total
  ! number of sources in the simulation (or vice versa). The second mask
  ! should normally be of smaller extent than the first, by a few multiples
  ! of the source PSF width, in order that the average flux from summed
  ! source wings doesn't decrease by too significant an amount near the
  ! edges of the field. This subroutine is intended to create mask 1 from
  ! mask 2. The result is saved (as a boolean) in another file.

    implicit none

    character(*), intent(in) :: inMaskFileName, paddedMaskName
    real(single), intent(in), optional :: padSizeArcsec_arg

    integer :: xsize, ysize, xiPad, yiPad, status,&
      xi, yi, startXi, finisXi, startYi, finisYi
    logical, pointer :: mask(:,:)
    logical, allocatable :: padArray(:,:), edgesImage(:,:)
    real(single) :: x, y, padSizeArcsec

    !-------------------------------------------------------------------
    ! Start

if (present(padSizeArcsec_arg)) then
  padSizeArcsec = padSizeArcsec_arg
else
  padSizeArcsec = psfTruncRadiusArcsec ! defined in my_psf_mod
end if

    ! AND together the input masks:

    call loadMaskImage(inMaskFileName, mask)
    xsize = size(mask, 1)
    ysize = size(mask, 2)

    xiPad = ceiling(padSizeArcsec / 3600.0 / abs(wcs%x%pixelSizeWorld))
    yiPad = ceiling(padSizeArcsec / 3600.0 / abs(wcs%y%pixelSizeWorld))

    allocate(padArray(-xiPad:xiPad,-yiPad:yiPad))
    padArray = .false. ! default
    do xi = -xiPad, xiPad
      x = xi / real(xiPad)
      do yi = -yiPad, yiPad
        y = yi / real(yiPad)
        if (x*x+y*y .lt. 1.0) padArray(xi, yi) = .true.
      end do
    end do

    allocate(edgesImage(xsize, ysize))

    call findEdges(mask, edgesImage, status=status)
if (status .ne. 0) call error('findEdgesStatus','')

    do xi = 1, xsize
    if (all(.not. edgesImage(xi, :))) cycle
      do yi = 1, ysize
      if (.not. edgesImage(xi, yi)) cycle
        startXi = max(1,     xi - xiPad)
        finisXi = min(xsize, xi + xiPad)
        startYi = max(1,     yi - yiPad)
        finisYi = min(ysize, yi + yiPad)

        mask(startXi:finisXi, startYi:finisYi)&
        = mask(startXi:finisXi, startYi:finisYi)&
        .or. padArray(startXi-xi:finisXi-xi,startYi-yi:finisYi-yi)
      end do
    end do
    deallocate(edgesImage, padArray)

    call saveMaskImage(mask, inMaskFileName, paddedMaskName)
    deallocate(mask)

  end subroutine padOutMask

  !---------------------------------------------------------------------
  subroutine loadMaskImage(maskSetName, mask)
  !---------------------------------------------------------------------
    
    implicit none

    character(*), intent(in) :: maskSetName
    logical, pointer ::&
      mask(:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray
    real(single), pointer ::&
      preMask(:,:)

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(maskSetName, READ)

!    call getCalInfo(set, instrumentId, expStartTimeStamp, scAttitude&
!    , instrumAttitude) ! also sets cal instrument state.

    inArray = array(set, 0, READ)
    call getImageWcsInfo(set, inArray, wcs)
    call loadImage(inArray, preMask)

    call release(set)

    allocate(mask(size(preMask, 1), size(preMask, 2)))
    mask = .false.
    where(preMask .gt. 0.0) mask = .true.
    deallocate(preMask)

  end subroutine loadMaskImage

  !---------------------------------------------------------------------
  subroutine loadImage(inArray, image)
  !---------------------------------------------------------------------
  ! Adapted from asmooth-2.5
  !
  ! Note that (up til 23 Apr 2002 anyway) the DAL call boolArray2Data(array)
  ! is not yet implemented.

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(single), dimension(:,:), pointer :: image

    integer(int32), dimension(:), pointer ::&
      dimVector
    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:), pointer ::&
      arrayDataReal64
!    logical(bool), dimension(:,:), pointer ::&
!      arrayDataBoolean
    integer(int16) :: xsize, ysize

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)
    xsize = dimVector(0)
    ysize = dimVector(1)

    allocate(image(xsize, ysize))

    if (dataType(inArray) .eq. INTEGER8) then
      arrayDataInt8 => int8Array2Data(inArray)
      image = real(arrayDataInt8, single)
    else if (dataType(inArray) .eq. INTEGER16) then
      arrayDataInt16 => int16Array2Data(inArray)
      image = real(arrayDataInt16, single)
    else if (dataType(inArray) .eq. INTEGER32) then
      arrayDataInt32 => int32Array2Data(inArray)
      image = real(arrayDataInt32, single)
    else if (dataType(inArray) .eq. REAL32) then
      arrayDataReal32 => real32Array2Data(inArray)
      image = arrayDataReal32
!    else if (dataType(inArray) .eq. BOOLEAN) then
!      arrayDataBoolean => boolArray2Data(inArray)
!      image = 0.0
!      where(arrayDataBoolean) image = 1.0
    else ! this gives a DAL error if data type is none of the above.
      arrayDataReal64 => real64Array2Data(inArray)
      image = real(arrayDataReal64, single)
    end if

  end subroutine loadImage

  !---------------------------------------------------------------------
  subroutine saveMaskImage(outMask, inSetName, outSetName)
  !---------------------------------------------------------------------
  ! Note that (up til 23 Apr 2002 anyway) the DAL call boolArray2Data(array)
  ! is not yet implemented.

    implicit none

    logical, intent(in) ::&
      outMask(:,:)
    character(*), intent(in) ::&
      inSetName, outSetName

    type(DataSetT) :: inSet, outSet
    type(ArrayT) :: inArray, outArray
!    logical(bool), dimension(:,:), pointer ::&
!      arrayDataBoolean
real(single), dimension(:,:), pointer ::&
arrayDataReal32
    integer(int32), dimension(:), pointer ::&
      dimVector

    !-------------------------------------------------------------------
    ! Start:

    inSet = dataSet(inSetName, READ)
    inArray = array(inSet, 0, READ)
    outSet = dataSet(outSetName, CREATE)
    allocate(dimVector(0:1))
    dimVector(0) = size(outMask, 1)
    dimVector(1) = size(outMask, 2)
!    outArray = addArray(outSet, 'PRIMARY', BOOLEAN, dimVector, position=0)
outArray = addArray(outSet, 'PRIMARY', REAL32, dimVector, position=0)
    call setAttributes(attributable(outSet), attributable(inSet))

    call copyVitalArrayKwds(inArray, outArray)
    ! This is done instead of a total copy of all array-header keywords because
    ! doing the latter stuffs up the DSS.

!    arrayDataBoolean => boolArray2Data(outArray)
!    arrayDataBoolean = outMask
arrayDataReal32 => real32Array2Data(outArray)
arrayDataReal32 = 0.0 ! default
where(outMask) arrayDataReal32 = 1.0

    call release(outSet)
    call release(inSet)

    deallocate(dimVector)

  end subroutine saveMaskImage

  !---------------------------------------------------------------------
  subroutine makeSimulatedSourceList(sourceListName, maskSetName, faintestRate&
  , numSources, brightestExpectedRate, sourceDensityArcsec, maxExtentArcsec)
  !---------------------------------------------------------------------
  ! This subroutine creates a list of simulated sources. Each source has
  ! a position in the sky (which must fall within the area specified by
  ! the input mask) and a brightness rate in counts per second. The
  ! positions are random with a uniform distribution and the brightness
  ! rates are random with an L^-5/2 distribution. Two parameters are
  ! necessary in order to produce this list: the rate of the faintest
  ! source (ie, where to truncate L^-5/2) and the sky density of sources.
  ! In the present interface there are three ways in which the user can
  ! specify the sky density of sources: this quantity can be specified
  ! directly; it can be specified implicitly by giving the total number
  ! of sources; or by giving the brightest expected rate (ie, the value
  ! of L_max such that the probability of a source appearing in the list
  ! with L >= L_max is equal to 1/2).

    implicit none

    character(*),   intent(in)           :: sourceListName, maskSetName
    real(single),   intent(in)           :: faintestRate
    integer(int32), intent(in), optional :: numSources
    real(single),   intent(in), optional ::&
      brightestExpectedRate, sourceDensityArcsec, maxExtentArcsec

    integer(int32) :: localNumSources, src
    type(localSrcDataType), allocatable :: srcData(:)
    real(single) ::&
      maxExtentDeg, localBrightestRate, localSourceDensityArcsec,&
      sourceDensityPixels
    logical, pointer :: maskImage(:,:)
    real(double) :: pixelAreaDeg
    integer :: status
    real(double), allocatable :: tanX(:), tanY(:), ra(:), dec(:)

    !-------------------------------------------------------------------
    ! Start

    call loadMaskImage(maskSetName, maskImage)
    if (all(.not. maskImage))&
call error('emptyMask','')

    if (present(maxExtentArcsec)) then
      maxExtentDeg = maxExtentArcsec / 3600.0
    else
      maxExtentDeg = 0.0
    end if

    pixelAreaDeg = abs(wcs%x%pixelSizeWorld * wcs%y%pixelSizeWorld)

    if (present(numSources)) then
      if (present(brightestExpectedRate) .or. present(sourceDensityArcsec))&
        call warning('','')

      localNumSources = numSources
      localSourceDensityArcsec = localNumSources / real(count(maskImage))&
        / pixelAreaDeg / 3600.0 / 3600.0
      localBrightestRate = faintestRate * (2.0 * localNumSources)**(2/3.0)

    else if (present(brightestExpectedRate)) then
      if (present(sourceDensityArcsec))&
        call warning('','')

      localBrightestRate = brightestExpectedRate
      localNumSources = 0.5 * (localBrightestRate / faintestRate) ** 1.5
      localSourceDensityArcsec = localNumSources / real(count(maskImage))&
        / pixelAreaDeg / 3600.0 / 3600.0

      !
      ! This is derived as follows: the total number density p(L) dL of sources
      ! between luminosity L and L + dL is given by
      !
      !		           3 N
      !		p(L) dL = ----- (L_min)^(3/2) dL L^(-5/2)
      !		           2 A
      !
      ! where N is the total number of sources in the field, A is the area of
      ! the field and L_min is the mininum luminosity (this of course can't
      ! be zero, otherwise we are into Olber's paradox territory - the
      ! integral doesn't converge). It is thus easily seen that the number
      ! N(>L) of sources of luminosity greater than L is given by
      !
      !		                            inf
      !		         3 N                /
      !		N(>L) = ----- (L_min)^(3/2) | dL L^(-5/2)
      !		          2                 /
      !		                           L
      ! or
      !		          ( L_min )^(3/2)
      !		N(>L) = N (-------)
      !		          (   L   ).
      !
      ! If N is not known beforehand, it can be estimated if the approximate
      ! maximum luminosity L_max is known. Let us define L_max as the
      ! luminosity such that the total expected number N(>L_Max) of sources
      ! brighter than this in the field (that is, N times the probability of
      ! any one source being brighter than L_max) equals 0.5. From the last
      ! equation this gives
      !
      !		        ( L_max )^(3/2)
      !		N = 0.5 (-------)
      !		        ( L_min ).

    else if (present(sourceDensityArcsec)) then
      localSourceDensityArcsec = sourceDensityArcsec
      localNumSources = localSourceDensityArcsec * count(maskImage)&
        * pixelAreaDeg * 3600.0 * 3600.0
      localBrightestRate = faintestRate * (2.0 * localNumSources)**(2/3.0)

    else
call error('insufficientInfo','')
    end if

if (localNumSources .le. 0.0) call error('noSources','')

    sourceDensityPixels = localNumSources / real(count(maskImage))

    print *, ''
    print *, '       num of mask pixels = ', count(maskImage)
    print *, '            brightestRate = ', localBrightestRate
    print *, '             faintestRate = ', faintestRate
    print *, 'sources per square arcsec = ', localSourceDensityArcsec
    print *, ''
    print *, '************************'
    print *, 'Number of sources:', localNumSources
    print *, '************************'
    print *, ''

    allocate(srcData(localNumSources))
    call makeSourceList(maskImage, wcs, faintestRate, maxExtentDeg, srcData)
    call srcImageDump(size(maskImage, 1), size(maskImage, 2), wcs, srcData)

    allocate(tanX(localNumSources), tanY(localNumSources)&
    , ra(localNumSources), dec(localNumSources))
    do src = 1, localNumSources
      tanX(src) = (srcData(src)%xPixels - wcs%x%refPixels)&
        * abs(wcs%x%pixelSizeWorld) * PiOn180
      tanY(src) = (srcData(src)%yPixels - wcs%y%refPixels)&
        * abs(wcs%y%pixelSizeWorld) * PiOn180
    end do

    call tanToRaDec(wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180&
    , tanX, tanY, ra, dec, status)
    if (status .ne. 0) then
print *, status
call error('tanToRaDecstatus','')
    end if

    srcData%raDeg  = ra  / PiOn180
    srcData%decDeg = dec / PiOn180

    call saveSourceList(srcData, maskSetName, sourceListName, wcs&
    , faintestRate, sourceDensityPixels)

    deallocate(maskImage)

  end subroutine makeSimulatedSourceList

  !---------------------------------------------------------------------
  subroutine makeSourceList(mask, maskWcs, minRate, maxExtentWorld, srcData)
  !---------------------------------------------------------------------
  ! Adapted from asmooth-2.5 (temp_ssclib)

    implicit none

    logical, intent(in) :: mask(:,:)
    type(wcsDataType), intent(in) :: maskWcs
    real(single), intent(in) :: minRate, maxExtentWorld
    type(localSrcDataType), intent(out) :: srcData(:)

    integer :: numSources, status, srcNum, trialSrc, xsize, ysize, xi, yi
    integer, parameter :: batchSize = 512
    real :: theta(batchSize), phi(batchSize), relativeFlux(batchSize),&
      x, y, maxR, extent(batchSize), coneHalfAngle
    logical :: edgeImage(size(mask, 1), size(mask, 2))
    real(single) ::&
      avePixelSizeWorld, xPixels, yPixels, xPixelsToRadians, yPixelsToRadians

    !-------------------------------------------------------------------
    ! Start

if (count(mask) .le. 0) call error('emptyMask','')

    xsize = size(mask, 1)
    ysize = size(mask, 2)

    call findEdges(mask, edgeImage, status=status)
if (status .ne. 0) call error('findEdgesStatus','')

    xPixelsToRadians = abs(maskWcs%x%pixelSizeWorld) * PiOn180
    yPixelsToRadians = abs(maskWcs%y%pixelSizeWorld) * PiOn180
    maxR = 0.0 ! default
    do xi = 1, xsize
    if (all(.not. edgeImage(xi, :))) cycle
      x = (xi - maskWcs%x%refPixels) * xPixelsToRadians
      do yi = 1, ysize
      if (.not. edgeImage(xi, yi)) cycle
        y = (yi - maskWcs%y%refPixels) * yPixelsToRadians
        maxR = max(maxR, x * x + y * y)
      end do
    end do
    coneHalfAngle = atan(sqrt(maxR))

    avePixelSizeWorld = (abs(maskWcs%X%pixelSizeWorld)&
    + abs(maskWcs%Y%pixelSizeWorld)) / 2.0

    numSources = size(srcData)

    srcNum = 0
    BATCH: do
      call getSineRandoms(coneHalfAngle, theta, status)
if (status .ne. 0) then
print *, 'getSineRandoms status:', status
call error('getSineRandomsStatus','')
end if
      call getRandomNumbers(phi)
      phi = TwoPi * phi
      call getMinus5on2Randoms(relativeFlux)
      call getRandomNumbers(extent)

      do trialSrc = 1, batchSize
        x = cos(phi(trialSrc)) * tan(theta(trialSrc))
        y = sin(phi(trialSrc)) * tan(theta(trialSrc))
        xPixels = maskWcs%x%refPixels + x / xPixelsToRadians
        yPixels = maskWcs%y%refPixels + y / yPixelsToRadians
        xi = nint(xPixels)
        yi = nint(yPixels)

        if (mask(xi, yi)) then
          srcNum = srcNum + 1
          srcData(srcNum)%xPixels = xPixels 
          srcData(srcNum)%yPixels = yPixels
          srcData(srcNum)%rate = relativeFlux(trialSrc) * minRate
          srcData(srcNum)%extentWorld = extent(trialSrc) * maxExtentWorld
          srcData(srcNum)%extentPixels = srcData(srcNum)%extentWorld&
          / avePixelSizeWorld

    if (srcNum .ge. numSources) exit BATCH
        end if
      end do
    end do BATCH

call test_dumpColumn(srcData%rate / minRate)

  end subroutine makeSourceList

  !---------------------------------------------------------------------
  subroutine srcImageDump(xsize, ysize, maskWcs, srcData)
  !---------------------------------------------------------------------
  ! This makes an image showing the location of the sources. A non-varying
  ! psf approximation is used, which means that it is much quicker to make
  ! this image than it is to make the 'proper' simulation image. This
  ! subroutine can thus be used as a reality check on the source list.

    use dump_mod
    use fitsutils_mod

    implicit none

    integer, intent(in) :: xsize, ysize
    type(wcsDataType), intent(in) :: maskWcs
    type(localSrcDataType), intent(in) :: srcData(:)

    real(single), pointer ::&
      coarsePSF(:,:)
    real(single) ::&
      coarseImage(xsize, ysize), xDeg, yDeg
    integer(int32) ::&
      src, numSources, xi, yi, psfLoXi, psfHiXi, psfLoYi, psfHiYi,&
      xiStart, xiFinis, yiStart, yiFinis

    !-------------------------------------------------------------------
    ! Start

    call makeCoarsePSF(coarsePSF)
!call dumpImageToFits(coarsePSF, 'coarse_psf.ds')
    psfLoXi = lbound(coarsePSF, 1)
    psfHiXi = ubound(coarsePSF, 1)
    psfLoYi = lbound(coarsePSF, 2)
    psfHiYi = ubound(coarsePSF, 2)

    numSources = size(srcData)
    coarseImage = 0.0 ! default
    do src = 1, numSources
      xDeg = cos(srcData(src)%phi)&
      * tan(srcData(src)%thetaArcsec / RadToArcsec) / PiOn180
      yDeg = sin(srcData(src)%phi)&
      * tan(srcData(src)%thetaArcsec / RadToArcsec) / PiOn180
!      xi = maskWcs%x%refPixels - xDeg / maskWcs%x%pixelSizeWorld
!      yi = maskWcs%y%refPixels - yDeg / maskWcs%y%pixelSizeWorld
! should be as follows?????
      xi = maskWcs%x%refPixels + xDeg / maskWcs%x%pixelSizeWorld
      yi = maskWcs%y%refPixels + yDeg / maskWcs%y%pixelSizeWorld

    if (xi + psfHiXi .lt. 1 .or. xi + psfLoXi .gt. xsize&
    .or. yi + psfHiYi .lt. 1 .or. yi + psfLoYi .gt. ysize) cycle

      xiStart = max(1,     xi + psfLoXi)
      xiFinis = min(xsize, xi + psfHiXi)
      yiStart = max(1,     yi + psfLoYi)
      yiFinis = min(ysize, yi + psfHiYi)

      coarseImage(xiStart:xiFinis, yiStart:yiFinis)&
      = coarseImage(xiStart:xiFinis, yiStart:yiFinis)&
      + srcData(src)%rate&
      * coarsePSF(xiStart-xi:xiFinis-xi, yiStart-yi:yiFinis-yi)
    end do

    deallocate(coarsePSF)

    call dumpImageToFits(coarseImage, 'coarse_src_map.ds')

  end subroutine srcImageDump

  !---------------------------------------------------------------------
  subroutine makeCoarsePSF(coarsePSF)
  !---------------------------------------------------------------------

    implicit none

    real(single), pointer :: coarsePSF(:,:)

    real(single), parameter :: sigmaInPixels = 1.5, sigmaMultiple = 7.0
    integer(int32) :: halfWidth, xi, yi
    real(single) :: x, y

    !-------------------------------------------------------------------
    ! Start

    halfWidth = ceiling(sigmaInPixels * sigmaMultiple)
    allocate(coarsePSF(-halfWidth:halfWidth, -halfWidth:halfWidth))

    do xi = -halfWidth, halfWidth
      x = xi / sigmaInPixels
      do yi = -halfWidth, halfWidth
        y = yi / sigmaInPixels
        coarsePSF(xi, yi) = exp(-(x * x + y * y) / 2.0)
      end do
    end do

  end subroutine makeCoarsePSF

  !---------------------------------------------------------------------
  subroutine saveSourceList(srcData, infoSetName, sourceListName, wcs&
  , faintestRate, sourceDensityPixels)
  !---------------------------------------------------------------------

    use fitsutils_mod

    implicit none

    type(localSrcDataType), intent(in) :: srcData(:)
    character(*), intent(in) :: infoSetName, sourceListName
    type(wcsDataType), intent(in) :: wcs
    real(single), intent(in) :: faintestRate, sourceDensityPixels

    type(DataSetT) :: srcSet, infoSet
    type(TableT)   :: tab
    type(ArrayT) :: infoArray
    integer(int32) :: int32Vector(size(srcData))
!    real(single) :: real32Vector(size(srcData))
!    real(double) :: real64Vector(size(srcData))
!    logical :: nullVector(size(srcData))
    integer :: status

    !-------------------------------------------------------------------
    ! Start

    srcSet = dataSet(sourceListName, CREATE)
    infoSet  = dataSet(infoSetName, READ)
    infoArray = array(infoSet, 0, READ)

!print *, 'infoSetName: ', infoSetName
!print *, 'hasAttribute(infoSet, CTYPE1)?', hasAttribute(dataSet(infoSetName, READ), 'CTYPE1')

    ! The subroutine copyAttribute() is in ssclib/fitsutils_mod.

    call copyAttribute(infoSet, 'INSTRUME', srcSet, status)
    call copyAttribute(infoSet, 'RA_PNT',   srcSet, status)
    call copyAttribute(infoSet, 'DEC_PNT',  srcSet, status)
    call copyAttribute(infoSet, 'PA_PNT',   srcSet, status)
    call copyAttribute(infoSet, 'DATE-OBS', srcSet, status)

    call setAttribute(srcSet, 'TCTYPE1', wcs%X%ctype)
    call setAttribute(srcSet, 'TCRPIX1', wcs%X%refPixels)
    call setAttribute(srcSet, 'TCRVAL1', wcs%X%refWorld)
    call setAttribute(srcSet, 'TCDELT1', wcs%X%pixelSizeWorld)
    call setAttribute(srcSet, 'TCTYPE2', wcs%Y%ctype)
    call setAttribute(srcSet, 'TCRPIX2', wcs%Y%refPixels)
    call setAttribute(srcSet, 'TCRVAL2', wcs%Y%refWorld)
    call setAttribute(srcSet, 'TCDELT2', wcs%Y%pixelSizeWorld)

    call release(infoSet)

    tab = addTable(srcSet, 'SRCLIST', size(srcData))

    call setAttribute(tab, 'NDENSITY', sourceDensityPixels, 'sources / pixel')
    call setAttribute(tab, 'L_FAINT',  faintestRate, 'counts / sec'&
      , 'Faint-end cutoff in the distribution of sources.')

!    int32Vector = indicize(1, size(srcData))
!!!!    call myAddColumn(tab, 'BOX_ID_SRC', int32Vector)
!    if (instrumentId .eq. EPN) then
!      int32Vector = 1
!    else if (instrumentId .eq. EMOS1) then
!      int32Vector = 2
!    else if (instrumentId .eq. EMOS2) then
!      int32Vector = 3
!    else
!call error('','')
!    end if
int32Vector = -1 ! indicates sim source
    call myAddColumn(tab, 'ID_INST', int32Vector)
!!!    int32Vector = 1
int32Vector = 10
    call myAddColumn(tab, 'ID_BAND', int32Vector)
    call myAddColumn(tab, 'EXT',       srcData%extentPixels, 'image pixels')
    call myAddColumn(tab, 'RATE',      srcData%rate, 'counts/pixel')
    call myAddColumn(tab, 'RA',        srcData%raDeg,  'deg')
    call myAddColumn(tab, 'DEC',       srcData%decDeg, 'deg')
    call myAddColumn(tab, 'X_IMA',     srcData%xPixels, 'image pix')
    call myAddColumn(tab, 'Y_IMA',     srcData%yPixels, 'image pix')
int32Vector = 0
    call myAddColumn(tab, 'FLAG',      int32Vector)
    call myAddColumn(tab, 'LOG10_SR',  log10(srcData%rate))

    call release(srcSet)

  end subroutine saveSourceList

  !---------------------------------------------------------------------
  subroutine makeSimulatedImage_withSet(expMapSetName, sourceListSetName&
  , sourceListTabName, rawImageName, psfModel)
  !---------------------------------------------------------------------

    use my_psf_mod ! for getWcsInfo

    implicit none

    character(*), intent(in) :: expMapSetName,&
      sourceListSetName, sourceListTabName, rawImageName, psfModel

    type(localSrcDataType), allocatable :: srcData(:)
    type(DataSetT) :: set
    integer(int32) ::&
      numSources
    type(TableT) :: tab
    type(ColumnT) ::&
      raDegCol, decDegCol, rateCol, extCol, ximaCol, yimaCol, flagCol
    real(single), pointer :: rate(:), extentPixels(:), xima(:), yima(:)
    real(double), pointer :: raDeg(:), decDeg(:)
    real(single) :: avePixelSizeWorld, sourceDensityPixels, faintestRate
    integer(int32), pointer :: flag(:)

    !-------------------------------------------------------------------
    ! Start

    set = dataSet(sourceListSetName, READ)
!    call getWcsInfo(set, wcs)
wcs%X%refPixels       = int32Attribute(set, 'TCRPIX1')
wcs%Y%refPixels       = int32Attribute(set, 'TCRPIX2')
wcs%X%pixelSizeWorld = real32Attribute(set, 'TCDELT1')
wcs%Y%pixelSizeWorld = real32Attribute(set, 'TCDELT2')
wcs%X%refWorld       = real32Attribute(set, 'TCRVAL1')
wcs%Y%refWorld       = real32Attribute(set, 'TCRVAL2')
wcs%X%ctype          = stringAttribute(set, 'TCTYPE1')
wcs%Y%ctype          = stringAttribute(set, 'TCTYPE2')

    avePixelSizeWorld = (abs(wcs%X%pixelSizeWorld)&
    + abs(wcs%Y%pixelSizeWorld)) / 2.0

    tab = table(set, sourceListTabName)

    numSources = numberOfRows(tab)
    allocate(srcData(numSources))

    raDegCol  = column(tab, 'RA')
    decDegCol = column(tab, 'DEC')
    rateCol   = column(tab, 'RATE')
    extCol    = column(tab, 'EXT')
    ximaCol   = column(tab, 'X_IMA')
    yimaCol   = column(tab, 'Y_IMA')
    flagCol   = column(tab, 'FLAG')

    raDeg        => real64Data(raDegCol)
    decDeg       => real64Data(decDegCol)
    rate         => real32Data(rateCol)
    extentPixels => real32Data(extCol)
    xima         => real32Data(ximaCol)
    yima         => real32Data(yimaCol)
    flag         => int32Data(flagCol)

    srcData%raDeg        = raDeg
    srcData%decDeg       = decDeg
    srcData%rate         = rate
    srcData%extentWorld  = extentPixels * avePixelSizeWorld
    srcData%xPixels      = xima
    srcData%yPixels      = yima
    srcData%flag         = flag

    call release(set)

    srcData%EeV = 1000.0 !***** temporary bodge. Should get band from expmap
    ! and average it.

    call makeSimulatedImage_withList(expMapSetName, srcData&
      , rawImageName, psfModel)

    set = dataSet(sourceListSetName, MODIFY)
    tab = table(set, sourceListTabName)
    flagCol   = column(tab, 'FLAG')
    flag => int32Data(flagCol)
    flag = srcData%flag

    sourceDensityPixels = real32Attribute(tab, 'NDENSITY')
    faintestRate        = real32Attribute(tab, 'L_FAINT')
    call release(set)

    set = dataSet(rawImageName, MODIFY)
    call setAttribute(set, 'NDENSITY', sourceDensityPixels, 'sources / pixel')
    call setAttribute(set, 'L_FAINT',  faintestRate, 'counts / sec'&
      , 'Faint-end cutoff in the distribution of sources.')
    call release(set)

  end subroutine makeSimulatedImage_withSet

  !---------------------------------------------------------------------
  subroutine makeSimulatedImage_withList(expMapSetName, srcData&
  , rawImageName, psfModel)
  !---------------------------------------------------------------------

    use my_psf_mod ! for getCalInfo
    use sscmisc

    implicit none

    character(*), intent(in) :: expMapSetName
    type(localSrcDataType), intent(inout) :: srcData(:)
    character(*), intent(in) :: rawImageName, psfModel

    integer(int32) :: numSources, src
    integer(int16) :: xsize, ysize
    real(single), allocatable :: image(:,:)
    real(single), pointer :: expMapImage(:,:), patch(:,:)
    real(double) :: tanX, tanY, energyFraction, pixelAreaDeg, pixelAreaArcsec
    real(single) ::&
      pixelSizeXmm, pixelSizeYmm, fovCentreXpixels, fovCentreYpixels,&
      sigmaArcsec = 5.0, sigmaXpixels, sigmaYpixels
    integer :: status
    type(PsfBinSizeT) :: psfPixelSizeMm, psfPixelSizePixels
logical :: testflag = .false.

    !-------------------------------------------------------------------
    ! Start

    numSources = size(srcData)

if (numSources .le. 0) call error('noSources','')

    call loadExposureMap(expMapSetName, expMapImage) ! sets cal->instrument

    xsize = size(expMapImage, 1)
    ysize = size(expMapImage, 2)
    allocate(image(xsize, ysize))
    image = 0.0 ! default

    if (psfModel == 'coarse') then
      pixelAreaDeg = abs(wcs%x%pixelSizeWorld * wcs%y%pixelSizeWorld)

      pixelAreaArcsec = pixelAreaDeg * 3600.0 * 3600.0
      sigmaXpixels = sigmaArcsec / abs(wcs%x%pixelSizeWorld) / 3600.0
      sigmaYpixels = sigmaArcsec / abs(wcs%y%pixelSizeWorld) / 3600.0

      do src = 1, numSources
        call getPsfPatch(srcData(src)%xPixels, srcData(src)%yPixels&
        , sigmaXpixels, sigmaYpixels, patch, testflag)

        srcData(src)%peakRate = srcData(src)%rate / sum(patch)
        call addPatch(image, patch, srcData(src)%peakRate)
        deallocate(patch)
      end do

    else if (psfModel == 'cal_acc_2') then
      ! Set up rest of cal (inst already set in call to loadExposureMap()):
      call CAL_setState(accuracyLevel=ACCURACY_MEDIUM)

      call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
      mmToArcsec = 3600.0 / PiOn180 / focalLengthMm

      ! Find pixel coords of theta=0 point (centre of fov):
      call raDecToTan(wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180&
        , instrumAttitude%ra, instrumAttitude%dec, tanX, tanY, status)

      if (status .ne. 0) then
print *, status
call error('tanToRaDecstatus','')
      end if

      fovCentreXpixels = (tanX / PiOn180) / abs(wcs%x%pixelSizeWorld)&
        + wcs%x%refPixels
      fovCentreYpixels = (tanY / PiOn180) / abs(wcs%y%pixelSizeWorld)&
        + wcs%y%refPixels

      pixelSizeXmm = abs(wcs%x%pixelSizeWorld) * 3600.0 / mmToArcsec
      pixelSizeYmm = abs(wcs%y%pixelSizeWorld) * 3600.0 / mmToArcsec

      energyFraction = 0.9d0
!*** make it changeable

      do src = 1, numSources
        call printEvery(int(src), int(numSources), fraction=0.05)

        if (expMapImage(floor(srcData(src)%xPixels), floor(srcData(src)%yPixels))&
        < 0.0) then
          select case(instrumentId)
          case(EPN)
            srcData(src)%flag = ibset(srcData(src)%flag, NOT_IN_PN_EXP)
          case(EMOS1)
            srcData(src)%flag = ibset(srcData(src)%flag, NOT_IN_M1_EXP)
          case(EMOS2)
            srcData(src)%flag = ibset(srcData(src)%flag, NOT_IN_M2_EXP)
          end select
        end if

        call getCalPsfPatch(srcData(src)%xPixels, srcData(src)%yPixels&
        , fovCentreXpixels, fovCentreYpixels, energyFraction&
        , pixelSizeXmm, pixelSizeYmm, patch, psfPixelSizeMm, testflag, status)

        if (status < 0) then
          select case(instrumentId)
          case(EPN)
            srcData(src)%flag = ibset(srcData(src)%flag, NO_PN_PSF)
          case(EMOS1)
            srcData(src)%flag = ibset(srcData(src)%flag, NO_M1_PSF)
          case(EMOS2)
            srcData(src)%flag = ibset(srcData(src)%flag, NO_M2_PSF)
          end select
      cycle
        end if

        psfPixelSizePixels%binX = psfPixelSizeMm%binX / pixelSizeXmm
        psfPixelSizePixels%binY = psfPixelSizeMm%binY / pixelSizeYmm

        call resizePatch(patch, srcData(src)%xPixels+0.5&
          , srcData(src)%yPixels+0.5, psfPixelSizePixels)
!** NOTE! The addition of 0.5 to both coords here is a fudge to remove an
!** offset of 0.5 pixels (probably originating in the way X_IMA etc are
!** defined) in the detected coordinates.

        srcData(src)%peakRate = srcData(src)%rate * energyFraction / sum(patch)
        call addPatch(image, patch, srcData(src)%peakRate)
        deallocate(patch)
      end do

    else
call error('badPsfModel','')
    end if ! psfModel

    call saveOutImage(image, expMapSetName, rawImageName)

  end subroutine makeSimulatedImage_withList

  !---------------------------------------------------------------------
  subroutine getPsfPatch(srcXpixels, srcYpixels&
  , sigmaXpixels, sigmaYpixels, patch, testflag)
  !---------------------------------------------------------------------

    use dump_mod
    use sscmisc

    implicit none

    real(single), intent(in) ::&
      srcXpixels, srcYpixels, sigmaXpixels, sigmaYpixels
    real(single), pointer :: patch(:,:)
logical, intent(in) :: testflag

    real(single) :: factor = 7.0
    integer(int16) :: xStart, xFinis, yStart, yFinis
    real(single), allocatable :: x(:), y(:), xExp(:), yExp(:)
integer(int16), pointer :: tempp(:), xi(:), yi(:)

    !-------------------------------------------------------------------
    ! Start
if (testflag) print *, 'aaa'

!sigmaXpixels = 10.0 * sigmaXpixels
!sigmaYpixels = 10.0 * sigmaYpixels

    xStart = floor(  srcXpixels - factor * sigmaXpixels)
if (testflag) print *, 'bbb'
    xFinis = ceiling(srcXpixels + factor * sigmaXpixels)
if (testflag) print *, 'ccc'
    yStart = floor(  srcYpixels - factor * sigmaYpixels)
if (testflag) print *, 'ddd'
    yFinis = ceiling(srcYpixels + factor * sigmaYpixels)
if (testflag) print *, 'eee'

    allocate(patch(xStart:xFinis,yStart:yFinis))
if (testflag) print *, 'fff'
    allocate(x(xStart:xFinis), xExp(xStart:xFinis))
if (testflag) print *, 'ggg'
    allocate(y(yStart:yFinis), yExp(yStart:yFinis))
if (testflag) print *, 'hhh'

if (testflag) print *, 'xStart, xFinis, srcXpixels, RootTwo, sigmaXpixels: '&
, xStart, xFinis, srcXpixels, RootTwo, sigmaXpixels
if (testflag) then
!  allocate(tempp(xStart:xFinis))
  tempp => indicize(xStart, xFinis, testflag)
  print *, tempp
  deallocate(tempp)
end if

xi => indicize(xStart, xFinis)
    x = real(xi, single)
if (testflag) print *, 'iii'
    x = x - srcXpixels
if (testflag) print *, 'jjj'
    x = x / RootTwo
if (testflag) print *, 'kkk'
    x = x / sigmaXpixels
if (testflag) print *, 'lll'

!!!    x = (real(indicize(xStart, xFinis), single) - srcXpixels)&
!!!    / RootTwo / sigmaXpixels
if (testflag) print *, 'mmm'
yi => indicize(yStart, yFinis)
    y = (real(yi, single) - srcYpixels)&
    / RootTwo / sigmaYpixels
if (testflag) print *, 'nnn'
    xExp = exp(-(x*x))
if (testflag) print *, 'ooo'
    yExp = exp(-(y*y))
if (testflag) print *, 'ppp'

    patch = spread(xExp, 2, (1 + yFinis - yStart))&
    * spread(yExp, 1, (1 + xFinis - xStart))
if (testflag) print *, 'mmm'

deallocate(xi, x, xExp, yi, y, yExp)
!call dumpImageToFits(patch, 'coarse_psf.ds')
!call error('','')

  end subroutine getPsfPatch

  !---------------------------------------------------------------------
  subroutine addPatch(image, patch, rate)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(inout) :: image(:,:)
    real(single), pointer :: patch(:,:)
    real(single), intent(in) :: rate

    integer :: xStart, xFinis, yStart, yFinis, xsize, ysize

    !-------------------------------------------------------------------
    ! Start

    xStart = lbound(patch, 1)
    xFinis = ubound(patch, 1)
    yStart = lbound(patch, 2)
    yFinis = ubound(patch, 2)
    xsize = size(image, 1)
    ysize = size(image, 2)

    if (xStart > xsize .or. xFinis < 1 .or. yStart > ysize .or. yFinis < 1) then
  return
    end if

    xStart = max(1, xStart)
    xFinis = min(xsize, xFinis)
    yStart = max(1, yStart)
    yFinis = min(ysize, yFinis)

    image(xStart:xFinis, yStart:yFinis) = image(xStart:xFinis, yStart:yFinis)&
    + patch(xStart:xFinis, yStart:yFinis) * rate

  end subroutine addPatch

  !---------------------------------------------------------------------
  subroutine getCalPsfPatch(srcXpixels, srcYpixels&
  , fovCentreXpixels, fovCentreYpixels, energyFraction&
  , pixelSizeXmm, pixelSizeYmm, patch, psfPixelSizeMm, testflag, status)
  !---------------------------------------------------------------------
  ! Requires cal to have been set up;
  !
  ! Makes use of the global variable mmToArcsec

    use dump_mod
    use sscmisc

    implicit none

    real(single), intent(in) ::&
      srcXpixels, srcYpixels,&
      fovCentreXpixels, fovCentreYpixels,&
      pixelSizeXmm, pixelSizeYmm
    real(double), intent(in) ::&
      energyFraction
    real(single), pointer :: patch(:,:)
    type(PsfBinSizeT), intent(out) :: psfPixelSizeMm
logical, intent(in) :: testflag
    integer, intent(out) :: status

    real(double) :: piMin, piMax, EeV, bogusThetaArcsec, bogusPhi, radius
    real(single) :: xMm, yMm
    type(PsfValidityRangesT) :: validPsfRanges
    type(PsfT) :: psfPointer
!    type(PsfBinSizeT) :: psfPixelSizeMm
    type(PsfWindowT) :: psfWindow

    !-------------------------------------------------------------------
    ! Start

    status = 0

piMin = 200.0
piMax = 12000.0

    EeV = (piMin + piMax) / 2.0 ! the same as is in emldetect.
    xMm = (srcXpixels - fovCentreXpixels) * pixelSizeXmm
    yMm = (srcYpixels - fovCentreYpixels) * pixelSizeYmm
    bogusThetaArcsec = sqrt(xMm * xMm + yMm * yMm) * mmToArcsec
    bogusPhi = atan2(-yMm, xMm)
    do while(bogusPhi < 0.0d0)
      bogusPhi = bogusPhi + TwoPi
    end do

    validPsfRanges = CAL_psfValidityRanges()

    if (EeV < validPsfRanges%eMin&
    .or. EeV > validPsfRanges%eMax&
    .or. bogusThetaArcsec < validPsfRanges%thetaMin&
    .or. bogusThetaArcsec > validPsfRanges%thetaMax&
    .or. bogusPhi < validPsfRanges%phiMin&
    .or. bogusPhi > validPsfRanges%phiMax) then
      status = -1
  return
    end if

    psfPointer = CAL_getPSF(EeV, bogusThetaArcsec, bogusPhi)
    psfPixelSizeMm = CAL_psfBinSize(psfPointer)

!    psfWindow = CAL_psfEnboxedEnergy(psfPointer, fraction=energyFraction)
radius = CAL_psfEncircledEnergy(psfPointer, fraction=energyFraction)
psfWindow%halfX = radius
psfWindow%halfY = radius

    call CAL_psfImage(psfPointer, psfWindow, patch)

  end subroutine getCalPsfPatch

  !---------------------------------------------------------------------
  subroutine resizePatch(patch, srcXpixels, srcYpixels, psfPixelSizePixels)
  !---------------------------------------------------------------------

    use dump_mod
    use temp_array_utils
    use mathutils_mod
    use sscmisc

    implicit none

    real(single), pointer :: patch(:,:)
    real(single), intent(in) :: srcXpixels, srcYpixels
    type(PsfBinSizeT), intent(in) :: psfPixelSizePixels

    integer(int16) :: newStartXi, newStartYi, newFinisXi, newFinisYi,&
      oldXsize, oldYsize
    real(single), pointer :: newPatch(:,:)
    real(single) :: oldHalfSizeX, oldHalfSizeY,&
      oldLoX, oldHiX, newLoX, newHiX, oldLoY, oldHiY, newLoY, newHiY
    integer(int16), pointer :: indices(:)
    real(single), allocatable ::&
      oldXedges(:), oldYedges(:), newXedges(:), newYedges(:)
    integer :: status

    !-------------------------------------------------------------------
    ! Start

    oldXsize = size(patch, 1)
    oldYsize = size(patch, 2)

    oldHalfSizeX = (1 + ubound(patch, 1) - lbound(patch, 1)) / 2.0
    oldHalfSizeY = (1 + ubound(patch, 2) - lbound(patch, 2)) / 2.0
    oldLoX = srcXpixels - psfPixelSizePixels%binX * oldHalfSizeX
    oldHiX = srcXpixels + psfPixelSizePixels%binX * oldHalfSizeX
    newStartXi = floor(oldLoX)
    newFinisXi = floor(oldHiX)
    newLoX = real(newStartXi)
    newHiX = real(newFinisXi + 1)
    oldLoY = srcYpixels - psfPixelSizePixels%binY * oldHalfSizeY
    oldHiY = srcYpixels + psfPixelSizePixels%binY * oldHalfSizeY
    newStartYi = floor(oldLoY)
    newFinisYi = floor(oldHiY)
    newLoY = real(newStartYi)
    newHiY = real(newFinisYi + 1)

    allocate(newPatch(newStartXi:newFinisXi,newStartYi:newFinisYi))
    allocate(oldXedges(size(   patch, 1)+1), oldYedges(size(   patch, 2)+1))
    allocate(newXedges(size(newPatch, 1)+1), newYedges(size(newPatch, 2)+1))

    indices => indicize(int(0, int16), oldXsize)
    oldXedges = indices / real(oldXsize)
    oldXedges = fracVal(oldLoX, oldHiX, oldXedges)
    deallocate(indices)

    indices => indicize(int(0, int16), oldYsize)
    oldYedges = indices / real(oldYsize)
    oldYedges = fracVal(oldLoY, oldHiY, oldYedges)
    deallocate(indices)

    indices => indicize(newStartXi, newFinisXi + int(1, int16))
    newXedges = real(indices, single)
    deallocate(indices)

    indices => indicize(newStartYi, newFinisYi + int(1, int16))
    newYedges = real(indices, single)
    deallocate(indices)

    call rebinImage(patch, oldXedges, oldYedges&
      , newPatch, newXedges, newYedges, status)
    if (status < 0) then
print *, status
call error('rebinStatus','')
    end if

    call CAL_releaseMemory(patch)
    patch => newPatch

  end subroutine resizePatch

!  !---------------------------------------------------------------------
!  subroutine blah(expMapSetName, expMapImage)
!  !---------------------------------------------------------------------
!  ! Adapted from asmooth-2.5
!
!    image = image * expMapImage
!    where(expMapImage .gt. 0.0) image = image + localOffset
!
!    call eventizeImage_slave(image)
!    call saveOutImage(image, expMapSetName, imageName)
!
!    deallocate(expMapImage)
!
!  end subroutine makeSimulatedImage_withList

  !---------------------------------------------------------------------
  subroutine loadExposureMap(expMapSetName, expMapImage)
  !---------------------------------------------------------------------
  ! Adapted from asmooth-2.5
    
    implicit none

    character(*), intent(in) :: expMapSetName
    real(single), pointer ::&
      expMapImage(:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(expMapSetName, READ)

    call getCalInfo(set, instrumentId, expStartTimeStamp, scAttitude&
    , instrumAttitude) ! also sets cal instrument state.

    inArray = array(set, 0, READ)
!    call getImageWcsInfo(set, inArray, wcs) ! now deprecated.
call getWcsInfo(set, inArray, wcs)
    call loadImage(inArray, expMapImage)

    call release(set)

  end subroutine loadExposureMap

  !---------------------------------------------------------------------
  subroutine eventizeImage(expMapSetName, rawImageName, imageName&
  , multiplier, nonVigBkgRate)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: expMapSetName, rawImageName, imageName
    real(single), intent(in) :: multiplier, nonVigBkgRate
! Try 2e-6 as a starting value for nonVigBkgRate.

    real(single), pointer :: image(:,:), expMapImage(:,:)
    real(single) :: offset
    type(DataSetT) :: set

    !-------------------------------------------------------------------
    ! Start

    call loadFitsArray(rawImageName, image)
    call loadExposureMap(expMapSetName, expMapImage) ! sets cal->instrument
    offset = nonVigBkgRate * maxval(expMapImage) * multiplier
    image = image * expMapImage * multiplier
    where(expMapImage .gt. 0.0) image = image + offset

    call eventizeImage_slave(image)
    call saveOutImage(image, expMapSetName, imageName)

    deallocate(image, expMapImage)

    set = dataSet(imageName, MODIFY)
    call setAttribute(set, 'EXP_MULT', multiplier, ''&
      , 'Factor by which exposure map was multiplied.')
    call setAttribute(set, 'NONVIGBR', nonVigBkgRate, 'counts/sec/pixel'&
      , 'Non-vignetted bkg rate which was added.')
    call release(set)

  end subroutine eventizeImage

  !---------------------------------------------------------------------
  subroutine eventizeImage_slave(image, status)
  !---------------------------------------------------------------------
  ! Adapted from asmooth-2.5 (temp_ssclib)

    implicit none

    real,    intent(out) :: image(:,:)
    integer, intent(out), optional :: status

    integer :: xi, yi, xsize, ysize, localStatus

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    xsize = size(image, 1)
    ysize = size(image, 2)

    do xi = 1, xsize
      do yi = 1, ysize
        image(xi, yi) = real(poissonRandom(image(xi, yi)))
      end do
    end do

    if (present(status)) status = localStatus

  end subroutine eventizeImage_slave

!  !---------------------------------------------------------------------
!  subroutine saveOutImage(outImage, inSetName, outSetName)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    real(single), intent(in) ::&
!      outImage(:,:)
!    character(*), intent(in) ::&
!      inSetName, outSetName
!
!    type(DataSetT) :: inSet, outSet
!    type(ArrayT) :: inArray, outArray
!!    type(BlockT) :: inBlock, outBlock
!!    type(DssT) ::&
!!      my_dss
!    real(single), dimension(:,:), pointer ::&
!      arrayDataReal32
!    integer(int32), dimension(:), pointer ::&
!      dimVector
!
!    !-------------------------------------------------------------------
!    ! Start:
!
!    inSet = dataSet(inSetName, READ)
!    inArray = array(inSet, 0, READ)
!!    inBlock = block(inArray)
!    outSet = dataSet(outSetName, CREATE)
!    allocate(dimVector(0:1))
!    dimVector(0) = size(outImage, 1)
!    dimVector(1) = size(outImage, 2)
!    outArray = addArray(outSet, 'PRIMARY', REAL32, dimVector, position=0)
!!    outBlock = block(outArray)
!    call setAttributes(attributable(outSet), attributable(inSet))
!
!    call copyVitalArrayKwds(inArray, outArray)
!    ! This is done instead of a total copy of all array-header keywords because
!    ! doing the latter stuffs up the DSS.
!
!    arrayDataReal32 => real32Array2Data(outArray)
!!    arrayDataReal32 = 0.0
!    arrayDataReal32 = outImage
!
!!    if (copyDss) then
!!      if (hasAttribute(inBlock, 'DSTYP1')) then
!!        my_dss = dataSubspace(inBlock)
!!        call dssWrite(my_dss, outBlock)
!!      else
!!        call warning('noDSSfound','No data subspace was found in the &
!!        &input image dataset.')
!!      end if
!!    end if
!
!    call release(outSet)
!    call release(inSet)
!
!    deallocate(dimVector)
!
!  end subroutine saveOutImage
!
!  !---------------------------------------------------------------------
!  subroutine copyVitalArrayKwds(inArray, outArray)
!  !---------------------------------------------------------------------
!  ! Adapted from asmooth-2.5
!
!    use fitsutils_mod
!
!    implicit none
!
!    type(ArrayT), intent(in) :: inArray, outArray
!    integer :: status
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    ! The subroutine copyAttribute() is in ssclib/fitsutils_mod.
!
!    call copyAttribute(inArray, 'SLCTEXPR', outArray, status)
!    call copyAttribute(inArray, 'HDUCLASS', outArray, status)
!    call copyAttribute(inArray, 'HDUCLAS1', outArray, status)
!    call copyAttribute(inArray, 'HDUCLAS2', outArray, status)
!    call copyAttribute(inArray, 'HDUVERS1', outArray, status)
!
!    call copyAttribute(inArray, 'CTYPE1', outArray, status)
!    call copyAttribute(inArray, 'CRPIX1', outArray, status)
!    call copyAttribute(inArray, 'CRVAL1', outArray, status)
!    call copyAttribute(inArray, 'CUNIT1', outArray, status)
!    call copyAttribute(inArray, 'CDELT1', outArray, status)
!    call copyAttribute(inArray, 'CTYPE2', outArray, status)
!    call copyAttribute(inArray, 'CRPIX2', outArray, status)
!    call copyAttribute(inArray, 'CRVAL2', outArray, status)
!    call copyAttribute(inArray, 'CUNIT2', outArray, status)
!    call copyAttribute(inArray, 'CDELT2', outArray, status)
!
!    call copyAttribute(inArray, 'EXPOSURE', outArray, status)
!
!  end subroutine copyVitalArrayKwds






  !---------------------------------------------------------------------
  subroutine findNearestSimSource(srcX, srcY, srcL, srcXuncert, srcYuncert&
  , srcLuncert, simSrcX, simSrcY, simSrcL, indexOfNearestSim, devOverStdDev&
  , status)
  !---------------------------------------------------------------------
  ! Given the position (X and Y) and brightness (L) of a detected source,
  ! plus the uncertainties in these values, and given a list of the simulated
  ! sources, determine that simulated source which is 'closest', that is,
  ! which gives the smallest value of the quantity
  !
  !	      (simSrcX - srcX)^2     (simSrcY - srcY)^2     (simSrcL - srcL)^2
  !	r2 = -------------------- + -------------------- + --------------------
  !	         srcXuncert^2           srcYuncert^2           srcLuncert^2
  !
  !.....................................................................

    implicit none

    real(single),   intent(in)            ::&
      srcL, srcLuncert, simSrcL(:), srcXuncert, srcYuncert
    real(double),   intent(in)            ::&
      srcX, srcY, simSrcX(:), simSrcY(:)
    integer(int32), intent(out)           ::&
      indexOfNearestSim
    real(single),   intent(out)           ::&
      devOverStdDev
    integer,        intent(out), optional ::&
      status

    integer(int32) :: numSrcs
    real(double), dimension(size(simSrcX)) :: dX, dY, dL, r2
    integer :: locVec(1)

    !-------------------------------------------------------------------
    ! Start

    status = 0

    indexOfNearestSim = 0 ! nonsensical default
    devOverStdDev = 0.0
    numSrcs = size(simSrcX)
    if (numSrcs .le. 0) then
      if (present(status)) status = -1
  return
    end if
    if (size(simSrcY) .ne. numSrcs .or. size(simSrcL) .ne. numSrcs) then
      if (present(status)) status = -2
  return
    end if
    if (srcXuncert .le. 0.0&
    .or. srcYuncert .le. 0.0&
    .or. srcLuncert .le. 0.0) then
      if (present(status)) status = -3
  return
    end if
    if (numSrcs .eq. 1) then ! trivial
      indexOfNearestSim = 1
      if (present(status)) status = 1
  return
    end if

    dX = simSrcX - srcX
    dY = simSrcY - srcY
    dL = simSrcL - srcL
    r2 = dX * dX / srcXuncert / srcXuncert&
    + dY * dY / srcYuncert / srcYuncert&
    + dL * dL / srcLuncert / srcLuncert

    locVec = minloc(r2)
    indexOfNearestSim = locVec(1)
    devOverStdDev = sqrt(r2(indexOfNearestSim))
!print *, 'simSrcL, srcL, minloc(r2), devOverStdDev: ', simSrcL(indexOfNearestSim), srcL, indexOfNearestSim, devOverStdDev

  end subroutine findNearestSimSource

  !---------------------------------------------------------------------
  subroutine probOfOneOrMoreByChance(srcL, srcXuncert&
  , srcYuncert, srcLuncert, devOverStdDev, simDensity, simFaintestL&
  , probability, status)
  !---------------------------------------------------------------------
  ! Where sources are randomly distributed in 3 dimensions r_ = (X,Y,L)
  ! with a probability density q(r_), the probability P that 1 or more
  ! sources will lie within a given volume V is given by
  !
  !	P = 1 - exp(-N_V)						(1)
  !
  ! where N_V is the number of sources that lies (on average) inside V; or,
  ! in other words,
  !
  !	               V
  !	           (  /           )
  !	P = 1 - exp( -| dr_ q[r_] ).
  !	           (  /           )
  !
  ! In the case of sources distributed uniformly through space, their
  ! distribution in the abstract space in which X and Y coordinates give
  ! the position on a projection plane and the third coordinate L gives
  ! the measured brightness is given, as is well known, by the probability
  ! density
  !
  !	                   3
  !	q(X,Y,L) = q(L) = --- p L_min^3/2 L^(-5/2),
  !	                   2
  !
  ! where p is the total number of sources per unit area of the projection
  ! plane. Hence the number N_V in equation (1) is given by
  !
  !	                   L_V_max
  !	       3              /
  !	N_V = --- p L_min^3/2 | dL A_V(L) L^(-5/2),
  !	       2              /
  !	                   L_V_min
  !
  ! where L_V_min and L_V_max are respectively the smallest and largest
  ! values of L enclosed within V, and A_V(L) is the area of the cross-
  ! section through V normal to the L axis at the specified value of L.
  !
  ! In the present subroutine, V is assumed to be an ellipsoid centred on
  ! srcX, srcY and srcL, with axes equal to devOverStdDev times
  ! respectively srcXuncert, srcYuncert and srcLuncert, and truncated at
  ! L = L_min. The resulting equation is as is given in the code below.
  !.....................................................................

    implicit none

    real(single), intent(in)            ::&
      srcL, srcLuncert, devOverStdDev, simDensity, simFaintestL,&
      srcXuncert, srcYuncert
    real(single), intent(out)           ::&
      probability
    integer,      intent(out), optional :: status

    real(single) ::&
      deltaX, deltaY, deltaL, L_V_max, L_V_min, deltaLsquared, numV
    real(double) ::&
      intgrlTerm1, intgrlTerm2, intgrlTerm3

!integer :: numPts, i
!real(single) :: dL, summ, frac, l, x, y, dVolume, density

    !-------------------------------------------------------------------
    ! Start

    status = 0

!print *, 'srcL: ',srcL
!print *, 'srcLuncert: ',srcLuncert
!print *, 'devOverStdDev: ',devOverStdDev
!print *, 'simDensity: ',simDensity
!print *, 'simFaintestL: ',simFaintestL
!print *, 'srcXuncert: ',srcXuncert
!print *, 'srcYuncert: ',srcYuncert

    probability = 0.0

    if (srcXuncert .le. 0.0&
    .or. srcYuncert .le. 0.0&
    .or. srcLuncert .le. 0.0&
    .or. devOverStdDev .le. 0.0) then
      if (present(status)) status = -1
  return
    end if

    deltaX = srcXuncert * devOverStdDev
    deltaY = srcYuncert * devOverStdDev
    deltaL = srcLuncert * devOverStdDev
    L_V_max = srcL + deltaL

    if (L_V_max .le. simFaintestL) then
      if (present(status)) status = -2
  return
    end if

    L_V_min = max(srcL - deltaL, simFaintestL)
    deltaLsquared = deltaL * deltaL

!print *, 'deltaX: ',deltaX
!print *, 'deltaY: ',deltaY
!print *, 'deltaL: ',deltaL
!print *, 'L_V_max: ',L_V_max
!print *, 'L_V_min: ',L_V_min
!print *, 'deltaLsquared: ',deltaLsquared

    intgrlTerm1 = -(2.0 / 3.0) * (1.0 - srcL * srcL / deltaLsquared)&
    * ((L_V_max)**(-1.5) - (L_V_min)**(-1.5))

    intgrlTerm2 = -4.0 * srcL / deltaLsquared&
    * ((L_V_max)**(-0.5) - (L_V_min)**(-0.5))

    intgrlTerm3 = -2.0 / deltaLsquared&
    * ((L_V_max)**(0.5)  - (L_V_min)**(0.5))

    numV = 1.5 * simDensity * simFaintestL**1.5&
    * Pi * deltaX * deltaY * (intgrlTerm1 + intgrlTerm2 + intgrlTerm3)

!print *, 'intgrlTerm1: ',intgrlTerm1
!print *, 'intgrlTerm2: ',intgrlTerm2
!print *, 'intgrlTerm3: ',intgrlTerm3
!print *, 'Pi: ',Pi
!print *, 'numV: ',numV

    if (numV <= 0.0) then
      probability = 0.0
      call warning('negativeProbability', 'This ought not to happen.')
    else if (numV .ge. -log(tiny(numV))) then
      probability = 1.0 ! saves the possibility of an underflow on the exp().
    else if (log(numV * numV / 2.0) .lt. (1 - precision(numV))) then
      probability = numV
!print *, 'log(numV * numV / 2.0): ', log(numV * numV / 2.0)
!print *, '   1 - precision(numV): ', 1 - precision(numV)
      call warning('probabilityUnderflow', 'No cause for concern. This &
      &is just to inform you that the probability has been calculated &
      &by use of the approximation prob = N rather than the exact &
      &prob = 1 - exp(-N).')
    else
      probability = 1.0 - exp(-numV)
    end if

!print *, 'probability: ', probability

!numPts = 100
!dL = (L_V_max - L_V_min) / real(numPts)
!summ = 0.0
!do i = 1, numPts
!  frac = (i - 0.5) / real(numPts)
!  l = fracVal(L_V_min, L_V_max, frac)
!  x = deltaX * sqrt(1 - (l - srcL) * (l - srcL) / deltaLsquared)
!  y = deltaY * sqrt(1 - (l - srcL) * (l - srcL) / deltaLsquared)
!  dVolume = Pi * x * y * dL
!  density = 1.5 * simDensity * simFaintestL**1.5 * l**(-2.5)
!
!  summ = summ + dVolume * density
!end do
!
!print *, 'approx numV: ',summ
!
!call error('','')

  end subroutine probOfOneOrMoreByChance

  !---------------------------------------------------------------------
  subroutine flagAlsoRans(indexOfNearestSim, devOverStdDev, alsoRan, status)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in)            :: indexOfNearestSim(:)
    real(single),   intent(in)            :: devOverStdDev(:)
    logical,        intent(out)           :: alsoRan(size(indexOfNearestSim))
    integer,        intent(out), optional :: status

    integer(int32) :: numSrcs, minIndex, maxIndex, index
    real(single)   :: minR
    logical        :: mask(size(indexOfNearestSim))
    integer        :: locVec(1)

    !-------------------------------------------------------------------
    ! Start

    status = 0

    alsoRan = .true. ! nonsensical default
    numSrcs = size(indexOfNearestSim)
    if (numSrcs .le. 0) then
      if (present(status)) status = -1
  return
    end if
    if (size(devOverStdDev) .ne. numSrcs) then
      if (present(status)) status = -2
  return
    end if

    minIndex = minval(indexOfNearestSim)
    maxIndex = maxval(indexOfNearestSim)

    if (minIndex .le. 0) then
      if (present(status)) status = -3
  return
    end if

    alsoRan = .false. ! sensible default
    do index = minIndex, maxIndex
      mask = (indexOfNearestSim .eq. index)
    if (count(mask) .le. 1) cycle
      minR = minval(devOverStdDev, mask)
      if (count(mask .and. devOverStdDev .eq. minR) .gt. 1) then ! there is
      ! more than 1 det source at the minimum 'distance' from the sim source
      ! of this index. Choose one of them to be the offical det source,
      ! setting alsoRan = .true. for all the others:
        where(mask) alsoRan = .true. ! covers everything.
        locVec = minloc(devOverStdDev, mask) ! just picks up 1 best location.
        alsoRan(locVec(1)) = .false.

      else ! there is only 1 det source at the minimum 'distance' from the
      ! sim source of this index. Leave alsoRan at .false. for this source,
      ! setting alsoRan = .true. for all the others at greater r:
        where(mask .and. devOverStdDev .gt. minR) alsoRan = .true.
      end if
    end do

  end subroutine flagAlsoRans

  !---------------------------------------------------------------------
  subroutine test_dumpColumn(vector)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: vector(:)

    type(DataSetT) :: set
    type(TableT) :: tab
    type(ColumnT) :: col
    real(single), pointer :: myData(:)
    integer(int32) :: numRows

    !-------------------------------------------------------------------
    ! Start

    numRows = size(vector)
    set = dataSet('junk.ds', CREATE)
    tab = addTable(set, 'SRCLIST', numRows)
    col = addColumn(tab, 'RATE', REAL32)
    myData => real32Data(col)
    myData = vector
    call release(set)

  end subroutine test_dumpColumn

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module temp_simulations
