!+
! Module: temp_ssclib
!
! Authors:  Ian Stewart - SSC/LUX
!
! Comments: TODO:
! 
! $Id: temp_ssclib.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module temp_ssclib

  use dal
  use errorHandling
  use fitsutils_mod ! copyAttribute()

  implicit none 

  private 

  public addTables
  public copyColumnBetweenTables
  public copyRowsBetweenTables
  public smallestIndex
  public largestIndex
  public setValuesToNull
  public temp_robustMean
  public checkAndCopyKeywords
  public attributeValuesAreEqual

  interface setValuesToNull
    module procedure setColValuesToNull
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine temp_robustMean(y, yUncert, largestDevianceRatio&
  , meanY, mask, meanYuncert, reducedChiSquared, status)
  !---------------------------------------------------------------------
  ! Usually when one has a sequence of values which is known to include some
  ! 'outliers', the quickest way to find a sort of average which doesn't
  ! include the outliers is to take the median of the sequence. However if
  ! most of the sequence values are either 0 or 1, there is not much use in
  ! taking a median since it too will be either 0 or 1. It is the purpose
  ! of this routine is to return a mean value of a masked subset of such a
  ! sequence, the mask being chosen such that no masked value departs from
  ! the mean of the masked values by more than largestDevianceRatio times the
  ! standard deviation of the masked values. In other words, the explicit
  ! variance of any data point i, calculated as (y_i - mean_y)^2, should not
  ! exceed the implicit variance, which is the square of the uncertainty in
  ! y_i, by more than largestVarianceRatio = largestDevianceRatio^2.
  !
  ! Note that the y values in the calculation of the mean are weighted
  ! by 1 / yUncert^2, which weighting is known to give the lowest uncertainty
  ! in the result.

    implicit none

    real,    intent(in)            :: y(:), yUncert(size(y)),&
                                      largestDevianceRatio
    real,    intent(out)           :: meanY
    real,    intent(out), optional :: meanYuncert, reducedChiSquared
    logical, intent(out), optional :: mask(size(y))
    integer, intent(out), optional :: status

    integer :: vsize, localStatus, degFree
    logical :: localMask(size(y))
    real :: largestVarianceRatio, deviance(size(y)),&
      variance(size(y)), scatterVariance(size(y)), meanYvariance

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    meanY = 0.0
    if (present(meanYuncert)) meanYuncert = 0.0
    if (present(reducedChiSquared)) reducedChiSquared = 0.0

    vsize = size(y)
    if (vsize <= 0) then
      if (present(status)) status = -1
  return
    end if
    if (largestDevianceRatio <= 1.0) then
      if (present(status)) status = -2
  return
    end if
    if (any(yUncert <= 0)) then
      if (present(status)) status = -3
  return
    end if

    localMask = .true.
    variance = yUncert * yUncert

    largestVarianceRatio = largestDevianceRatio * largestDevianceRatio
    do
      if (count(localMask) <= 0) then
        if (present(status)) status = -4
  return
      end if

      meanY = sum(y / variance, localMask) / sum(1.0 / variance, localMask)
      deviance = y - meanY
      scatterVariance = deviance * deviance

      if (.not. any(scatterVariance / variance  > largestVarianceRatio&
      .and. localMask))&
    exit ! because no new pixels have been masked in this iteration:
         ! hence the algorithm has converged.

      where(scatterVariance / variance > largestVarianceRatio)&
        localMask = .false.
    end do

    meanYvariance = 1.0 / sum(1.0 / variance, localMask)

    if (present(mask)) mask = localMask
    if (present(meanYuncert)) meanYuncert = sqrt(meanYvariance)

    if (present(reducedChiSquared)) then
      degFree = count(localMask) - 1
      if (degFree <= 0) then
        if (present(status)) status = -5
  return
      end if

      reducedChiSquared = sum(scatterVariance / variance, localMask)&
        / real(degFree)
    end if

    if(present(status)) status = localStatus

  end subroutine temp_robustMean

  !---------------------------------------------------------------------
  subroutine addTables(masterTab, slaveTab, indexColName, indexValue)
  !---------------------------------------------------------------------
!*** put into fits_utils

    implicit none

    type(TableT), intent(in) :: masterTab, slaveTab
    character(*), intent(in) :: indexColName
    integer(int16), intent(in) :: indexValue

    integer(int32) :: numMasterRows, numSlaveRows, colNum
    type(ColumnT) :: slaveCol, masterCol, indexCol
    integer(int16), pointer :: index(:)
logical :: testflag = .true.

    !-------------------------------------------------------------------
    ! Start:

    numMasterRows = numberOfRows(masterTab)
    numSlaveRows  = numberOfRows(slaveTab)
    call insertRows(masterTab, numMasterRows, numSlaveRows)
if (testflag) print *, 'inserted', numSlaveRows, 'rows at starting row', numMasterRows, 'in master table.'

!**** now call copyRowsBetweenTables instead of the below???

    do colNum = 0, numberOfColumns(slaveTab) - 1
      slaveCol = column(slaveTab, colNum)
    if (name(slaveCol) == indexColName) cycle

      if (hasColumn(masterTab, name(slaveCol))) then
        masterCol = column(masterTab, name(slaveCol))

        if (dataType(slaveCol) .ne. dataType(masterCol))&
call error('mixedDataTypes','')

      else ! master table doesn't have such a column.
        masterCol = addColumn(masterTab, name(slaveCol), dataType(slaveCol))
        call setValuesToNull(masterCol) ! sets everything to null.
      end if

!if (dataType(slaveCol) .ne. dataType(masterCol)) then
!  print *, 'name(slaveCol), dataType(slaveCol), dataType(masterCol): '&
!  , trim(name(slaveCol)), dataType(slaveCol), dataType(masterCol)
!call error('mixedDataTypes','')
!end if
      call copyColumnBetweenTables(slaveCol, 0, numSlaveRows&
        , masterCol, numMasterRows)

      call release(slaveCol)
      call release(masterCol)
    end do

    ! If master has a col that slave does not, add nulls to relevant
    ! rows in master:
    do colNum = 0, numberOfColumns(masterTab) - 1
      masterCol = column(masterTab, colNum)
    if (name(masterCol) == indexColName) cycle

      if (.not. hasColumn(slaveTab, name(masterCol))) then
        call setValuesToNull(masterCol, numMasterRows, numSlaveRows)
        ! sets everything to null.
      end if
    end do

    ! If index col doesn't exist in master, create it
    if (.not. hasColumn(masterTab, indexColName)) then
      indexCol = addColumn(masterTab, indexColName, INTEGER16)
!*** set to null or blank ??? Prob not nec here.
    else if (dataType(column(masterTab, indexColName)) .ne. INTEGER16) then
      call deleteColumn(masterTab, indexColName)
      indexCol = addColumn(masterTab, indexColName, INTEGER16)
!*** set to null or blank ??? Prob not nec here.
    else
      indexCol = column(masterTab, indexColName)
    end if

    ! Load index col with indexValue
    index => int16Data(indexCol)
    index(numMasterRows:numMasterRows + numSlaveRows - 1) = indexValue
    call release(indexCol)

  end subroutine addTables

  !---------------------------------------------------------------------
  subroutine copyRowsBetweenTables(fromTab, fromRowStart, numFromRowsToCopy&
  , toTab, toRowStart)
  !---------------------------------------------------------------------
!*** put into fits_utils

    implicit none

    type(TableT),   intent(in) ::&
      fromTab, toTab
    integer(int32), intent(in) ::&
      fromRowStart, numFromRowsToCopy, toRowStart

    integer(int32) :: numFromRows, numToRows, colNum
    type(ColumnT) :: fromCol, toCol
logical :: testflag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testflag) print *, 'copy rows:'
if (testflag) print *, '  fromRowStart, numFromRowsToCopy, toRowStart: '&
, fromRowStart, numFromRowsToCopy, toRowStart

    numFromRows = numberOfRows(fromTab)
    numToRows   = numberOfRows(toTab)
if (fromRowStart < 0) call error('fromStartRowTooLow','')
if (toRowStart   < 0) call error('toStartRowTooLow','')
    if (fromRowStart + numFromRowsToCopy > numFromRows)&
call error('tooFewFromRows','')
if (toRowStart + numFromRowsToCopy > numToRows) call error('tooFewToRows','')

    ! If fromTab has columns that toTab does not, add these columns to toTab:

    do colNum = 0, numberOfColumns(fromTab) - 1
      fromCol = column(fromTab, colNum)

      if (hasColumn(toTab, name(fromCol))) then
        toCol = column(toTab, name(fromCol))

if (dataType(fromCol) .ne. dataType(toCol)) call error('mixedDataTypes','')

      else ! toTab doesn't have such a column.
        toCol = addColumn(toTab, name(fromCol), dataType(fromCol))
        call setValuesToNull(toCol) ! sets everything to null.
      end if

      call copyColumnBetweenTables(fromCol, fromRowStart, numFromRowsToCopy&
        , toCol, toRowStart)

      call release(fromCol)
      call release(toCol)
    end do

  end subroutine copyRowsBetweenTables

  !---------------------------------------------------------------------
  subroutine copyColumnBetweenTables(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
!*** put into fits_utils

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow
logical :: testflag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testflag) print *, '  copy row', fromColStartRow, 'to row'&
, fromColStartRow+numFromRows-1, ' of column name '//trim(name(fromCol))&
//' of table A to rows', toColStartRow, '-', toColStartRow+numFromRows-1&
, 'of table B.'

    if (fromColStartRow < 0)&
call error('fromStartRowTooLow','')
    if (toColStartRow < 0)&
call error('toStartRowTooLow','')
    if (numberOfRows(fromCol) < fromColStartRow + numFromRows)&
call error('tooFewFromRows','')
    if (numberOfRows(toCol) < toColStartRow + numFromRows)&
call error('tooFewToRows','')

    select case(dataType(fromCol))
    case(INTEGER8)
      call copyColsTweenInt8(fromCol, fromColStartRow, numFromRows&
        , toCol, toColStartRow)
    case(INTEGER16)
      call copyColsTweenInt16(fromCol, fromColStartRow, numFromRows&
        , toCol, toColStartRow)
    case(INTEGER32)
      call copyColsTweenInt32(fromCol, fromColStartRow, numFromRows&
        , toCol, toColStartRow)
    case(REAL32)
      call copyColsTweenSingle(fromCol, fromColStartRow, numFromRows&
        , toCol, toColStartRow)
    case(REAL64)
      call copyColsTweenDouble(fromCol, fromColStartRow, numFromRows&
        , toCol, toColStartRow)
    case(STRING)
      call copyColsTweenString(fromCol, fromColStartRow, numFromRows&
        , toCol, toColStartRow)
    case(BOOLEAN)
      call copyColsTweenBool(fromCol, fromColStartRow, numFromRows&
        , toCol, toColStartRow)
    case default
call error('badColDataType','')
    end select

  end subroutine copyColumnBetweenTables

  !---------------------------------------------------------------------
  subroutine copyColsTweenInt8(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
  ! This should NOT be made a public subroutine.
  !
  ! The checks on size of the columns have already been done in
  ! copyColumnBetweenTables().

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow

    integer(int8), pointer :: toVector(:), fromVector(:)

    !-------------------------------------------------------------------
    ! Start

    fromVector => int8Data(fromCol)
    toVector   => int8Data(toCol)
    toVector(toColStartRow:toColStartRow + numFromRows - 1)&
      = fromVector(fromColStartRow:fromColStartRow + numFromRows - 1)
    call release(fromCol)
    call release(toCol)

  end subroutine copyColsTweenInt8

  !---------------------------------------------------------------------
  subroutine copyColsTweenInt16(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
  ! This should NOT be made a public subroutine.
  !
  ! The checks on size of the columns have already been done in
  ! copyColumnBetweenTables().

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow

    integer(int16), pointer :: toVector(:), fromVector(:)

    !-------------------------------------------------------------------
    ! Start

    fromVector => int16Data(fromCol)
    toVector   => int16Data(toCol)
    toVector(toColStartRow:toColStartRow + numFromRows - 1)&
      = fromVector(fromColStartRow:fromColStartRow + numFromRows - 1)
    call release(fromCol)
    call release(toCol)

  end subroutine copyColsTweenInt16

  !---------------------------------------------------------------------
  subroutine copyColsTweenInt32(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
  ! This should NOT be made a public subroutine.
  !
  ! The checks on size of the columns have already been done in
  ! copyColumnBetweenTables().

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow

    integer(int32), pointer :: toVector(:), fromVector(:)

    !-------------------------------------------------------------------
    ! Start

    fromVector => int32Data(fromCol)
    toVector   => int32Data(toCol)
    toVector(toColStartRow:toColStartRow + numFromRows - 1)&
      = fromVector(fromColStartRow:fromColStartRow + numFromRows - 1)
    call release(fromCol)
    call release(toCol)

  end subroutine copyColsTweenInt32

  !---------------------------------------------------------------------
  subroutine copyColsTweenSingle(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
  ! This should NOT be made a public subroutine.
  !
  ! The checks on size of the columns have already been done in
  ! copyColumnBetweenTables().

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow

    real(single), pointer :: toVector(:), fromVector(:)

    !-------------------------------------------------------------------
    ! Start

    fromVector => real32Data(fromCol)
    toVector   => real32Data(toCol)
    toVector(toColStartRow:toColStartRow + numFromRows - 1)&
      = fromVector(fromColStartRow:fromColStartRow + numFromRows - 1)
    call release(fromCol)
    call release(toCol)

  end subroutine copyColsTweenSingle

  !---------------------------------------------------------------------
  subroutine copyColsTweenDouble(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
  ! This should NOT be made a public subroutine.
  !
  ! The checks on size of the columns have already been done in
  ! copyColumnBetweenTables().

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow

    real(double), pointer :: toVector(:), fromVector(:)

    !-------------------------------------------------------------------
    ! Start

    fromVector => real64Data(fromCol)
    toVector   => real64Data(toCol)
    toVector(toColStartRow:toColStartRow + numFromRows - 1)&
      = fromVector(fromColStartRow:fromColStartRow + numFromRows - 1)
    call release(fromCol)
    call release(toCol)

  end subroutine copyColsTweenDouble

  !---------------------------------------------------------------------
  subroutine copyColsTweenString(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
  ! This should NOT be made a public subroutine.
  !
  ! The checks on size of the columns have already been done in
  ! copyColumnBetweenTables().

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow

    integer(int32) :: row

    !-------------------------------------------------------------------
    ! Start

    do row = 0, numFromRows - 1
      call setStringCell(toCol, row + toColStartRow&
        , stringCell(fromCol, row + fromColStartRow))
    end do

  end subroutine copyColsTweenString

  !---------------------------------------------------------------------
  subroutine copyColsTweenBool(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
  ! This should NOT be made a public subroutine.
  !
  ! The checks on size of the columns have already been done in
  ! copyColumnBetweenTables().

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow

    logical(bool), pointer :: toVector(:), fromVector(:)

    !-------------------------------------------------------------------
    ! Start

    fromVector => boolData(fromCol)
    toVector   => boolData(toCol)
    toVector(toColStartRow:toColStartRow + numFromRows - 1)&
      = fromVector(fromColStartRow:fromColStartRow + numFromRows - 1)
    call release(fromCol)
    call release(toCol)

  end subroutine copyColsTweenBool

  !---------------------------------------------------------------------
  integer function smallestIndex(maskVector)
  !---------------------------------------------------------------------
!*** put into sscmisc

    implicit none

    logical, intent(in) :: maskVector(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    smallestIndex = 0
    do i = 1, size(maskVector)
    if (maskVector(i)) then
        smallestIndex = i
    exit
      end if
    end do

  end function smallestIndex

  !---------------------------------------------------------------------
  integer function largestIndex(maskVector)
  !---------------------------------------------------------------------
!*** put into sscmisc

    implicit none

    logical, intent(in) :: maskVector(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    largestIndex = size(maskVector) + 1
    do i = size(maskVector), 1, -1
      if (maskVector(i)) then
        largestIndex = i
    exit
      end if
    end do

  end function largestIndex

  !---------------------------------------------------------------------
  subroutine setColValuesToNull(col, startRow, numRows)
  !---------------------------------------------------------------------
  ! Sets the column entries to nulls where possible, blanks or zeroes otherwise.

!*** put into fits_utils

    implicit none

    type(ColumnT), intent(in) :: col
    integer(int32), intent(in), optional :: startRow, numRows

    integer(int32) :: finisRow

    !-------------------------------------------------------------------
    ! Start
!****

  end subroutine setColValuesToNull

  !---------------------------------------------------------------------
  subroutine checkAndCopyKeywords(fromTabs, toTab, keywordNames&
  , keywordIsUniform)
  !---------------------------------------------------------------------
  ! This sub checks each keyword in the list to see whether it is present and of the same value in all the input tables. If so, it copies it into the output table.

!*** put into fits_utils

    implicit none

    type(TableT), intent(in) :: fromTabs(:), toTab
    character(*), intent(in) :: keywordNames(:)
    logical, intent(out), optional ::&
      keywordIsUniform(size(keywordNames))

    integer(int16) :: numTabs, numKeywords, keywordNum, tabNum
    logical :: localKwdIsUniform(size(keywordNames))!!! = .false.
    type(AttributeT) :: attA, attB

    !-------------------------------------------------------------------
    ! Start

localKwdIsUniform = .false.
    numTabs = size(fromTabs)
    numKeywords = size(keywordNames)

    KEYWORD: do keywordNum = 1, numKeywords
      localKwdIsUniform(keywordNum) = .true. ! temporary default.

      ! Note that there are two ways in which a keyword may be uniform:
      ! either it is present in all tables, and of the same value; or it
      ! is not present in all tables. These two cases are tested separately
      ! in the two branches of the following 'if'.

      if (hasAttribute(fromTabs(1), keywordNames(keywordNum))) then
        do tabNum = 2, numTabs
          if (.not. hasAttribute(fromTabs(tabNum), keywordNames(keywordNum)))&
          then
            localKwdIsUniform(keywordNum) = .false.
    cycle KEYWORD
          end if

          attA = attribute(fromTabs(1),      keywordNames(keywordNum))
          attB = attribute(fromTabs(tabNum), keywordNames(keywordNum))
          if (.not. attributeValuesAreEqual(attA, attB)) then
            localKwdIsUniform(keywordNum) = .false.
    cycle KEYWORD
          end if
        end do

        if (localKwdIsUniform(keywordNum)) then ! copy it:
          call copyAttribute(fromTabs(1), keywordNames(keywordNum), toTab)
        end if

      else ! table 1 doesn't have the attribute.
        do tabNum = 2, numTabs
          if (hasAttribute(fromTabs(tabNum), keywordNames(keywordNum))) then
            localKwdIsUniform(keywordNum) = .false.
    cycle KEYWORD
          end if
        end do
      end if
    end do KEYWORD

    if (present(keywordIsUniform)) keywordIsUniform = localKwdIsUniform

  end subroutine checkAndCopyKeywords

  !---------------------------------------------------------------------
  logical function attributeValuesAreEqual(attA, attB)
  !---------------------------------------------------------------------

!*** put into fits_utils

    implicit none

    type(AttributeT), intent(in) :: attA, attB

    !-------------------------------------------------------------------
    ! Start

    attributeValuesAreEqual = .false. ! default

  if (dataType(attA) .ne. dataType(attB)) return

    select case(dataType(attA))
    case(INTEGER8)
      if (int8Attribute(attA) == int8Attribute(attB))&
        attributeValuesAreEqual = .true.
    case(INTEGER16)
      if (int16Attribute(attA) == int16Attribute(attB))&
        attributeValuesAreEqual = .true.
    case(INTEGER32)
      if (int32Attribute(attA) == int32Attribute(attB))&
        attributeValuesAreEqual = .true.
    case(REAL32)
      if (real32Attribute(attA) == real32Attribute(attB))&
        attributeValuesAreEqual = .true.
    case(REAL64)
      if (real64Attribute(attA) == real64Attribute(attB))&
        attributeValuesAreEqual = .true.
    case(STRING)
      if (stringAttribute(attA) == stringAttribute(attB))&
        attributeValuesAreEqual = .true.
    case(BOOLEAN)
      if (booleanAttribute(attA) .eqv. booleanAttribute(attB))&
        attributeValuesAreEqual = .true.
    end select

  end function attributeValuesAreEqual

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module temp_ssclib
