!
! Module: temp_ssclib
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some routines for dealing with keywords and columns of FITS files.
! 
! Comments: TODO:
!*********** Look through ALL software for unnecessary pointer allocations!
!*********** Look through ALL software and declare all ptrs => null()!
! 
! $Id: temp_ssclib.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- fred
module temp_ssclib

  use dal
  use errorHandling

  use constants	! PiOn2
  use array_utils
  use sort_mod
!use coordinate
use dump

  implicit none 

  private 

  public calcSmallestConePair
  public calcSmallestConeTriplet
  public calcSmallestCone
  public myAddColumn
  public copyVitalArrayKwds

  interface myAddColumn
    module procedure myAddColumnInt32
    module procedure myAddColumnSingle
    module procedure myAddColumnDouble
  end interface

  contains

  !---------------------------------------------------------------------
  subroutine calcSmallestConePair(ray1, ray2, centralRay, coneAngle)
  !---------------------------------------------------------------------
  ! This subroutine takes two directions in 3-space, defined by their
  ! vectors of direction cosines, and finds that conical locus which contains
  ! both vectors and has the smallest opening angle.
!*** put into array_utils?

    implicit none

    real(double), intent(in) :: ray1(3), ray2(3)
    real(double), intent(out) :: centralRay(3), coneAngle

    real(double) ::&
      diffVector(3), axisRay(3),&
      tempRay(3), centralRayNorm, sinConeAngle, cosConeAngle
    integer(int8) :: axis

    !-------------------------------------------------------------------
    ! Start

    centralRay = (ray1 + ray2) / 2.0
!print *, 'centralRay(X), centralRay(Y), centralRay(Z):'&
!, centralRay(1), centralRay(2), centralRay(3)
    centralRayNorm = sqrt(sum(centralRay * centralRay))
!print *, 'centralRayNorm:'&
!, centralRayNorm

    if (centralRayNorm <= 0.0) then ! trouble, the rays are
    ! diametrically opposed and there is therefore an infinity of
    ! rays which are the same distance from both. To best solve this,
    ! first find the axis which is furthest from ray1 (this will be
    ! the axis for which ray1 has the smallest magnitude component):

      axis = minLoc1d(abs(ray1)) ! in ssclib/array_utils

      ! Then find the ray which is (i) orthogonal to ray1
      ! and (ii) close as possible to this axis. This is done by
      ! taking ray1 cross axis to get an intermediate vector
      ! (guaranteed to be of decent length because we have chosen that
      ! axis _farthest_ from ray1), then crossing the
      ! intermediate vector with ray1:

      axisRay = 0.0d0
      axisRay(axis) = 1.0
      tempRay = vectorCross(ray1, axisRay)

      centralRay = vectorCross(tempRay, ray1)
      centralRayNorm = sqrt(sum(centralRay * centralRay))
if (centralRayNorm <= 0.0) call error('bug',"")
      centralRay = centralRay / centralRayNorm
      coneAngle = PiOn2

    else
      centralRay = centralRay / centralRayNorm
!print *, 'centralRay(X), centralRay(Y), centralRay(Z):'&
!, centralRay(1), centralRay(2), centralRay(3)

      ! We could find the angle between centralRay and the
      ! rays it bisects simply enough
      ! by taking the dot product of either ray and the central one
      ! then taking arccos, but if the angle is smallish that becomes 
      ! inaccurate. Similar considerations apply with largish angles
      ! and cross products. Better to have belt and braces and use
      ! that nice function atan2:

      diffVector = ray1 - ray2
      sinConeAngle = sqrt(sum(diffVector * diffVector)) / 2.0
      cosConeAngle = sum(ray1 * centralRay)
      coneAngle = atan2(sinConeAngle, cosConeAngle)
    end if

  end subroutine calcSmallestConePair

  !---------------------------------------------------------------------
  subroutine calcSmallestConeTriplet(rays, centralRay, coneAngle&
  , numStretchPoints)
  !---------------------------------------------------------------------
  ! This subroutine takes three directions in 3-space, defined by their
  ! vectors of direction cosines, and finds that conical locus which contains
  ! all three vectors and has the smallest opening angle. This central
  ! vector either bisects the angle between two out of the three rays or
  ! it is equidistant from all three.

!*** put into array_utils?

    implicit none

    real(double), intent(in) :: rays(3,3)
    real(double), intent(out) :: centralRay(3), coneAngle
    integer(int16), intent(out), optional :: numStretchPoints

    type :: pairType
      integer(int8) :: first, second
    end type pairType

    integer(int8), parameter :: X=1, Y=2, Z=3
    real(double) ::&
      diffVector(3,X:Z), diffNormSqrd(3)
    integer(int8) :: widestPairIndex, oddManOut
    type(pairType) :: pairs(3)
!integer :: i
!real(double) :: angle
    !-------------------------------------------------------------------
    ! Start

    if (present(numStretchPoints)) numStretchPoints = 2 ! default

    ! Check first to see if a cone [with centralRay which bisects the
    ! angle between the two furthest-apart rays and a coneAngle which is
    ! half the angle between the two furthest-apart rays] also includes
    ! the third ray. If not, have to find a centralRay equidistant from
    ! all 3 rays.

    pairs(1)%first = 1
    pairs(1)%second = 2
    pairs(2)%first = 2
    pairs(2)%second = 3
    pairs(3)%first = 3
    pairs(3)%second = 1

    diffVector(1:3,X:Z) = rays(pairs(1:3)%first,X:Z)&
      - rays(pairs(1:3)%second,X:Z)

    diffNormSqrd(1) = sum(diffVector(1,X:Z) * diffVector(1,X:Z))
    diffNormSqrd(2) = sum(diffVector(2,X:Z) * diffVector(2,X:Z))
    diffNormSqrd(3) = sum(diffVector(3,X:Z) * diffVector(3,X:Z))

    widestPairIndex = maxLoc1d(diffNormSqrd)
    oddManOut = 1 + modulo(pairs(widestPairIndex)%second, 3_int8)

    call calcSmallestConePair(rays(pairs(widestPairIndex)%first,X:Z)&
      , rays(pairs(widestPairIndex)%second,X:Z), centralRay, coneAngle)

  if (angleBetween(centralRay, rays(oddManOut,X:Z)) <= coneAngle) return
  ! angleBetween is in ssclib/array_utils.

    if (present(numStretchPoints)) numStretchPoints = 3

    call calcEquidistantRay(rays, centralRay, coneAngle)

  end subroutine calcSmallestConeTriplet

  !---------------------------------------------------------------------
  subroutine calcEquidistantRay(rays, centralRay, coneAngle)
  !---------------------------------------------------------------------
  ! This subroutine takes three directions in 3-space, defined by their
  ! vectors of direction cosines, and finds that vector which is equidistant
  ! from all three. This (angular) distance is also returned.
  !
  ! The algorithm is as follows. First, for each pair of rays, imagine a
  ! plane that bisects them; the desired vector must lie at the
  ! intersection of the three planes. Thus the first step in the
  ! algorithm is to find a secondary vector which is in the same plane
  ! as both and orthogonal to their average. The sign doesn't matter at
  ! this stage. The plane mentioned must be orthogonal to this secondary
  ! vector.
  !
  ! The second step is, for at least 1 pair of planes, to find the ray at
  ! which they intersect. This can be done just by taking the cross product
  ! of the secondary vectors orthogonal to each plane. The result must be
  ! orthogonal to both secondary vectors, thus within both planes, thus
  ! at the intersection between the planes.
  !
  ! There is some reduncancy here because we only need 2 not the full 3
  ! of these planes to define their intersection: so we take the two
  ! secondary vectors which are most nearly orthogonal - this can be done
  ! by taking cross products for all pairs of secondary vectors and
  ! discarding all but the largest. There is also some ambiguity because
  ! in fact there are (in general) 2 rays which are equidistant from the
  ! 3 input rays. We want the one which minimizes the distance rather
  ! than maximizes it however, and it is easy to distinguish between
  ! these alternatives.
!*** put into array_utils?

    implicit none

    real(double), intent(in) :: rays(3,3)
    real(double), intent(out) :: centralRay(3), coneAngle

    type :: pairType
      integer(int8) :: first, second
    end type pairType

    integer(int8), parameter :: X=1, Y=2, Z=3
    real(double) ::&
      diffVector(3,X:Z), diffNorm(3),&
      sinConeAngle, cosConeAngle,&
      secondaryRays(3,X:Z), crossRays(3,X:Z), crossNormSquared(3),&
      lastCross(X:Z)
    integer(int8) :: i
    type(pairType) :: pairs(3)

    !-------------------------------------------------------------------
    ! Start

    ! First, for each pair of rays, find a vector which is in the same
    ! plane as both and orthogonal to their average. The sign doesn't
    ! matter at this stage.

    pairs(1)%first = 1
    pairs(1)%second = 2
    pairs(2)%first = 1
    pairs(2)%second = 3
    pairs(3)%first = 2
    pairs(3)%second = 3

    diffVector(1:3,X:Z) = rays(pairs(1:3)%first,X:Z)&
      - rays(pairs(1:3)%second,X:Z)

    diffNorm(1) = sqrt(sum(diffVector(1,X:Z) * diffVector(1,X:Z)))
    diffNorm(2) = sqrt(sum(diffVector(2,X:Z) * diffVector(2,X:Z)))
    diffNorm(3) = sqrt(sum(diffVector(3,X:Z) * diffVector(3,X:Z)))

    ! Deal with special cases:

    if (any(diffNorm <= 0.0)) then ! must be at least 2 rays which
    ! are identical.
      if (all(diffNorm <= 0.0)) then ! all the rays are identical.
      ! This one's easy.
        centralRay = rays(1,X:Z)
        coneAngle = 0.0d0

      else ! take your pick of the other 2 diffNorms, they must be
      ! both the same:
        i = smallestTrueIndex(logical(diffNorm > 0.0, bool))
        call calcSmallestConePair(rays(pairs(i)%first,X:Z)&
          , rays(pairs(i)%second,X:Z), centralRay, coneAngle)
      end if

  return
    end if

    ! That's the end of the awkward squad, now we can turn to the nice
    ! cases. First we complete step 1 by calculating the secondary vectors:

    secondaryRays(1,X:Z) = diffVector(1,X:Z) / diffNorm(1)
    secondaryRays(2,X:Z) = diffVector(2,X:Z) / diffNorm(2)
    secondaryRays(3,X:Z) = diffVector(3,X:Z) / diffNorm(3)

    ! The second step in the algorithm is to take cross products between
    ! pairs of these secondary vectors:

    crossRays(1,X:Z) = vectorCross(secondaryRays(pairs(1)%first,X:Z)&
      , secondaryRays(pairs(1)%second,X:Z))
    crossRays(2,X:Z) = vectorCross(secondaryRays(pairs(2)%first,X:Z)&
      , secondaryRays(pairs(2)%second,X:Z))
    crossRays(3,X:Z) = vectorCross(secondaryRays(pairs(3)%first,X:Z)&
      , secondaryRays(pairs(3)%second,X:Z))

    ! These _ought_ to be in the same direction (modulo sign), therefore
    ! they are redundant. We'll keep the biggest (and therefore most
    ! accurate) and discard the rest.

    crossNormSquared(1) = sum(crossRays(1,X:Z) * crossRays(1,X:Z))
    crossNormSquared(2) = sum(crossRays(2,X:Z) * crossRays(2,X:Z))
    crossNormSquared(1) = sum(crossRays(3,X:Z) * crossRays(3,X:Z))
    ! No need to take sqrt

    i = maxLoc1d(crossNormSquared) ! in ssclib/array_utils

    centralRay = crossRays(i,X:Z) / sqrt(crossNormSquared(i))
    cosConeAngle = sum(rays(1,X:Z) * centralRay) ! any of rays() would do,
    ! since they are all the same angle away from centralRay.

    if (cosConeAngle < 0.0) then ! wrong sign.
      centralRay   = -centralRay
      cosConeAngle = -cosConeAngle
    end if

    lastCross = vectorCross(rays(1,X:Z), centralRay)
    sinConeAngle = sqrt(sum(lastCross * lastCross))
    coneAngle = atan2(sinConeAngle, cosConeAngle)

  end subroutine calcEquidistantRay

  !---------------------------------------------------------------------
  subroutine calcSmallestCone(rays, smallestConeCentralRay, smallestConeAngle)
  !---------------------------------------------------------------------
  ! This subroutine takes N directions in 3-space, defined by their
  ! vectors of direction cosines, and finds that conical locus which contains
  ! all N vectors and has the smallest opening angle.
  !
  ! There is already a routine for calculating this cone in the case of
  ! three vectors. A brute-force algorithm would therefore be to simply
  ! work through all the combinations of three vectors and keep the cone
  ! that kept everything. Two improvements come to mind immediately: if
  ! one starts at triplets which include the pair with the largest
  ! separation, and discard any triplets whose members are all inside a
  ! previously-tried cone, then matters ought to go a bit quicker.

!*** put into array_utils?

    implicit none

    real(double), intent(in)  :: rays(:,:)
    real(double), intent(out) :: smallestConeCentralRay(3), smallestConeAngle

    integer(int8), parameter :: X=1, Y=2, Z=3
    integer(int16) ::&
      numRays, i, j, k, ii, jj, kk, numWithinCone, numStretchPoints
    logical ::&
      tripletAvailable(size(rays,1),size(rays,1),size(rays,1)),&
      withinConeMask(size(rays,1))
    real(double) ::&
      trialCentralRay(X:Z), tripletRays(3,X:Z), trialConeAngle,&
      coneDiffNorm, coneDiffNormSquared, diffVector(X:Z),&
      diffNormSquared(size(rays,1))
!    integer(int16), pointer :: indices(:), indicesWithinCone(:)
integer(int16), pointer :: indicesWithinCone(:)
integer(int16) :: indices(size(rays, 1))

!integer :: blub
    !-------------------------------------------------------------------
    ! Start

if (size(rays, 2) .ne. 3) call error('',"")

    numRays = size(rays, 1)
!*** make sure rays are of unit length? err if any of zero length?

    ! Special cases:

    if (numRays < 4) then
      if (numRays <= 0) then
call error('','')
      else if (numRays == 1) then
        smallestConeCentralRay = rays(1,X:Z)
        smallestConeAngle = 0.0d0

      else if (numRays == 2) then
        call calcSmallestConePair(rays(1,X:Z), rays(2,X:Z)&
          , smallestConeCentralRay, smallestConeAngle)

      else if (numRays == 3) then
        call calcSmallestConeTriplet(rays, smallestConeCentralRay&
          , smallestConeAngle)
      end if

  return
    end if

    ! Set up triplet availability cube:

    tripletAvailable = .true. ! default

    ! Knock out all triplets for which the indices are not in increasing order:

    do i = 1, numRays
      do j = 1, i
        tripletAvailable(i, j, :) = .false.
      end do
      do j = i + 1, numRays
        do k = 1, j
          tripletAvailable(i, j, k) = .false.
        end do
      end do
    end do

!    smallestConeAngle = TwoPi ! an impossibly large starting value. Any
!    ! real value must be smaller than this.
smallestConeAngle = -1.0

!    indices => indicize(numRays) ! in ssclib/array_utils
indices = (/(i,i=1,numRays)/)

!blub = 0
    OUTER: do i = 1, numRays - 2
      do j = i + 1, numRays - 1
        do k = j + 1, numRays
        if (.not. tripletAvailable(i, j, k)) cycle

!print *, 'i, j, k:', i, j, k

          tripletAvailable(i, j, k) = .false.

tripletRays(:,X:Z) = rays((/i, j, k/),X:Z)
          call calcSmallestConeTriplet(tripletRays, trialCentralRay&
            , trialConeAngle, numStretchPoints)
!blub = blub + 1
!if (blub == 3) then
!call test_doConePlot(tripletRays, trialCentralRay, trialConeAngle)
!call error('','')
!end if
!          if (trialConeAngle < smallestConeAngle) then
!            smallestConeAngle = trialConeAngle
!            smallestConeCentralRay = trialCentralRay
!          end if 
if (trialConeAngle >= smallestConeAngle) then
  smallestConeAngle = trialConeAngle
  smallestConeCentralRay = trialCentralRay
end if 

          ! Now we want to flag all triplets inside the cone. (Because
          ! any triplet inside this cone must fit inside a still smaller
          ! cone, which therefore could not include the rays that defined
          ! the present cone.) The obvious first step here would be to
          ! calculate the angles between trialCentralRay and the other
          ! rays, then identify those which were smaller than
          ! trialConeAngle. A faster procedure is to find the square of
          ! the norm of the difference between trialCentralRay and any
          ! ray lying in the surface of the cone, then compare this with
          ! the similar number for all the rays. So let's do that.
          !
          ! NOTE though that this is just a means of (hopefully) quickening
          ! the process up. The actual solution cone is that which is
          ! LARGEST yet still is tightly stretched over 3 of the rays.
          ! Whether we manage to bust out of the loop because we find
          ! (see below) that all the rays are now enclosed, or whether we
          ! just run out of triangles, the solution will be contained in
          ! the variables smallestConeCentralRay and smallestConeAngle.

          coneDiffNorm = 2.0 * sin(trialConeAngle / 2.0)
          coneDiffNormSquared = coneDiffNorm * coneDiffNorm

          do ii = 1, numRays
            diffVector(X:Z) = rays(ii,X:Z) - trialCentralRay(X:Z)
            diffNormSquared(ii) = sum(diffVector(X:Z) * diffVector(X:Z))
!print *, 'ii, diffNormSquared(ii):', ii, diffNormSquared(ii)
          end do

          withinConeMask = (diffNormSquared <= coneDiffNormSquared * 0.999d0)
!            * (1.0d0 + epsilon(1.0d0)))
!print *, 'coneDiffNormSquared:', coneDiffNormSquared * (1.0d0 + epsilon(1.0d0))

          numWithinCone = count(withinConeMask)
!print *, 'numWithinCone, numRays:', numWithinCone, numRays
!print *, ''

    if (numWithinCone >= numRays - numStretchPoints) exit OUTER

        if (numWithinCone <= 0) cycle

          allocate(indicesWithinCone(numWithinCone))
          indicesWithinCone = pack(indices, withinConeMask)
          call quickSort(indicesWithinCone) ! in ssclib/sort

          do ii = 1, numWithinCone - 2
            do jj = ii + 1, numWithinCone - 1
              do kk = jj + 1, numWithinCone
                tripletAvailable(&
                  indicesWithinCone(ii),&
                  indicesWithinCone(jj),&
                  indicesWithinCone(kk)&
                ) = .false.
              end do
            end do
          end do

          deallocate(indicesWithinCone)

        end do ! k
      end do ! j
    end do OUTER ! i

!    deallocate(indices)

  end subroutine calcSmallestCone

  !---------------------------------------------------------------------
  subroutine myAddColumnInt32(tab, name, inData, unit, comment, nulls&
  , nullValue)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: name
    character(*), intent(in), optional :: unit, comment
    logical, intent(in), optional :: nulls(:)
    integer(int32), intent(in), optional :: nullValue
    integer(int32), intent(in) :: inData(:)

    character(255) :: localUnit, localComment
    logical :: localNulls(size(inData))
    type(ColumnT) :: col
    integer(int32), pointer :: colData(:)
    integer(int32) :: i, localNullValue

    !-------------------------------------------------------------------
    ! Start

    if (present(unit)) then
      localUnit = unit
    else
      localUnit = ''
    end if
    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if
    if (present(nullValue)) then
      localNullValue = nullValue
    else
      localNullValue = -999
    end if
    if (present(nulls)) then
      localNulls = .true.
      if (size(nulls) .eq. size(inData)) then
        localNulls = nulls
      else if (size(nulls) .gt. size(inData)) then
        localNulls = nulls(1:size(inData))
      else
        localNulls(1:size(nulls)) = nulls
      end if
    else
      localNulls = .false.
    end if

    col = addColumn(tab, name, INTEGER32, trim(localUnit)&
    , comment=trim(localComment))

    colData => int32Data(col)
    colData = inData(1:size(inData))

    if (any(localNulls)) then
      call setNullValue(col, localNullValue)
      do i = 1, size(inData)
        if (localNulls(i)) call setToNull(col, i - 1)
      end do
    end if

    call release(col)

  end subroutine myAddColumnInt32

  !---------------------------------------------------------------------
  subroutine myAddColumnSingle(tab, name, inData, unit, comment, nulls)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: name
    character(*), intent(in), optional :: unit, comment
    logical, intent(in), optional :: nulls(:)
    real(single), intent(in) :: inData(:)

    character(255) :: localUnit, localComment
    logical :: localNulls(size(inData))
    type(ColumnT) :: col
    real(single), pointer :: colData(:)
    integer(int32) :: i

    !-------------------------------------------------------------------
    ! Start

    if (present(unit)) then
      localUnit = unit
    else
      localUnit = ''
    end if
    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if
    if (present(nulls)) then
      localNulls = .true.
      if (size(nulls) .eq. size(inData)) then
        localNulls = nulls
      else if (size(nulls) .gt. size(inData)) then
        localNulls = nulls(1:size(inData))
      else
        localNulls(1:size(nulls)) = nulls
      end if
    else
      localNulls = .false.
    end if

    col = addColumn(tab, name, REAL32, trim(localUnit)&
    , comment=trim(localComment))

    colData => real32Data(col)
    colData = inData(1:size(inData))

    if (any(localNulls)) then
      do i = 1, size(inData)
        if (localNulls(i)) call setToNull(col, i - 1)
      end do
    end if

    call release(col)

  end subroutine myAddColumnSingle

  !---------------------------------------------------------------------
  subroutine myAddColumnDouble(tab, name, inData, unit, comment, nulls)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: name
    character(*), intent(in), optional :: unit, comment
    logical, intent(in), optional :: nulls(:)
    real(double), intent(in) :: inData(:)

    character(255) :: localUnit, localComment
    logical :: localNulls(size(inData))
    type(ColumnT) :: col
    real(double), pointer :: colData(:)
    integer(int32) :: i

    !-------------------------------------------------------------------
    ! Start

    if (present(unit)) then
      localUnit = unit
    else
      localUnit = ''
    end if
    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if
    if (present(nulls)) then
      localNulls = .true.
      if (size(nulls) .eq. size(inData)) then
        localNulls = nulls
      else if (size(nulls) .gt. size(inData)) then
        localNulls = nulls(1:size(inData))
      else
        localNulls(1:size(nulls)) = nulls
      end if
    else
      localNulls = .false.
    end if

    col = addColumn(tab, name, REAL64, trim(localUnit)&
    , comment=trim(localComment))

    colData => real64Data(col)
    colData = inData(1:size(inData))

    if (any(localNulls)) then
      do i = 1, size(inData)
        if (localNulls(i)) call setToNull(col, i - 1)
      end do
    end if

    call release(col)

  end subroutine myAddColumnDouble

  !---------------------------------------------------------------------
  subroutine copyVitalArrayKwds(inArray, outArray)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray, outArray

!    integer :: status

    !-------------------------------------------------------------------
    ! Start

!    call copyAttribute(inArray, 'SLCTEXPR', outArray, status)
!    call copyAttribute(inArray, 'HDUCLASS', outArray, status)
!    call copyAttribute(inArray, 'HDUCLAS1', outArray, status)
!    call copyAttribute(inArray, 'HDUCLAS2', outArray, status)
!    call copyAttribute(inArray, 'HDUVERS1', outArray, status)
call setAttribute(outArray, attribute(inArray, 'SLCTEXPR'))
call setAttribute(outArray, attribute(inArray, 'HDUCLASS'))
call setAttribute(outArray, attribute(inArray, 'HDUCLAS1'))
call setAttribute(outArray, attribute(inArray, 'HDUCLAS2'))
call setAttribute(outArray, attribute(inArray, 'HDUVERS1'))

!    call copyAttribute(inArray, 'CTYPE1', outArray, status)
!    call copyAttribute(inArray, 'CRPIX1', outArray, status)
!    call copyAttribute(inArray, 'CRVAL1', outArray, status)
!    call copyAttribute(inArray, 'CUNIT1', outArray, status)
!    call copyAttribute(inArray, 'CDELT1', outArray, status)
!    call copyAttribute(inArray, 'CTYPE2', outArray, status)
!    call copyAttribute(inArray, 'CRPIX2', outArray, status)
!    call copyAttribute(inArray, 'CRVAL2', outArray, status)
!    call copyAttribute(inArray, 'CUNIT2', outArray, status)
!    call copyAttribute(inArray, 'CDELT2', outArray, status)
call setAttribute(outArray, attribute(inArray, 'CTYPE1'))
call setAttribute(outArray, attribute(inArray, 'CRPIX1'))
call setAttribute(outArray, attribute(inArray, 'CRVAL1'))
call setAttribute(outArray, attribute(inArray, 'CUNIT1'))
call setAttribute(outArray, attribute(inArray, 'CDELT1'))
call setAttribute(outArray, attribute(inArray, 'CTYPE2'))
call setAttribute(outArray, attribute(inArray, 'CRPIX2'))
call setAttribute(outArray, attribute(inArray, 'CRVAL2'))
call setAttribute(outArray, attribute(inArray, 'CUNIT2'))
call setAttribute(outArray, attribute(inArray, 'CDELT2'))

!    call copyAttribute(inArray, 'EXPOSURE', outArray, status)
call setAttribute(outArray, attribute(inArray, 'EXPOSURE'))

  end subroutine copyVitalArrayKwds

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module temp_ssclib
